/* Generated by Cython 3.1.3 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [],
        "include_dirs": [
            "pyflink/fn_execution/table/"
        ],
        "name": "pyflink.fn_execution.table.window_aggregate_fast",
        "sources": [
            "pyflink/fn_execution/table/window_aggregate_fast.pyx"
        ]
    },
    "module_name": "pyflink.fn_execution.table.window_aggregate_fast"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
/* InitLimitedAPI */
#if defined(Py_LIMITED_API) && !defined(CYTHON_LIMITED_API)
  #define CYTHON_LIMITED_API 1
#endif

#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x03080000
    #error Cython requires Python 3.8+.
#else
#define __PYX_ABI_VERSION "3_1_3"
#define CYTHON_HEX_VERSION 0x030103F0
#define CYTHON_FUTURE_DIVISION 1
/* CModulePreamble */
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(_WIN32) && !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #define HAVE_LONG_LONG
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#define __PYX_LIMITED_VERSION_HEX PY_VERSION_HEX
#if defined(GRAALVM_PYTHON)
  /* For very preliminary testing purposes. Most variables are set the same as PyPy.
     The existence of this section does not imply that anything works or is even tested */
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 1
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 1
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(PYPY_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 1
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #if PY_VERSION_HEX < 0x03090000
    #undef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #elif !defined(CYTHON_PEP489_MULTI_PHASE_INIT)
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #undef CYTHON_USE_MODULE_STATE
  #define CYTHON_USE_MODULE_STATE 0
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PYPY_VERSION_NUM >= 0x07030C00)
  #endif
  #undef CYTHON_USE_AM_SEND
  #define CYTHON_USE_AM_SEND 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC (PYPY_VERSION_NUM >= 0x07031100)
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#elif defined(CYTHON_LIMITED_API)
  #ifdef Py_LIMITED_API
    #undef __PYX_LIMITED_VERSION_HEX
    #define __PYX_LIMITED_VERSION_HEX Py_LIMITED_API
  #endif
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #define CYTHON_COMPILING_IN_LIMITED_API 1
  #define CYTHON_COMPILING_IN_GRAAL 0
  #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #undef CYTHON_CLINE_IN_TRACEBACK
  #define CYTHON_CLINE_IN_TRACEBACK 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_TYPE_SPECS
  #define CYTHON_USE_TYPE_SPECS 1
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #endif
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_ASSUME_SAFE_SIZE
  #define CYTHON_ASSUME_SAFE_SIZE 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_GIL
  #define CYTHON_FAST_GIL 0
  #undef CYTHON_METH_FASTCALL
  #define CYTHON_METH_FASTCALL (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #undef CYTHON_USE_SYS_MONITORING
  #define CYTHON_USE_SYS_MONITORING 0
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 0
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND (__PYX_LIMITED_VERSION_HEX >= 0x030A0000)
  #endif
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 0
  #endif
  #undef CYTHON_USE_FREELISTS
  #define CYTHON_USE_FREELISTS 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #define CYTHON_COMPILING_IN_LIMITED_API 0
  #define CYTHON_COMPILING_IN_GRAAL 0
  #ifdef Py_GIL_DISABLED
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 1
  #else
    #define CYTHON_COMPILING_IN_CPYTHON_FREETHREADING 0
  #endif
  #if PY_VERSION_HEX < 0x030A0000
    #undef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #elif !defined(CYTHON_USE_TYPE_SLOTS)
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #ifndef CYTHON_USE_TYPE_SPECS
    #define CYTHON_USE_TYPE_SPECS 0
  #endif
  #ifndef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #ifndef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLIST_INTERNALS)
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING || PY_VERSION_HEX >= 0x030B00A2
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 1
  #elif !defined(CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS)
    #define CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_ASSUME_SAFE_SIZE
    #define CYTHON_ASSUME_SAFE_SIZE 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_FAST_GIL
    #define CYTHON_FAST_GIL 0
  #elif !defined(CYTHON_FAST_GIL)
    #define CYTHON_FAST_GIL (PY_VERSION_HEX < 0x030C00A6)
  #endif
  #ifndef CYTHON_METH_FASTCALL
    #define CYTHON_METH_FASTCALL 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP487_INIT_SUBCLASS
    #define CYTHON_PEP487_INIT_SUBCLASS 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT 1
  #endif
  #ifndef CYTHON_USE_MODULE_STATE
    #define CYTHON_USE_MODULE_STATE 0
  #endif
  #ifndef CYTHON_USE_SYS_MONITORING
    #define CYTHON_USE_SYS_MONITORING (PY_VERSION_HEX >= 0x030d00B1)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE 1
  #endif
  #ifndef CYTHON_USE_AM_SEND
    #define CYTHON_USE_AM_SEND 1
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    #undef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS 0
  #elif !defined(CYTHON_USE_DICT_VERSIONS)
    #define CYTHON_USE_DICT_VERSIONS  (PY_VERSION_HEX < 0x030C00A5 && !CYTHON_USE_MODULE_STATE)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK 1
  #endif
  #ifndef CYTHON_UPDATE_DESCRIPTOR_DOC
    #define CYTHON_UPDATE_DESCRIPTOR_DOC 1
  #endif
  #ifndef CYTHON_USE_FREELISTS
    #define CYTHON_USE_FREELISTS (!CYTHON_COMPILING_IN_CPYTHON_FREETHREADING)
  #endif
#endif
#ifndef CYTHON_FAST_PYCCALL
#define CYTHON_FAST_PYCCALL  CYTHON_FAST_PYCALL
#endif
#ifndef CYTHON_VECTORCALL
#if CYTHON_COMPILING_IN_LIMITED_API
#define CYTHON_VECTORCALL  (__PYX_LIMITED_VERSION_HEX >= 0x030C0000)
#else
#define CYTHON_VECTORCALL  (CYTHON_FAST_PYCCALL && PY_VERSION_HEX >= 0x030800B1)
#endif
#endif
#define CYTHON_BACKPORT_VECTORCALL (CYTHON_METH_FASTCALL && PY_VERSION_HEX < 0x030800B1)
#if CYTHON_USE_PYLONG_INTERNALS
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef CYTHON_LOCK_AND_GIL_DEADLOCK_AVOIDANCE_TIME
  #define CYTHON_LOCK_AND_GIL_DEADLOCK_AVOIDANCE_TIME 100
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(maybe_unused) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(maybe_unused)
        #define CYTHON_UNUSED [[maybe_unused]]
      #endif
    #endif
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
  #define CYTHON_MAYBE_UNUSED_VAR(x) CYTHON_UNUSED_VAR(x)
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON && !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_USE_CPP_STD_MOVE
  #if defined(__cplusplus) && (\
    __cplusplus >= 201103L || (defined(_MSC_VER) && _MSC_VER >= 1600))
    #define CYTHON_USE_CPP_STD_MOVE 1
  #else
    #define CYTHON_USE_CPP_STD_MOVE 0
  #endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
            typedef unsigned char     uint8_t;
            typedef unsigned short    uint16_t;
            typedef unsigned int      uint32_t;
        #else
            typedef unsigned __int8   uint8_t;
            typedef unsigned __int16  uint16_t;
            typedef unsigned __int32  uint32_t;
        #endif
    #endif
    #if _MSC_VER < 1300
        #ifdef _WIN64
            typedef unsigned long long  __pyx_uintptr_t;
        #else
            typedef unsigned int        __pyx_uintptr_t;
        #endif
    #else
        #ifdef _WIN64
            typedef unsigned __int64    __pyx_uintptr_t;
        #else
            typedef unsigned __int32    __pyx_uintptr_t;
        #endif
    #endif
#else
    #include <stdint.h>
    typedef uintptr_t  __pyx_uintptr_t;
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus)
    /* for clang __has_cpp_attribute(fallthrough) is true even before C++17
     * but leads to warnings with -pedantic, since it is a C++17 feature */
    #if ((defined(_MSVC_LANG) && _MSVC_LANG >= 201703L) || __cplusplus >= 201703L)
      #if __has_cpp_attribute(fallthrough)
        #define CYTHON_FALLTHROUGH [[fallthrough]]
      #endif
    #endif
    #ifndef CYTHON_FALLTHROUGH
      #if __has_cpp_attribute(clang::fallthrough)
        #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
      #elif __has_cpp_attribute(gnu::fallthrough)
        #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
      #endif
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif
#ifndef Py_UNREACHABLE
  #define Py_UNREACHABLE()  assert(0); abort()
#endif
#ifdef __cplusplus
  template <typename T>
  struct __PYX_IS_UNSIGNED_IMPL {static const bool value = T(0) < T(-1);};
  #define __PYX_IS_UNSIGNED(type) (__PYX_IS_UNSIGNED_IMPL<type>::value)
#else
  #define __PYX_IS_UNSIGNED(type) (((type)-1) > 0)
#endif
#if CYTHON_COMPILING_IN_PYPY == 1
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x030A0000)
#else
  #define __PYX_NEED_TP_PRINT_SLOT  (PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000)
#endif
#define __PYX_REINTERPRET_FUNCION(func_pointer, other_pointer) ((func_pointer)(void(*)(void))(other_pointer))

/* CInitCode */
#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

/* PythonCompatibility */
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#define __Pyx_BUILTIN_MODULE_NAME "builtins"
#define __Pyx_DefaultClassType PyType_Type
#if CYTHON_COMPILING_IN_LIMITED_API
    #ifndef CO_OPTIMIZED
    static int CO_OPTIMIZED;
    #endif
    #ifndef CO_NEWLOCALS
    static int CO_NEWLOCALS;
    #endif
    #ifndef CO_VARARGS
    static int CO_VARARGS;
    #endif
    #ifndef CO_VARKEYWORDS
    static int CO_VARKEYWORDS;
    #endif
    #ifndef CO_ASYNC_GENERATOR
    static int CO_ASYNC_GENERATOR;
    #endif
    #ifndef CO_GENERATOR
    static int CO_GENERATOR;
    #endif
    #ifndef CO_COROUTINE
    static int CO_COROUTINE;
    #endif
#else
    #ifndef CO_COROUTINE
      #define CO_COROUTINE 0x80
    #endif
    #ifndef CO_ASYNC_GENERATOR
      #define CO_ASYNC_GENERATOR 0x200
    #endif
#endif
static int __Pyx_init_co_variables(void);
#if PY_VERSION_HEX >= 0x030900A4 || defined(Py_IS_TYPE)
  #define __Pyx_IS_TYPE(ob, type) Py_IS_TYPE(ob, type)
#else
  #define __Pyx_IS_TYPE(ob, type) (((const PyObject*)ob)->ob_type == (type))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_Is)
  #define __Pyx_Py_Is(x, y)  Py_Is(x, y)
#else
  #define __Pyx_Py_Is(x, y) ((x) == (y))
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsNone)
  #define __Pyx_Py_IsNone(ob) Py_IsNone(ob)
#else
  #define __Pyx_Py_IsNone(ob) __Pyx_Py_Is((ob), Py_None)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsTrue)
  #define __Pyx_Py_IsTrue(ob) Py_IsTrue(ob)
#else
  #define __Pyx_Py_IsTrue(ob) __Pyx_Py_Is((ob), Py_True)
#endif
#if PY_VERSION_HEX >= 0x030A00B1 || defined(Py_IsFalse)
  #define __Pyx_Py_IsFalse(ob) Py_IsFalse(ob)
#else
  #define __Pyx_Py_IsFalse(ob) __Pyx_Py_Is((ob), Py_False)
#endif
#define __Pyx_NoneAsNull(obj)  (__Pyx_Py_IsNone(obj) ? NULL : (obj))
#if PY_VERSION_HEX >= 0x030900F0 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyObject_GC_IsFinalized(o) PyObject_GC_IsFinalized(o)
#else
  #define __Pyx_PyObject_GC_IsFinalized(o) _PyGC_FINALIZED(o)
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef Py_TPFLAGS_SEQUENCE
  #define Py_TPFLAGS_SEQUENCE 0
#endif
#ifndef Py_TPFLAGS_MAPPING
  #define Py_TPFLAGS_MAPPING 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#ifndef METH_FASTCALL
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #if PY_VERSION_HEX >= 0x030d00A4
  #  define __Pyx_PyCFunctionFast PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords PyCFunctionFastWithKeywords
  #else
  #  define __Pyx_PyCFunctionFast _PyCFunctionFast
  #  define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
  #endif
#endif
#if CYTHON_METH_FASTCALL
  #define __Pyx_METH_FASTCALL METH_FASTCALL
  #define __Pyx_PyCFunction_FastCall __Pyx_PyCFunctionFast
  #define __Pyx_PyCFunction_FastCallWithKeywords __Pyx_PyCFunctionFastWithKeywords
#else
  #define __Pyx_METH_FASTCALL METH_VARARGS
  #define __Pyx_PyCFunction_FastCall PyCFunction
  #define __Pyx_PyCFunction_FastCallWithKeywords PyCFunctionWithKeywords
#endif
#if CYTHON_VECTORCALL
  #define __pyx_vectorcallfunc vectorcallfunc
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  PY_VECTORCALL_ARGUMENTS_OFFSET
  #define __Pyx_PyVectorcall_NARGS(n)  PyVectorcall_NARGS((size_t)(n))
#elif CYTHON_BACKPORT_VECTORCALL
  typedef PyObject *(*__pyx_vectorcallfunc)(PyObject *callable, PyObject *const *args,
                                            size_t nargsf, PyObject *kwnames);
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  ((size_t)1 << (8 * sizeof(size_t) - 1))
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(((size_t)(n)) & ~__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET))
#else
  #define __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET  0
  #define __Pyx_PyVectorcall_NARGS(n)  ((Py_ssize_t)(n))
#endif
#if PY_VERSION_HEX >= 0x030900B1
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_CheckExact(func)
#else
#define __Pyx_PyCFunction_CheckExact(func)  PyCFunction_Check(func)
#endif
#define __Pyx_CyOrPyCFunction_Check(func)  PyCFunction_Check(func)
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  (((PyCFunctionObject*)(func))->m_ml->ml_meth)
#elif !CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyOrPyCFunction_GET_FUNCTION(func)  PyCFunction_GET_FUNCTION(func)
#endif
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_CyOrPyCFunction_GET_FLAGS(func)  (((PyCFunctionObject*)(func))->m_ml->ml_flags)
static CYTHON_INLINE PyObject* __Pyx_CyOrPyCFunction_GET_SELF(PyObject *func) {
    return (__Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_STATIC) ? NULL : ((PyCFunctionObject*)func)->m_self;
}
#endif
static CYTHON_INLINE int __Pyx__IsSameCFunction(PyObject *func, void (*cfunc)(void)) {
#if CYTHON_COMPILING_IN_LIMITED_API
    return PyCFunction_Check(func) && PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
#else
    return PyCFunction_Check(func) && PyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
#endif
}
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCFunction(func, cfunc)
#if __PYX_LIMITED_VERSION_HEX < 0x03090000
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  ((void)m, PyType_FromSpecWithBases(s, b))
  typedef PyObject *(*__Pyx_PyCMethod)(PyObject *, PyTypeObject *, PyObject *const *, size_t, PyObject *);
#else
  #define __Pyx_PyType_FromModuleAndSpec(m, s, b)  PyType_FromModuleAndSpec(m, s, b)
  #define __Pyx_PyCMethod  PyCMethod
#endif
#ifndef METH_METHOD
  #define METH_METHOD 0x200
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)
#elif CYTHON_COMPILING_IN_GRAAL
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) _PyFrame_SetLineNumber((frame), (lineno))
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyThreadState_Current PyThreadState_Get()
#elif !CYTHON_FAST_THREAD_STATE
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x030d00A1
  #define __Pyx_PyThreadState_Current PyThreadState_GetUnchecked()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#endif
#if CYTHON_USE_MODULE_STATE
static CYTHON_INLINE void *__Pyx__PyModule_GetState(PyObject *op)
{
    void *result;
    result = PyModule_GetState(op);
    if (!result)
        Py_FatalError("Couldn't find the module state");
    return result;
}
#define __Pyx_PyModule_GetState(o) (__pyx_mstatetype *)__Pyx__PyModule_GetState(o)
#else
#define __Pyx_PyModule_GetState(op) ((void)op,__pyx_mstate_global)
#endif
#define __Pyx_PyObject_GetSlot(obj, name, func_ctype)  __Pyx_PyType_GetSlot(Py_TYPE((PyObject *) obj), name, func_ctype)
#define __Pyx_PyObject_TryGetSlot(obj, name, func_ctype) __Pyx_PyType_TryGetSlot(Py_TYPE(obj), name, func_ctype)
#define __Pyx_PyObject_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#define __Pyx_PyObject_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSubSlot(Py_TYPE(obj), sub, name, func_ctype)
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((type)->name)
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype) __Pyx_PyType_GetSlot(type, name, func_ctype)
  #define __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype) (((type)->sub) ? ((type)->sub->name) : NULL)
  #define __Pyx_PyType_TryGetSubSlot(type, sub, name, func_ctype) __Pyx_PyType_GetSubSlot(type, sub, name, func_ctype)
#else
  #define __Pyx_PyType_GetSlot(type, name, func_ctype)  ((func_ctype) PyType_GetSlot((type), Py_##name))
  #define __Pyx_PyType_TryGetSlot(type, name, func_ctype)\
    ((__PYX_LIMITED_VERSION_HEX >= 0x030A0000 ||\
     (PyType_GetFlags(type) & Py_TPFLAGS_HEAPTYPE) || __Pyx_get_runtime_version() >= 0x030A0000) ?\
     __Pyx_PyType_GetSlot(type, name, func_ctype) : NULL)
  #define __Pyx_PyType_GetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_GetSlot(obj, name, func_ctype)
  #define __Pyx_PyType_TryGetSubSlot(obj, sub, name, func_ctype) __Pyx_PyType_TryGetSlot(obj, name, func_ctype)
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
#define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStrWithError(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStr(PyObject *dict, PyObject *name) {
    PyObject *res = __Pyx_PyDict_GetItemStrWithError(dict, name);
    if (res == NULL) PyErr_Clear();
    return res;
}
#elif !CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07020000
#define __Pyx_PyDict_GetItemStrWithError  PyDict_GetItemWithError
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#else
static CYTHON_INLINE PyObject * __Pyx_PyDict_GetItemStrWithError(PyObject *dict, PyObject *name) {
#if CYTHON_COMPILING_IN_PYPY
    return PyDict_GetItem(dict, name);
#else
    PyDictEntry *ep;
    PyDictObject *mp = (PyDictObject*) dict;
    long hash = ((PyStringObject *) name)->ob_shash;
    assert(hash != -1);
    ep = (mp->ma_lookup)(mp, name, hash);
    if (ep == NULL) {
        return NULL;
    }
    return ep->me_value;
#endif
}
#define __Pyx_PyDict_GetItemStr           PyDict_GetItem
#endif
#if CYTHON_USE_TYPE_SLOTS
  #define __Pyx_PyType_GetFlags(tp)   (((PyTypeObject *)tp)->tp_flags)
  #define __Pyx_PyType_HasFeature(type, feature)  ((__Pyx_PyType_GetFlags(type) & (feature)) != 0)
#else
  #define __Pyx_PyType_GetFlags(tp)   (PyType_GetFlags((PyTypeObject *)tp))
  #define __Pyx_PyType_HasFeature(type, feature)  PyType_HasFeature(type, feature)
#endif
#define __Pyx_PyObject_GetIterNextFunc(iterator)  __Pyx_PyObject_GetSlot(iterator, tp_iternext, iternextfunc)
#if CYTHON_USE_TYPE_SPECS && PY_VERSION_HEX >= 0x03080000
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  {\
    PyTypeObject *type = Py_TYPE((PyObject*)obj);\
    assert(__Pyx_PyType_HasFeature(type, Py_TPFLAGS_HEAPTYPE));\
    PyObject_GC_Del(obj);\
    Py_DECREF(type);\
}
#else
#define __Pyx_PyHeapTypeObject_GC_Del(obj)  PyObject_GC_Del(obj)
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_ReadChar(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((void)u, 1114111U)
  #define __Pyx_PyUnicode_KIND(u)         ((void)u, (0))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)u)
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)k, PyUnicode_ReadChar((PyObject*)(d), i))
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GetLength(u))
#else
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_READY(op)       (0)
  #else
    #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                                0 : _PyUnicode_Ready((PyObject *)(op)))
  #endif
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         ((int)PyUnicode_KIND(u))
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, (Py_UCS4) ch)
  #if PY_VERSION_HEX >= 0x030C0000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #else
    #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x03090000
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : ((PyCompactUnicodeObject *)(u))->wstr_length))
    #else
    #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
    #endif
  #endif
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #if !defined(PyUnicode_DecodeUnicodeEscape)
    #define PyUnicode_DecodeUnicodeEscape(s, size, errors)  PyUnicode_Decode(s, size, "unicode_escape", errors)
  #endif
  #if !defined(PyUnicode_Contains)
    #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
  #endif
  #if !defined(PyByteArray_Check)
    #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
  #endif
  #if !defined(PyObject_Format)
    #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
  #endif
#endif
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if CYTHON_COMPILING_IN_CPYTHON
  #define __Pyx_PySequence_ListKeepNew(obj)\
    (likely(PyList_CheckExact(obj) && Py_REFCNT(obj) == 1) ? __Pyx_NewRef(obj) : PySequence_List(obj))
#else
  #define __Pyx_PySequence_ListKeepNew(obj)  PySequence_List(obj)
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        __Pyx_IS_TYPE(obj, &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyList_GetItemRef(o, i) (likely((i) >= 0) ? PySequence_GetItem(o, i) : (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
  #else
    #define __Pyx_PyList_GetItemRef(o, i) PySequence_ITEM(o, i)
  #endif
#elif CYTHON_COMPILING_IN_LIMITED_API || !CYTHON_ASSUME_SAFE_MACROS
  #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    #define __Pyx_PyList_GetItemRef(o, i) PyList_GetItemRef(o, i)
  #else
    #define __Pyx_PyList_GetItemRef(o, i) __Pyx_XNewRef(PyList_GetItem(o, i))
  #endif
#else
  #define __Pyx_PyList_GetItemRef(o, i) __Pyx_NewRef(PyList_GET_ITEM(o, i))
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyDict_GetItemRef(dict, key, result) PyDict_GetItemRef(dict, key, result)
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyObject_GetItem(dict, key);
  if (*result == NULL) {
    if (PyErr_ExceptionMatches(PyExc_KeyError)) {
      PyErr_Clear();
      return 0;
    }
    return -1;
  }
  return 1;
}
#else
static CYTHON_INLINE int __Pyx_PyDict_GetItemRef(PyObject *dict, PyObject *key, PyObject **result) {
  *result = PyDict_GetItemWithError(dict, key);
  if (*result == NULL) {
    return PyErr_Occurred() ? -1 : 0;
  }
  Py_INCREF(*result);
  return 1;
}
#endif
#if defined(CYTHON_DEBUG_VISIT_CONST) && CYTHON_DEBUG_VISIT_CONST
  #define __Pyx_VISIT_CONST(obj)  Py_VISIT(obj)
#else
  #define __Pyx_VISIT_CONST(obj)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_ITEM(o, i) PySequence_ITEM(o, i)
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) (PyTuple_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GET_ITEM(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) (PyList_SET_ITEM(o, i, v), (0))
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GET_ITEM(o, i)
#else
  #define __Pyx_PySequence_ITEM(o, i) PySequence_GetItem(o, i)
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
  #define __Pyx_PyTuple_SET_ITEM(o, i, v) PyTuple_SetItem(o, i, v)
  #define __Pyx_PyTuple_GET_ITEM(o, i) PyTuple_GetItem(o, i)
  #define __Pyx_PyList_SET_ITEM(o, i, v) PyList_SetItem(o, i, v)
  #define __Pyx_PyList_GET_ITEM(o, i) PyList_GetItem(o, i)
#endif
#if CYTHON_ASSUME_SAFE_SIZE
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_GET_SIZE(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_GET_SIZE(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_GET_SIZE(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_GET_SIZE(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_GET_SIZE(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GET_LENGTH(o)
#else
  #define __Pyx_PyTuple_GET_SIZE(o) PyTuple_Size(o)
  #define __Pyx_PyList_GET_SIZE(o) PyList_Size(o)
  #define __Pyx_PySet_GET_SIZE(o) PySet_Size(o)
  #define __Pyx_PyBytes_GET_SIZE(o) PyBytes_Size(o)
  #define __Pyx_PyByteArray_GET_SIZE(o) PyByteArray_Size(o)
  #define __Pyx_PyUnicode_GET_LENGTH(o) PyUnicode_GetLength(o)
#endif
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
  #define __Pyx_PyImport_AddModuleRef(name) PyImport_AddModuleRef(name)
#else
  static CYTHON_INLINE PyObject *__Pyx_PyImport_AddModuleRef(const char *name) {
      PyObject *module = PyImport_AddModule(name);
      Py_XINCREF(module);
      return module;
  }
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_InternFromString)
  #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
#endif
#define __Pyx_PyLong_FromHash_t PyLong_FromSsize_t
#define __Pyx_PyLong_AsHash_t   __Pyx_PyIndex_AsSsize_t
#if __PYX_LIMITED_VERSION_HEX >= 0x030A0000
    #define __Pyx_PySendResult PySendResult
#else
    typedef enum {
        PYGEN_RETURN = 0,
        PYGEN_ERROR = -1,
        PYGEN_NEXT = 1,
    } __Pyx_PySendResult;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX < 0x030A00A3
  typedef __Pyx_PySendResult (*__Pyx_pyiter_sendfunc)(PyObject *iter, PyObject *value, PyObject **result);
#else
  #define __Pyx_pyiter_sendfunc sendfunc
#endif
#if !CYTHON_USE_AM_SEND
#define __PYX_HAS_PY_AM_SEND 0
#elif __PYX_LIMITED_VERSION_HEX >= 0x030A0000
#define __PYX_HAS_PY_AM_SEND 1
#else
#define __PYX_HAS_PY_AM_SEND 2  // our own backported implementation
#endif
#if __PYX_HAS_PY_AM_SEND < 2
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
#else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
        __Pyx_pyiter_sendfunc am_send;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_SlotTpAsAsync(s) ((PyAsyncMethods*)(s))
#endif
#if CYTHON_USE_AM_SEND && PY_VERSION_HEX < 0x030A00F0
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (1UL << 21)
#else
    #define __Pyx_TPFLAGS_HAVE_AM_SEND (0)
#endif
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyInterpreterState_Get() PyInterpreterState_Get()
#else
#define __Pyx_PyInterpreterState_Get() PyThreadState_Get()->interp
#endif
#if CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030A0000
#ifdef __cplusplus
extern "C"
#endif
PyAPI_FUNC(void *) PyMem_Calloc(size_t nelem, size_t elsize);
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static int __Pyx_init_co_variable(PyObject *inspect, const char* name, int *write_to) {
    int value;
    PyObject *py_value = PyObject_GetAttrString(inspect, name);
    if (!py_value) return 0;
    value = (int) PyLong_AsLong(py_value);
    Py_DECREF(py_value);
    *write_to = value;
    return value != -1 || !PyErr_Occurred();
}
static int __Pyx_init_co_variables(void) {
    PyObject *inspect;
    int result;
    inspect = PyImport_ImportModule("inspect");
    result =
#if !defined(CO_OPTIMIZED)
        __Pyx_init_co_variable(inspect, "CO_OPTIMIZED", &CO_OPTIMIZED) &&
#endif
#if !defined(CO_NEWLOCALS)
        __Pyx_init_co_variable(inspect, "CO_NEWLOCALS", &CO_NEWLOCALS) &&
#endif
#if !defined(CO_VARARGS)
        __Pyx_init_co_variable(inspect, "CO_VARARGS", &CO_VARARGS) &&
#endif
#if !defined(CO_VARKEYWORDS)
        __Pyx_init_co_variable(inspect, "CO_VARKEYWORDS", &CO_VARKEYWORDS) &&
#endif
#if !defined(CO_ASYNC_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_ASYNC_GENERATOR", &CO_ASYNC_GENERATOR) &&
#endif
#if !defined(CO_GENERATOR)
        __Pyx_init_co_variable(inspect, "CO_GENERATOR", &CO_GENERATOR) &&
#endif
#if !defined(CO_COROUTINE)
        __Pyx_init_co_variable(inspect, "CO_COROUTINE", &CO_COROUTINE) &&
#endif
        1;
    Py_DECREF(inspect);
    return result ? 0 : -1;
}
#else
static int __Pyx_init_co_variables(void) {
    return 0;  // It's a limited API-only feature
}
#endif

/* MathInitCode */
#if defined(_WIN32) || defined(WIN32) || defined(MS_WINDOWS)
  #ifndef _USE_MATH_DEFINES
    #define _USE_MATH_DEFINES
  #endif
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#ifndef CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#define CYTHON_CLINE_IN_TRACEBACK_RUNTIME 0
#endif
#ifndef CYTHON_CLINE_IN_TRACEBACK
#define CYTHON_CLINE_IN_TRACEBACK CYTHON_CLINE_IN_TRACEBACK_RUNTIME
#endif
#if CYTHON_CLINE_IN_TRACEBACK
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; __pyx_clineno = __LINE__; (void) __pyx_clineno; }
#else
#define __PYX_MARK_ERR_POS(f_index, lineno)  { __pyx_filename = __pyx_f[f_index]; (void) __pyx_filename; __pyx_lineno = lineno; (void) __pyx_lineno; (void) __pyx_clineno; }
#endif
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pyflink__fn_execution__table__window_aggregate_fast
#define __PYX_HAVE_API__pyflink__fn_execution__table__window_aggregate_fast
/* Early includes */
#include <stdint.h>
#include <string.h>
#include <stdlib.h>
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s);
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char*);
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AS_STRING(s)
#else
    #define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AsString(s))
    #define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AsString(s))
    #define __Pyx_PyByteArray_AsString(s) PyByteArray_AsString(s)
#endif
#define __Pyx_PyObject_AsWritableString(s)    ((char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*)(__pyx_uintptr_t) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#define __Pyx_PyUnicode_FromOrdinal(o)       PyUnicode_FromOrdinal((int)o)
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
static CYTHON_INLINE PyObject *__Pyx_NewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_NewRef)
    return Py_NewRef(obj);
#else
    Py_INCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_XNewRef(PyObject *obj) {
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030a0000 || defined(Py_XNewRef)
    return Py_XNewRef(obj);
#else
    Py_XINCREF(obj);
    return obj;
#endif
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b);
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t);
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject*);
#if CYTHON_ASSUME_SAFE_MACROS
#define __Pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AS_DOUBLE(x)
#else
#define __Pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#define __Pyx_PyFloat_AS_DOUBLE(x) PyFloat_AsDouble(x)
#endif
#define __Pyx_PyFloat_AsFloat(x) ((float) __Pyx_PyFloat_AsDouble(x))
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#if CYTHON_USE_PYLONG_INTERNALS
  #if PY_VERSION_HEX >= 0x030C00A7
  #ifndef _PyLong_SIGN_MASK
    #define _PyLong_SIGN_MASK 3
  #endif
  #ifndef _PyLong_NON_SIZE_BITS
    #define _PyLong_NON_SIZE_BITS 3
  #endif
  #define __Pyx_PyLong_Sign(x)  (((PyLongObject*)x)->long_value.lv_tag & _PyLong_SIGN_MASK)
  #define __Pyx_PyLong_IsNeg(x)  ((__Pyx_PyLong_Sign(x) & 2) != 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (!__Pyx_PyLong_IsNeg(x))
  #define __Pyx_PyLong_IsZero(x)  (__Pyx_PyLong_Sign(x) & 1)
  #define __Pyx_PyLong_IsPos(x)  (__Pyx_PyLong_Sign(x) == 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  (__Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  ((Py_ssize_t) (((PyLongObject*)x)->long_value.lv_tag >> _PyLong_NON_SIZE_BITS))
  #define __Pyx_PyLong_SignedDigitCount(x)\
        ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * __Pyx_PyLong_DigitCount(x))
  #if defined(PyUnstable_Long_IsCompact) && defined(PyUnstable_Long_CompactValue)
    #define __Pyx_PyLong_IsCompact(x)     PyUnstable_Long_IsCompact((PyLongObject*) x)
    #define __Pyx_PyLong_CompactValue(x)  PyUnstable_Long_CompactValue((PyLongObject*) x)
  #else
    #define __Pyx_PyLong_IsCompact(x)     (((PyLongObject*)x)->long_value.lv_tag < (2 << _PyLong_NON_SIZE_BITS))
    #define __Pyx_PyLong_CompactValue(x)  ((1 - (Py_ssize_t) __Pyx_PyLong_Sign(x)) * (Py_ssize_t) __Pyx_PyLong_Digits(x)[0])
  #endif
  typedef Py_ssize_t  __Pyx_compact_pylong;
  typedef size_t  __Pyx_compact_upylong;
  #else
  #define __Pyx_PyLong_IsNeg(x)  (Py_SIZE(x) < 0)
  #define __Pyx_PyLong_IsNonNeg(x)  (Py_SIZE(x) >= 0)
  #define __Pyx_PyLong_IsZero(x)  (Py_SIZE(x) == 0)
  #define __Pyx_PyLong_IsPos(x)  (Py_SIZE(x) > 0)
  #define __Pyx_PyLong_CompactValueUnsigned(x)  ((Py_SIZE(x) == 0) ? 0 : __Pyx_PyLong_Digits(x)[0])
  #define __Pyx_PyLong_DigitCount(x)  __Pyx_sst_abs(Py_SIZE(x))
  #define __Pyx_PyLong_SignedDigitCount(x)  Py_SIZE(x)
  #define __Pyx_PyLong_IsCompact(x)  (Py_SIZE(x) == 0 || Py_SIZE(x) == 1 || Py_SIZE(x) == -1)
  #define __Pyx_PyLong_CompactValue(x)\
        ((Py_SIZE(x) == 0) ? (sdigit) 0 : ((Py_SIZE(x) < 0) ? -(sdigit)__Pyx_PyLong_Digits(x)[0] : (sdigit)__Pyx_PyLong_Digits(x)[0]))
  typedef sdigit  __Pyx_compact_pylong;
  typedef digit  __Pyx_compact_upylong;
  #endif
  #if PY_VERSION_HEX >= 0x030C00A5
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->long_value.ob_digit)
  #else
  #define __Pyx_PyLong_Digits(x)  (((PyLongObject*)x)->ob_digit)
  #endif
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#elif __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeASCII(c_str, size, NULL)
#else
  #define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
/* PretendToInitialize */
#ifdef __cplusplus
#if __cplusplus > 201103L
#include <type_traits>
#endif
template <typename T>
static void __Pyx_pretend_to_initialize(T* ptr) {
#if __cplusplus > 201103L
    if ((std::is_trivially_default_constructible<T>::value))
#endif
        *ptr = T();
    (void)ptr;
}
#else
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }
#endif


#if !CYTHON_USE_MODULE_STATE
static PyObject *__pyx_m = NULL;
#endif
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * const __pyx_cfilenm = __FILE__;
static const char *__pyx_filename;

/* #### Code section: filename_table ### */

static const char* const __pyx_f[] = {
  "pyflink/fn_execution/table/window_aggregate_fast.pyx",
  "<stringsource>",
  "pyflink/fn_execution/stream_fast.pxd",
  "pyflink/fn_execution/coder_impl_fast.pxd",
  "pyflink/fn_execution/table/aggregate_fast.pxd",
};
/* #### Code section: utility_code_proto_before_types ### */
/* Profile_config.proto */
#ifndef CYTHON_PROFILE
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
  #define CYTHON_PROFILE 0
#else
  #define CYTHON_PROFILE 1
#endif
#endif
#ifndef CYTHON_TRACE_NOGIL
  #define CYTHON_TRACE_NOGIL 0
#else
  #if CYTHON_TRACE_NOGIL && !defined(CYTHON_TRACE)
    #define CYTHON_TRACE 1
  #endif
#endif
#ifndef CYTHON_TRACE
  #define CYTHON_TRACE 0
#endif
#if CYTHON_PROFILE || CYTHON_TRACE
#if CYTHON_USE_SYS_MONITORING
    typedef enum {
        __Pyx_Monitoring_PY_START = 0,
        __Pyx_Monitoring_PY_RETURN,
        __Pyx_Monitoring_PY_UNWIND,
        __Pyx_Monitoring_LINE,
        __Pyx_Monitoring_RAISE,
        __Pyx_Monitoring_RERAISE,
        __Pyx_Monitoring_EXCEPTION_HANDLED,
        __Pyx_Monitoring_PY_RESUME,
        __Pyx_Monitoring_PY_YIELD,
        __Pyx_Monitoring_STOP_ITERATION,
    } __Pyx_Monitoring_Event_Index;
    static const unsigned char __Pyx_MonitoringEventTypes[] = {
        PY_MONITORING_EVENT_PY_START,
        PY_MONITORING_EVENT_PY_RETURN,
        PY_MONITORING_EVENT_PY_UNWIND,
        PY_MONITORING_EVENT_LINE,
        PY_MONITORING_EVENT_RAISE,
        PY_MONITORING_EVENT_RERAISE,
        PY_MONITORING_EVENT_EXCEPTION_HANDLED,
        PY_MONITORING_EVENT_PY_RESUME,
        PY_MONITORING_EVENT_PY_YIELD,
        PY_MONITORING_EVENT_STOP_ITERATION,
    };
    #define __Pyx_MonitoringEventTypes_CyFunc_count (sizeof(__Pyx_MonitoringEventTypes) - 3)
    #define __Pyx_MonitoringEventTypes_CyGen_count (sizeof(__Pyx_MonitoringEventTypes))
#endif
#endif

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __PYX_CYTHON_ATOMICS_ENABLED() CYTHON_ATOMICS
#define __PYX_GET_CYTHON_COMPILING_IN_CPYTHON_FREETHREADING() CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __pyx_atomic_int_type int
#define __pyx_nonatomic_int_type int
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__))
    #include <stdatomic.h>
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)))
    #include <atomic>
#endif
#if CYTHON_ATOMICS && (defined(__STDC_VERSION__) &&\
                        (__STDC_VERSION__ >= 201112L) &&\
                        !defined(__STDC_NO_ATOMICS__) &&\
                       ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type atomic_int
    #define __pyx_atomic_ptr_type atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) atomic_fetch_add_explicit(value, 1, memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) atomic_fetch_add_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) atomic_fetch_sub_explicit(value, 1, memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) atomic_load(value)
    #define __pyx_atomic_store(value, new_value) atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) atomic_load_explicit(value, memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) atomic_load_explicit(value, memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C atomics"
    #endif
#elif CYTHON_ATOMICS && (defined(__cplusplus) && (\
                    (__cplusplus >= 201103L) ||\
\
                    (defined(_MSC_VER) && _MSC_VER >= 1700)) &&\
                    ATOMIC_INT_LOCK_FREE == 2)
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type std::atomic_int
    #define __pyx_atomic_ptr_type std::atomic_uintptr_t
    #define __pyx_nonatomic_ptr_type uintptr_t
    #define __pyx_atomic_incr_relaxed(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_relaxed)
    #define __pyx_atomic_incr_acq_rel(value) std::atomic_fetch_add_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_decr_acq_rel(value) std::atomic_fetch_sub_explicit(value, 1, std::memory_order_acq_rel)
    #define __pyx_atomic_sub(value, arg) std::atomic_fetch_sub(value, arg)
    #define __pyx_atomic_int_cmp_exchange(value, expected, desired) std::atomic_compare_exchange_strong(value, expected, desired)
    #define __pyx_atomic_load(value) std::atomic_load(value)
    #define __pyx_atomic_store(value, new_value) std::atomic_store(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) std::atomic_load_explicit(value, std::memory_order_relaxed)
    #define __pyx_atomic_pointer_load_acquire(value) std::atomic_load_explicit(value, std::memory_order_acquire)
    #define __pyx_atomic_pointer_exchange(value, new_value) std::atomic_exchange(value, (__pyx_nonatomic_ptr_type)new_value)
    #if defined(__PYX_DEBUG_ATOMICS) && defined(_MSC_VER)
        #pragma message ("Using standard C++ atomics")
    #elif defined(__PYX_DEBUG_ATOMICS)
        #warning "Using standard C++ atomics"
    #endif
#elif CYTHON_ATOMICS && (__GNUC__ >= 5 || (__GNUC__ == 4 &&\
                    (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL__ >= 2))))
    #define __pyx_atomic_ptr_type void*
    #define __pyx_atomic_incr_relaxed(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) __sync_fetch_and_sub(value, 1)
    #define __pyx_atomic_sub(value, arg) __sync_fetch_and_sub(value, arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = __sync_val_compare_and_swap(value, *expected, desired);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_store(value, new_value) __sync_lock_test_and_set(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_load_acquire(value) __sync_fetch_and_add(value, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) __sync_lock_test_and_set(value, (__pyx_atomic_ptr_type)new_value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER)
    #include <intrin.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type long
    #define __pyx_atomic_ptr_type void*
    #undef __pyx_nonatomic_int_type
    #define __pyx_nonatomic_int_type long
    #pragma intrinsic (_InterlockedExchangeAdd, _InterlockedExchange, _InterlockedCompareExchange, _InterlockedCompareExchangePointer, _InterlockedExchangePointer)
    #define __pyx_atomic_incr_relaxed(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_incr_acq_rel(value) _InterlockedExchangeAdd(value, 1)
    #define __pyx_atomic_decr_acq_rel(value) _InterlockedExchangeAdd(value, -1)
    #define __pyx_atomic_sub(value, arg) _InterlockedExchangeAdd(value, -arg)
    static CYTHON_INLINE int __pyx_atomic_int_cmp_exchange(__pyx_atomic_int_type* value, __pyx_nonatomic_int_type* expected, __pyx_nonatomic_int_type desired) {
        __pyx_nonatomic_int_type old = _InterlockedCompareExchange(value, desired, *expected);
        int result = old == *expected;
        *expected = old;
        return result;
    }
    #define __pyx_atomic_load(value) _InterlockedExchangeAdd(value, 0)
    #define __pyx_atomic_store(value, new_value) _InterlockedExchange(value, new_value)
    #define __pyx_atomic_pointer_load_relaxed(value) *(void * volatile *)value
    #define __pyx_atomic_pointer_load_acquire(value) _InterlockedCompareExchangePointer(value, 0, 0)
    #define __pyx_atomic_pointer_exchange(value, new_value) _InterlockedExchangePointer(value, (__pyx_atomic_ptr_type)new_value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_relaxed(__pyx_get_slice_count_pointer(memview))
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_acq_rel(__pyx_get_slice_count_pointer(memview))
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* IncludeStructmemberH.proto */
#include <structmember.h>

/* CriticalSections.proto */
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
#define __Pyx_PyCriticalSection void*
#define __Pyx_PyCriticalSection2 void*
#define __Pyx_PyCriticalSection_Begin1(cs, arg) (void)cs
#define __Pyx_PyCriticalSection_Begin2(cs, arg1, arg2) (void)cs
#define __Pyx_PyCriticalSection_End1(cs)
#define __Pyx_PyCriticalSection_End2(cs)
#else
#define __Pyx_PyCriticalSection PyCriticalSection
#define __Pyx_PyCriticalSection2 PyCriticalSection2
#define __Pyx_PyCriticalSection_Begin1 PyCriticalSection_Begin
#define __Pyx_PyCriticalSection_Begin2 PyCriticalSection2_Begin
#define __Pyx_PyCriticalSection_End1 PyCriticalSection_End
#define __Pyx_PyCriticalSection_End2 PyCriticalSection2_End
#endif
#if PY_VERSION_HEX < 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_BEGIN_CRITICAL_SECTION(o) {
#define __Pyx_END_CRITICAL_SECTION() }
#else
#define __Pyx_BEGIN_CRITICAL_SECTION Py_BEGIN_CRITICAL_SECTION
#define __Pyx_END_CRITICAL_SECTION Py_END_CRITICAL_SECTION
#endif

/* #### Code section: numeric_typedefs ### */
/* #### Code section: complex_type_declarations ### */
/* #### Code section: type_declarations ### */

/*--- Type declarations ---*/
struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream;
struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream;
struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream;
struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction;
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction;
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction;
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction;

/* "pyflink/fn_execution/coder_impl_fast.pxd":25
 *     InputStream, OutputStream
 * 
 * cdef enum InternalRowKind:             # <<<<<<<<<<<<<<
 *     INSERT = 0
 *     UPDATE_BEFORE = 1
*/
enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind {
  __pyx_e_7pyflink_12fn_execution_15coder_impl_fast_INSERT = 0,
  __pyx_e_7pyflink_12fn_execution_15coder_impl_fast_UPDATE_BEFORE = 1,
  __pyx_e_7pyflink_12fn_execution_15coder_impl_fast_UPDATE_AFTER = 2,
  __pyx_e_7pyflink_12fn_execution_15coder_impl_fast_DELETE = 3
};

/* "pyflink/fn_execution/stream_fast.pxd":21
 * from libc.stdint cimport int8_t, int16_t, int32_t, int64_t
 * 
 * cdef class LengthPrefixInputStream:             # <<<<<<<<<<<<<<
 *     cdef size_t read(self, char**data)
 *     cdef size_t available(self)
*/
struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream *__pyx_vtab;
};


/* "pyflink/fn_execution/stream_fast.pxd":25
 *     cdef size_t available(self)
 * 
 * cdef class LengthPrefixOutputStream:             # <<<<<<<<<<<<<<
 *     cdef void write(self, char*data, size_t length)
 *     cpdef void flush(self)
*/
struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream *__pyx_vtab;
};


/* "pyflink/fn_execution/stream_fast.pxd":30
 *     cpdef void close(self)
 * 
 * cdef class InputStream:             # <<<<<<<<<<<<<<
 *     cdef char*_input_data
 *     cdef size_t _input_pos
*/
struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_InputStream *__pyx_vtab;
  char *_input_data;
  size_t _input_pos;
};


/* "pyflink/fn_execution/stream_fast.pxd":44
 *     cdef bytes read_bytes(self)
 * 
 * cdef class OutputStream:             # <<<<<<<<<<<<<<
 *     cdef char*buffer
 *     cdef size_t buffer_size
*/
struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_OutputStream *__pyx_vtab;
  char *buffer;
  size_t buffer_size;
  size_t pos;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":31
 *     DELETE = 3
 * 
 * cdef class InternalRow:             # <<<<<<<<<<<<<<
 *     cdef readonly list values
 *     cdef readonly InternalRowKind row_kind
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_vtab;
  PyObject *values;
  enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind row_kind;
  PyObject *field_names;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":39
 *     cdef bint is_accumulate_msg(self)
 * 
 * cdef class MaskUtils:             # <<<<<<<<<<<<<<
 *     cdef bint*_mask
 *     cdef unsigned char*_mask_byte_search_table
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils *__pyx_vtab;
  int *_mask;
  unsigned char *_mask_byte_search_table;
  unsigned char *_row_kind_byte_table;
  size_t _field_count;
  size_t _leading_complete_bytes_num;
  size_t _remaining_bits_num;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":50
 *     cdef bint*read_mask(self, InputStream input_stream)
 * 
 * cdef class LengthPrefixBaseCoderImpl:             # <<<<<<<<<<<<<<
 *     cdef FieldCoderImpl _field_coder
 *     cdef OutputStream _data_out_stream
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl *__pyx_vtab;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *_field_coder;
  struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *_data_out_stream;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":58
 *     cdef void _write_data_to_output_stream(self, LengthPrefixOutputStream output_stream)
 * 
 * cdef class FieldCoderImpl:             # <<<<<<<<<<<<<<
 *     cpdef encode_to_stream(self, value, OutputStream out_stream)
 *     cpdef decode_from_stream(self, InputStream in_stream, size_t size)
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *__pyx_vtab;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":64
 *     cpdef decode(self, encoded)
 * 
 * cdef class InputStreamWrapper:             # <<<<<<<<<<<<<<
 *     cdef ValueCoderImpl _value_coder
 *     cdef LengthPrefixInputStream _input_stream
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper *__pyx_vtab;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl *_value_coder;
  struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream *_input_stream;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":71
 *     cpdef next(self)
 * 
 * cdef class IterableCoderImpl(LengthPrefixBaseCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef char*_end_message
 *     cdef bint _separated_with_end_message
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl __pyx_base;
  char *_end_message;
  int _separated_with_end_message;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":75
 *     cdef bint _separated_with_end_message
 * 
 * cdef class ValueCoderImpl(LengthPrefixBaseCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":80
 * cdef unsigned char ROW_KIND_BIT_SIZE
 * 
 * cdef class FlattenRowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef list _field_coders
 *     cdef list _reuse_flatten_row
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *_field_coders;
  PyObject *_reuse_flatten_row;
  size_t _field_count;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils *_mask_utils;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":86
 *     cdef MaskUtils _mask_utils
 * 
 * cdef class RowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef list _field_coders
 *     cdef size_t _field_count
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *_field_coders;
  size_t _field_count;
  PyObject *_field_names;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils *_mask_utils;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":92
 *     cdef MaskUtils _mask_utils
 * 
 * cdef class ArrowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _schema
 *     cdef list _field_types
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *_schema;
  PyObject *_field_types;
  PyObject *_timezone;
  PyObject *_resettable_io;
  PyObject *_batch_reader;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":101
 *     cdef list decode_one_batch_from_stream(self, InputStream in_stream, size_t size)
 * 
 * cdef class OverWindowArrowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef ArrowCoderImpl _arrow_coder
 *     cdef IntCoderImpl _int_coder
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl *_arrow_coder;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl *_int_coder;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":105
 *     cdef IntCoderImpl _int_coder
 * 
 * cdef class TinyIntCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":108
 *     pass
 * 
 * cdef class SmallIntCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":111
 *     pass
 * 
 * cdef class IntCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":114
 *     pass
 * 
 * cdef class BigIntCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":117
 *     pass
 * 
 * cdef class BooleanCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":120
 *     pass
 * 
 * cdef class FloatCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":123
 *     pass
 * 
 * cdef class DoubleCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":126
 *     pass
 * 
 * cdef class BinaryCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":129
 *     pass
 * 
 * cdef class CharCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":132
 *     pass
 * 
 * cdef class BigDecimalCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef CharCoderImpl _value_coder
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl *_value_coder;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":135
 *     cdef CharCoderImpl _value_coder
 * 
 * cdef class DecimalCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _context
 *     cdef object _scale_format
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *_context;
  PyObject *_scale_format;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl *_value_coder;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":140
 *     cdef CharCoderImpl _value_coder
 * 
 * cdef class DateCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef int _EPOCH_ORDINAL
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  int _EPOCH_ORDINAL;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":143
 *     cdef int _EPOCH_ORDINAL
 * 
 * cdef class TimeCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":146
 *     pass
 * 
 * cdef class TimestampCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef bint _is_compact
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  int _is_compact;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":151
 *     cdef _decode_timestamp_data_from_stream(self, InputStream in_stream)
 * 
 * cdef class LocalZonedTimestampCoderImpl(TimestampCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _timezone
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl __pyx_base;
  PyObject *_timezone;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":154
 *     cdef object _timezone
 * 
 * cdef class InstantCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef int64_t _null_seconds
 *     cdef int32_t _null_nanos
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  int64_t _null_seconds;
  int32_t _null_nanos;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":158
 *     cdef int32_t _null_nanos
 * 
 * cdef class CloudPickleCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":161
 *     pass
 * 
 * cdef class PickleCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":164
 *     pass
 * 
 * cdef class GenericArrayCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef FieldCoderImpl _elem_coder
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *_elem_coder;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":167
 *     cdef FieldCoderImpl _elem_coder
 * 
 * cdef class PrimitiveArrayCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef FieldCoderImpl _elem_coder
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *_elem_coder;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":170
 *     cdef FieldCoderImpl _elem_coder
 * 
 * cdef class MapCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef FieldCoderImpl _key_coder
 *     cdef FieldCoderImpl _value_coder
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *_key_coder;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *_value_coder;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":174
 *     cdef FieldCoderImpl _value_coder
 * 
 * cdef class TupleCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef list _field_coders
 *     cdef size_t _field_count
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *_field_coders;
  size_t _field_count;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":178
 *     cdef size_t _field_count
 * 
 * cdef class TimeWindowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":181
 *     pass
 * 
 * cdef class CountWindowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":184
 *     pass
 * 
 * cdef class DataViewFilterCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _udf_data_view_specs
 *     cdef PickleCoderImpl _pickle_coder
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *_udf_data_view_specs;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl *_pickle_coder;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":188
 *     cdef PickleCoderImpl _pickle_coder
 * 
 * cdef class AvroCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _buffer_wrapper
 *     cdef object _schema
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *_buffer_wrapper;
  PyObject *_schema;
  PyObject *_decoder;
  PyObject *_encoder;
  PyObject *_reader;
  PyObject *_writer;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":196
 *     cdef object _writer
 * 
 * cdef class LocalDateCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":199
 *     pass
 * 
 * cdef class LocalTimeCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/coder_impl_fast.pxd":202
 *     pass
 * 
 * cdef class LocalDateTimeCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
*/
struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":21
 * from pyflink.fn_execution.coder_impl_fast cimport InternalRow, InternalRowKind
 * 
 * cdef class DistinctViewDescriptor:             # <<<<<<<<<<<<<<
 *     cdef object input_extractor
 *     cdef int*filter_args
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor {
  PyObject_HEAD
  PyObject *input_extractor;
  int *filter_args;
  size_t filter_length;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":26
 *     cdef size_t filter_length
 * 
 * cdef class RowKeySelector:             # <<<<<<<<<<<<<<
 *     cdef size_t*grouping
 *     cdef size_t length
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *__pyx_vtab;
  size_t *grouping;
  size_t length;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":32
 *     cdef list get_key(self, list data)
 * 
 * cdef class AggsHandleFunctionBase:             # <<<<<<<<<<<<<<
 *     cdef void open(self, object state_data_view_store)
 *     cdef void accumulate(self, InternalRow input_data)
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *__pyx_vtab;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":45
 *     cdef list emit_value(self, list current_key, bint is_retract)
 * 
 * cdef class SimpleAggsHandleFunctionBase(AggsHandleFunctionBase):             # <<<<<<<<<<<<<<
 *     cdef list _udfs
 *     cdef size_t _udf_num
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase {
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase __pyx_base;
  PyObject *_udfs;
  size_t _udf_num;
  PyObject *_input_extractors;
  PyObject *_accumulators;
  PyObject *_udf_data_view_specs;
  PyObject *_udf_data_views;
  int *_filter_args;
  int *_distinct_indexes;
  PyObject *_distinct_view_descriptors;
  PyObject *_distinct_data_views;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":57
 *     cdef dict _distinct_data_views
 * 
 * cdef class SimpleAggsHandleFunction(SimpleAggsHandleFunctionBase):             # <<<<<<<<<<<<<<
 *     cdef size_t*_get_value_indexes
 *     cdef size_t _get_value_indexes_length
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction {
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase __pyx_base;
  size_t *_get_value_indexes;
  size_t _get_value_indexes_length;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":61
 *     cdef size_t _get_value_indexes_length
 * 
 * cdef class SimpleTableAggsHandleFunction(SimpleAggsHandleFunctionBase):             # <<<<<<<<<<<<<<
 *     cdef list _convert_to_row(self, data)
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction {
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase __pyx_base;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":64
 *     cdef list _convert_to_row(self, data)
 * 
 * cdef class RecordCounter:             # <<<<<<<<<<<<<<
 *     cdef bint record_count_is_zero(self, list acc)
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter *__pyx_vtab;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":70
 *     cdef RecordCounter of(int index_of_count_star)
 * 
 * cdef class AccumulationRecordCounter(RecordCounter):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter {
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter __pyx_base;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":73
 *     pass
 * 
 * cdef class RetractionRecordCounter(RecordCounter):             # <<<<<<<<<<<<<<
 *     cdef int _index_of_count_star
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter {
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter __pyx_base;
  int _index_of_count_star;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":76
 *     cdef int _index_of_count_star
 * 
 * cdef class GroupAggFunctionBase:             # <<<<<<<<<<<<<<
 *     cdef AggsHandleFunctionBase aggs_handle
 *     cdef bint generate_update_before
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase *__pyx_vtab;
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *aggs_handle;
  int generate_update_before;
  int state_cleaning_enabled;
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *key_selector;
  PyObject *state_value_coder;
  PyObject *state_backend;
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter *record_counter;
  PyObject *buffer;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":92
 *     cpdef void on_timer(self, InternalRow key)
 * 
 * cdef class GroupAggFunction(GroupAggFunctionBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction {
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase __pyx_base;
};


/* "pyflink/fn_execution/table/aggregate_fast.pxd":95
 *     pass
 * 
 * cdef class GroupTableAggFunction(GroupAggFunctionBase):             # <<<<<<<<<<<<<<
 *     pass
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction {
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase __pyx_base;
};


/* "pyflink/fn_execution/table/window_aggregate_fast.pxd":22
 * from pyflink.fn_execution.coder_impl_fast cimport InternalRow
 * 
 * cdef class NamespaceAggsHandleFunctionBase:             # <<<<<<<<<<<<<<
 *     cdef void open(self, object state_data_view_store)
 *     cdef void accumulate(self, InternalRow input_data)
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_vtab;
};


/* "pyflink/fn_execution/table/window_aggregate_fast.pxd":33
 *     cdef void close(self)
 * 
 * cdef class NamespaceAggsHandleFunction(NamespaceAggsHandleFunctionBase):             # <<<<<<<<<<<<<<
 *     cpdef list get_value(self, object namespace)
 * 
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase __pyx_base;
};


/* "pyflink/fn_execution/table/window_aggregate_fast.pxd":36
 *     cpdef list get_value(self, object namespace)
 * 
 * cdef class SimpleNamespaceAggsHandleFunction(NamespaceAggsHandleFunction):             # <<<<<<<<<<<<<<
 *     cdef list _udfs
 *     cdef size_t _udf_num
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction __pyx_base;
  PyObject *_udfs;
  size_t _udf_num;
  PyObject *_input_extractors;
  PyObject *_named_property_extractor;
  PyObject *_accumulators;
  PyObject *_udf_data_view_specs;
  PyObject *_udf_data_views;
  int *_filter_args;
  int *_distinct_indexes;
  PyObject *_distinct_view_descriptors;
  PyObject *_distinct_data_views;
  PyObject *named_property_extractor;
  size_t *_get_value_indexes;
  size_t _get_value_indexes_length;
};


/* "pyflink/fn_execution/table/window_aggregate_fast.pxd":52
 *     cdef size_t _get_value_indexes_length
 * 
 * cdef class GroupWindowAggFunctionBase:             # <<<<<<<<<<<<<<
 *     cdef int64_t _allowed_lateness
 *     cdef int _rowtime_index
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_vtab;
  int64_t _allowed_lateness;
  int _rowtime_index;
  PyObject *_shift_timezone;
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *_key_selector;
  PyObject *_state_value_coder;
  PyObject *_state_backend;
  PyObject *_window_assigner;
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *_window_aggregator;
  PyObject *_trigger;
  PyObject *_window_function;
  PyObject *_internal_timer_service;
  PyObject *_window_context;
  PyObject *_trigger_context;
  PyObject *_window_state;
};


/* "pyflink/fn_execution/table/window_aggregate_fast.pxd":80
 *     cdef InternalRow _emit_window_result(self, list key, object window)
 * 
 * cdef class GroupWindowAggFunction(GroupWindowAggFunctionBase):             # <<<<<<<<<<<<<<
 *     pass
*/
struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase __pyx_base;
};



/* "pyflink/fn_execution/stream_fast.pxd":21
 * from libc.stdint cimport int8_t, int16_t, int32_t, int64_t
 * 
 * cdef class LengthPrefixInputStream:             # <<<<<<<<<<<<<<
 *     cdef size_t read(self, char**data)
 *     cdef size_t available(self)
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream {
  size_t (*read)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream *, char **);
  size_t (*available)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream *__pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream;


/* "pyflink/fn_execution/stream_fast.pxd":25
 *     cdef size_t available(self)
 * 
 * cdef class LengthPrefixOutputStream:             # <<<<<<<<<<<<<<
 *     cdef void write(self, char*data, size_t length)
 *     cpdef void flush(self)
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream {
  void (*write)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream *, char *, size_t);
  void (*flush)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream *, int __pyx_skip_dispatch);
  void (*close)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream *__pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream;


/* "pyflink/fn_execution/stream_fast.pxd":30
 *     cpdef void close(self)
 * 
 * cdef class InputStream:             # <<<<<<<<<<<<<<
 *     cdef char*_input_data
 *     cdef size_t _input_pos
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_InputStream {
  PyObject *(*read)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *, size_t, int __pyx_skip_dispatch);
  long (*read_byte)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
  int8_t (*read_int8)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
  int16_t (*read_int16)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
  int32_t (*read_int32)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
  int64_t (*read_int64)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
  float (*read_float)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
  double (*read_double)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
  PyObject *(*read_bytes)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_InputStream *__pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_InputStream;


/* "pyflink/fn_execution/stream_fast.pxd":44
 *     cdef bytes read_bytes(self)
 * 
 * cdef class OutputStream:             # <<<<<<<<<<<<<<
 *     cdef char*buffer
 *     cdef size_t buffer_size
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_OutputStream {
  void (*write)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, PyObject *, int __pyx_skip_dispatch);
  void (*write_byte)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, unsigned char);
  void (*write_int8)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, int8_t);
  void (*write_int16)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, int16_t);
  void (*write_int32)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, int32_t);
  void (*write_int64)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, int64_t);
  void (*write_float)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, float);
  void (*write_double)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, double);
  void (*write_bytes)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, char *, size_t);
  void (*_extend)(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, size_t);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_OutputStream *__pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_OutputStream;


/* "pyflink/fn_execution/coder_impl_fast.pxd":31
 *     DELETE = 3
 * 
 * cdef class InternalRow:             # <<<<<<<<<<<<<<
 *     cdef readonly list values
 *     cdef readonly InternalRowKind row_kind
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InternalRow {
  PyObject *(*to_row)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *, int __pyx_skip_dispatch);
  int (*is_retract_msg)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *);
  int (*is_accumulate_msg)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InternalRow;


/* "pyflink/fn_execution/coder_impl_fast.pxd":39
 *     cdef bint is_accumulate_msg(self)
 * 
 * cdef class MaskUtils:             # <<<<<<<<<<<<<<
 *     cdef bint*_mask
 *     cdef unsigned char*_mask_byte_search_table
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils {
  void (*write_mask)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils *, PyObject *, unsigned char, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *);
  int *(*read_mask)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils *, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils;


/* "pyflink/fn_execution/coder_impl_fast.pxd":50
 *     cdef bint*read_mask(self, InputStream input_stream)
 * 
 * cdef class LengthPrefixBaseCoderImpl:             # <<<<<<<<<<<<<<
 *     cdef FieldCoderImpl _field_coder
 *     cdef OutputStream _data_out_stream
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl {
  PyObject *(*encode_to_stream)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl *, PyObject *, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream *, int __pyx_skip_dispatch);
  PyObject *(*decode_from_stream)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl *, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream *, int __pyx_skip_dispatch);
  void (*_write_data_to_output_stream)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl *, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":58
 *     cdef void _write_data_to_output_stream(self, LengthPrefixOutputStream output_stream)
 * 
 * cdef class FieldCoderImpl:             # <<<<<<<<<<<<<<
 *     cpdef encode_to_stream(self, value, OutputStream out_stream)
 *     cpdef decode_from_stream(self, InputStream in_stream, size_t size)
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl {
  PyObject *(*encode_to_stream)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *, PyObject *, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream *, int __pyx_skip_dispatch);
  PyObject *(*decode_from_stream)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *, size_t, int __pyx_skip_dispatch);
  PyObject *(*encode)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*decode)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":64
 *     cpdef decode(self, encoded)
 * 
 * cdef class InputStreamWrapper:             # <<<<<<<<<<<<<<
 *     cdef ValueCoderImpl _value_coder
 *     cdef LengthPrefixInputStream _input_stream
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper {
  int (*has_next)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper *, int __pyx_skip_dispatch);
  PyObject *(*next)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper;


/* "pyflink/fn_execution/coder_impl_fast.pxd":71
 *     cpdef next(self)
 * 
 * cdef class IterableCoderImpl(LengthPrefixBaseCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef char*_end_message
 *     cdef bint _separated_with_end_message
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":75
 *     cdef bint _separated_with_end_message
 * 
 * cdef class ValueCoderImpl(LengthPrefixBaseCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":80
 * cdef unsigned char ROW_KIND_BIT_SIZE
 * 
 * cdef class FlattenRowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef list _field_coders
 *     cdef list _reuse_flatten_row
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":86
 *     cdef MaskUtils _mask_utils
 * 
 * cdef class RowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef list _field_coders
 *     cdef size_t _field_count
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":92
 *     cdef MaskUtils _mask_utils
 * 
 * cdef class ArrowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _schema
 *     cdef list _field_types
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *(*decode_one_batch_from_stream)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl *, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *, size_t);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":101
 *     cdef list decode_one_batch_from_stream(self, InputStream in_stream, size_t size)
 * 
 * cdef class OverWindowArrowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef ArrowCoderImpl _arrow_coder
 *     cdef IntCoderImpl _int_coder
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":105
 *     cdef IntCoderImpl _int_coder
 * 
 * cdef class TinyIntCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":108
 *     pass
 * 
 * cdef class SmallIntCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":111
 *     pass
 * 
 * cdef class IntCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":114
 *     pass
 * 
 * cdef class BigIntCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":117
 *     pass
 * 
 * cdef class BooleanCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":120
 *     pass
 * 
 * cdef class FloatCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":123
 *     pass
 * 
 * cdef class DoubleCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":126
 *     pass
 * 
 * cdef class BinaryCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":129
 *     pass
 * 
 * cdef class CharCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":132
 *     pass
 * 
 * cdef class BigDecimalCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef CharCoderImpl _value_coder
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":135
 *     cdef CharCoderImpl _value_coder
 * 
 * cdef class DecimalCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _context
 *     cdef object _scale_format
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":140
 *     cdef CharCoderImpl _value_coder
 * 
 * cdef class DateCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef int _EPOCH_ORDINAL
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":143
 *     cdef int _EPOCH_ORDINAL
 * 
 * cdef class TimeCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":146
 *     pass
 * 
 * cdef class TimestampCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef bint _is_compact
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
  PyObject *(*_decode_timestamp_data_from_stream)(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl *, struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":151
 *     cdef _decode_timestamp_data_from_stream(self, InputStream in_stream)
 * 
 * cdef class LocalZonedTimestampCoderImpl(TimestampCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _timezone
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":154
 *     cdef object _timezone
 * 
 * cdef class InstantCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef int64_t _null_seconds
 *     cdef int32_t _null_nanos
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":158
 *     cdef int32_t _null_nanos
 * 
 * cdef class CloudPickleCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":161
 *     pass
 * 
 * cdef class PickleCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":164
 *     pass
 * 
 * cdef class GenericArrayCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef FieldCoderImpl _elem_coder
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":167
 *     cdef FieldCoderImpl _elem_coder
 * 
 * cdef class PrimitiveArrayCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef FieldCoderImpl _elem_coder
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":170
 *     cdef FieldCoderImpl _elem_coder
 * 
 * cdef class MapCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef FieldCoderImpl _key_coder
 *     cdef FieldCoderImpl _value_coder
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":174
 *     cdef FieldCoderImpl _value_coder
 * 
 * cdef class TupleCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef list _field_coders
 *     cdef size_t _field_count
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":178
 *     cdef size_t _field_count
 * 
 * cdef class TimeWindowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":181
 *     pass
 * 
 * cdef class CountWindowCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":184
 *     pass
 * 
 * cdef class DataViewFilterCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _udf_data_view_specs
 *     cdef PickleCoderImpl _pickle_coder
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":188
 *     cdef PickleCoderImpl _pickle_coder
 * 
 * cdef class AvroCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     cdef object _buffer_wrapper
 *     cdef object _schema
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":196
 *     cdef object _writer
 * 
 * cdef class LocalDateCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":199
 *     pass
 * 
 * cdef class LocalTimeCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl;


/* "pyflink/fn_execution/coder_impl_fast.pxd":202
 *     pass
 * 
 * cdef class LocalDateTimeCoderImpl(FieldCoderImpl):             # <<<<<<<<<<<<<<
 *     pass
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl *__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":26
 *     cdef size_t filter_length
 * 
 * cdef class RowKeySelector:             # <<<<<<<<<<<<<<
 *     cdef size_t*grouping
 *     cdef size_t length
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector {
  PyObject *(*get_key)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *, PyObject *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":32
 *     cdef list get_key(self, list data)
 * 
 * cdef class AggsHandleFunctionBase:             # <<<<<<<<<<<<<<
 *     cdef void open(self, object state_data_view_store)
 *     cdef void accumulate(self, InternalRow input_data)
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase {
  void (*open)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *, PyObject *);
  void (*accumulate)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *);
  void (*retract)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *);
  void (*merge)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *, PyObject *);
  void (*set_accumulators)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *, PyObject *);
  PyObject *(*get_accumulators)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *);
  PyObject *(*create_accumulators)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *);
  void (*cleanup)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *);
  void (*close)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *);
  PyObject *(*get_value)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *);
  PyObject *(*emit_value)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *, PyObject *, int);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":45
 *     cdef list emit_value(self, list current_key, bint is_retract)
 * 
 * cdef class SimpleAggsHandleFunctionBase(AggsHandleFunctionBase):             # <<<<<<<<<<<<<<
 *     cdef list _udfs
 *     cdef size_t _udf_num
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":57
 *     cdef dict _distinct_data_views
 * 
 * cdef class SimpleAggsHandleFunction(SimpleAggsHandleFunctionBase):             # <<<<<<<<<<<<<<
 *     cdef size_t*_get_value_indexes
 *     cdef size_t _get_value_indexes_length
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":61
 *     cdef size_t _get_value_indexes_length
 * 
 * cdef class SimpleTableAggsHandleFunction(SimpleAggsHandleFunctionBase):             # <<<<<<<<<<<<<<
 *     cdef list _convert_to_row(self, data)
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase __pyx_base;
  PyObject *(*_convert_to_row)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction *, PyObject *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":64
 *     cdef list _convert_to_row(self, data)
 * 
 * cdef class RecordCounter:             # <<<<<<<<<<<<<<
 *     cdef bint record_count_is_zero(self, list acc)
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter {
  int (*record_count_is_zero)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter *, PyObject *);
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter *(*of)(int);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":70
 *     cdef RecordCounter of(int index_of_count_star)
 * 
 * cdef class AccumulationRecordCounter(RecordCounter):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":73
 *     pass
 * 
 * cdef class RetractionRecordCounter(RecordCounter):             # <<<<<<<<<<<<<<
 *     cdef int _index_of_count_star
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":76
 *     cdef int _index_of_count_star
 * 
 * cdef class GroupAggFunctionBase:             # <<<<<<<<<<<<<<
 *     cdef AggsHandleFunctionBase aggs_handle
 *     cdef bint generate_update_before
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase {
  void (*open)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase *, PyObject *, int __pyx_skip_dispatch);
  void (*close)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase *, int __pyx_skip_dispatch);
  void (*process_element)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *, int __pyx_skip_dispatch);
  PyObject *(*finish_bundle)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase *, int __pyx_skip_dispatch);
  void (*on_timer)(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":92
 *     cpdef void on_timer(self, InternalRow key)
 * 
 * cdef class GroupAggFunction(GroupAggFunctionBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction;


/* "pyflink/fn_execution/table/aggregate_fast.pxd":95
 *     pass
 * 
 * cdef class GroupTableAggFunction(GroupAggFunctionBase):             # <<<<<<<<<<<<<<
 *     pass
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction *__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction;


/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":48
 *     return InternalRow(left.__add__(right), row_kind)
 * 
 * cdef class NamespaceAggsHandleFunctionBase:             # <<<<<<<<<<<<<<
 *     cdef void open(self, object state_data_view_store):
 *         """
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase {
  void (*open)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *);
  void (*accumulate)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *);
  void (*retract)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *);
  void (*merge)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  void (*set_accumulators)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_accumulators)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *);
  PyObject *(*create_accumulators)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, int __pyx_skip_dispatch);
  void (*cleanup)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, int __pyx_skip_dispatch);
  void (*close)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;


/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":114
 *         pass
 * 
 * cdef class NamespaceAggsHandleFunction(NamespaceAggsHandleFunctionBase):             # <<<<<<<<<<<<<<
 *     cpdef list get_value(self, object namespace):
 *         """
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase __pyx_base;
  PyObject *(*get_value)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *, PyObject *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;


/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":124
 *         pass
 * 
 * cdef class SimpleNamespaceAggsHandleFunction(NamespaceAggsHandleFunction):             # <<<<<<<<<<<<<<
 *     def __init__(self,
 *                  udfs: List[ImperativeAggregateFunction],
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction;


/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":325
 *             free(self._distinct_indexes)
 * 
 * cdef class GroupWindowAggFunctionBase:             # <<<<<<<<<<<<<<
 *     def __init__(self,
 *                  allowed_lateness: int,
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase {
  void (*open)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, int __pyx_skip_dispatch);
  void (*close)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, int __pyx_skip_dispatch);
  PyObject *(*process_element)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *, int __pyx_skip_dispatch);
  void (*process_watermark)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, int64_t, int __pyx_skip_dispatch);
  PyObject *(*on_event_time)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*on_processing_time)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, int __pyx_skip_dispatch);
  PyObject *(*get_timers)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, int __pyx_skip_dispatch);
  int64_t (*to_utc_timestamp_mills)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, int64_t, int __pyx_skip_dispatch);
  int64_t (*cleanup_time)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, int __pyx_skip_dispatch);
  void (*_register_cleanup_timer)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *);
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *(*_emit_window_result)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, PyObject *);
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;


/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":506
 *         pass
 * 
 * cdef class GroupWindowAggFunction(GroupWindowAggFunctionBase):             # <<<<<<<<<<<<<<
 *     def __init__(self,
 *                  allowed_lateness: int,
*/

struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction {
  struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase __pyx_base;
};
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction;
/* #### Code section: utility_code_proto ### */

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, Py_ssize_t);
    void (*DECREF)(void*, PyObject*, Py_ssize_t);
    void (*GOTREF)(void*, PyObject*, Py_ssize_t);
    void (*GIVEREF)(void*, PyObject*, Py_ssize_t);
    void* (*SetupContext)(const char*, Py_ssize_t, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), (__LINE__), (__FILE__));\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContextNogil() {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __Pyx_RefNannyFinishContext();\
              PyGILState_Release(__pyx_gilstate_save);\
          }
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), (__LINE__))
  #define __Pyx_XINCREF(r)  do { if((r) == NULL); else {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) == NULL); else {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) == NULL); else {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) == NULL); else {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContextNogil()
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_Py_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; Py_XDECREF(tmp);\
    } while (0)
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#if PY_VERSION_HEX >= 0x030C00A6
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->current_exception != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->current_exception ? (PyObject*) Py_TYPE(__pyx_tstate->current_exception) : (PyObject*) NULL)
#else
#define __Pyx_PyErr_Occurred()  (__pyx_tstate->curexc_type != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  (__pyx_tstate->curexc_type)
#endif
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  (PyErr_Occurred() != NULL)
#define __Pyx_PyErr_CurrentExceptionType()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A6
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* Profile.proto */
#if CYTHON_TRACE
  #undef CYTHON_PROFILE_REUSE_FRAME
#endif
#if CYTHON_USE_MODULE_STATE
  #undef CYTHON_PROFILE_REUSE_CODEOBJ
  #define CYTHON_PROFILE_REUSE_CODEOBJ 0
  #undef CYTHON_PROFILE_REUSE_FRAME
#endif
#ifndef CYTHON_PROFILE_REUSE_CODEOBJ
  #define CYTHON_PROFILE_REUSE_CODEOBJ 1
#endif
#ifndef CYTHON_PROFILE_REUSE_FRAME
  #define CYTHON_PROFILE_REUSE_FRAME 0
#endif
#if CYTHON_USE_SYS_MONITORING && (CYTHON_PROFILE || CYTHON_TRACE)
  #define __PYX_MONITORING_ABI_SUFFIX  "_mon"
#else
  #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_PROFILE || CYTHON_TRACE
#if CYTHON_USE_SYS_MONITORING
  typedef uint64_t __pyx_monitoring_version_type;
  #define __Pyx_TraceDeclarationsFunc\
      PyObject *__pyx_frame_code = NULL;\
      PyMonitoringState __pyx_pymonitoring_state[__Pyx_MonitoringEventTypes_CyFunc_count];\
      int __pyx_exception_already_reported = 0;\
      const int __pyx_sys_monitoring_disabled_in_parallel = 0; CYTHON_UNUSED_VAR(__pyx_sys_monitoring_disabled_in_parallel);
  #define __Pyx_TraceDeclarationsGen\
      PyObject *__pyx_frame_code = Py_NewRef(__pyx_generator->gi_code);\
      PyMonitoringState* __pyx_pymonitoring_state = __pyx_generator->__pyx_pymonitoring_state;\
      __pyx_monitoring_version_type __pyx_pymonitoring_version = __pyx_generator->__pyx_pymonitoring_version;\
      int __pyx_exception_already_reported = 0;\
      const int __pyx_sys_monitoring_disabled_in_parallel = 0; CYTHON_UNUSED_VAR(__pyx_sys_monitoring_disabled_in_parallel);
  #define __Pyx_IsTracing(event_id)  ((!__pyx_sys_monitoring_disabled_in_parallel) && (__pyx_pymonitoring_state[event_id]).active)
  #define __Pyx_TraceFrameInit(codeobj)\
      if (codeobj) __pyx_frame_code = codeobj;
  #define __Pyx_TurnOffSysMonitoringInParallel\
    const int __pyx_sys_monitoring_disabled_in_parallel = 1;\
    CYTHON_UNUSED_VAR(__pyx_sys_monitoring_disabled_in_parallel);
  CYTHON_UNUSED static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno);
  CYTHON_UNUSED static int __Pyx__TraceStartFunc(PyMonitoringState *state_array, PyObject *code_obj, int offset, int skip_event);
  CYTHON_UNUSED static int __Pyx__TraceStartGen(PyMonitoringState *state_array, __pyx_monitoring_version_type *version, PyObject *code_obj, int offset);
  CYTHON_UNUSED static int __Pyx__TraceResumeGen(PyMonitoringState *state_array, __pyx_monitoring_version_type *version, PyObject *code_obj, int offset);
  CYTHON_UNUSED static void __Pyx__TraceException(PyMonitoringState *monitoring_state, PyObject *code_obj, int offset, int reraised);
  #define __Pyx_PyMonitoring_ExitScope(nogil)\
    if (nogil) {\
        (void) __pyx_exception_already_reported;\
        if (CYTHON_TRACE_NOGIL) {\
            PyGILState_STATE state = PyGILState_Ensure();\
            PyMonitoring_ExitScope();\
            Py_XDECREF(__pyx_frame_code);\
            PyGILState_Release(state);\
        }\
    } else {\
        PyMonitoring_ExitScope();\
        Py_XDECREF(__pyx_frame_code);\
    }
  #define __Pyx_TraceStartFunc(funcname, srcfile, firstlineno, offset, nogil, skip_event, goto_error)\
  if ((0) ); else {\
      int ret = 0;\
      memset(__pyx_pymonitoring_state, 0, sizeof(__pyx_pymonitoring_state));\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyGILState_STATE state = PyGILState_Ensure();\
              if (!__Pyx_PyThreadState_Current->tracing) {\
                  if (likely(__pyx_frame_code)) Py_INCREF(__pyx_frame_code);\
                  else __pyx_frame_code = (PyObject*) __Pyx_createFrameCodeObject(funcname, srcfile, firstlineno);\
                  if (unlikely(!__pyx_frame_code)) goto_error;\
                  ret = __Pyx__TraceStartFunc(__pyx_pymonitoring_state, __pyx_frame_code, offset, skip_event);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          if (!__Pyx_PyThreadState_Current->tracing) {\
              if (likely(__pyx_frame_code)) Py_INCREF(__pyx_frame_code);\
              else __pyx_frame_code = (PyObject*) __Pyx_createFrameCodeObject(funcname, srcfile, firstlineno);\
              if (unlikely(!__pyx_frame_code)) goto_error;\
              ret = __Pyx__TraceStartFunc(__pyx_pymonitoring_state, __pyx_frame_code, offset, skip_event);\
          }\
      }\
      if (unlikely(ret == -1)) goto_error;\
  }
  #define __Pyx_TraceStartGen(funcname, srcfile, firstlineno, offset, nogil, skip_event, goto_error)\
  if ((0) ); else {\
      int ret = __Pyx__TraceStartGen(__pyx_pymonitoring_state, &__pyx_pymonitoring_version, __pyx_frame_code, offset);\
      if (unlikely(ret == -1)) goto_error;\
  }
  #define __Pyx_TraceResumeGen(funcname, srcfile, firstlineno, offset, goto_error)\
  if ((0) ); else {\
      int ret = __Pyx__TraceResumeGen(__pyx_pymonitoring_state, &__pyx_pymonitoring_version, __pyx_frame_code, offset);\
      if (unlikely(ret == -1)) goto_error;\
  }
  #define __Pyx_TraceYield(result, offset, goto_error)\
  if (!__Pyx_IsTracing(__Pyx_Monitoring_PY_YIELD)); else {\
      int ret = PyMonitoring_FirePyYieldEvent(&__pyx_pymonitoring_state[__Pyx_Monitoring_PY_RETURN], __pyx_frame_code, offset, result);\
      PyMonitoring_ExitScope();\
      if (unlikely(ret == -1)) goto_error;\
  }
  #define __Pyx_TraceException(offset, reraised, fresh)\
  if (!__Pyx_IsTracing((reraised) ? __Pyx_Monitoring_RERAISE : __Pyx_Monitoring_RAISE)); else {\
      if (fresh || reraised || !__pyx_exception_already_reported) {\
          __Pyx__TraceException(&__pyx_pymonitoring_state[(reraised) ? __Pyx_Monitoring_RERAISE : __Pyx_Monitoring_RAISE], __pyx_frame_code, offset, reraised);\
      }\
      __pyx_exception_already_reported = 1;\
  }
  #define __Pyx_TraceExceptionDone()  __pyx_exception_already_reported = 0
  #define __Pyx_TraceExceptionHandled(offset)\
  if (!__Pyx_IsTracing(__Pyx_Monitoring_EXCEPTION_HANDLED)); else {\
      (void) PyMonitoring_FireExceptionHandledEvent(&__pyx_pymonitoring_state[__Pyx_Monitoring_EXCEPTION_HANDLED], __pyx_frame_code, offset);\
      __pyx_exception_already_reported = 0;\
  }
  #define __Pyx_TraceReturnValue(result, offset, nogil, goto_error)\
  if (!__Pyx_IsTracing(__Pyx_Monitoring_PY_RETURN)); else {\
      int ret = 0;\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyGILState_STATE state = PyGILState_Ensure();\
              ret = PyMonitoring_FirePyReturnEvent(&__pyx_pymonitoring_state[__Pyx_Monitoring_PY_RETURN], __pyx_frame_code, offset, result);\
              PyGILState_Release(state);\
          }\
      } else {\
          ret = PyMonitoring_FirePyReturnEvent(&__pyx_pymonitoring_state[__Pyx_Monitoring_PY_RETURN], __pyx_frame_code, offset, result);\
      }\
      if (unlikely(ret == -1)) goto_error;\
  }
  #define __Pyx_TraceReturnCValue(cresult, convert_function, offset, nogil, goto_error)\
  if (!__Pyx_IsTracing(__Pyx_Monitoring_PY_RETURN)); else {\
      int ret = 0;\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyGILState_STATE state = PyGILState_Ensure();\
              PyObject *pyvalue = convert_function(cresult);\
              if (unlikely(!pyvalue)) {\
                  PyErr_Clear();\
                  pyvalue = Py_None; Py_INCREF(Py_None);\
              }\
              ret = PyMonitoring_FirePyReturnEvent(&__pyx_pymonitoring_state[__Pyx_Monitoring_PY_RETURN], __pyx_frame_code, offset, pyvalue);\
              Py_DECREF(pyvalue);\
              PyGILState_Release(state);\
          }\
      } else {\
          PyObject *pyvalue = convert_function(cresult);\
          if (unlikely(!pyvalue)) {\
              PyErr_Clear();\
              pyvalue = Py_None; Py_INCREF(Py_None);\
          }\
          ret = PyMonitoring_FirePyReturnEvent(&__pyx_pymonitoring_state[__Pyx_Monitoring_PY_RETURN], __pyx_frame_code, offset, pyvalue);\
          Py_DECREF(pyvalue);\
      }\
      if (unlikely(ret == -1)) goto_error;\
  }
  #define __Pyx_TraceExceptionUnwind(offset, nogil)\
  if (!__Pyx_IsTracing(__Pyx_Monitoring_PY_UNWIND)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyGILState_STATE state = PyGILState_Ensure();\
              (void) PyMonitoring_FirePyUnwindEvent(&__pyx_pymonitoring_state[__Pyx_Monitoring_PY_UNWIND], __pyx_frame_code, offset);\
              PyGILState_Release(state);\
          }\
      } else {\
          (void) PyMonitoring_FirePyUnwindEvent(&__pyx_pymonitoring_state[__Pyx_Monitoring_PY_UNWIND], __pyx_frame_code, offset);\
      }\
  }
  #if CYTHON_TRACE
  CYTHON_UNUSED static int __Pyx__TraceLine(PyMonitoringState *monitoring_state, PyObject *code_obj, int line, int offset);
  #define __Pyx_TraceLine(line, offset, nogil, goto_error)\
  if (!__Pyx_IsTracing(__Pyx_Monitoring_LINE)); else {\
      int ret = 0;\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyGILState_STATE state = PyGILState_Ensure();\
              ret = __Pyx__TraceLine(&__pyx_pymonitoring_state[__Pyx_Monitoring_LINE], __pyx_frame_code, line, offset);\
              PyGILState_Release(state);\
          }\
      } else {\
          ret = __Pyx__TraceLine(&__pyx_pymonitoring_state[__Pyx_Monitoring_LINE], __pyx_frame_code, line, offset);\
      }\
      if (unlikely(ret == -1)) goto_error;\
  }
  #endif
#else
  #include "compile.h"
  #include "frameobject.h"
  #include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
  #if CYTHON_PROFILE_REUSE_FRAME
    #define CYTHON_FRAME_MODIFIER static
    #define CYTHON_FRAME_DEL(frame)
  #else
    #define CYTHON_FRAME_MODIFIER
    #define CYTHON_FRAME_DEL(frame) Py_CLEAR(frame)
  #endif
  #if CYTHON_PROFILE_REUSE_CODEOBJ
    #define CYTHON_CODEOBJ_MODIFIER static
  #else
    #define CYTHON_CODEOBJ_MODIFIER
  #endif
  #define __Pyx_TraceDeclarationsFunc\
      CYTHON_CODEOBJ_MODIFIER PyCodeObject *__pyx_frame_code = NULL;\
      CYTHON_FRAME_MODIFIER PyFrameObject *__pyx_frame = NULL;\
      int __Pyx_use_tracing = 0;
  #define __Pyx_TraceDeclarationsGen\
      PyObject *__pyx_frame_code = __pyx_generator->gi_code;\
      CYTHON_FRAME_MODIFIER PyFrameObject *__pyx_frame = NULL;\
      int __Pyx_use_tracing = 0;
  #define __Pyx_TraceFrameInit(codeobj)\
      if (codeobj) __pyx_frame_code = (PyCodeObject*) codeobj;
  #define __Pyx_PyMonitoring_ExitScope(nogil)  {}
  #define __Pyx_TraceException(offset, reraised, fresh)  {}
  #define __Pyx_TraceExceptionHandled(offset)  {}
  #define __Pyx_TraceExceptionDone()  {}
  #define __Pyx_TurnOffSysMonitoringInParallel {} // Only needed for freethreading
#if PY_VERSION_HEX >= 0x030b00a2
  #if PY_VERSION_HEX >= 0x030C00b1
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     ((!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #else
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->cframe->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #endif
  #define __Pyx_EnterTracing(tstate)  PyThreadState_EnterTracing(tstate)
  #define __Pyx_LeaveTracing(tstate)  PyThreadState_LeaveTracing(tstate)
#elif PY_VERSION_HEX >= 0x030a00b1
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->cframe->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_EnterTracing(tstate)\
      do { tstate->tracing++; tstate->cframe->use_tracing = 0; } while (0)
  #define __Pyx_LeaveTracing(tstate)\
      do {\
          tstate->tracing--;\
          tstate->cframe->use_tracing = ((CYTHON_TRACE && tstate->c_tracefunc != NULL)\
                                 || tstate->c_profilefunc != NULL);\
      } while (0)
#else
  #define __Pyx_IsTracing(tstate, check_tracing, check_funcs)\
     (unlikely((tstate)->use_tracing) &&\
         (!(check_tracing) || !(tstate)->tracing) &&\
         (!(check_funcs) || (tstate)->c_profilefunc || (CYTHON_TRACE && (tstate)->c_tracefunc)))
  #define __Pyx_EnterTracing(tstate)\
      do { tstate->tracing++; tstate->use_tracing = 0; } while (0)
  #define __Pyx_LeaveTracing(tstate)\
      do {\
          tstate->tracing--;\
          tstate->use_tracing = ((CYTHON_TRACE && tstate->c_tracefunc != NULL)\
                                         || tstate->c_profilefunc != NULL);\
      } while (0)
#endif
  #define __Pyx_TraceStartFunc(funcname, srcfile, firstlineno, offset, nogil, skip_event, goto_error)\
  if (nogil) {\
      if (CYTHON_TRACE_NOGIL) {\
          PyThreadState *tstate;\
          PyGILState_STATE state = PyGILState_Ensure();\
          tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 1, 1)) {\
              __Pyx_use_tracing = __Pyx_TraceSetupAndCall((PyCodeObject**)&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno, skip_event);\
          }\
          PyGILState_Release(state);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  } else {\
      PyThreadState* tstate = PyThreadState_GET();\
      if (__Pyx_IsTracing(tstate, 1, 1)) {\
          __Pyx_use_tracing = __Pyx_TraceSetupAndCall((PyCodeObject**)&__pyx_frame_code, &__pyx_frame, tstate, funcname, srcfile, firstlineno, skip_event);\
          if (unlikely(__Pyx_use_tracing < 0)) goto_error;\
      }\
  }
  #define __Pyx_TraceStartGen __Pyx_TraceStartFunc
  #define __Pyx_TraceYield(result, offset, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      PyThreadState* tstate = __Pyx_PyThreadState_Current;\
      if (__Pyx_IsTracing(tstate, 0, 0)) {\
          __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result, 0);\
      }\
      if ((1)); else goto_error;\
  }
  #define __Pyx_TraceResumeGen(funcname, srcfile, firstlineno, offset, goto_error)\
      __Pyx_TraceStartFunc(funcname, srcfile, firstlineno, offset, 0, 0, goto_error)
  CYTHON_UNUSED static void __Pyx_call_return_trace_func(PyThreadState *tstate, PyFrameObject *frame, PyObject *result, int delete_frame) {
      PyObject *type, *value, *traceback;
      __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
      __Pyx_EnterTracing(tstate);
      if (CYTHON_TRACE && tstate->c_tracefunc)
          tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_RETURN, result);
      if (tstate->c_profilefunc)
          tstate->c_profilefunc(tstate->c_profileobj, frame, PyTrace_RETURN, result);
      if (delete_frame) {
          CYTHON_FRAME_DEL(frame);
      }
      __Pyx_LeaveTracing(tstate);
      __Pyx_ErrRestoreInState(tstate, type, value, traceback);
  }
  #define __Pyx_TraceReturnValue(result, offset, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0)) {\
                  __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result, 1);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0)) {\
              __Pyx_call_return_trace_func(tstate, __pyx_frame, (PyObject*)result, 1);\
          }\
      }\
      if ((1)); else goto_error;\
  }
  #define __Pyx_TraceReturnCValue(cresult, convert_function, offset, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0)) {\
                  PyObject *pyvalue = convert_function(cresult);\
                  if (unlikely(!pyvalue)) goto_error;\
                  __Pyx_call_return_trace_func(tstate, __pyx_frame, pyvalue, 1);\
                  Py_DECREF(pyvalue);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0)) {\
              PyObject *pyvalue = convert_function(cresult);\
              if (unlikely(!pyvalue)) goto_error;\
              __Pyx_call_return_trace_func(tstate, __pyx_frame, pyvalue, 1);\
              Py_DECREF(pyvalue);\
          }\
      }\
  }
  #define __Pyx_TraceExceptionUnwind(offset, nogil)\
  if (likely(!__Pyx_use_tracing)); else {\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0)) {\
                  __Pyx_call_return_trace_func(tstate, __pyx_frame, Py_None, 1);\
              }\
              PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0)) {\
              __Pyx_call_return_trace_func(tstate, __pyx_frame, Py_None, 1);\
          }\
      }\
  }
  static int __Pyx_TraceSetupAndCall(PyCodeObject** code, PyFrameObject** frame, PyThreadState* tstate, const char *funcname, const char *srcfile, int firstlineno, int skip_event);
#if CYTHON_TRACE
  CYTHON_UNUSED static int __Pyx_call_line_trace_func(PyThreadState *tstate, PyFrameObject *frame, int line);
  #define __Pyx_TraceLine(line, offset, nogil, goto_error)\
  if (likely(!__Pyx_use_tracing)); else {\
      int ret = 0;\
      if (nogil) {\
          if (CYTHON_TRACE_NOGIL) {\
              PyThreadState *tstate;\
              PyGILState_STATE state = __Pyx_PyGILState_Ensure();\
              tstate = __Pyx_PyThreadState_Current;\
              if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
                  ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, line);\
              }\
              __Pyx_PyGILState_Release(state);\
          }\
      } else {\
          PyThreadState* tstate = __Pyx_PyThreadState_Current;\
          if (__Pyx_IsTracing(tstate, 0, 0) && tstate->c_tracefunc && __pyx_frame->f_trace) {\
              ret = __Pyx_call_line_trace_func(tstate, __pyx_frame, line);\
          }\
      }\
      if (unlikely(ret)) goto_error;\
  }
#endif
#endif
#else
  #define __Pyx_TraceDeclarationsFunc
  #define __Pyx_TraceDeclarationsGen
  #define __Pyx_TraceExceptionDone()  {}
  #define __Pyx_TraceFrameInit(codeobj)  {}
  #define __Pyx_TurnOffSysMonitoringInParallel {}
  #define __Pyx_PyMonitoring_ExitScope(nogil)  {}
  #define __Pyx_TraceException(offset, reraised, fresh)  {}
  #define __Pyx_TraceExceptionUnwind(offset, nogil)  {}
  #define __Pyx_TraceExceptionHandled(offset)  {}
  #define __Pyx_TraceStartFunc(funcname, srcfile, firstlineno, offset, nogil, skip_event, goto_error)   if ((1)); else goto_error;
  #define __Pyx_TraceStartGen __Pyx_TraceStartFunc
  #define __Pyx_TraceResumeGen(funcname, srcfile, firstlineno, offset, goto_error)   if ((1)); else goto_error;
  #define __Pyx_TraceYield(result, offset, goto_error)   if ((1)); else goto_error;
  #define __Pyx_TraceReturnValue(result, offset, nogil, goto_error)\
      if ((1)); else goto_error;
  #define __Pyx_TraceReturnCValue(cresult, convert_function, offset, nogil, goto_error)\
      if ((1)); else { (void) convert_function; goto_error }
#endif
#if !CYTHON_TRACE
  #define __Pyx_TraceLine(line, offset, nogil, goto_error)   if ((1)); else goto_error;
#endif

/* CallCFunction.proto */
#define __Pyx_CallCFunction(cfunc, self, args)\
    ((PyCFunction)(void(*)(void))(cfunc)->func)(self, args)
#define __Pyx_CallCFunctionWithKeywords(cfunc, self, args, kwargs)\
    ((PyCFunctionWithKeywords)(void(*)(void))(cfunc)->func)(self, args, kwargs)
#define __Pyx_CallCFunctionFast(cfunc, self, args, nargs)\
    ((__Pyx_PyCFunctionFast)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs)
#define __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, nargs, kwnames)\
    ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))(PyCFunction)(cfunc)->func)(self, args, nargs, kwnames)

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#if !CYTHON_VECTORCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject *const *args, Py_ssize_t nargs, PyObject *kwargs);
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
#if !CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03080000
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()
  #define __Pyx_PyFrame_GetLocalsplus(frame)  ((frame)->f_localsplus)
#else
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectFastCall.proto */
#define __Pyx_PyObject_FastCall(func, args, nargs)  __Pyx_PyObject_FastCallDict(func, args, (size_t)(nargs), NULL)
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs);

/* PyObjectCall2Args.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* UnpackUnboundCMethod.proto */
typedef struct {
    PyObject *type;
    PyObject **method_name;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && CYTHON_ATOMICS
    __pyx_atomic_int_type initialized;
#endif
    PyCFunction func;
    PyObject *method;
    int flag;
} __Pyx_CachedCFunction;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
static CYTHON_INLINE int __Pyx_CachedCFunction_GetAndSetInitializing(__Pyx_CachedCFunction *cfunc) {
#if !CYTHON_ATOMICS
    return 1;
#else
    __pyx_nonatomic_int_type expected = 0;
    if (__pyx_atomic_int_cmp_exchange(&cfunc->initialized, &expected, 1)) {
        return 0;
    }
    return expected;
#endif
}
static CYTHON_INLINE void __Pyx_CachedCFunction_SetFinishedInitializing(__Pyx_CachedCFunction *cfunc) {
#if CYTHON_ATOMICS
    __pyx_atomic_store(&cfunc->initialized, 2);
#endif
}
#else
#define __Pyx_CachedCFunction_GetAndSetInitializing(cfunc) 2
#define __Pyx_CachedCFunction_SetFinishedInitializing(cfunc)
#endif

/* CallUnboundCMethod1.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg);
#else
#define __Pyx_CallUnboundCMethod1(cfunc, self, arg)  __Pyx__CallUnboundCMethod1(cfunc, self, arg)
#endif

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* TupleAndListFromArray.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n);
#endif
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject* __Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n);
#endif

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* fastcall.proto */
#if CYTHON_AVOID_BORROWED_REFS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_PySequence_ITEM(args, i)
#elif CYTHON_ASSUME_SAFE_MACROS
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_NewRef(__Pyx_PyTuple_GET_ITEM(args, i))
#else
    #define __Pyx_ArgRef_VARARGS(args, i) __Pyx_XNewRef(PyTuple_GetItem(args, i))
#endif
#define __Pyx_NumKwargs_VARARGS(kwds) PyDict_Size(kwds)
#define __Pyx_KwValues_VARARGS(args, nargs) NULL
#define __Pyx_GetKwValue_VARARGS(kw, kwvalues, s) __Pyx_PyDict_GetItemStrWithError(kw, s)
#define __Pyx_KwargsAsDict_VARARGS(kw, kwvalues) PyDict_Copy(kw)
#if CYTHON_METH_FASTCALL
    #define __Pyx_ArgRef_FASTCALL(args, i) __Pyx_NewRef(args[i])
    #define __Pyx_NumKwargs_FASTCALL(kwds) __Pyx_PyTuple_GET_SIZE(kwds)
    #define __Pyx_KwValues_FASTCALL(args, nargs) ((args) + (nargs))
    static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s);
  #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
    CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues);
  #else
    #define __Pyx_KwargsAsDict_FASTCALL(kw, kwvalues) _PyStack_AsDict(kwvalues, kw)
  #endif
#else
    #define __Pyx_ArgRef_FASTCALL __Pyx_ArgRef_VARARGS
    #define __Pyx_NumKwargs_FASTCALL __Pyx_NumKwargs_VARARGS
    #define __Pyx_KwValues_FASTCALL __Pyx_KwValues_VARARGS
    #define __Pyx_GetKwValue_FASTCALL __Pyx_GetKwValue_VARARGS
    #define __Pyx_KwargsAsDict_FASTCALL __Pyx_KwargsAsDict_VARARGS
#endif
#define __Pyx_ArgsSlice_VARARGS(args, start, stop) PyTuple_GetSlice(args, start, stop)
#if CYTHON_METH_FASTCALL || (CYTHON_COMPILING_IN_CPYTHON && CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) __Pyx_PyTuple_FromArray(args + start, stop - start)
#else
#define __Pyx_ArgsSlice_FASTCALL(args, start, stop) PyTuple_GetSlice(args, start, stop)
#endif

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static CYTHON_INLINE int __Pyx_ParseKeywords(
    PyObject *kwds, PyObject *const *kwvalues, PyObject ** const argnames[],
    PyObject *kwds2, PyObject *values[],
    Py_ssize_t num_pos_args, Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs
);

/* CallUnboundCMethod2.proto */
CYTHON_UNUSED
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2);
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2);
#else
#define __Pyx_CallUnboundCMethod2(cfunc, self, arg1, arg2)  __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2)
#endif

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely(__Pyx_IS_TYPE(obj, type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* RaiseUnexpectedTypeError.proto */
static int __Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj);

/* RejectKeywords.proto */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  do {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_mstate_global->__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  do {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
} while(0)
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* PyObjectFastCallMethod.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_PyObject_FastCallMethod(name, args, nargsf) PyObject_VectorcallMethod(name, args, nargsf, NULL)
#else
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf);
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        #if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000
        L->ob_item[len] = x;
        #else
        PyList_SET_ITEM(list, len, x);
        #endif
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectCallNoArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
    PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple);
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyDictContains.proto */
static CYTHON_INLINE int __Pyx_PyDict_ContainsTF(PyObject* item, PyObject* dict, int eq) {
    int result = PyDict_Contains(dict, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* DictGetItem.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key);
#define __Pyx_PyObject_Dict_GetItem(obj, name)\
    (likely(PyDict_CheckExact(obj)) ?\
     __Pyx_PyDict_GetItem(obj, name) : PyObject_GetItem(obj, name))
#else
#define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#define __Pyx_PyObject_Dict_GetItem(obj, name)  PyObject_GetItem(obj, name)
#endif

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* SetItemInt.proto */
#define __Pyx_SetItemInt(o, i, v, type, is_signed, to_py_func, is_list, wraparound, boundscheck, has_gil)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_SetItemInt_Fast(o, (Py_ssize_t)i, v, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list assignment index out of range"), -1) :\
               __Pyx_SetItemInt_Generic(o, to_py_func(i), v)))
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v);
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v,
                                               int is_list, int wraparound, int boundscheck);

/* PyLongBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE PyObject* __Pyx_PyLong_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyLong_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyLongCompare.proto */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace);

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00a2
    return PyList_Extend(L, v);
#elif CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030d0000
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyObjectVectorCallKwBuilder.proto */
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#if CYTHON_VECTORCALL
#if PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_Vectorcall_CallFromBuilder PyObject_Vectorcall
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder _PyObject_Vectorcall
#endif
#define __Pyx_MakeVectorcallBuilderKwds(n) PyTuple_New(n)
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n);
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n);
#else
#define __Pyx_Object_Vectorcall_CallFromBuilder __Pyx_PyObject_FastCallDict
#define __Pyx_MakeVectorcallBuilderKwds(n) __Pyx_PyDict_NewPresized(n)
#define __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n) PyDict_SetItem(builder, key, value)
#define __Pyx_VectorcallBuilder_AddArgStr(key, value, builder, args, n) PyDict_SetItemString(builder, key, value)
#endif

/* PyObjectVectorCallMethodKwBuilder.proto */
#if CYTHON_VECTORCALL && PY_VERSION_HEX >= 0x03090000
#define __Pyx_Object_VectorcallMethod_CallFromBuilder PyObject_VectorcallMethod
#else
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#endif

/* PyObjectDelAttr.proto */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
#define __Pyx_PyObject_DelAttr(o, n) PyObject_SetAttr(o, n, NULL)
#else
#define __Pyx_PyObject_DelAttr(o, n) PyObject_DelAttr(o, n)
#endif

/* PyObjectSetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
#define __Pyx_PyObject_DelAttrStr(o,n) __Pyx_PyObject_SetAttrStr(o, n, NULL)
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value);
#else
#define __Pyx_PyObject_DelAttrStr(o,n)   __Pyx_PyObject_DelAttr(o,n)
#define __Pyx_PyObject_SetAttrStr(o,n,v) PyObject_SetAttr(o,n,v)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* HasAttr.proto */
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_HasAttr(o, n)  PyObject_HasAttrWithError(o, n)
#else
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);
#endif

/* CallNextTpTraverse.proto */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse);

/* CallTypeTraverse.proto */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#define __Pyx_call_type_traverse(o, always_call, visit, arg) 0
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg);
#endif

/* CallNextTpClear.proto */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear);

/* LimitedApiGetTypeDict.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp);
#endif

/* SetItemOnTypeDict.proto */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v);
#define __Pyx_SetItemOnTypeDict(tp, k, v) __Pyx__SetItemOnTypeDict((PyTypeObject*)tp, k, v)

/* FixUpExtensionType.proto */
static CYTHON_INLINE int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type);

/* ValidateBasesTuple.proto */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases);
#endif

/* PyType_Ready.proto */
CYTHON_UNUSED static int __Pyx_PyType_Ready(PyTypeObject *t);

/* SetVTable.proto */
static int __Pyx_SetVtable(PyTypeObject* typeptr , void* vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyTypeObject *type);

/* MergeVTables.proto */
static int __Pyx_MergeVtables(PyTypeObject *type);

/* DelItemOnTypeDict.proto */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k);
#define __Pyx_DelItemOnTypeDict(tp, k) __Pyx__DelItemOnTypeDict((PyTypeObject*)tp, k)

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto_3_1_3
#define __PYX_HAVE_RT_ImportType_proto_3_1_3
#if defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L
#include <stdalign.h>
#endif
#if (defined (__STDC_VERSION__) && __STDC_VERSION__ >= 201112L) || __cplusplus >= 201103L
#define __PYX_GET_STRUCT_ALIGNMENT_3_1_3(s) alignof(s)
#else
#define __PYX_GET_STRUCT_ALIGNMENT_3_1_3(s) sizeof(void*)
#endif
enum __Pyx_ImportType_CheckSize_3_1_3 {
   __Pyx_ImportType_CheckSize_Error_3_1_3 = 0,
   __Pyx_ImportType_CheckSize_Warn_3_1_3 = 1,
   __Pyx_ImportType_CheckSize_Ignore_3_1_3 = 2
};
static PyTypeObject *__Pyx_ImportType_3_1_3(PyObject* module, const char *module_name, const char *class_name, size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_1_3 check_size);
#endif

/* ImportDottedModule.proto */
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple);
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple);

/* ListPack.proto */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...);

/* FetchSharedCythonModule.proto */
static PyObject *__Pyx_FetchSharedCythonABIModule(void);

/* dict_setdefault.proto */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value, int is_safe_type);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases);

/* CommonTypesMetaclass.proto */
static int __pyx_CommonTypesMetaclass_init(PyObject *module);
#define __Pyx_CommonTypesMetaclass_USED

/* PyMethodNew.proto */
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ);

/* PyVectorcallFastCallDict.proto */
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw);
#endif

/* CythonFunctionShared.proto */
#define __Pyx_CyFunction_USED
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CYFUNCTION_COROUTINE     0x08
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
  #define __Pyx_CyFunction_GetClassObj(f)\
      (((__pyx_CyFunctionObject *) (f))->func_classobj)
#else
  #define __Pyx_CyFunction_GetClassObj(f)\
      ((PyObject*) ((PyCMethodObject *) (f))->mm_class)
#endif
#define __Pyx_CyFunction_SetClassObj(f, classobj)\
    __Pyx__CyFunction_SetClassObj((__pyx_CyFunctionObject *) (f), (classobj))
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject_HEAD
    PyObject *func;
#elif PY_VERSION_HEX < 0x030900B1
    PyCFunctionObject func;
#else
    PyCMethodObject func;
#endif
#if CYTHON_BACKPORT_VECTORCALL ||\
        (CYTHON_COMPILING_IN_LIMITED_API && CYTHON_METH_FASTCALL)
    __pyx_vectorcallfunc func_vectorcall;
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    PyObject *func_classobj;
#endif
    PyObject *defaults;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
    PyObject *func_is_coroutine;
} __pyx_CyFunctionObject;
#undef __Pyx_CyOrPyCFunction_Check
#define __Pyx_CyFunction_Check(obj)  __Pyx_TypeCheck(obj, __pyx_mstate_global->__pyx_CyFunctionType)
#define __Pyx_CyOrPyCFunction_Check(obj)  __Pyx_TypeCheck2(obj, __pyx_mstate_global->__pyx_CyFunctionType, &PyCFunction_Type)
#define __Pyx_CyFunction_CheckExact(obj)  __Pyx_IS_TYPE(obj, __pyx_mstate_global->__pyx_CyFunctionType)
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void));
#undef __Pyx_IsSameCFunction
#define __Pyx_IsSameCFunction(func, cfunc)   __Pyx__IsSameCyOrCFunction(func, cfunc)
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject* op, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj);
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func,
                                                         PyTypeObject *defaults_type);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(PyObject *module);
#if CYTHON_METH_FASTCALL
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames);
#if CYTHON_BACKPORT_VECTORCALL || CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_func_vectorcall(f) (((__pyx_CyFunctionObject*)f)->func_vectorcall)
#else
#define __Pyx_CyFunction_func_vectorcall(f) (((PyCFunctionObject*)f)->vectorcall)
#endif
#endif

/* CythonFunction.proto */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *closure,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);

/* CLineInTraceback.proto */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#else
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#endif

/* CodeObjectCache.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject __Pyx_CachedCodeObjectType;
#else
typedef PyCodeObject __Pyx_CachedCodeObjectType;
#endif
typedef struct {
    __Pyx_CachedCodeObjectType* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_int_type accessor_count;
  #endif
};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if !defined(__INTEL_COMPILER) && defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value);

/* CIntFromPy.proto */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum____pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind value);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value);

/* FormatTypeName.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
typedef PyObject *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%U"
#define __Pyx_DECREF_TypeName(obj) Py_XDECREF(obj)
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
#define __Pyx_PyType_GetFullyQualifiedName PyType_GetFullyQualifiedName
#else
static __Pyx_TypeName __Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp);
#endif
#else  // !LIMITED_API
typedef const char *__Pyx_TypeName;
#define __Pyx_FMT_TYPENAME "%.200s"
#define __Pyx_PyType_GetFullyQualifiedName(tp) ((tp)->tp_name)
#define __Pyx_DECREF_TypeName(obj)
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) __Pyx_IsAnySubtype2(Py_TYPE(obj), (PyTypeObject *)type1, (PyTypeObject *)type2)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_TypeCheck2(obj, type1, type2) (PyObject_TypeCheck(obj, (PyTypeObject *)type1) || PyObject_TypeCheck(obj, (PyTypeObject *)type2))
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2) {
    return PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2);
}
#endif
#define __Pyx_PyErr_ExceptionMatches2(err1, err2)  __Pyx_PyErr_GivenExceptionMatches2(__Pyx_PyErr_CurrentExceptionType(), err1, err2)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#ifdef PyExceptionInstance_Check
  #define __Pyx_PyBaseException_Check(obj) PyExceptionInstance_Check(obj)
#else
  #define __Pyx_PyBaseException_Check(obj) __Pyx_TypeCheck(obj, PyExc_BaseException)
#endif

/* GetRuntimeVersion.proto */
static unsigned long __Pyx_get_runtime_version(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer);

/* VoidPtrImport.proto */
static int __Pyx_ImportVoidPtr_3_1_3(PyObject *module, const char *name, void **p, const char *sig);

/* MultiPhaseInitModuleState.proto */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
static PyObject *__Pyx_State_FindModule(void*);
static int __Pyx_State_AddModule(PyObject* module, void*);
static int __Pyx_State_RemoveModule(void*);
#elif CYTHON_USE_MODULE_STATE
#define __Pyx_State_FindModule PyState_FindModule
#define __Pyx_State_AddModule PyState_AddModule
#define __Pyx_State_RemoveModule PyState_RemoveModule
#endif

/* #### Code section: module_declarations ### */
/* CythonABIVersion.proto */
#if CYTHON_COMPILING_IN_LIMITED_API
    #if CYTHON_METH_FASTCALL
        #define __PYX_FASTCALL_ABI_SUFFIX  "_fastcall"
    #else
        #define __PYX_FASTCALL_ABI_SUFFIX
    #endif
    #define __PYX_LIMITED_ABI_SUFFIX "limited" __PYX_FASTCALL_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#else
    #define __PYX_LIMITED_ABI_SUFFIX
#endif
#if __PYX_HAS_PY_AM_SEND == 1
    #define __PYX_AM_SEND_ABI_SUFFIX
#elif __PYX_HAS_PY_AM_SEND == 2
    #define __PYX_AM_SEND_ABI_SUFFIX "amsendbackport"
#else
    #define __PYX_AM_SEND_ABI_SUFFIX "noamsend"
#endif
#ifndef __PYX_MONITORING_ABI_SUFFIX
    #define __PYX_MONITORING_ABI_SUFFIX
#endif
#if CYTHON_USE_TP_FINALIZE
    #define __PYX_TP_FINALIZE_ABI_SUFFIX
#else
    #define __PYX_TP_FINALIZE_ABI_SUFFIX "nofinalize"
#endif
#if CYTHON_USE_FREELISTS || !defined(__Pyx_AsyncGen_USED)
    #define __PYX_FREELISTS_ABI_SUFFIX
#else
    #define __PYX_FREELISTS_ABI_SUFFIX "nofreelists"
#endif
#define CYTHON_ABI  __PYX_ABI_VERSION __PYX_LIMITED_ABI_SUFFIX __PYX_MONITORING_ABI_SUFFIX __PYX_TP_FINALIZE_ABI_SUFFIX __PYX_FREELISTS_ABI_SUFFIX __PYX_AM_SEND_ABI_SUFFIX
#define __PYX_ABI_MODULE_NAME "_cython_" CYTHON_ABI
#define __PYX_TYPE_MODULE_PREFIX __PYX_ABI_MODULE_NAME "."

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_open(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_state_data_view_store); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_accumulate(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_data); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_retract(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_data); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_namespace, CYTHON_UNUSED PyObject *__pyx_v_accumulators, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_set_accumulators(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_namespace, CYTHON_UNUSED PyObject *__pyx_v_accumulators, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_get_accumulators(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_create_accumulators(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_cleanup(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_namespace, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_close(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_namespace, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_open(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_state_data_view_store); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_accumulate(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_data); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_retract(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_data); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_merge(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_set_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_get_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_create_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_cleanup(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_close(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_get_value(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_open(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_function_context, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_process_element(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_row, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_process_watermark(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int64_t __pyx_v_watermark, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_on_event_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_timer, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_on_processing_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_timer, int __pyx_skip_dispatch); /* proto*/
static int64_t __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_to_utc_timestamp_mills(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int64_t __pyx_v_epoch_mills, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_get_timers(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_close(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase__register_cleanup_timer(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_window); /* proto*/
static int64_t __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_cleanup_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_window, int __pyx_skip_dispatch); /* proto*/
static struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase__emit_window_result(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_key, CYTHON_UNUSED PyObject *__pyx_v_window); /* proto*/
static struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction__emit_window_result(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_window); /* proto*/

/* Module declarations from "libc.stdint" */

/* Module declarations from "pyflink.fn_execution.stream_fast" */

/* Module declarations from "pyflink.fn_execution.coder_impl_fast" */
static unsigned char *__pyx_vp_7pyflink_12fn_execution_15coder_impl_fast_ROW_KIND_BIT_SIZE = 0;
#define __pyx_v_7pyflink_12fn_execution_15coder_impl_fast_ROW_KIND_BIT_SIZE (*__pyx_vp_7pyflink_12fn_execution_15coder_impl_fast_ROW_KIND_BIT_SIZE)

/* Module declarations from "pyflink.fn_execution.table.aggregate_fast" */

/* Module declarations from "libc.string" */

/* Module declarations from "libc.stdlib" */

/* Module declarations from "pyflink.fn_execution.table.window_aggregate_fast" */
static struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_join_row(PyObject *, PyObject *, enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunction__set_state(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *, PyObject *); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_GroupWindowAggFunctionBase__set_state(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_GroupWindowAggFunction__set_state(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *, PyObject *); /*proto*/
/* #### Code section: typeinfo ### */
/* #### Code section: before_global_var ### */
#define __Pyx_MODULE_NAME "pyflink.fn_execution.table.window_aggregate_fast"
extern int __pyx_module_is_main_pyflink__fn_execution__table__window_aggregate_fast;
int __pyx_module_is_main_pyflink__fn_execution__table__window_aggregate_fast = 0;

/* Implementation of "pyflink.fn_execution.table.window_aggregate_fast" */
/* #### Code section: global_var ### */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_super;
/* #### Code section: string_decls ### */
static const char __pyx_k_[] = ".";
static const char __pyx_k_9[] = "\200\001\330\0049\270\021\270&\300\001";
static const char __pyx_k_A[] = "\200A";
static const char __pyx_k_K[] = "K";
static const char __pyx_k_Q[] = "\200\001\330\004\n\210+\220Q";
static const char __pyx_k_W[] = "W";
static const char __pyx_k_i[] = "i";
static const char __pyx_k__2[] = "?";
static const char __pyx_k_gc[] = "gc";
static const char __pyx_k_8_q[] = "\200\001\330\0048\270\001\270\026\270q";
static const char __pyx_k_QfA[] = "\200\001\330\004=\270Q\270f\300A";
static const char __pyx_k_UTC[] = "UTC";
static const char __pyx_k_add[] = "__add__";
static const char __pyx_k_key[] = "key";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_pop[] = "pop";
static const char __pyx_k_utc[] = "utc";
static const char __pyx_k_4AV1[] = "\200\001\330\0044\260A\260V\2701";
static const char __pyx_k_A_t1[] = "\200A\330\010\017\210t\2201";
static const char __pyx_k_A_vQ[] = "\200A\330\010\014\320\014\037\230v\240Q";
static const char __pyx_k_Dict[] = "Dict";
static const char __pyx_k_List[] = "List";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_func[] = "__func__";
static const char __pyx_k_init[] = "__init__";
static const char __pyx_k_keys[] = "keys";
static const char __pyx_k_left[] = "left";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_open[] = "open";
static const char __pyx_k_pytz[] = "pytz";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_spec[] = "__spec__";
static const char __pyx_k_temp[] = "temp";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_udfs[] = "udfs";
static const char __pyx_k_A_6aq[] = "\200A\330\010\014\320\014$\320$6\260a\260q";
static const char __pyx_k_clear[] = "clear";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_merge[] = "merge";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_right[] = "right";
static const char __pyx_k_state[] = "state";
static const char __pyx_k_super[] = "super";
static const char __pyx_k_timer[] = "timer";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_at81HA[] = "\200\001\330\004\013\210;\220a\220t\2308\2401\240H\250A";
static const char __pyx_k_dict_2[] = "_dict";
static const char __pyx_k_enable[] = "enable";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_remove[] = "remove";
static const char __pyx_k_timers[] = "_timers";
static const char __pyx_k_typing[] = "typing";
static const char __pyx_k_tzinfo[] = "tzinfo";
static const char __pyx_k_udfs_2[] = "_udfs";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_values[] = "values";
static const char __pyx_k_window[] = "window";
static const char __pyx_k_Trigger[] = "Trigger";
static const char __pyx_k_cleanup[] = "cleanup";
static const char __pyx_k_dealloc[] = "__dealloc__";
static const char __pyx_k_disable[] = "disable";
static const char __pyx_k_get_key[] = "get_key";
static const char __pyx_k_replace[] = "replace";
static const char __pyx_k_retract[] = "retract";
static const char __pyx_k_trigger[] = "trigger";
static const char __pyx_k_udf_num[] = "_udf_num";
static const char __pyx_k_A_G4q_vQ[] = "\200A\330\010\014\210G\2204\220q\330\014\017\210v\220Q";
static const char __pyx_k_A_q_c_WD[] = "\200A\330\010\017\210q\220\003\320\023&\240c\250\024\250W\260D\270\001";
static const char __pyx_k_add_note[] = "add_note";
static const char __pyx_k_datetime[] = "datetime";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_join_row[] = "join_row";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_row_kind[] = "row_kind";
static const char __pyx_k_set_name[] = "__set_name__";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_state_id[] = "state_id";
static const char __pyx_k_timezone[] = "timezone";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_get_value[] = "get_value";
static const char __pyx_k_input_row[] = "input_row";
static const char __pyx_k_isenabled[] = "isenabled";
static const char __pyx_k_key_coder[] = "key_coder";
static const char __pyx_k_namespace[] = "namespace";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_timestamp[] = "timestamp";
static const char __pyx_k_trigger_2[] = "_trigger";
static const char __pyx_k_watermark[] = "watermark";
static const char __pyx_k_accumulate[] = "accumulate";
static const char __pyx_k_get_timers[] = "get_timers";
static const char __pyx_k_input_data[] = "input_data";
static const char __pyx_k_on_element[] = "on_element";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_A_4q_Q_4q_Q[] = "\200A\330\010\013\2104\210q\330\014\020\220\001\220\024\220Q\330\010\013\2104\210q\330\014\020\220\001\220\024\220Q";
static const char __pyx_k_MapViewSpec[] = "MapViewSpec";
static const char __pyx_k_PickleCoder[] = "PickleCoder";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_epoch_mills[] = "epoch_mills";
static const char __pyx_k_field_index[] = "field_index";
static const char __pyx_k_filter_args[] = "filter_args";
static const char __pyx_k_microsecond[] = "microsecond";
static const char __pyx_k_s_D_q_a_IWA[] = "\200\001\330\004\007\200s\210!\210=\230\002\230\"\230D\240\007\240q\250\016\260a\330\010\024\220I\230W\240A\240[\260\001\260\021";
static const char __pyx_k_value_coder[] = "value_coder";
static const char __pyx_k_A_t_1_4_a_1A[] = "\200A\340\010\027\220t\230=\250\001\250\021\330\010\013\210=\230\003\2301\330\014\r\340\010\013\2104\320\017 \240\016\250a\330\014\020\320\020!\320!;\2701\270A\340\014\020\320\020!\320!@\300\001\300\021";
static const char __pyx_k_DataViewSpec[] = "DataViewSpec";
static const char __pyx_k_ListViewSpec[] = "ListViewSpec";
static const char __pyx_k_accumulators[] = "accumulators";
static const char __pyx_k_cleanup_time[] = "cleanup_time";
static const char __pyx_k_initializing[] = "_initializing";
static const char __pyx_k_is_coroutine[] = "_is_coroutine";
static const char __pyx_k_key_selector[] = "key_selector";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "<stringsource>";
static const char __pyx_k_use_setstate[] = "use_setstate";
static const char __pyx_k_window_state[] = "window_state";
static const char __pyx_k_A_IT_1_a_HF_q[] = "\200A\360\006\000\t\022\220\021\330\010\014\210I\220T\320\0311\260\030\270\025\270a\330\014\022\220'\230\021\230!\330\010\014\320\014$\240H\250F\260!\330\010\017\210q";
static const char __pyx_k_WindowContext[] = "WindowContext";
static const char __pyx_k_agg_ddistinct[] = "agg%ddistinct";
static const char __pyx_k_class_getitem[] = "__class_getitem__";
static const char __pyx_k_element_coder[] = "element_coder";
static const char __pyx_k_filter_args_2[] = "_filter_args";
static const char __pyx_k_fromtimestamp[] = "fromtimestamp";
static const char __pyx_k_get_namespace[] = "get_namespace";
static const char __pyx_k_get_timestamp[] = "get_timestamp";
static const char __pyx_k_is_event_time[] = "is_event_time";
static const char __pyx_k_max_timestamp[] = "max_timestamp";
static const char __pyx_k_on_event_time[] = "on_event_time";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_rowtime_index[] = "rowtime_index";
static const char __pyx_k_state_backend[] = "state_backend";
static const char __pyx_k_total_seconds[] = "total_seconds";
static const char __pyx_k_MAX_LONG_VALUE[] = "MAX_LONG_VALUE";
static const char __pyx_k_TriggerContext[] = "TriggerContext";
static const char __pyx_k_WindowAssigner[] = "WindowAssigner";
static const char __pyx_k_accumulators_2[] = "_accumulators";
static const char __pyx_k_key_selector_2[] = "_key_selector";
static const char __pyx_k_shift_timezone[] = "shift_timezone";
static const char __pyx_k_udf_data_views[] = "_udf_data_views";
static const char __pyx_k_window_context[] = "_window_context";
static const char __pyx_k_window_state_2[] = "_window_state";
static const char __pyx_k_get_value_state[] = "get_value_state";
static const char __pyx_k_process_element[] = "process_element";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_rowtime_index_2[] = "_rowtime_index";
static const char __pyx_k_set_current_key[] = "set_current_key";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_state_backend_2[] = "_state_backend";
static const char __pyx_k_trigger_context[] = "_trigger_context";
static const char __pyx_k_window_assigner[] = "window_assigner";
static const char __pyx_k_window_function[] = "_window_function";
static const char __pyx_k_allowed_lateness[] = "allowed_lateness";
static const char __pyx_k_distinct_indexes[] = "distinct_indexes";
static const char __pyx_k_function_context[] = "function_context";
static const char __pyx_k_get_accumulators[] = "get_accumulators";
static const char __pyx_k_input_extractors[] = "input_extractors";
static const char __pyx_k_set_accumulators[] = "set_accumulators";
static const char __pyx_k_shift_timezone_2[] = "_shift_timezone";
static const char __pyx_k_utcfromtimestamp[] = "utcfromtimestamp";
static const char __pyx_k_A_Q_e9A_nO_Na_y_q[] = "\200A\330#$\330\037 \360\006\000#$\330$%\330\032\033\330 !\330!\"\330\010\r\210Q\320\016&\240e\2509\260A\330\014\036\230n\250O\320;N\310a\330\014\037\230y\250\017\260q\330\010\014\320\014\"\240!";
static const char __pyx_k_advance_watermark[] = "advance_watermark";
static const char __pyx_k_get_value_indexes[] = "_get_value_indexes";
static const char __pyx_k_process_watermark[] = "process_watermark";
static const char __pyx_k_pyflink_table_udf[] = "pyflink.table.udf";
static const char __pyx_k_state_value_coder[] = "state_value_coder";
static const char __pyx_k_window_aggregator[] = "window_aggregator";
static const char __pyx_k_window_assigner_2[] = "_window_assigner";
static const char __pyx_k_allowed_lateness_2[] = "_allowed_lateness";
static const char __pyx_k_asyncio_coroutines[] = "asyncio.coroutines";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_create_accumulator[] = "create_accumulator";
static const char __pyx_k_distinct_indexes_2[] = "_distinct_indexes";
static const char __pyx_k_emit_window_result[] = "_emit_window_result";
static const char __pyx_k_get_state_map_view[] = "get_state_map_view";
static const char __pyx_k_input_extractors_2[] = "_input_extractors";
static const char __pyx_k_on_processing_time[] = "on_processing_time";
static const char __pyx_k_A_DAQ_T_Jaq_XQe_a_q[] = "\200A\360\006\000\t\r\320\014\035\320\035D\300A\300Q\330\010\025\220T\320\031,\250J\260a\260q\330\010\025\220X\230Q\230e\320#>\270a\330\010\017\210q";
static const char __pyx_k_PanedWindowAssigner[] = "PanedWindowAssigner";
static const char __pyx_k_count_star_inserted[] = "count_star_inserted";
static const char __pyx_k_create_accumulators[] = "create_accumulators";
static const char __pyx_k_distinct_data_views[] = "_distinct_data_views";
static const char __pyx_k_get_runtime_context[] = "get_runtime_context";
static const char __pyx_k_get_state_list_view[] = "get_state_list_view";
static const char __pyx_k_index_of_count_star[] = "index_of_count_star";
static const char __pyx_k_state_value_coder_2[] = "_state_value_coder";
static const char __pyx_k_udf_data_view_specs[] = "udf_data_view_specs";
static const char __pyx_k_window_aggregator_2[] = "_window_aggregator";
static const char __pyx_k_A_E_as_4q_T_1_G1_q_q[] = "\200A\340\010\014\210E\220\025\220a\220s\230!\2304\230q\330\014\020\220\r\230T\320!1\260\021\260\"\260G\2701\330\020\031\320\031/\250q\260\001\330\020\031\230\026\230q";
static const char __pyx_k_hk_A_1_a_7_8_9RR_a_1[] = "\200\001\360\006\000\005\010\200\177\220h\230k\250\033\260A\330\010\r\210^\2301\330\010\016\320\016!\360\000\000\"^\005\360\000\000^\005`\005\360\000\000`\005a\005\330\004\023\320\023)\250\030\260\021\260!\330\004\007\200|\2207\230!\330\0108\270\001\3209R\320R`\320`a\330\004\013\2101";
static const char __pyx_k_MergingWindowAssigner[] = "MergingWindowAssigner";
static const char __pyx_k_assign_actual_windows[] = "assign_actual_windows";
static const char __pyx_k_set_current_namespace[] = "set_current_namespace";
static const char __pyx_k_state_data_view_store[] = "state_data_view_store";
static const char __pyx_k_udf_data_view_specs_2[] = "_udf_data_view_specs";
static const char __pyx_k_GroupWindowAggFunction[] = "GroupWindowAggFunction";
static const char __pyx_k_assign_state_namespace[] = "assign_state_namespace";
static const char __pyx_k_clean_window_if_needed[] = "clean_window_if_needed";
static const char __pyx_k_internal_timer_service[] = "_internal_timer_service";
static const char __pyx_k_register_cleanup_timer[] = "_register_cleanup_timer";
static const char __pyx_k_to_utc_timestamp_mills[] = "to_utc_timestamp_mills";
static const char __pyx_k_A_D_a_Jat_J_U_4q_4q_8_q[] = "\200A\360\010\000\t\035\230D\240\001\330\010\021\220\021\330\014\020\220\006\220a\320\027(\250\001\250\023\250J\260a\260t\270>\310\021\320J[\320[\\\320\\]\330\014\020\220\005\220U\230!\2304\230q\330\010\013\2104\210q\330\014\022\220'\230\021\230$\320\0368\270\001\270\021\330\010\017\210q";
static const char __pyx_k_current_processing_time[] = "current_processing_time";
static const char __pyx_k_A_Q_a_a_q_q_q_L_T_0_AQQR[] = "\200A\330#$\330\037 \360\006\000#$\330$%\330\032\033\330 !\330!\"\330\010\014\320\014!\240\021\330\010\014\320\014\035\230Q\330\010\014\320\014\036\230a\330\010\014\320\014\"\240!\330\010\014\320\014 \240\001\330\010\014\320\014\"\240!\330\010\014\320\014\036\230a\330\010\014\320\014\037\230q\330\010\014\320\014 \240\001\330\010\014\320\014'\240q\330\010\014\320\014\037\230q\330\010\014\210L\230\001\330\010\014\320\014 \240\001\330\010\014\320\014\035\230T\240\037\3200@\300\001\320AQ\320QR";
static const char __pyx_k_get_value_indexes_length[] = "_get_value_indexes_length";
static const char __pyx_k_hk_A_1_a_XQa_7_A_ZZhhi_1[] = "\200\001\360\006\000\005\010\200\177\220h\230k\250\033\260A\330\010\r\210^\2301\330\010\016\320\016!\360\000\000\"^\005\360\000\000^\005`\005\360\000\000`\005a\005\330\004\023\320\023-\250X\260Q\260a\330\004\007\200|\2207\230!\330\010<\270A\320=Z\320Zh\320hi\330\004\013\2101";
static const char __pyx_k_named_property_extractor[] = "named_property_extractor";
static const char __pyx_k_pyflink_common_constants[] = "pyflink.common.constants";
static const char __pyx_k_distinct_view_descriptors[] = "distinct_view_descriptors";
static const char __pyx_k_register_event_time_timer[] = "register_event_time_timer";
static const char __pyx_k_A_4_a_as_c_4q_Cv_1_q_q_s_a[] = "\200A\340\010\013\2104\320\017 \240\016\250a\330\014\036\230a\230s\240&\250\016\260c\270\022\2704\270q\330\014\017\210}\230C\230v\240^\2601\330\020\027\220q\340\020\027\220q\340\014\027\220s\230&\240\016\250a";
static const char __pyx_k_GroupWindowAggFunctionBase[] = "GroupWindowAggFunctionBase";
static const char __pyx_k_PanedWindowProcessFunction[] = "PanedWindowProcessFunction";
static const char __pyx_k_hk_A_1_kkmmn_haq_7_Q_jjk_1[] = "\200\001\360\006\000\005\010\200\177\220h\230k\250\033\260A\330\010\r\210^\2301\330\010\016\320\016!\320!k\320km\320mn\330\004\023\320\023.\250h\260a\260q\330\004\007\200|\2207\230!\330\010=\270Q\320>\\\320\\j\320jk\330\004\013\2101";
static const char __pyx_k_named_property_extractor_2[] = "_named_property_extractor";
static const char __pyx_k_ImperativeAggregateFunction[] = "ImperativeAggregateFunction";
static const char __pyx_k_NamespaceAggsHandleFunction[] = "NamespaceAggsHandleFunction";
static const char __pyx_k_PerWindowStateDataViewStore[] = "PerWindowStateDataViewStore";
static const char __pyx_k_distinct_view_descriptors_2[] = "_distinct_view_descriptors";
static const char __pyx_k_pyflink_fn_execution_coders[] = "pyflink.fn_execution.coders";
static const char __pyx_k_A_G1F_a_vWA_q_q_q_D_SZZ_D_ST[] = "\200\001\360\010\000\005\r\210A\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220A\330\010\022\220!\330\010\027\220q\340\010\027\220q\330\004\007\200q\330\010\017\320\017<\270D\300\001\300\027\310\013\320SZ\320Z[\340\010\017\320\017<\270D\300\001\300\027\310\013\320ST";
static const char __pyx_k_GeneralWindowProcessFunction[] = "GeneralWindowProcessFunction";
static const char __pyx_k_MergingWindowProcessFunction[] = "MergingWindowProcessFunction";
static const char __pyx_k_InternalWindowProcessFunction[] = "InternalWindowProcessFunction";
static const char __pyx_k_hk_A_1_kkmmn_2_1_7_A_Bddrrs_1[] = "\200\001\360\006\000\005\010\200\177\220h\230k\250\033\260A\330\010\r\210^\2301\330\010\016\320\016!\320!k\320km\320mn\330\004\023\320\0232\260(\270!\2701\330\004\007\200|\2207\230!\330\010A\300\021\320Bd\320dr\320rs\330\004\013\2101";
static const char __pyx_k_A_G1F_a_vWA_q_q_q_AWKW___AWKWX[] = "\200\001\360\010\000\005\r\210A\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220A\330\010\022\220!\330\010\027\220q\340\010\027\220q\330\004\007\200q\330\010\017\320\017@\300\004\300A\300W\310K\320W^\320^_\340\010\017\320\017@\300\004\300A\300W\310K\320WX";
static const char __pyx_k_LegacyInternalTimerServiceImpl[] = "LegacyInternalTimerServiceImpl";
static const char __pyx_k_pyx_unpickle_GroupWindowAggFun[] = "__pyx_unpickle_GroupWindowAggFunctionBase";
static const char __pyx_k_pyx_unpickle_NamespaceAggsHand[] = "__pyx_unpickle_NamespaceAggsHandleFunctionBase";
static const char __pyx_k_register_processing_time_timer[] = "register_processing_time_timer";
static const char __pyx_k_A_4_JgQ_U_3at1_G_ARvQ_31A_1IQ_Q[] = "\200A\340\010\013\2104\320\017 \240\004\240J\250g\260Q\330\014\020\220\005\220U\230!\2303\230a\230t\2401\330\020\024\220G\230=\250\004\320,<\270A\270R\270v\300Q\330\024\035\320\0353\2601\260A\330\024 \240\001\240\022\2401\240I\250Q\330\010\014\320\014\035\230Q";
static const char __pyx_k_A_EQd_uA_q_a_Qd_Q_a_t1_q_q_t1_A[] = "\200A\330\010\014\320\014'\320'E\300Q\300d\310!\330\010\014\320\014\037\230u\240A\330\014'\240q\320(:\270$\270a\340\010\013\210:\220Q\220d\320\032-\250Q\330\014\020\320\020$\320$>\270a\330\020\024\320\024(\250\004\320,?\270t\3001\330\r\027\220q\230\004\320\034/\250q\330\014\020\320\020$\320$@\300\001\330\020\024\320\024(\250\004\320,?\270t\3001\330\020\024\220A\340\014\020\320\020$\320$@\300\001\330\020\024\320\024(\250\004\320,?\270t\3001\330\010\014\320\014 \240\016\250a\330\014\020\220\013\2304\320\0379\270\024\270Q\330\010\014\320\014\035\230U\240!\330\010\014\320\014\037\230}\250A\330\014\022\220$\320\026)\250\024\320->\270d\300!\330\014\020\320\020*\250$\320.?\270~\310Q\330\010\014\320\014\035\230U\240!\2404\240q";
static const char __pyx_k_A_E_q_e81_O_1A_nA_Zq_4_aq_2_5_4[] = "\200A\360\010\000\t\022\220\021\330\010\024\220E\230\036\240q\330\010\016\210e\2208\2301\330\010\014\210O\320\033+\2501\250A\330\010\021\220\025\220n\240A\330\010\014\320\014\035\230Z\240q\330\010\013\2104\320\017 \240\016\250a\250q\340\014\022\220'\230\021\230$\320\0362\260!\2605\270\001\340\010\013\2104\320\017 \240\016\250a\330\014\020\320\020!\320!8\270\001\270\030\300\021\330\010\017\210q";
static const char __pyx_k_A_G4q_uA_a_t1_Q_a_a_Q_A_q_0_a_Q[] = "\200A\340\010\014\210G\2204\220q\330\014\017\210u\220A\320\025*\320*>\270a\330\010\014\320\014\037\230t\2401\330\014\031\230\021\330\014\020\320\020\"\240!\330\020\023\220:\230Q\320\036.\250a\330\024\036\230a\230~\250Q\330\030-\320-A\300\021\330\034*\250!\330\034*\250!\330\025\037\230q\320 0\260\001\330\024\036\230a\230~\250Q\330\030-\320-@\300\001\330\034*\250!\330\034*\250!\330\034*\250!\330\014\020\320\020 \240\007\240q\250\001\330\010\014\210G\2204\320\0272\260%\260q\330\014\020\320\020%\240Q\240g\320-B\320BU\320UV\330\020 \240\002\240!\330\020\033\2301\330\020\033\2301";
static const char __pyx_k_A_gghhllx_y_J_J_U_U_V_V_Z_Z_f_f[] = "\200\001\330\004\020\320\020%\240[\260\001\260\024\260\\\320A\\\320\\g\320gh\320hl\320lx\360\000\000y\001J\002\360\000\000J\002U\002\360\000\000U\002V\002\360\000\000V\002Z\002\360\000\000Z\002f\002\360\000\000f\002x\002\360\000\000x\002C\003\360\000\000C\003D\003\360\000\000D\003H\003\360\000\000H\003T\003\360\000\000T\003g\003\360\000\000g\003r\003\360\000\000r\003s\003\360\000\000s\003w\003\360\000\000w\003C\004\360\000\000C\004U\004\360\000\000U\004`\004\360\000\000`\004a\004\360\000\000a\004e\004\360\000\000e\004q\004\360\000\000q\004G\005\360\000\000G\005R\005\360\000\000R\005S\005\360\000\000S\005W\005\360\000\000W\005c\005\360\000\000c\005o\005\360\000\000o\005z\005\360\000\000z\005{\005\360\000\000{\005\177\005\360\000\000\177\005K\006\360\000\000K\006_\006\360\000\000_\006j\006\360\000\000j\006k\006\360\000\000k\006o\006\360\000\000o\006{\006\360\000\000{\006Q\007\360\000\000Q\007\\\007\360\000\000\\\007]\007\360\000\000]\007a\007\360\000\000a\007m\007\360\000\000m\007A\010\360\000\000A\010L\010\360\000\000L\010M\010\360\000\000M\010R\010\360\000\000R\010^\010\360\000\000^\010q\010\360\000\000q\010|\010\360\000\000|\010}\010\360\000\000}\010B\t\360\000\000B\tN\t\360\000\000N\tb\t\360\000\000b\tm\t\360\000\000m\tn\t\360\000\000n\ts\t\360\000\000s\t\177\t\360\000\000\177\tP\n\360\000\000P\n[\n\360\000\000[\n\\\n\360\000\000\\\n]\n\330\004\007\200s\210!\210=\230\002\230#\230T\240\027\250\001\250\036\260q\330\010\024\220I\230W\240A\240[\260\001\260\021";
static const char __pyx_k_GroupWindowAggFunctionBase_open[] = "GroupWindowAggFunctionBase.open";
static const char __pyx_k_GroupWindowAggFunction___reduce[] = "GroupWindowAggFunction.__reduce_cython__";
static const char __pyx_k_NamespaceAggsHandleFunctionBase[] = "NamespaceAggsHandleFunctionBase";
static const char __pyx_k_NamespaceAggsHandleFunction_get[] = "NamespaceAggsHandleFunction.get_value";
static const char __pyx_k_T_T_C4GWW_llp_q_C_C_G_G_X_X_q_q[] = "\200\001\360\010\000\005\016\210T\320\021%\240T\320)C\3004\320GW\320W[\320[l\320lp\360\000\000q\001C\002\360\000\000C\002G\002\360\000\000G\002X\002\360\000\000X\002\\\002\360\000\000\\\002q\002\360\000\000q\002u\002\360\000\000u\002@\003\360\000\000@\003D\003\360\000\000D\003W\003\360\000\000W\003[\003\360\000\000[\003p\003\360\000\000p\003t\003\360\000\000t\003G\004\360\000\000G\004K\004\360\000\000K\004]\004\360\000\000]\004a\004\360\000\000a\004t\004\360\000\000t\004x\004\360\000\000x\004y\004\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220A\330\010\022\220!\330\010\027\220q\340\010\027\220t\320\0334\260G\2705\300\003\3004\300\177\320V]\320]b\320be\320ei\320iz\360\000\000{\001B\002\360\000\000B\002G\002\360\000\000G\002J\002\360\000\000J\002N\002\360\000\000N\002^\002\360\000\000^\002e\002\360\000\000e\002j\002\360\000\000j\002m\002\360\000\000m\002q\002\360\000\000q\002E\003\360\000\000E\003L\003\360\000\000L\003Q\003\360\000\000Q\003T\003\360\000\000T\003X\003\360\000\000X\003b\003\360\000\000b\003i\003\360\000\000i\003n\003\360\000\000n\003q\003\360\000\000q\003u\003\360\000\000u\003G\004\360\000\000G\004N\004\360\000\000N\004S\004\360\000\000S\004V\004\360\000\000V\004Z\004\360\000\000Z\004n\004\360\000\000n\004u\004\360\000\000u\004z\004\360\000\000z\004}\004\360\000\000}\004A\005\360\000\000A\005S\005\360\000\000S\005Z\005\360\000\000Z\005_\005\360\000\000_\005b\005\360\000\000b\005f\005\360\000\000f\005w\005\360\000\000w\005~\005\360\000\000~\005C\006\360\000\000C\006F\006\360\000\000F\006J\006\360\000\000J\006\\\006\360\000\000\\\006c\006\360\000\000c\006h\006\360\000\000h\006k\006\360\000\000k\006o\006\360\000\000o\006~\006\360\000\000~\006E\007\360\000\000E\007F\007\330\004\007\200q\330\010\017\320\017;\2704\270q\300\007\300{\320RY\320YZ\340\010\017\320\017;\2704\270q\300\007\300{\320RS";
static const char __pyx_k_pyflink_fn_execution_datastream[] = "pyflink.fn_execution.datastream.process.timerservice_impl";
static const char __pyx_k_A_4_1_1_t9AT_hi_Ql_GST_HI_6aq_3b[] = "\200A\330\010\013\2104\320\017 \240\003\2401\330\014\023\2201\340\014\027\220t\2309\240A\240T\250\021\330\014\036\230h\240i\250~\270Q\270l\310\"\310G\320ST\330\020\030\230\001\230\027\240\001\330\014\024\220H\230I\320%6\260a\260q\330\014\023\2203\220b\320\030(\250\002\250&\260\016\270c\300\022\3001";
static const char __pyx_k_A_4q_U_3at1_G_ARvQ_31A_1IQ_E_at1[] = "\200A\340\010\013\2104\210q\330\014\020\220\005\220U\230!\2303\230a\230t\2401\330\020\024\220G\230=\250\004\320,<\270A\270R\270v\300Q\330\024\035\320\0353\2601\260A\330\024 \240\001\240\022\2401\240I\250Q\330\010\014\210E\220\025\220a\220t\2301\330\014\020\220\006\220a\220r\230\026\230q\240\004\240N\260!\2604\260q\270\014\300A\300Q";
static const char __pyx_k_A_E_at1_r_D_4_J_1_8_A_q_q_U_1_Qa[] = "\200A\360\020\000\t\r\210E\220\025\220a\220t\2301\330\014\017\210r\220\023\220D\230\001\330\020+\2504\320/J\310!\3101\330\020 \320 8\270\001\330\020\023\220>\240\023\240A\330\024\037\230q\340\024\037\230q\330\024\"\320\":\270!\330\024\030\230\005\230U\240!\2401\330\030%\240[\260\001\260\021\330\030\033\230:\240Q\240a\330\034'\240q\330\034\035\330\020\023\2204\220q\330\024\033\320\0333\3203C\3001\300A\330\024)\250\024\320-B\300!\3001\330\024\027\220u\230C\230q\330\030*\250!\2509\260A\340\030*\250!\2508\2601\330\014\031\230\024\230]\250!\2501\330\014\017\210{\230#\230R\230t\2404\240z\260\021\260!\330\020\021\330\014\023\2204\320\027)\250\021\250\"\250A\250Q\330\014\035\230T\320!3\2601\260A\330\014\017\210\177\230c\240\021\330\020%\240T\320)>\270a\270q\330\020\023\2205\230\003\2301\330\024\027\320\027)\250\021\250&\260\002\260!\330\030\031\340\024\032\230)\2401\330\030\031\340\014\017\210s\220!\2206\230\023\230B\230d\240*\250A\250T\260\021\260$\260a\330\020\037\230~\250T\260\021\260!\330\020\024\220A\220U\230,\240g\250Q\330\014\020\220\006\220a\220r\230\033\240A\240T\250\036\260q\270\005\270Q";
static const char __pyx_k_A_E_q_e81_O_1A_nA_Zq_4_31A_2_5_4[] = "\200A\360\010\000\t\022\220\021\330\010\024\220E\230\036\240q\330\010\016\210e\2208\2301\330\010\014\210O\320\033+\2501\250A\330\010\021\220\025\220n\240A\330\010\014\320\014\035\230Z\240q\330\010\013\2104\320\017 \320 3\2601\260A\340\014\022\220'\230\021\230$\320\0362\260!\2605\270\001\340\010\013\2104\210t\320\023$\240N\260!\330\014\020\320\020!\320!8\270\001\270\030\300\021\330\010\017\210q";
static const char __pyx_k_A_IQ_L_1D_Q_A_q_G6_j_at_Rq_E_at1[] = "\200A\330\027\030\330#$\330&'\330&'\340&'\330\036\037\330#$\330,-\330\010\014\210I\220Q\330\010\014\210L\230\003\2301\230D\240\001\330\010\014\320\014!\240\021\330\010\014\320\014)\250\021\330\010\014\320\014\035\230Q\330\010\014\320\014$\240A\330\010\014\320\014\037\230q\330\010\014\320\014\034\230G\2406\250\021\250$\250j\270\002\270!\330\010\014\320\014!\240\027\250\006\250a\250t\260:\270R\270q\330\010\014\210E\220\025\220a\220t\2301\330\014\020\220\r\230Q\230e\240;\250a\250q\330\014\020\320\020\"\240!\2405\320(8\270\001\270\021\330\010\014\320\014*\250!\330\010\014\320\014$\240A\330\010\017\210q\220\002\220$\220e\2305\240\001\240\023\240A\240Q\330\010\013\320\013\037\230s\240\"\240D\250\001\340\014\020\220\007\220q\230\001\330\010\014\320\014)\250\023\250A\250Q\330\010\014\320\014\"\240*\250F\260!\260?\300\"\300D\310\001\330\010\014\210E\220\025\220a\220t\2301\330\014\020\320\020#\2401\240E\250\024\250Q\250a";
static const char __pyx_k_A_iq_d_O_1A_4_a_1D_c_81G89E_ST_Y[] = "\200A\360\n\000\t\027\220i\230q\330\010\026\220d\230.\250\010\260\001\260\021\330\010\014\210O\320\033+\2501\250A\330\010\013\2104\320\017 \240\016\250a\330\014\030\230\013\2401\240D\250\001\330\014\026\220c\230\021\230)\2408\2501\250G\2608\2709\300E\310\032\320ST\330\014%\240Y\250a\330\014\033\2308\2402\240U\250\"\320,C\3003\300a\330\014\030\230\001\340\014\030\230\004\320\0344\3204L\310A\340\010\024\220D\320\030/\250q\260\001\360\006\000\t\034\2304\320\0370\3200G\300q\310\r\320UV\330\010\014\210J\220a\330\014\020\220\016\320\0364\260A\260Q\330\014\022\220$\220n\240F\250!\330\014\017\210t\2203\220a\330\020\026\220d\320\032-\320-A\300\021\330\014\020\320\020#\320#4\260A\260X\270Q\340\014\017\210y\320\030*\250!\330\020\024\320\024'\240{\260!\2601\340\020\024\320\024'\240x\250q\260\001\330\014\022\220$\320\026)\320):\270!\330\014\020\220\016\230g\240Q\240a\360\006\000\t\032\230\024\320\035.\320.D\300A\300]\320RS\330\010\021\220\021\330\010\014\210J\220a\330\014\020\320\020!\240\032\2501\330\014\035\230T\320!2\260+\270Q\270k\310\021\330\014\017\210q\330\020\026\220g\230Q\230d\320\"6\260a\260}\300A\330\014\020\320\020(\250\001\250\021\330\010\017\210q";
static const char __pyx_k_GroupWindowAggFunctionBase___red[] = "GroupWindowAggFunctionBase.__reduce_cython__";
static const char __pyx_k_GroupWindowAggFunctionBase___set[] = "GroupWindowAggFunctionBase.__setstate_cython__";
static const char __pyx_k_GroupWindowAggFunctionBase_clean[] = "GroupWindowAggFunctionBase.cleanup_time";
static const char __pyx_k_GroupWindowAggFunctionBase_close[] = "GroupWindowAggFunctionBase.close";
static const char __pyx_k_GroupWindowAggFunctionBase_get_t[] = "GroupWindowAggFunctionBase.get_timers";
static const char __pyx_k_GroupWindowAggFunctionBase_on_ev[] = "GroupWindowAggFunctionBase.on_event_time";
static const char __pyx_k_GroupWindowAggFunctionBase_on_pr[] = "GroupWindowAggFunctionBase.on_processing_time";
static const char __pyx_k_GroupWindowAggFunctionBase_proce[] = "GroupWindowAggFunctionBase.process_element";
static const char __pyx_k_GroupWindowAggFunctionBase_to_ut[] = "GroupWindowAggFunctionBase.to_utc_timestamp_mills";
static const char __pyx_k_GroupWindowAggFunction___setstat[] = "GroupWindowAggFunction.__setstate_cython__";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0[] = "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())";
static const char __pyx_k_NamespaceAggsHandleFunction___re[] = "NamespaceAggsHandleFunction.__reduce_cython__";
static const char __pyx_k_NamespaceAggsHandleFunction___se[] = "NamespaceAggsHandleFunction.__setstate_cython__";
static const char __pyx_k_Note_that_Cython_is_deliberately[] = "Note that Cython is deliberately stricter than PEP-484 and rejects subclasses of builtin types. If you need to pass subclasses then set the 'annotation_typing' directive to False.";
static const char __pyx_k_SimpleNamespaceAggsHandleFunctio[] = "SimpleNamespaceAggsHandleFunction";
static const char __pyx_k_The_args_are_not_in_the_distinct[] = "The args are not in the distinct data view, this should not happen.";
static const char __pyx_k_prepare_aggregate_accumulator_fo[] = "prepare_aggregate_accumulator_for_emit";
static const char __pyx_k_pyflink_fn_execution_table_state[] = "pyflink.fn_execution.table.state_data_view";
static const char __pyx_k_pyflink_fn_execution_table_windo[] = "pyflink.fn_execution.table.window_assigner";
static const char __pyx_k_pyx_unpickle_GroupWindowAggFun_2[] = "__pyx_unpickle_GroupWindowAggFunction";
static const char __pyx_k_pyx_unpickle_GroupWindowAggFun_3[] = "__pyx_unpickle_GroupWindowAggFunctionBase__set_state";
static const char __pyx_k_pyx_unpickle_GroupWindowAggFun_4[] = "__pyx_unpickle_GroupWindowAggFunction__set_state";
static const char __pyx_k_pyx_unpickle_NamespaceAggsHand_2[] = "__pyx_unpickle_NamespaceAggsHandleFunction";
static const char __pyx_k_pyx_unpickle_NamespaceAggsHand_3[] = "__pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state";
static const char __pyx_k_pyx_unpickle_NamespaceAggsHand_4[] = "__pyx_unpickle_NamespaceAggsHandleFunction__set_state";
static const char __pyx_k_self__distinct_indexes_self__fil[] = "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling";
static const char __pyx_k_NamespaceAggsHandleFunctionBase_2[] = "NamespaceAggsHandleFunctionBase.merge";
static const char __pyx_k_NamespaceAggsHandleFunctionBase_3[] = "NamespaceAggsHandleFunctionBase.set_accumulators";
static const char __pyx_k_NamespaceAggsHandleFunctionBase_4[] = "NamespaceAggsHandleFunctionBase.create_accumulators";
static const char __pyx_k_NamespaceAggsHandleFunctionBase_5[] = "NamespaceAggsHandleFunctionBase.cleanup";
static const char __pyx_k_NamespaceAggsHandleFunctionBase_6[] = "NamespaceAggsHandleFunctionBase.__reduce_cython__";
static const char __pyx_k_NamespaceAggsHandleFunctionBase_7[] = "NamespaceAggsHandleFunctionBase.__setstate_cython__";
static const char __pyx_k_T_T_C4GWW_llp_q_C_C_G_G_X_X_q_q_2[] = "\200\001\360\010\000\005\016\210T\320\021%\240T\320)C\3004\320GW\320W[\320[l\320lp\360\000\000q\001C\002\360\000\000C\002G\002\360\000\000G\002X\002\360\000\000X\002\\\002\360\000\000\\\002q\002\360\000\000q\002u\002\360\000\000u\002@\003\360\000\000@\003D\003\360\000\000D\003W\003\360\000\000W\003[\003\360\000\000[\003p\003\360\000\000p\003t\003\360\000\000t\003G\004\360\000\000G\004K\004\360\000\000K\004]\004\360\000\000]\004a\004\360\000\000a\004t\004\360\000\000t\004x\004\360\000\000x\004y\004\330\004\014\210G\2201\220F\230,\240a\330\004\007\200v\210W\220A\330\010\022\220!\330\010\027\220q\340\010\027\220t\320\0334\260G\2705\300\003\3004\300\177\320V]\320]b\320be\320ei\320iz\360\000\000{\001B\002\360\000\000B\002G\002\360\000\000G\002J\002\360\000\000J\002N\002\360\000\000N\002^\002\360\000\000^\002e\002\360\000\000e\002j\002\360\000\000j\002m\002\360\000\000m\002q\002\360\000\000q\002E\003\360\000\000E\003L\003\360\000\000L\003Q\003\360\000\000Q\003T\003\360\000\000T\003X\003\360\000\000X\003b\003\360\000\000b\003i\003\360\000\000i\003n\003\360\000\000n\003q\003\360\000\000q\003u\003\360\000\000u\003G\004\360\000\000G\004N\004\360\000\000N\004S\004\360\000\000S\004V\004\360\000\000V\004Z\004\360\000\000Z\004n\004\360\000\000n\004u\004\360\000\000u\004z\004\360\000\000z\004}\004\360\000\000}\004A\005\360\000\000A\005S\005\360\000\000S\005Z\005\360\000\000Z\005_\005\360\000\000_\005b\005\360\000\000b\005f\005\360\000\000f\005w\005\360\000\000w\005~\005\360\000\000~\005C\006\360\000\000C\006F\006\360\000\000F\006J\006\360\000\000J\006\\\006\360\000\000\\\006c\006\360\000\000c\006h\006\360\000\000h\006k\006\360\000\000k\006o\006\360\000\000o\006~\006\360\000\000~\006E\007\360\000\000E\007F\007\330\004\007\200q\330\010\017\320\0177\260t\2701\270G\300;\310g\320UV\340\010\017\320\0177\260t\2701\270G\300;\310a";
static const char __pyx_k_A_E_at1_r_D_4_J_1_8_A_q_q_U_1_Qa_2[] = "\200A\360\016\000\t\r\210E\220\025\220a\220t\2301\330\014\017\210r\220\023\220D\230\001\330\020+\2504\320/J\310!\3101\330\020 \320 8\270\001\330\020\023\220>\240\023\240A\330\024\037\230q\340\024\037\230q\330\024\030\230\005\230U\240!\2401\330\030%\320%=\270\\\310\021\310!\330\030\033\230:\240Q\240a\330\034'\240q\330\034\035\330\020\023\2204\220q\330\024\033\320\0333\3203C\3001\300A\330\024)\250\024\320-B\300!\3001\330\024\027\220u\230C\230q\330\030*\250!\2509\260A\330\030\033\320\033-\250Q\250f\260C\260q\330\034 \320 2\260!\2601\330\014\031\230\024\230]\250!\2501\330\014\017\210{\230#\230R\230t\2404\240z\260\021\260!\330\020\021\330\014\023\2204\320\027)\250\021\250\"\250A\250Q\330\014\035\230T\320!3\2601\260A\330\014\017\210\177\230c\240\022\2404\240u\250C\250t\3203H\310\001\310\021\330\020\021\340\014\017\210s\220!\2206\230\023\230B\230d\240*\250A\250T\260\021\260$\260a\330\020\037\230~\250T\260\021\260!\330\020\024\220A\220U\230,\240g\250Q\330\014\020\220\006\220a\220r\230\030\240\021\240$\240n\260A\260U\270!";
static const char __pyx_k_GroupWindowAggFunctionBase_proce_2[] = "GroupWindowAggFunctionBase.process_watermark";
static const char __pyx_k_Incompatible_checksums_0x_x_vs_0_2[] = "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))";
static const char __pyx_k_SimpleNamespaceAggsHandleFunctio_2[] = "SimpleNamespaceAggsHandleFunction.merge";
static const char __pyx_k_SimpleNamespaceAggsHandleFunctio_3[] = "SimpleNamespaceAggsHandleFunction.set_accumulators";
static const char __pyx_k_SimpleNamespaceAggsHandleFunctio_4[] = "SimpleNamespaceAggsHandleFunction.create_accumulators";
static const char __pyx_k_SimpleNamespaceAggsHandleFunctio_5[] = "SimpleNamespaceAggsHandleFunction.cleanup";
static const char __pyx_k_SimpleNamespaceAggsHandleFunctio_6[] = "SimpleNamespaceAggsHandleFunction.get_value";
static const char __pyx_k_SimpleNamespaceAggsHandleFunctio_7[] = "SimpleNamespaceAggsHandleFunction.__reduce_cython__";
static const char __pyx_k_SimpleNamespaceAggsHandleFunctio_8[] = "SimpleNamespaceAggsHandleFunction.__setstate_cython__";
static const char __pyx_k_pyflink_fn_execution_table_windo_2[] = "pyflink.fn_execution.table.window_context";
static const char __pyx_k_pyflink_fn_execution_table_windo_3[] = "pyflink.fn_execution.table.window_process_function";
static const char __pyx_k_pyflink_fn_execution_table_windo_4[] = "pyflink.fn_execution.table.window_trigger";
static const char __pyx_k_pyflink_fn_execution_table_windo_5[] = "pyflink.fn_execution.table.window_aggregate_fast";
static const char __pyx_k_pyflink_fn_execution_table_windo_6[] = "pyflink/fn_execution/table/window_aggregate_fast.pyx";
/* #### Code section: decls ### */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_2set_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_4create_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_6cleanup(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_8__reduce_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_10__setstate_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_2__reduce_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_4__setstate_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction___init__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_udfs, PyObject *__pyx_v_input_extractors, PyObject *__pyx_v_index_of_count_star, PyObject *__pyx_v_count_star_inserted, PyObject *__pyx_v_named_property_extractor, PyObject *__pyx_v_udf_data_view_specs, PyObject *__pyx_v_filter_args, PyObject *__pyx_v_distinct_indexes, PyObject *__pyx_v_distinct_view_descriptors); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_2merge(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_4set_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_6create_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_8cleanup(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_10get_value(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace); /* proto */
static void __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_12__dealloc__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase___init__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_allowed_lateness, struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *__pyx_v_key_selector, PyObject *__pyx_v_state_backend, PyObject *__pyx_v_state_value_coder, PyObject *__pyx_v_window_assigner, struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_window_aggregator, PyObject *__pyx_v_trigger, PyObject *__pyx_v_rowtime_index, PyObject *__pyx_v_shift_timezone); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_2open(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_function_context); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_4process_element(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_row); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_6process_watermark(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int64_t __pyx_v_watermark); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_8on_event_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_timer); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_10on_processing_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_timer); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_12to_utc_timestamp_mills(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int64_t __pyx_v_epoch_mills); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_14get_timers(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_16close(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_18cleanup_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_window); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_20__reduce_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_22__setstate_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction___init__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v_self, PyObject *__pyx_v_allowed_lateness, struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *__pyx_v_key_selector, PyObject *__pyx_v_state_backend, PyObject *__pyx_v_state_value_coder, PyObject *__pyx_v_window_assigner, PyObject *__pyx_v_window_aggregator, PyObject *__pyx_v_trigger, PyObject *__pyx_v_rowtime_index, PyObject *__pyx_v_shift_timezone); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_2__reduce_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_4__setstate_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunctionBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_2__pyx_unpickle_NamespaceAggsHandleFunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_4__pyx_unpickle_GroupWindowAggFunctionBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_6__pyx_unpickle_GroupWindowAggFunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
/* #### Code section: late_includes ### */
/* #### Code section: module_state ### */
/* SmallCodeConfig */
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

typedef struct {
  PyObject *__pyx_d;
  PyObject *__pyx_b;
  PyObject *__pyx_cython_runtime;
  PyObject *__pyx_empty_tuple;
  PyObject *__pyx_empty_bytes;
  PyObject *__pyx_empty_unicode;
  #ifdef __Pyx_CyFunction_USED
  PyTypeObject *__pyx_CyFunctionType;
  #endif
  #ifdef __Pyx_FusedFunction_USED
  PyTypeObject *__pyx_FusedFunctionType;
  #endif
  #ifdef __Pyx_Generator_USED
  PyTypeObject *__pyx_GeneratorType;
  #endif
  #ifdef __Pyx_IterableCoroutine_USED
  PyTypeObject *__pyx_IterableCoroutineType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineAwaitType;
  #endif
  #ifdef __Pyx_Coroutine_USED
  PyTypeObject *__pyx_CoroutineType;
  #endif
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_11stream_fast_InputStream;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_11stream_fast_OutputStream;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction;
  PyObject *__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;
  PyObject *__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;
  PyObject *__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction;
  PyObject *__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;
  PyObject *__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;
  PyTypeObject *__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction;
  __Pyx_CachedCFunction __pyx_umethod_PyDict_Type_pop;
  __Pyx_CachedCFunction __pyx_umethod_PyList_Type____add;
  __Pyx_CachedCFunction __pyx_umethod_PyList_Type__remove;
  PyObject *__pyx_tuple[2];
  PyObject *__pyx_codeobj_tab[55];
  PyObject *__pyx_string_tab[258];
  PyObject *__pyx_float_1000_0;
  PyObject *__pyx_int_0;
  PyObject *__pyx_int_1;
  PyObject *__pyx_int_179515037;
  PyObject *__pyx_int_185799962;
  PyObject *__pyx_int_209029984;
  PyObject *__pyx_int_222419149;
  PyObject *__pyx_int_228825662;
  PyObject *__pyx_int_238750788;
/* #### Code section: module_state_contents ### */
/* CommonTypesMetaclass.module_state_decls */
PyTypeObject *__pyx_CommonTypesMetaclassType;

/* CachedMethodType.module_state_decls */
#if CYTHON_COMPILING_IN_LIMITED_API
PyObject *__Pyx_CachedMethodType;
#endif

/* CodeObjectCache.module_state_decls */
struct __Pyx_CodeObjectCache __pyx_code_cache;

/* #### Code section: module_state_end ### */
} __pyx_mstatetype;

#if CYTHON_USE_MODULE_STATE
#ifdef __cplusplus
namespace {
extern struct PyModuleDef __pyx_moduledef;
} /* anonymous namespace */
#else
static struct PyModuleDef __pyx_moduledef;
#endif

#define __pyx_mstate_global (__Pyx_PyModule_GetState(__Pyx_State_FindModule(&__pyx_moduledef)))

#define __pyx_m (__Pyx_State_FindModule(&__pyx_moduledef))
#else
static __pyx_mstatetype __pyx_mstate_global_static =
#ifdef __cplusplus
    {};
#else
    {0};
#endif
static __pyx_mstatetype * const __pyx_mstate_global = &__pyx_mstate_global_static;
#endif
/* #### Code section: constant_name_defines ### */
#define __pyx_kp_u_ __pyx_string_tab[0]
#define __pyx_n_u_DataViewSpec __pyx_string_tab[1]
#define __pyx_n_u_Dict __pyx_string_tab[2]
#define __pyx_n_u_GeneralWindowProcessFunction __pyx_string_tab[3]
#define __pyx_n_u_GroupWindowAggFunction __pyx_string_tab[4]
#define __pyx_n_u_GroupWindowAggFunctionBase __pyx_string_tab[5]
#define __pyx_n_u_GroupWindowAggFunctionBase___red __pyx_string_tab[6]
#define __pyx_n_u_GroupWindowAggFunctionBase___set __pyx_string_tab[7]
#define __pyx_n_u_GroupWindowAggFunctionBase_clean __pyx_string_tab[8]
#define __pyx_n_u_GroupWindowAggFunctionBase_close __pyx_string_tab[9]
#define __pyx_n_u_GroupWindowAggFunctionBase_get_t __pyx_string_tab[10]
#define __pyx_n_u_GroupWindowAggFunctionBase_on_ev __pyx_string_tab[11]
#define __pyx_n_u_GroupWindowAggFunctionBase_on_pr __pyx_string_tab[12]
#define __pyx_n_u_GroupWindowAggFunctionBase_open __pyx_string_tab[13]
#define __pyx_n_u_GroupWindowAggFunctionBase_proce __pyx_string_tab[14]
#define __pyx_n_u_GroupWindowAggFunctionBase_proce_2 __pyx_string_tab[15]
#define __pyx_n_u_GroupWindowAggFunctionBase_to_ut __pyx_string_tab[16]
#define __pyx_n_u_GroupWindowAggFunction___reduce __pyx_string_tab[17]
#define __pyx_n_u_GroupWindowAggFunction___setstat __pyx_string_tab[18]
#define __pyx_n_u_ImperativeAggregateFunction __pyx_string_tab[19]
#define __pyx_kp_u_Incompatible_checksums_0x_x_vs_0 __pyx_string_tab[20]
#define __pyx_kp_u_Incompatible_checksums_0x_x_vs_0_2 __pyx_string_tab[21]
#define __pyx_n_u_InternalWindowProcessFunction __pyx_string_tab[22]
#define __pyx_n_u_K __pyx_string_tab[23]
#define __pyx_n_u_LegacyInternalTimerServiceImpl __pyx_string_tab[24]
#define __pyx_n_u_List __pyx_string_tab[25]
#define __pyx_n_u_ListViewSpec __pyx_string_tab[26]
#define __pyx_n_u_MAX_LONG_VALUE __pyx_string_tab[27]
#define __pyx_n_u_MapViewSpec __pyx_string_tab[28]
#define __pyx_n_u_MergingWindowAssigner __pyx_string_tab[29]
#define __pyx_n_u_MergingWindowProcessFunction __pyx_string_tab[30]
#define __pyx_n_u_NamespaceAggsHandleFunction __pyx_string_tab[31]
#define __pyx_n_u_NamespaceAggsHandleFunctionBase __pyx_string_tab[32]
#define __pyx_n_u_NamespaceAggsHandleFunctionBase_2 __pyx_string_tab[33]
#define __pyx_n_u_NamespaceAggsHandleFunctionBase_3 __pyx_string_tab[34]
#define __pyx_n_u_NamespaceAggsHandleFunctionBase_4 __pyx_string_tab[35]
#define __pyx_n_u_NamespaceAggsHandleFunctionBase_5 __pyx_string_tab[36]
#define __pyx_n_u_NamespaceAggsHandleFunctionBase_6 __pyx_string_tab[37]
#define __pyx_n_u_NamespaceAggsHandleFunctionBase_7 __pyx_string_tab[38]
#define __pyx_n_u_NamespaceAggsHandleFunction___re __pyx_string_tab[39]
#define __pyx_n_u_NamespaceAggsHandleFunction___se __pyx_string_tab[40]
#define __pyx_n_u_NamespaceAggsHandleFunction_get __pyx_string_tab[41]
#define __pyx_kp_u_Note_that_Cython_is_deliberately __pyx_string_tab[42]
#define __pyx_n_u_PanedWindowAssigner __pyx_string_tab[43]
#define __pyx_n_u_PanedWindowProcessFunction __pyx_string_tab[44]
#define __pyx_n_u_PerWindowStateDataViewStore __pyx_string_tab[45]
#define __pyx_n_u_PickleCoder __pyx_string_tab[46]
#define __pyx_n_u_PickleError __pyx_string_tab[47]
#define __pyx_n_u_SimpleNamespaceAggsHandleFunctio __pyx_string_tab[48]
#define __pyx_n_u_SimpleNamespaceAggsHandleFunctio_2 __pyx_string_tab[49]
#define __pyx_n_u_SimpleNamespaceAggsHandleFunctio_3 __pyx_string_tab[50]
#define __pyx_n_u_SimpleNamespaceAggsHandleFunctio_4 __pyx_string_tab[51]
#define __pyx_n_u_SimpleNamespaceAggsHandleFunctio_5 __pyx_string_tab[52]
#define __pyx_n_u_SimpleNamespaceAggsHandleFunctio_6 __pyx_string_tab[53]
#define __pyx_n_u_SimpleNamespaceAggsHandleFunctio_7 __pyx_string_tab[54]
#define __pyx_n_u_SimpleNamespaceAggsHandleFunctio_8 __pyx_string_tab[55]
#define __pyx_kp_u_The_args_are_not_in_the_distinct __pyx_string_tab[56]
#define __pyx_n_u_Trigger __pyx_string_tab[57]
#define __pyx_n_u_TriggerContext __pyx_string_tab[58]
#define __pyx_n_u_TypeError __pyx_string_tab[59]
#define __pyx_n_u_UTC __pyx_string_tab[60]
#define __pyx_n_u_W __pyx_string_tab[61]
#define __pyx_n_u_WindowAssigner __pyx_string_tab[62]
#define __pyx_n_u_WindowContext __pyx_string_tab[63]
#define __pyx_kp_u__2 __pyx_string_tab[64]
#define __pyx_n_u_accumulate __pyx_string_tab[65]
#define __pyx_n_u_accumulators __pyx_string_tab[66]
#define __pyx_n_u_accumulators_2 __pyx_string_tab[67]
#define __pyx_n_u_add __pyx_string_tab[68]
#define __pyx_kp_u_add_note __pyx_string_tab[69]
#define __pyx_n_u_advance_watermark __pyx_string_tab[70]
#define __pyx_kp_u_agg_ddistinct __pyx_string_tab[71]
#define __pyx_n_u_allowed_lateness __pyx_string_tab[72]
#define __pyx_n_u_allowed_lateness_2 __pyx_string_tab[73]
#define __pyx_n_u_assign_actual_windows __pyx_string_tab[74]
#define __pyx_n_u_assign_state_namespace __pyx_string_tab[75]
#define __pyx_n_u_asyncio_coroutines __pyx_string_tab[76]
#define __pyx_n_u_class_getitem __pyx_string_tab[77]
#define __pyx_n_u_clean_window_if_needed __pyx_string_tab[78]
#define __pyx_n_u_cleanup __pyx_string_tab[79]
#define __pyx_n_u_cleanup_time __pyx_string_tab[80]
#define __pyx_n_u_clear __pyx_string_tab[81]
#define __pyx_n_u_cline_in_traceback __pyx_string_tab[82]
#define __pyx_n_u_close __pyx_string_tab[83]
#define __pyx_n_u_count_star_inserted __pyx_string_tab[84]
#define __pyx_n_u_create_accumulator __pyx_string_tab[85]
#define __pyx_n_u_create_accumulators __pyx_string_tab[86]
#define __pyx_n_u_current_processing_time __pyx_string_tab[87]
#define __pyx_n_u_datetime __pyx_string_tab[88]
#define __pyx_n_u_dealloc __pyx_string_tab[89]
#define __pyx_n_u_dict __pyx_string_tab[90]
#define __pyx_n_u_dict_2 __pyx_string_tab[91]
#define __pyx_kp_u_disable __pyx_string_tab[92]
#define __pyx_n_u_distinct_data_views __pyx_string_tab[93]
#define __pyx_n_u_distinct_indexes __pyx_string_tab[94]
#define __pyx_n_u_distinct_indexes_2 __pyx_string_tab[95]
#define __pyx_n_u_distinct_view_descriptors __pyx_string_tab[96]
#define __pyx_n_u_distinct_view_descriptors_2 __pyx_string_tab[97]
#define __pyx_n_u_element_coder __pyx_string_tab[98]
#define __pyx_n_u_emit_window_result __pyx_string_tab[99]
#define __pyx_kp_u_enable __pyx_string_tab[100]
#define __pyx_n_u_epoch_mills __pyx_string_tab[101]
#define __pyx_n_u_field_index __pyx_string_tab[102]
#define __pyx_n_u_filter_args __pyx_string_tab[103]
#define __pyx_n_u_filter_args_2 __pyx_string_tab[104]
#define __pyx_n_u_fromtimestamp __pyx_string_tab[105]
#define __pyx_n_u_func __pyx_string_tab[106]
#define __pyx_n_u_function_context __pyx_string_tab[107]
#define __pyx_kp_u_gc __pyx_string_tab[108]
#define __pyx_n_u_get_accumulators __pyx_string_tab[109]
#define __pyx_n_u_get_key __pyx_string_tab[110]
#define __pyx_n_u_get_namespace __pyx_string_tab[111]
#define __pyx_n_u_get_runtime_context __pyx_string_tab[112]
#define __pyx_n_u_get_state_list_view __pyx_string_tab[113]
#define __pyx_n_u_get_state_map_view __pyx_string_tab[114]
#define __pyx_n_u_get_timers __pyx_string_tab[115]
#define __pyx_n_u_get_timestamp __pyx_string_tab[116]
#define __pyx_n_u_get_value __pyx_string_tab[117]
#define __pyx_n_u_get_value_indexes __pyx_string_tab[118]
#define __pyx_n_u_get_value_indexes_length __pyx_string_tab[119]
#define __pyx_n_u_get_value_state __pyx_string_tab[120]
#define __pyx_n_u_getstate __pyx_string_tab[121]
#define __pyx_n_u_i __pyx_string_tab[122]
#define __pyx_n_u_index_of_count_star __pyx_string_tab[123]
#define __pyx_n_u_init __pyx_string_tab[124]
#define __pyx_n_u_initializing __pyx_string_tab[125]
#define __pyx_n_u_input_data __pyx_string_tab[126]
#define __pyx_n_u_input_extractors __pyx_string_tab[127]
#define __pyx_n_u_input_extractors_2 __pyx_string_tab[128]
#define __pyx_n_u_input_row __pyx_string_tab[129]
#define __pyx_n_u_internal_timer_service __pyx_string_tab[130]
#define __pyx_n_u_is_coroutine __pyx_string_tab[131]
#define __pyx_n_u_is_event_time __pyx_string_tab[132]
#define __pyx_kp_u_isenabled __pyx_string_tab[133]
#define __pyx_n_u_items __pyx_string_tab[134]
#define __pyx_n_u_join_row __pyx_string_tab[135]
#define __pyx_n_u_key __pyx_string_tab[136]
#define __pyx_n_u_key_coder __pyx_string_tab[137]
#define __pyx_n_u_key_selector __pyx_string_tab[138]
#define __pyx_n_u_key_selector_2 __pyx_string_tab[139]
#define __pyx_n_u_keys __pyx_string_tab[140]
#define __pyx_n_u_left __pyx_string_tab[141]
#define __pyx_n_u_main __pyx_string_tab[142]
#define __pyx_n_u_max_timestamp __pyx_string_tab[143]
#define __pyx_n_u_merge __pyx_string_tab[144]
#define __pyx_n_u_microsecond __pyx_string_tab[145]
#define __pyx_n_u_module __pyx_string_tab[146]
#define __pyx_n_u_name __pyx_string_tab[147]
#define __pyx_n_u_named_property_extractor __pyx_string_tab[148]
#define __pyx_n_u_named_property_extractor_2 __pyx_string_tab[149]
#define __pyx_n_u_namespace __pyx_string_tab[150]
#define __pyx_n_u_new __pyx_string_tab[151]
#define __pyx_n_u_on_element __pyx_string_tab[152]
#define __pyx_n_u_on_event_time __pyx_string_tab[153]
#define __pyx_n_u_on_processing_time __pyx_string_tab[154]
#define __pyx_n_u_open __pyx_string_tab[155]
#define __pyx_n_u_pickle __pyx_string_tab[156]
#define __pyx_n_u_pop __pyx_string_tab[157]
#define __pyx_n_u_prepare_aggregate_accumulator_fo __pyx_string_tab[158]
#define __pyx_n_u_process_element __pyx_string_tab[159]
#define __pyx_n_u_process_watermark __pyx_string_tab[160]
#define __pyx_n_u_pyflink_common_constants __pyx_string_tab[161]
#define __pyx_n_u_pyflink_fn_execution_coders __pyx_string_tab[162]
#define __pyx_n_u_pyflink_fn_execution_datastream __pyx_string_tab[163]
#define __pyx_n_u_pyflink_fn_execution_table_state __pyx_string_tab[164]
#define __pyx_n_u_pyflink_fn_execution_table_windo __pyx_string_tab[165]
#define __pyx_n_u_pyflink_fn_execution_table_windo_2 __pyx_string_tab[166]
#define __pyx_n_u_pyflink_fn_execution_table_windo_3 __pyx_string_tab[167]
#define __pyx_n_u_pyflink_fn_execution_table_windo_4 __pyx_string_tab[168]
#define __pyx_n_u_pyflink_fn_execution_table_windo_5 __pyx_string_tab[169]
#define __pyx_kp_u_pyflink_fn_execution_table_windo_6 __pyx_string_tab[170]
#define __pyx_n_u_pyflink_table_udf __pyx_string_tab[171]
#define __pyx_n_u_pytz __pyx_string_tab[172]
#define __pyx_n_u_pyx_PickleError __pyx_string_tab[173]
#define __pyx_n_u_pyx_checksum __pyx_string_tab[174]
#define __pyx_n_u_pyx_result __pyx_string_tab[175]
#define __pyx_n_u_pyx_state __pyx_string_tab[176]
#define __pyx_n_u_pyx_type __pyx_string_tab[177]
#define __pyx_n_u_pyx_unpickle_GroupWindowAggFun __pyx_string_tab[178]
#define __pyx_n_u_pyx_unpickle_GroupWindowAggFun_2 __pyx_string_tab[179]
#define __pyx_n_u_pyx_unpickle_GroupWindowAggFun_3 __pyx_string_tab[180]
#define __pyx_n_u_pyx_unpickle_GroupWindowAggFun_4 __pyx_string_tab[181]
#define __pyx_n_u_pyx_unpickle_NamespaceAggsHand __pyx_string_tab[182]
#define __pyx_n_u_pyx_unpickle_NamespaceAggsHand_2 __pyx_string_tab[183]
#define __pyx_n_u_pyx_unpickle_NamespaceAggsHand_3 __pyx_string_tab[184]
#define __pyx_n_u_pyx_unpickle_NamespaceAggsHand_4 __pyx_string_tab[185]
#define __pyx_n_u_pyx_vtable __pyx_string_tab[186]
#define __pyx_n_u_qualname __pyx_string_tab[187]
#define __pyx_n_u_range __pyx_string_tab[188]
#define __pyx_n_u_reduce __pyx_string_tab[189]
#define __pyx_n_u_reduce_cython __pyx_string_tab[190]
#define __pyx_n_u_reduce_ex __pyx_string_tab[191]
#define __pyx_n_u_register_cleanup_timer __pyx_string_tab[192]
#define __pyx_n_u_register_event_time_timer __pyx_string_tab[193]
#define __pyx_n_u_register_processing_time_timer __pyx_string_tab[194]
#define __pyx_n_u_remove __pyx_string_tab[195]
#define __pyx_n_u_replace __pyx_string_tab[196]
#define __pyx_n_u_retract __pyx_string_tab[197]
#define __pyx_n_u_right __pyx_string_tab[198]
#define __pyx_n_u_row_kind __pyx_string_tab[199]
#define __pyx_n_u_rowtime_index __pyx_string_tab[200]
#define __pyx_n_u_rowtime_index_2 __pyx_string_tab[201]
#define __pyx_n_u_self __pyx_string_tab[202]
#define __pyx_kp_u_self__distinct_indexes_self__fil __pyx_string_tab[203]
#define __pyx_n_u_set_accumulators __pyx_string_tab[204]
#define __pyx_n_u_set_current_key __pyx_string_tab[205]
#define __pyx_n_u_set_current_namespace __pyx_string_tab[206]
#define __pyx_n_u_set_name __pyx_string_tab[207]
#define __pyx_n_u_setstate __pyx_string_tab[208]
#define __pyx_n_u_setstate_cython __pyx_string_tab[209]
#define __pyx_n_u_shift_timezone __pyx_string_tab[210]
#define __pyx_n_u_shift_timezone_2 __pyx_string_tab[211]
#define __pyx_n_u_spec __pyx_string_tab[212]
#define __pyx_n_u_state __pyx_string_tab[213]
#define __pyx_n_u_state_backend __pyx_string_tab[214]
#define __pyx_n_u_state_backend_2 __pyx_string_tab[215]
#define __pyx_n_u_state_data_view_store __pyx_string_tab[216]
#define __pyx_n_u_state_id __pyx_string_tab[217]
#define __pyx_n_u_state_value_coder __pyx_string_tab[218]
#define __pyx_n_u_state_value_coder_2 __pyx_string_tab[219]
#define __pyx_kp_u_stringsource __pyx_string_tab[220]
#define __pyx_n_u_super __pyx_string_tab[221]
#define __pyx_n_u_temp __pyx_string_tab[222]
#define __pyx_n_u_test __pyx_string_tab[223]
#define __pyx_n_u_timer __pyx_string_tab[224]
#define __pyx_n_u_timers __pyx_string_tab[225]
#define __pyx_n_u_timestamp __pyx_string_tab[226]
#define __pyx_n_u_timezone __pyx_string_tab[227]
#define __pyx_n_u_to_utc_timestamp_mills __pyx_string_tab[228]
#define __pyx_n_u_total_seconds __pyx_string_tab[229]
#define __pyx_n_u_trigger __pyx_string_tab[230]
#define __pyx_n_u_trigger_2 __pyx_string_tab[231]
#define __pyx_n_u_trigger_context __pyx_string_tab[232]
#define __pyx_n_u_typing __pyx_string_tab[233]
#define __pyx_n_u_tzinfo __pyx_string_tab[234]
#define __pyx_n_u_udf_data_view_specs __pyx_string_tab[235]
#define __pyx_n_u_udf_data_view_specs_2 __pyx_string_tab[236]
#define __pyx_n_u_udf_data_views __pyx_string_tab[237]
#define __pyx_n_u_udf_num __pyx_string_tab[238]
#define __pyx_n_u_udfs __pyx_string_tab[239]
#define __pyx_n_u_udfs_2 __pyx_string_tab[240]
#define __pyx_n_u_update __pyx_string_tab[241]
#define __pyx_n_u_use_setstate __pyx_string_tab[242]
#define __pyx_n_u_utc __pyx_string_tab[243]
#define __pyx_n_u_utcfromtimestamp __pyx_string_tab[244]
#define __pyx_n_u_value __pyx_string_tab[245]
#define __pyx_n_u_value_coder __pyx_string_tab[246]
#define __pyx_n_u_values __pyx_string_tab[247]
#define __pyx_n_u_watermark __pyx_string_tab[248]
#define __pyx_n_u_window __pyx_string_tab[249]
#define __pyx_n_u_window_aggregator __pyx_string_tab[250]
#define __pyx_n_u_window_aggregator_2 __pyx_string_tab[251]
#define __pyx_n_u_window_assigner __pyx_string_tab[252]
#define __pyx_n_u_window_assigner_2 __pyx_string_tab[253]
#define __pyx_n_u_window_context __pyx_string_tab[254]
#define __pyx_n_u_window_function __pyx_string_tab[255]
#define __pyx_n_u_window_state __pyx_string_tab[256]
#define __pyx_n_u_window_state_2 __pyx_string_tab[257]
/* #### Code section: module_state_clear ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_clear(PyObject *m) {
  __pyx_mstatetype *clear_module_state = __Pyx_PyModule_GetState(m);
  if (!clear_module_state) return 0;
  Py_CLEAR(clear_module_state->__pyx_d);
  Py_CLEAR(clear_module_state->__pyx_b);
  Py_CLEAR(clear_module_state->__pyx_cython_runtime);
  Py_CLEAR(clear_module_state->__pyx_empty_tuple);
  Py_CLEAR(clear_module_state->__pyx_empty_bytes);
  Py_CLEAR(clear_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_CLEAR(clear_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_CLEAR(clear_module_state->__pyx_FusedFunctionType);
  #endif
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __Pyx_State_RemoveModule(NULL);
  #endif
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_InputStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_OutputStream);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase);
  Py_CLEAR(clear_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction);
  Py_CLEAR(clear_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction);
  Py_CLEAR(clear_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase);
  Py_CLEAR(clear_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase);
  Py_CLEAR(clear_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction);
  Py_CLEAR(clear_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction);
  for (int i=0; i<2; ++i) { Py_CLEAR(clear_module_state->__pyx_tuple[i]); }
  for (int i=0; i<55; ++i) { Py_CLEAR(clear_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<258; ++i) { Py_CLEAR(clear_module_state->__pyx_string_tab[i]); }
  Py_CLEAR(clear_module_state->__pyx_float_1000_0);
  Py_CLEAR(clear_module_state->__pyx_int_0);
  Py_CLEAR(clear_module_state->__pyx_int_1);
  Py_CLEAR(clear_module_state->__pyx_int_179515037);
  Py_CLEAR(clear_module_state->__pyx_int_185799962);
  Py_CLEAR(clear_module_state->__pyx_int_209029984);
  Py_CLEAR(clear_module_state->__pyx_int_222419149);
  Py_CLEAR(clear_module_state->__pyx_int_228825662);
  Py_CLEAR(clear_module_state->__pyx_int_238750788);
  return 0;
}
#endif
/* #### Code section: module_state_traverse ### */
#if CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __pyx_m_traverse(PyObject *m, visitproc visit, void *arg) {
  __pyx_mstatetype *traverse_module_state = __Pyx_PyModule_GetState(m);
  if (!traverse_module_state) return 0;
  Py_VISIT(traverse_module_state->__pyx_d);
  Py_VISIT(traverse_module_state->__pyx_b);
  Py_VISIT(traverse_module_state->__pyx_cython_runtime);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_tuple);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_bytes);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_empty_unicode);
  #ifdef __Pyx_CyFunction_USED
  Py_VISIT(traverse_module_state->__pyx_CyFunctionType);
  #endif
  #ifdef __Pyx_FusedFunction_USED
  Py_VISIT(traverse_module_state->__pyx_FusedFunctionType);
  #endif
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_InputStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_OutputStream);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase);
  Py_VISIT(traverse_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction);
  Py_VISIT(traverse_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction);
  Py_VISIT(traverse_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase);
  Py_VISIT(traverse_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase);
  Py_VISIT(traverse_module_state->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction);
  Py_VISIT(traverse_module_state->__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction);
  for (int i=0; i<2; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_tuple[i]); }
  for (int i=0; i<55; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_codeobj_tab[i]); }
  for (int i=0; i<258; ++i) { __Pyx_VISIT_CONST(traverse_module_state->__pyx_string_tab[i]); }
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_float_1000_0);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_0);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_1);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_179515037);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_185799962);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_209029984);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_222419149);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_228825662);
  __Pyx_VISIT_CONST(traverse_module_state->__pyx_int_238750788);
  return 0;
}
#endif
/* #### Code section: module_code ### */

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":45
 * from pyflink.table.udf import ImperativeAggregateFunction
 * 
 * cdef InternalRow join_row(list left, list right, InternalRowKind row_kind):             # <<<<<<<<<<<<<<
 *     return InternalRow(left.__add__(right), row_kind)
 * 
*/

static struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_join_row(PyObject *__pyx_v_left, PyObject *__pyx_v_right, enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind __pyx_v_row_kind) {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[0]))
  __Pyx_RefNannySetupContext("join_row", 0);
  __Pyx_TraceStartFunc("join_row", __pyx_f[0], 45, 0, 0, 0, __PYX_ERR(0, 45, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":46
 * 
 * cdef InternalRow join_row(list left, list right, InternalRowKind row_kind):
 *     return InternalRow(left.__add__(right), row_kind)             # <<<<<<<<<<<<<<
 * 
 * cdef class NamespaceAggsHandleFunctionBase:
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __pyx_t_2 = NULL;
  __Pyx_INCREF((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow);
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow); 
  __pyx_t_4 = __Pyx_CallUnboundCMethod1(&__pyx_mstate_global->__pyx_umethod_PyList_Type____add, __pyx_v_left, __pyx_v_right); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyLong_From_enum____pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind(__pyx_v_row_kind); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_t_4, __pyx_t_5};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 46, __pyx_L1_error)
    __Pyx_GOTREF((PyObject *)__pyx_t_1);
  }
  __pyx_r = ((struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_TraceReturnValue((PyObject *)__pyx_r, 1, 0, __PYX_ERR(0, 46, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":45
 * from pyflink.table.udf import ImperativeAggregateFunction
 * 
 * cdef InternalRow join_row(list left, list right, InternalRowKind row_kind):             # <<<<<<<<<<<<<<
 *     return InternalRow(left.__add__(right), row_kind)
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 45, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.join_row", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":49
 * 
 * cdef class NamespaceAggsHandleFunctionBase:
 *     cdef void open(self, object state_data_view_store):             # <<<<<<<<<<<<<<
 *         """
 *         Initialization method for the function. It is called before the actual working methods.
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_open(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_state_data_view_store) {
  __Pyx_TraceDeclarationsFunc
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[1]))
  __Pyx_TraceStartFunc("open", __pyx_f[0], 49, 0, 0, 0, __PYX_ERR(0, 49, __pyx_L1_error));

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 49, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 49, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":57
 *         pass
 * 
 *     cdef void accumulate(self, InternalRow input_data):             # <<<<<<<<<<<<<<
 *         """
 *         Accumulates the input values to the accumulators.
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_accumulate(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_data) {
  __Pyx_TraceDeclarationsFunc
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[2]))
  __Pyx_TraceStartFunc("accumulate", __pyx_f[0], 57, 0, 0, 0, __PYX_ERR(0, 57, __pyx_L1_error));

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 57, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 57, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.accumulate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":65
 *         pass
 * 
 *     cdef void retract(self, InternalRow input_data):             # <<<<<<<<<<<<<<
 *         """
 *         Retracts the input values from the accumulators.
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_retract(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_data) {
  __Pyx_TraceDeclarationsFunc
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[3]))
  __Pyx_TraceStartFunc("retract", __pyx_f[0], 65, 0, 0, 0, __PYX_ERR(0, 65, __pyx_L1_error));

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 65, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 65, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.retract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":72
 *         """
 * 
 *     cpdef void merge(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         """
 *         Merges the other accumulators into current accumulators.
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_1merge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_namespace, CYTHON_UNUSED PyObject *__pyx_v_accumulators, int __pyx_skip_dispatch) {
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4]))
  __Pyx_RefNannySetupContext("merge", 0);
  __Pyx_TraceStartFunc("merge", __pyx_f[0], 72, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 72, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_merge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_1merge)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_namespace, __pyx_v_accumulators};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":73
 * 
 *     cpdef void merge(self, object namespace, list accumulators):
 *         """             # <<<<<<<<<<<<<<
 *         Merges the other accumulators into current accumulators.
 *         """
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 72, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 72, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.merge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":72
 *         """
 * 
 *     cpdef void merge(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         """
 *         Merges the other accumulators into current accumulators.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_1merge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge, "\n        Merges the other accumulators into current accumulators.\n        ");
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_1merge = {"merge", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_1merge, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_1merge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  PyObject *__pyx_v_accumulators = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("merge (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_namespace,&__pyx_mstate_global->__pyx_n_u_accumulators,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 72, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 72, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 72, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "merge", 0) < 0) __PYX_ERR(0, 72, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("merge", 1, 2, 2, i); __PYX_ERR(0, 72, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 72, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 72, __pyx_L3_error)
    }
    __pyx_v_namespace = values[0];
    __pyx_v_accumulators = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("merge", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 72, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.merge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_accumulators), (&PyList_Type), 1, "accumulators", 1))) __PYX_ERR(0, 72, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self), __pyx_v_namespace, __pyx_v_accumulators);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4]))
  __Pyx_RefNannySetupContext("merge", 0);
  __Pyx_TraceStartFunc("merge (wrapper)", __pyx_f[0], 72, 0, 0, 0, __PYX_ERR(0, 72, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge(__pyx_v_self, __pyx_v_namespace, __pyx_v_accumulators, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 72, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 72, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.merge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":78
 *         pass
 * 
 *     cpdef void set_accumulators(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         """
 *         Set the current accumulators (saved in a row) which contains the current aggregated results.
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_3set_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_set_accumulators(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_namespace, CYTHON_UNUSED PyObject *__pyx_v_accumulators, int __pyx_skip_dispatch) {
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5]))
  __Pyx_RefNannySetupContext("set_accumulators", 0);
  __Pyx_TraceStartFunc("set_accumulators", __pyx_f[0], 78, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 78, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_set_accumulators); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_3set_accumulators)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_namespace, __pyx_v_accumulators};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":79
 * 
 *     cpdef void set_accumulators(self, object namespace, list accumulators):
 *         """             # <<<<<<<<<<<<<<
 *         Set the current accumulators (saved in a row) which contains the current aggregated results.
 *         """
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 78, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 78, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.set_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":78
 *         pass
 * 
 *     cpdef void set_accumulators(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         """
 *         Set the current accumulators (saved in a row) which contains the current aggregated results.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_3set_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_2set_accumulators, "\n        Set the current accumulators (saved in a row) which contains the current aggregated results.\n        ");
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_3set_accumulators = {"set_accumulators", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_3set_accumulators, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_2set_accumulators};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_3set_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  PyObject *__pyx_v_accumulators = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_accumulators (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_namespace,&__pyx_mstate_global->__pyx_n_u_accumulators,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 78, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 78, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 78, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_accumulators", 0) < 0) __PYX_ERR(0, 78, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_accumulators", 1, 2, 2, i); __PYX_ERR(0, 78, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 78, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 78, __pyx_L3_error)
    }
    __pyx_v_namespace = values[0];
    __pyx_v_accumulators = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_accumulators", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 78, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.set_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_accumulators), (&PyList_Type), 1, "accumulators", 1))) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_2set_accumulators(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self), __pyx_v_namespace, __pyx_v_accumulators);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_2set_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5]))
  __Pyx_RefNannySetupContext("set_accumulators", 0);
  __Pyx_TraceStartFunc("set_accumulators (wrapper)", __pyx_f[0], 78, 0, 0, 0, __PYX_ERR(0, 78, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_set_accumulators(__pyx_v_self, __pyx_v_namespace, __pyx_v_accumulators, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 78, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 78, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.set_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":84
 *         pass
 * 
 *     cdef list get_accumulators(self):             # <<<<<<<<<<<<<<
 *         """
 *         Gets the current accumulators (saved in a list) which contains the current
*/

static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_get_accumulators(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[6]))
  __Pyx_RefNannySetupContext("get_accumulators", 0);
  __Pyx_TraceStartFunc("get_accumulators", __pyx_f[0], 84, 0, 0, 0, __PYX_ERR(0, 84, __pyx_L1_error));

  /* function exit code */
  __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 84, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 84, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.get_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":93
 *         pass
 * 
 *     cpdef list create_accumulators(self):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes the accumulators and save them to an accumulators List.
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_5create_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_create_accumulators(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7]))
  __Pyx_RefNannySetupContext("create_accumulators", 0);
  __Pyx_TraceStartFunc("create_accumulators", __pyx_f[0], 93, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 93, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_create_accumulators); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_5create_accumulators)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 93, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 93, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":94
 * 
 *     cpdef list create_accumulators(self):
 *         """             # <<<<<<<<<<<<<<
 *         Initializes the accumulators and save them to an accumulators List.
 * 
*/

  /* function exit code */
  __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 93, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 93, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.create_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":93
 *         pass
 * 
 *     cpdef list create_accumulators(self):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes the accumulators and save them to an accumulators List.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_5create_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_4create_accumulators, "\n        Initializes the accumulators and save them to an accumulators List.\n\n        :return: A List of accumulators which contains the aggregated results.\n        ");
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_5create_accumulators = {"create_accumulators", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_5create_accumulators, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_4create_accumulators};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_5create_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_accumulators (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("create_accumulators", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("create_accumulators", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_4create_accumulators(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_4create_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7]))
  __Pyx_RefNannySetupContext("create_accumulators", 0);
  __Pyx_TraceStartFunc("create_accumulators (wrapper)", __pyx_f[0], 93, 0, 0, 0, __PYX_ERR(0, 93, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_create_accumulators(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 93, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.create_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":101
 *         pass
 * 
 *     cpdef void cleanup(self, object namespace):             # <<<<<<<<<<<<<<
 *         """
 *         Cleanup for the retired accumulators state.
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_7cleanup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_cleanup(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_namespace, int __pyx_skip_dispatch) {
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8]))
  __Pyx_RefNannySetupContext("cleanup", 0);
  __Pyx_TraceStartFunc("cleanup", __pyx_f[0], 101, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 101, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_cleanup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_7cleanup)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_namespace};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":102
 * 
 *     cpdef void cleanup(self, object namespace):
 *         """             # <<<<<<<<<<<<<<
 *         Cleanup for the retired accumulators state.
 *         """
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 101, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 101, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.cleanup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":101
 *         pass
 * 
 *     cpdef void cleanup(self, object namespace):             # <<<<<<<<<<<<<<
 *         """
 *         Cleanup for the retired accumulators state.
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_7cleanup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_6cleanup, "\n        Cleanup for the retired accumulators state.\n        ");
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_7cleanup = {"cleanup", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_7cleanup, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_6cleanup};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_7cleanup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cleanup (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_namespace,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 101, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 101, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cleanup", 0) < 0) __PYX_ERR(0, 101, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cleanup", 1, 1, 1, i); __PYX_ERR(0, 101, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 101, __pyx_L3_error)
    }
    __pyx_v_namespace = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cleanup", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 101, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.cleanup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_6cleanup(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self), __pyx_v_namespace);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_6cleanup(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, PyObject *__pyx_v_namespace) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8]))
  __Pyx_RefNannySetupContext("cleanup", 0);
  __Pyx_TraceStartFunc("cleanup (wrapper)", __pyx_f[0], 101, 0, 0, 0, __PYX_ERR(0, 101, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_cleanup(__pyx_v_self, __pyx_v_namespace, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 101, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 101, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.cleanup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":107
 *         pass
 * 
 *     cdef void close(self):             # <<<<<<<<<<<<<<
 *         """
 *         Tear-down method for this function. It can be used for clean up work.
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_close(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self) {
  __Pyx_TraceDeclarationsFunc
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[9]))
  __Pyx_TraceStartFunc("close", __pyx_f[0], 107, 0, 0, 0, __PYX_ERR(0, 107, __pyx_L1_error));

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 107, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 107, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_9__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_9__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_8__reduce_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_8__reduce_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10]))
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceStartFunc("__reduce_cython__", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
  __pyx_v_state = __pyx_mstate_global->__pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, None), state
*/
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NamespaceAggsHand); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_TraceReturnValue(__pyx_r, 21, 0, __PYX_ERR(1, 13, __pyx_L1_error));
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NamespaceAggsHand); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_TraceReturnValue(__pyx_r, 29, 0, __PYX_ERR(1, 15, __pyx_L1_error));
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_11__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_11__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_10__setstate_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_10__setstate_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11]))
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceStartFunc("__setstate_cython__", __pyx_f[1], 16, 0, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));

  /* "(tree fragment)":17
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":115
 * 
 * cdef class NamespaceAggsHandleFunction(NamespaceAggsHandleFunctionBase):
 *     cpdef list get_value(self, object namespace):             # <<<<<<<<<<<<<<
 *         """
 *         Gets the result of the aggregation from the current accumulators and namespace properties
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_1get_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_namespace, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12]))
  __Pyx_RefNannySetupContext("get_value", 0);
  __Pyx_TraceStartFunc("get_value", __pyx_f[0], 115, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 115, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_get_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_1get_value)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_namespace};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 115, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 115, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":116
 * cdef class NamespaceAggsHandleFunction(NamespaceAggsHandleFunctionBase):
 *     cpdef list get_value(self, object namespace):
 *         """             # <<<<<<<<<<<<<<
 *         Gets the result of the aggregation from the current accumulators and namespace properties
 *         (like window start).
*/

  /* function exit code */
  __pyx_r = ((PyObject*)Py_None); __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 115, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 115, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunction.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":115
 * 
 * cdef class NamespaceAggsHandleFunction(NamespaceAggsHandleFunctionBase):
 *     cpdef list get_value(self, object namespace):             # <<<<<<<<<<<<<<
 *         """
 *         Gets the result of the aggregation from the current accumulators and namespace properties
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_1get_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
PyDoc_STRVAR(__pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value, "\n        Gets the result of the aggregation from the current accumulators and namespace properties\n        (like window start).\n        :param namespace: the namespace properties which should be calculated, such window start\n        :return: the final result (saved in a List) of the current accumulators.\n        ");
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_1get_value = {"get_value", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_1get_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, __pyx_doc_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_1get_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_namespace,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 115, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 115, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_value", 0) < 0) __PYX_ERR(0, 115, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_value", 1, 1, 1, i); __PYX_ERR(0, 115, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 115, __pyx_L3_error)
    }
    __pyx_v_namespace = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 115, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunction.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *)__pyx_v_self), __pyx_v_namespace);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12]))
  __Pyx_RefNannySetupContext("get_value", 0);
  __Pyx_TraceStartFunc("get_value (wrapper)", __pyx_f[0], 115, 0, 0, 0, __PYX_ERR(0, 115, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value(__pyx_v_self, __pyx_v_namespace, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 115, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunction.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_2__reduce_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_2__reduce_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13]))
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceStartFunc("__reduce_cython__", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = ()             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
*/
  __Pyx_INCREF(__pyx_mstate_global->__pyx_empty_tuple);
  __pyx_v_state = __pyx_mstate_global->__pyx_empty_tuple;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
*/
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = False
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = ()
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = False             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, None), state
*/
  /*else*/ {
    __pyx_v_use_setstate = 0;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = False
 *     if use_setstate:
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NamespaceAggsHand_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_3 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    __Pyx_TraceReturnValue(__pyx_r, 21, 0, __PYX_ERR(1, 13, __pyx_L1_error));
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = False
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, None), state
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NamespaceAggsHandleFunction__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NamespaceAggsHand_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_238750788);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_238750788);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_mstate_global->__pyx_int_238750788) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_4) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_TraceReturnValue(__pyx_r, 29, 0, __PYX_ERR(1, 15, __pyx_L1_error));
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunction.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NamespaceAggsHandleFunction__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_4__setstate_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_4__setstate_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14]))
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceStartFunc("__setstate_cython__", __pyx_f[1], 16, 0, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));

  /* "(tree fragment)":17
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_NamespaceAggsHandleFunction__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunction__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NamespaceAggsHandleFunction__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":125
 * 
 * cdef class SimpleNamespaceAggsHandleFunction(NamespaceAggsHandleFunction):
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  udfs: List[ImperativeAggregateFunction],
 *                  input_extractors: List,
*/

/* Python wrapper */
static int __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_udfs = 0;
  PyObject *__pyx_v_input_extractors = 0;
  PyObject *__pyx_v_index_of_count_star = 0;
  PyObject *__pyx_v_count_star_inserted = 0;
  PyObject *__pyx_v_named_property_extractor = 0;
  PyObject *__pyx_v_udf_data_view_specs = 0;
  PyObject *__pyx_v_filter_args = 0;
  PyObject *__pyx_v_distinct_indexes = 0;
  PyObject *__pyx_v_distinct_view_descriptors = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_udfs,&__pyx_mstate_global->__pyx_n_u_input_extractors,&__pyx_mstate_global->__pyx_n_u_index_of_count_star,&__pyx_mstate_global->__pyx_n_u_count_star_inserted,&__pyx_mstate_global->__pyx_n_u_named_property_extractor,&__pyx_mstate_global->__pyx_n_u_udf_data_view_specs,&__pyx_mstate_global->__pyx_n_u_filter_args,&__pyx_mstate_global->__pyx_n_u_distinct_indexes,&__pyx_mstate_global->__pyx_n_u_distinct_view_descriptors,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 125, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 125, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < 0) __PYX_ERR(0, 125, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 9; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 9, 9, i); __PYX_ERR(0, 125, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 9)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 125, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 125, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 125, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 125, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 125, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 125, __pyx_L3_error)
      values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 125, __pyx_L3_error)
      values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 125, __pyx_L3_error)
      values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 125, __pyx_L3_error)
    }
    __pyx_v_udfs = ((PyObject*)values[0]);
    __pyx_v_input_extractors = ((PyObject*)values[1]);
    __pyx_v_index_of_count_star = ((PyObject*)values[2]);
    __pyx_v_count_star_inserted = values[3];
    __pyx_v_named_property_extractor = values[4];
    __pyx_v_udf_data_view_specs = ((PyObject*)values[5]);
    __pyx_v_filter_args = ((PyObject*)values[6]);
    __pyx_v_distinct_indexes = ((PyObject*)values[7]);
    __pyx_v_distinct_view_descriptors = ((PyObject*)values[8]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 9, 9, __pyx_nargs); __PYX_ERR(0, 125, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_udfs), (&PyList_Type), 0, "udfs", 2))) __PYX_ERR(0, 126, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_input_extractors), (&PyList_Type), 0, "input_extractors", 2))) __PYX_ERR(0, 127, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_index_of_count_star), (&PyLong_Type), 0, "index_of_count_star", 2))) __PYX_ERR(0, 128, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_udf_data_view_specs), (&PyList_Type), 0, "udf_data_view_specs", 2))) __PYX_ERR(0, 131, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_filter_args), (&PyList_Type), 0, "filter_args", 2))) __PYX_ERR(0, 132, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_distinct_indexes), (&PyList_Type), 0, "distinct_indexes", 2))) __PYX_ERR(0, 133, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_distinct_view_descriptors), (&PyDict_Type), 0, "distinct_view_descriptors", 2))) __PYX_ERR(0, 134, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction___init__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self), __pyx_v_udfs, __pyx_v_input_extractors, __pyx_v_index_of_count_star, __pyx_v_count_star_inserted, __pyx_v_named_property_extractor, __pyx_v_udf_data_view_specs, __pyx_v_filter_args, __pyx_v_distinct_indexes, __pyx_v_distinct_view_descriptors);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction___init__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_udfs, PyObject *__pyx_v_input_extractors, PyObject *__pyx_v_index_of_count_star, PyObject *__pyx_v_count_star_inserted, PyObject *__pyx_v_named_property_extractor, PyObject *__pyx_v_udf_data_view_specs, PyObject *__pyx_v_filter_args, PyObject *__pyx_v_distinct_indexes, PyObject *__pyx_v_distinct_view_descriptors) {
  size_t __pyx_v_i;
  PyObject *__pyx_v_temp = NULL;
  Py_ssize_t __pyx_7genexpr__pyx_v_i;
  int __pyx_r;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  size_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  int __pyx_t_11;
  size_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[15]))
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_TraceStartFunc("__init__", __pyx_f[0], 125, 0, 0, 0, __PYX_ERR(0, 125, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":135
 *                  distinct_indexes: List[int],
 *                  distinct_view_descriptors: Dict[int, DistinctViewDescriptor]):
 *         self._udfs = udfs             # <<<<<<<<<<<<<<
 *         self._udf_num = len(self._udfs)
 *         self._input_extractors = input_extractors
*/
  __Pyx_INCREF(__pyx_v_udfs);
  __Pyx_GIVEREF(__pyx_v_udfs);
  __Pyx_GOTREF(__pyx_v_self->_udfs);
  __Pyx_DECREF(__pyx_v_self->_udfs);
  __pyx_v_self->_udfs = __pyx_v_udfs;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":136
 *                  distinct_view_descriptors: Dict[int, DistinctViewDescriptor]):
 *         self._udfs = udfs
 *         self._udf_num = len(self._udfs)             # <<<<<<<<<<<<<<
 *         self._input_extractors = input_extractors
 *         self._named_property_extractor = named_property_extractor
*/
  __pyx_t_1 = __pyx_v_self->_udfs;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 136, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_self->_udf_num = __pyx_t_2;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":137
 *         self._udfs = udfs
 *         self._udf_num = len(self._udfs)
 *         self._input_extractors = input_extractors             # <<<<<<<<<<<<<<
 *         self._named_property_extractor = named_property_extractor
 *         self._accumulators = None
*/
  __Pyx_INCREF(__pyx_v_input_extractors);
  __Pyx_GIVEREF(__pyx_v_input_extractors);
  __Pyx_GOTREF(__pyx_v_self->_input_extractors);
  __Pyx_DECREF(__pyx_v_self->_input_extractors);
  __pyx_v_self->_input_extractors = __pyx_v_input_extractors;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":138
 *         self._udf_num = len(self._udfs)
 *         self._input_extractors = input_extractors
 *         self._named_property_extractor = named_property_extractor             # <<<<<<<<<<<<<<
 *         self._accumulators = None
 *         self._udf_data_view_specs = udf_data_view_specs
*/
  __Pyx_INCREF(__pyx_v_named_property_extractor);
  __Pyx_GIVEREF(__pyx_v_named_property_extractor);
  __Pyx_GOTREF(__pyx_v_self->_named_property_extractor);
  __Pyx_DECREF(__pyx_v_self->_named_property_extractor);
  __pyx_v_self->_named_property_extractor = __pyx_v_named_property_extractor;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":139
 *         self._input_extractors = input_extractors
 *         self._named_property_extractor = named_property_extractor
 *         self._accumulators = None             # <<<<<<<<<<<<<<
 *         self._udf_data_view_specs = udf_data_view_specs
 *         self._udf_data_views = []
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_accumulators);
  __Pyx_DECREF(__pyx_v_self->_accumulators);
  __pyx_v_self->_accumulators = ((PyObject*)Py_None);

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":140
 *         self._named_property_extractor = named_property_extractor
 *         self._accumulators = None
 *         self._udf_data_view_specs = udf_data_view_specs             # <<<<<<<<<<<<<<
 *         self._udf_data_views = []
 *         self._filter_args = <int*> malloc(self._udf_num * sizeof(int))
*/
  __Pyx_INCREF(__pyx_v_udf_data_view_specs);
  __Pyx_GIVEREF(__pyx_v_udf_data_view_specs);
  __Pyx_GOTREF(__pyx_v_self->_udf_data_view_specs);
  __Pyx_DECREF(__pyx_v_self->_udf_data_view_specs);
  __pyx_v_self->_udf_data_view_specs = __pyx_v_udf_data_view_specs;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":141
 *         self._accumulators = None
 *         self._udf_data_view_specs = udf_data_view_specs
 *         self._udf_data_views = []             # <<<<<<<<<<<<<<
 *         self._filter_args = <int*> malloc(self._udf_num * sizeof(int))
 *         self._distinct_indexes = <int*> malloc(self._udf_num * sizeof(int))
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_udf_data_views);
  __Pyx_DECREF(__pyx_v_self->_udf_data_views);
  __pyx_v_self->_udf_data_views = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":142
 *         self._udf_data_view_specs = udf_data_view_specs
 *         self._udf_data_views = []
 *         self._filter_args = <int*> malloc(self._udf_num * sizeof(int))             # <<<<<<<<<<<<<<
 *         self._distinct_indexes = <int*> malloc(self._udf_num * sizeof(int))
 *         for i in range(self._udf_num):
*/
  __pyx_v_self->_filter_args = ((int *)malloc((__pyx_v_self->_udf_num * (sizeof(int)))));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":143
 *         self._udf_data_views = []
 *         self._filter_args = <int*> malloc(self._udf_num * sizeof(int))
 *         self._distinct_indexes = <int*> malloc(self._udf_num * sizeof(int))             # <<<<<<<<<<<<<<
 *         for i in range(self._udf_num):
 *             self._filter_args[i] = filter_args[i]
*/
  __pyx_v_self->_distinct_indexes = ((int *)malloc((__pyx_v_self->_udf_num * (sizeof(int)))));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":144
 *         self._filter_args = <int*> malloc(self._udf_num * sizeof(int))
 *         self._distinct_indexes = <int*> malloc(self._udf_num * sizeof(int))
 *         for i in range(self._udf_num):             # <<<<<<<<<<<<<<
 *             self._filter_args[i] = filter_args[i]
 *             self._distinct_indexes[i] = distinct_indexes[i]
*/
  __pyx_t_3 = __pyx_v_self->_udf_num;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":145
 *         self._distinct_indexes = <int*> malloc(self._udf_num * sizeof(int))
 *         for i in range(self._udf_num):
 *             self._filter_args[i] = filter_args[i]             # <<<<<<<<<<<<<<
 *             self._distinct_indexes[i] = distinct_indexes[i]
 *         self._distinct_view_descriptors = distinct_view_descriptors
*/
    __pyx_t_6 = __Pyx_PyLong_As_int(__Pyx_PyList_GET_ITEM(__pyx_v_filter_args, __pyx_v_i)); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 145, __pyx_L1_error)
    (__pyx_v_self->_filter_args[__pyx_v_i]) = __pyx_t_6;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":146
 *         for i in range(self._udf_num):
 *             self._filter_args[i] = filter_args[i]
 *             self._distinct_indexes[i] = distinct_indexes[i]             # <<<<<<<<<<<<<<
 *         self._distinct_view_descriptors = distinct_view_descriptors
 *         self._distinct_data_views = {}
*/
    __pyx_t_6 = __Pyx_PyLong_As_int(__Pyx_PyList_GET_ITEM(__pyx_v_distinct_indexes, __pyx_v_i)); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 146, __pyx_L1_error)
    (__pyx_v_self->_distinct_indexes[__pyx_v_i]) = __pyx_t_6;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":147
 *             self._filter_args[i] = filter_args[i]
 *             self._distinct_indexes[i] = distinct_indexes[i]
 *         self._distinct_view_descriptors = distinct_view_descriptors             # <<<<<<<<<<<<<<
 *         self._distinct_data_views = {}
 *         temp = [i for i in range(len(udfs))]
*/
  __Pyx_INCREF(__pyx_v_distinct_view_descriptors);
  __Pyx_GIVEREF(__pyx_v_distinct_view_descriptors);
  __Pyx_GOTREF(__pyx_v_self->_distinct_view_descriptors);
  __Pyx_DECREF(__pyx_v_self->_distinct_view_descriptors);
  __pyx_v_self->_distinct_view_descriptors = __pyx_v_distinct_view_descriptors;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":148
 *             self._distinct_indexes[i] = distinct_indexes[i]
 *         self._distinct_view_descriptors = distinct_view_descriptors
 *         self._distinct_data_views = {}             # <<<<<<<<<<<<<<
 *         temp = [i for i in range(len(udfs))]
 *         if index_of_count_star >= 0 and count_star_inserted:
*/
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_distinct_data_views);
  __Pyx_DECREF(__pyx_v_self->_distinct_data_views);
  __pyx_v_self->_distinct_data_views = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":149
 *         self._distinct_view_descriptors = distinct_view_descriptors
 *         self._distinct_data_views = {}
 *         temp = [i for i in range(len(udfs))]             # <<<<<<<<<<<<<<
 *         if index_of_count_star >= 0 and count_star_inserted:
 *             # The record count is used internally, should be ignored by the get_value method.
*/
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 149, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_udfs); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 149, __pyx_L1_error)
    __pyx_t_7 = __pyx_t_2;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_7genexpr__pyx_v_i = __pyx_t_8;
      __pyx_t_9 = PyLong_FromSsize_t(__pyx_7genexpr__pyx_v_i); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 149, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_9))) __PYX_ERR(0, 149, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
  } /* exit inner scope */
  __pyx_v_temp = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":150
 *         self._distinct_data_views = {}
 *         temp = [i for i in range(len(udfs))]
 *         if index_of_count_star >= 0 and count_star_inserted:             # <<<<<<<<<<<<<<
 *             # The record count is used internally, should be ignored by the get_value method.
 *             temp.remove(index_of_count_star)
*/
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_index_of_count_star, __pyx_mstate_global->__pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_11) {
  } else {
    __pyx_t_10 = __pyx_t_11;
    goto __pyx_L8_bool_binop_done;
  }
  __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_v_count_star_inserted); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 150, __pyx_L1_error)
  __pyx_t_10 = __pyx_t_11;
  __pyx_L8_bool_binop_done:;
  if (__pyx_t_10) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":152
 *         if index_of_count_star >= 0 and count_star_inserted:
 *             # The record count is used internally, should be ignored by the get_value method.
 *             temp.remove(index_of_count_star)             # <<<<<<<<<<<<<<
 *         self._get_value_indexes_length = len(temp)
 *         self._get_value_indexes = <size_t*> malloc(sizeof(size_t) * self._get_value_indexes_length)
*/
    __pyx_t_1 = __Pyx_CallUnboundCMethod1(&__pyx_mstate_global->__pyx_umethod_PyList_Type__remove, __pyx_v_temp, __pyx_v_index_of_count_star); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":150
 *         self._distinct_data_views = {}
 *         temp = [i for i in range(len(udfs))]
 *         if index_of_count_star >= 0 and count_star_inserted:             # <<<<<<<<<<<<<<
 *             # The record count is used internally, should be ignored by the get_value method.
 *             temp.remove(index_of_count_star)
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":153
 *             # The record count is used internally, should be ignored by the get_value method.
 *             temp.remove(index_of_count_star)
 *         self._get_value_indexes_length = len(temp)             # <<<<<<<<<<<<<<
 *         self._get_value_indexes = <size_t*> malloc(sizeof(size_t) * self._get_value_indexes_length)
 *         for i in range(self._get_value_indexes_length):
*/
  __pyx_t_2 = __Pyx_PyList_GET_SIZE(__pyx_v_temp); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(0, 153, __pyx_L1_error)
  __pyx_v_self->_get_value_indexes_length = __pyx_t_2;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":154
 *             temp.remove(index_of_count_star)
 *         self._get_value_indexes_length = len(temp)
 *         self._get_value_indexes = <size_t*> malloc(sizeof(size_t) * self._get_value_indexes_length)             # <<<<<<<<<<<<<<
 *         for i in range(self._get_value_indexes_length):
 *             self._get_value_indexes[i] = temp[i]
*/
  __pyx_v_self->_get_value_indexes = ((size_t *)malloc(((sizeof(size_t)) * __pyx_v_self->_get_value_indexes_length)));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":155
 *         self._get_value_indexes_length = len(temp)
 *         self._get_value_indexes = <size_t*> malloc(sizeof(size_t) * self._get_value_indexes_length)
 *         for i in range(self._get_value_indexes_length):             # <<<<<<<<<<<<<<
 *             self._get_value_indexes[i] = temp[i]
 * 
*/
  __pyx_t_3 = __pyx_v_self->_get_value_indexes_length;
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":156
 *         self._get_value_indexes = <size_t*> malloc(sizeof(size_t) * self._get_value_indexes_length)
 *         for i in range(self._get_value_indexes_length):
 *             self._get_value_indexes[i] = temp[i]             # <<<<<<<<<<<<<<
 * 
 *     cdef void open(self, object state_data_view_store):
*/
    __pyx_t_12 = __Pyx_PyLong_As_size_t(__Pyx_PyList_GET_ITEM(__pyx_v_temp, __pyx_v_i)); if (unlikely((__pyx_t_12 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 156, __pyx_L1_error)
    (__pyx_v_self->_get_value_indexes[__pyx_v_i]) = __pyx_t_12;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":125
 * 
 * cdef class SimpleNamespaceAggsHandleFunction(NamespaceAggsHandleFunction):
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  udfs: List[ImperativeAggregateFunction],
 *                  input_extractors: List,
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_TraceReturnCValue(__pyx_r, __Pyx_Owned_Py_None, 0, 0, __PYX_ERR(0, 125, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 125, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_temp);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":158
 *             self._get_value_indexes[i] = temp[i]
 * 
 *     cdef void open(self, object state_data_view_store):             # <<<<<<<<<<<<<<
 *         cdef dict data_views
 *         for udf in self._udfs:
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_open(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_state_data_view_store) {
  PyObject *__pyx_v_data_views = 0;
  PyObject *__pyx_v_udf = NULL;
  PyObject *__pyx_v_data_view_specs = NULL;
  PyObject *__pyx_v_data_view_spec = NULL;
  PyObject *__pyx_v_key = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  int __pyx_t_14;
  int __pyx_t_15;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[16]))
  __Pyx_RefNannySetupContext("open", 0);
  __Pyx_TraceStartFunc("open", __pyx_f[0], 158, 0, 0, 0, __PYX_ERR(0, 158, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":160
 *     cdef void open(self, object state_data_view_store):
 *         cdef dict data_views
 *         for udf in self._udfs:             # <<<<<<<<<<<<<<
 *             udf.open(state_data_view_store.get_runtime_context())
 *         for data_view_specs in self._udf_data_view_specs:
*/
  if (unlikely(__pyx_v_self->_udfs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 160, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->_udfs; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 160, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    __pyx_t_3 = __Pyx_PyList_GetItemRef(__pyx_t_1, __pyx_t_2);
    ++__pyx_t_2;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_udf, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":161
 *         cdef dict data_views
 *         for udf in self._udfs:
 *             udf.open(state_data_view_store.get_runtime_context())             # <<<<<<<<<<<<<<
 *         for data_view_specs in self._udf_data_view_specs:
 *             data_views = {}
*/
    __pyx_t_4 = __pyx_v_udf;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_6 = __pyx_v_state_data_view_store;
    __Pyx_INCREF(__pyx_t_6);
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_runtime_context, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
    }
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_t_5};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_open, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 161, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":160
 *     cdef void open(self, object state_data_view_store):
 *         cdef dict data_views
 *         for udf in self._udfs:             # <<<<<<<<<<<<<<
 *             udf.open(state_data_view_store.get_runtime_context())
 *         for data_view_specs in self._udf_data_view_specs:
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":162
 *         for udf in self._udfs:
 *             udf.open(state_data_view_store.get_runtime_context())
 *         for data_view_specs in self._udf_data_view_specs:             # <<<<<<<<<<<<<<
 *             data_views = {}
 *             for data_view_spec in data_view_specs:
*/
  if (unlikely(__pyx_v_self->_udf_data_view_specs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 162, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->_udf_data_view_specs; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 162, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    __pyx_t_3 = __Pyx_PyList_GetItemRef(__pyx_t_1, __pyx_t_2);
    ++__pyx_t_2;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 162, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_data_view_specs, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":163
 *             udf.open(state_data_view_store.get_runtime_context())
 *         for data_view_specs in self._udf_data_view_specs:
 *             data_views = {}             # <<<<<<<<<<<<<<
 *             for data_view_spec in data_view_specs:
 *                 if isinstance(data_view_spec, ListViewSpec):
*/
    __pyx_t_3 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 163, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_data_views, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":164
 *         for data_view_specs in self._udf_data_view_specs:
 *             data_views = {}
 *             for data_view_spec in data_view_specs:             # <<<<<<<<<<<<<<
 *                 if isinstance(data_view_spec, ListViewSpec):
 *                     data_views[data_view_spec.field_index] = \
*/
    if (likely(PyList_CheckExact(__pyx_v_data_view_specs)) || PyTuple_CheckExact(__pyx_v_data_view_specs)) {
      __pyx_t_3 = __pyx_v_data_view_specs; __Pyx_INCREF(__pyx_t_3);
      __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_data_view_specs); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_9 = (CYTHON_COMPILING_IN_LIMITED_API) ? PyIter_Next : __Pyx_PyObject_GetIterNextFunc(__pyx_t_3); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 164, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_3))) {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 164, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          __pyx_t_5 = __Pyx_PyList_GetItemRef(__pyx_t_3, __pyx_t_8);
          ++__pyx_t_8;
        } else {
          {
            Py_ssize_t __pyx_temp = __Pyx_PyTuple_GET_SIZE(__pyx_t_3);
            #if !CYTHON_ASSUME_SAFE_SIZE
            if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 164, __pyx_L1_error)
            #endif
            if (__pyx_t_8 >= __pyx_temp) break;
          }
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_5 = __Pyx_NewRef(PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_8));
          #else
          __pyx_t_5 = __Pyx_PySequence_ITEM(__pyx_t_3, __pyx_t_8);
          #endif
          ++__pyx_t_8;
        }
        if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 164, __pyx_L1_error)
      } else {
        __pyx_t_5 = __pyx_t_9(__pyx_t_3);
        if (unlikely(!__pyx_t_5)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) __PYX_ERR(0, 164, __pyx_L1_error)
            PyErr_Clear();
          }
          break;
        }
      }
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_XDECREF_SET(__pyx_v_data_view_spec, __pyx_t_5);
      __pyx_t_5 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":165
 *             data_views = {}
 *             for data_view_spec in data_view_specs:
 *                 if isinstance(data_view_spec, ListViewSpec):             # <<<<<<<<<<<<<<
 *                     data_views[data_view_spec.field_index] = \
 *                         state_data_view_store.get_state_list_view(
*/
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_ListViewSpec); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 165, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PyObject_IsInstance(__pyx_v_data_view_spec, __pyx_t_5); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 165, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_10) {

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":167
 *                 if isinstance(data_view_spec, ListViewSpec):
 *                     data_views[data_view_spec.field_index] = \
 *                         state_data_view_store.get_state_list_view(             # <<<<<<<<<<<<<<
 *                             data_view_spec.state_id,
 *                             data_view_spec.element_coder)
*/
        __pyx_t_4 = __pyx_v_state_data_view_store;
        __Pyx_INCREF(__pyx_t_4);

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":168
 *                     data_views[data_view_spec.field_index] = \
 *                         state_data_view_store.get_state_list_view(
 *                             data_view_spec.state_id,             # <<<<<<<<<<<<<<
 *                             data_view_spec.element_coder)
 *                 elif isinstance(data_view_spec, MapViewSpec):
*/
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_view_spec, __pyx_mstate_global->__pyx_n_u_state_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 168, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":169
 *                         state_data_view_store.get_state_list_view(
 *                             data_view_spec.state_id,
 *                             data_view_spec.element_coder)             # <<<<<<<<<<<<<<
 *                 elif isinstance(data_view_spec, MapViewSpec):
 *                     data_views[data_view_spec.field_index] = \
*/
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_view_spec, __pyx_mstate_global->__pyx_n_u_element_coder); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 169, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __pyx_t_7 = 0;
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_6, __pyx_t_11};
          __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_state_list_view, __pyx_callargs+__pyx_t_7, (3-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 167, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
        }

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":166
 *             for data_view_spec in data_view_specs:
 *                 if isinstance(data_view_spec, ListViewSpec):
 *                     data_views[data_view_spec.field_index] = \             # <<<<<<<<<<<<<<
 *                         state_data_view_store.get_state_list_view(
 *                             data_view_spec.state_id,
*/
        __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_view_spec, __pyx_mstate_global->__pyx_n_u_field_index); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        if (unlikely((PyDict_SetItem(__pyx_v_data_views, __pyx_t_11, __pyx_t_5) < 0))) __PYX_ERR(0, 166, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":165
 *             data_views = {}
 *             for data_view_spec in data_view_specs:
 *                 if isinstance(data_view_spec, ListViewSpec):             # <<<<<<<<<<<<<<
 *                     data_views[data_view_spec.field_index] = \
 *                         state_data_view_store.get_state_list_view(
*/
        goto __pyx_L10;
      }

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":170
 *                             data_view_spec.state_id,
 *                             data_view_spec.element_coder)
 *                 elif isinstance(data_view_spec, MapViewSpec):             # <<<<<<<<<<<<<<
 *                     data_views[data_view_spec.field_index] = \
 *                         state_data_view_store.get_state_map_view(
*/
      __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_mstate_global->__pyx_n_u_MapViewSpec); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_10 = PyObject_IsInstance(__pyx_v_data_view_spec, __pyx_t_5); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 170, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (__pyx_t_10) {

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":172
 *                 elif isinstance(data_view_spec, MapViewSpec):
 *                     data_views[data_view_spec.field_index] = \
 *                         state_data_view_store.get_state_map_view(             # <<<<<<<<<<<<<<
 *                             data_view_spec.state_id,
 *                             data_view_spec.key_coder,
*/
        __pyx_t_11 = __pyx_v_state_data_view_store;
        __Pyx_INCREF(__pyx_t_11);

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":173
 *                     data_views[data_view_spec.field_index] = \
 *                         state_data_view_store.get_state_map_view(
 *                             data_view_spec.state_id,             # <<<<<<<<<<<<<<
 *                             data_view_spec.key_coder,
 *                             data_view_spec.value_coder)
*/
        __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_view_spec, __pyx_mstate_global->__pyx_n_u_state_id); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 173, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":174
 *                         state_data_view_store.get_state_map_view(
 *                             data_view_spec.state_id,
 *                             data_view_spec.key_coder,             # <<<<<<<<<<<<<<
 *                             data_view_spec.value_coder)
 *             self._udf_data_views.append(data_views)
*/
        __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_view_spec, __pyx_mstate_global->__pyx_n_u_key_coder); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 174, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":175
 *                             data_view_spec.state_id,
 *                             data_view_spec.key_coder,
 *                             data_view_spec.value_coder)             # <<<<<<<<<<<<<<
 *             self._udf_data_views.append(data_views)
 *         for key in self._distinct_view_descriptors.keys():
*/
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_view_spec, __pyx_mstate_global->__pyx_n_u_value_coder); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 175, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_7 = 0;
        {
          PyObject *__pyx_callargs[4] = {__pyx_t_11, __pyx_t_6, __pyx_t_4, __pyx_t_12};
          __pyx_t_5 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_state_map_view, __pyx_callargs+__pyx_t_7, (4-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
          if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 172, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_5);
        }

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":171
 *                             data_view_spec.element_coder)
 *                 elif isinstance(data_view_spec, MapViewSpec):
 *                     data_views[data_view_spec.field_index] = \             # <<<<<<<<<<<<<<
 *                         state_data_view_store.get_state_map_view(
 *                             data_view_spec.state_id,
*/
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_data_view_spec, __pyx_mstate_global->__pyx_n_u_field_index); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 171, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        if (unlikely((PyDict_SetItem(__pyx_v_data_views, __pyx_t_12, __pyx_t_5) < 0))) __PYX_ERR(0, 171, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":170
 *                             data_view_spec.state_id,
 *                             data_view_spec.element_coder)
 *                 elif isinstance(data_view_spec, MapViewSpec):             # <<<<<<<<<<<<<<
 *                     data_views[data_view_spec.field_index] = \
 *                         state_data_view_store.get_state_map_view(
*/
      }
      __pyx_L10:;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":164
 *         for data_view_specs in self._udf_data_view_specs:
 *             data_views = {}
 *             for data_view_spec in data_view_specs:             # <<<<<<<<<<<<<<
 *                 if isinstance(data_view_spec, ListViewSpec):
 *                     data_views[data_view_spec.field_index] = \
*/
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":176
 *                             data_view_spec.key_coder,
 *                             data_view_spec.value_coder)
 *             self._udf_data_views.append(data_views)             # <<<<<<<<<<<<<<
 *         for key in self._distinct_view_descriptors.keys():
 *             self._distinct_data_views[key] = state_data_view_store.get_state_map_view(
*/
    if (unlikely(__pyx_v_self->_udf_data_views == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "append");
      __PYX_ERR(0, 176, __pyx_L1_error)
    }
    __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_self->_udf_data_views, __pyx_v_data_views); if (unlikely(__pyx_t_13 == ((int)-1))) __PYX_ERR(0, 176, __pyx_L1_error)

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":162
 *         for udf in self._udfs:
 *             udf.open(state_data_view_store.get_runtime_context())
 *         for data_view_specs in self._udf_data_view_specs:             # <<<<<<<<<<<<<<
 *             data_views = {}
 *             for data_view_spec in data_view_specs:
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":177
 *                             data_view_spec.value_coder)
 *             self._udf_data_views.append(data_views)
 *         for key in self._distinct_view_descriptors.keys():             # <<<<<<<<<<<<<<
 *             self._distinct_data_views[key] = state_data_view_store.get_state_map_view(
 *                 "agg%ddistinct" % key,
*/
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v_self->_distinct_view_descriptors == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(0, 177, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_dict_iterator(__pyx_v_self->_distinct_view_descriptors, 1, __pyx_mstate_global->__pyx_n_u_keys, (&__pyx_t_8), (&__pyx_t_14)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_3;
  __pyx_t_3 = 0;
  while (1) {
    __pyx_t_15 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_8, &__pyx_t_2, &__pyx_t_3, NULL, NULL, __pyx_t_14);
    if (unlikely(__pyx_t_15 == 0)) break;
    if (unlikely(__pyx_t_15 == -1)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_key, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":178
 *             self._udf_data_views.append(data_views)
 *         for key in self._distinct_view_descriptors.keys():
 *             self._distinct_data_views[key] = state_data_view_store.get_state_map_view(             # <<<<<<<<<<<<<<
 *                 "agg%ddistinct" % key,
 *                 PickleCoder(),
*/
    __pyx_t_5 = __pyx_v_state_data_view_store;
    __Pyx_INCREF(__pyx_t_5);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":179
 *         for key in self._distinct_view_descriptors.keys():
 *             self._distinct_data_views[key] = state_data_view_store.get_state_map_view(
 *                 "agg%ddistinct" % key,             # <<<<<<<<<<<<<<
 *                 PickleCoder(),
 *                 PickleCoder())
*/
    __pyx_t_12 = __Pyx_PyUnicode_FormatSafe(__pyx_mstate_global->__pyx_kp_u_agg_ddistinct, __pyx_v_key); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_12);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":180
 *             self._distinct_data_views[key] = state_data_view_store.get_state_map_view(
 *                 "agg%ddistinct" % key,
 *                 PickleCoder(),             # <<<<<<<<<<<<<<
 *                 PickleCoder())
 * 
*/
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_11, __pyx_mstate_global->__pyx_n_u_PickleCoder); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 180, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_11);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 180, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":181
 *                 "agg%ddistinct" % key,
 *                 PickleCoder(),
 *                 PickleCoder())             # <<<<<<<<<<<<<<
 * 
 *     cdef void accumulate(self, InternalRow input_data):
*/
    __pyx_t_6 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_mstate_global->__pyx_n_u_PickleCoder); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_16))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_16);
      assert(__pyx_t_6);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_16);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_16, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_6, NULL};
      __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_16, __pyx_callargs+__pyx_t_7, (1-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
    }
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_5, __pyx_t_12, __pyx_t_4, __pyx_t_11};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_state_map_view, __pyx_callargs+__pyx_t_7, (4-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":178
 *             self._udf_data_views.append(data_views)
 *         for key in self._distinct_view_descriptors.keys():
 *             self._distinct_data_views[key] = state_data_view_store.get_state_map_view(             # <<<<<<<<<<<<<<
 *                 "agg%ddistinct" % key,
 *                 PickleCoder(),
*/
    if (unlikely(__pyx_v_self->_distinct_data_views == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 178, __pyx_L1_error)
    }
    if (unlikely((PyDict_SetItem(__pyx_v_self->_distinct_data_views, __pyx_v_key, __pyx_t_3) < 0))) __PYX_ERR(0, 178, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":158
 *             self._get_value_indexes[i] = temp[i]
 * 
 *     cdef void open(self, object state_data_view_store):             # <<<<<<<<<<<<<<
 *         cdef dict data_views
 *         for udf in self._udfs:
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 158, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 158, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data_views);
  __Pyx_XDECREF(__pyx_v_udf);
  __Pyx_XDECREF(__pyx_v_data_view_specs);
  __Pyx_XDECREF(__pyx_v_data_view_spec);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":183
 *                 PickleCoder())
 * 
 *     cdef void accumulate(self, InternalRow input_data):             # <<<<<<<<<<<<<<
 *         cdef size_t i, j, filter_length
 *         cdef int distinct_index, filter_arg
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_accumulate(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_data) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  size_t __pyx_v_filter_length;
  int __pyx_v_distinct_index;
  int __pyx_v_filter_arg;
  int *__pyx_v_filter_args;
  int __pyx_v_filtered;
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor *__pyx_v_distinct_view_descriptor = 0;
  PyObject *__pyx_v_distinct_data_view = 0;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_internal_row = 0;
  PyObject *__pyx_v_args = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int *__pyx_t_8;
  size_t __pyx_t_9;
  size_t __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_t_12;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[17]))
  __Pyx_RefNannySetupContext("accumulate", 0);
  __Pyx_TraceStartFunc("accumulate", __pyx_f[0], 183, 0, 0, 0, __PYX_ERR(0, 183, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":191
 *         cdef object distinct_data_view
 *         cdef InternalRow internal_row
 *         for i in range(self._udf_num):             # <<<<<<<<<<<<<<
 *             if i in self._distinct_data_views:
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
*/
  __pyx_t_1 = __pyx_v_self->_udf_num;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":192
 *         cdef InternalRow internal_row
 *         for i in range(self._udf_num):
 *             if i in self._distinct_data_views:             # <<<<<<<<<<<<<<
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length
*/
    __pyx_t_4 = __Pyx_PyLong_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__pyx_v_self->_distinct_data_views == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 192, __pyx_L1_error)
    }
    __pyx_t_5 = (__Pyx_PyDict_ContainsTF(__pyx_t_4, __pyx_v_self->_distinct_data_views, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":193
 *         for i in range(self._udf_num):
 *             if i in self._distinct_data_views:
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]             # <<<<<<<<<<<<<<
 *                 filter_length = distinct_view_descriptor.filter_length
 *                 if filter_length == 0:
*/
      if (unlikely(__pyx_v_self->_distinct_view_descriptors == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 193, __pyx_L1_error)
      }
      __pyx_t_4 = __Pyx_PyLong_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_self->_distinct_view_descriptors, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor))))) __PYX_ERR(0, 193, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_distinct_view_descriptor, ((struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor *)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":194
 *             if i in self._distinct_data_views:
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length             # <<<<<<<<<<<<<<
 *                 if filter_length == 0:
 *                     filtered = False
*/
      __pyx_t_7 = __pyx_v_distinct_view_descriptor->filter_length;
      __pyx_v_filter_length = __pyx_t_7;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":195
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length
 *                 if filter_length == 0:             # <<<<<<<<<<<<<<
 *                     filtered = False
 *                 else:
*/
      __pyx_t_5 = (__pyx_v_filter_length == 0);
      if (__pyx_t_5) {

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":196
 *                 filter_length = distinct_view_descriptor.filter_length
 *                 if filter_length == 0:
 *                     filtered = False             # <<<<<<<<<<<<<<
 *                 else:
 *                     filtered = True
*/
        __pyx_v_filtered = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":195
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length
 *                 if filter_length == 0:             # <<<<<<<<<<<<<<
 *                     filtered = False
 *                 else:
*/
        goto __pyx_L6;
      }

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":198
 *                     filtered = False
 *                 else:
 *                     filtered = True             # <<<<<<<<<<<<<<
 *                     filter_args = distinct_view_descriptor.filter_args
 *                     for j in range(filter_length):
*/
      /*else*/ {
        __pyx_v_filtered = 1;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":199
 *                 else:
 *                     filtered = True
 *                     filter_args = distinct_view_descriptor.filter_args             # <<<<<<<<<<<<<<
 *                     for j in range(filter_length):
 *                         filter_arg = filter_args[j]
*/
        __pyx_t_8 = __pyx_v_distinct_view_descriptor->filter_args;
        __pyx_v_filter_args = __pyx_t_8;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":200
 *                     filtered = True
 *                     filter_args = distinct_view_descriptor.filter_args
 *                     for j in range(filter_length):             # <<<<<<<<<<<<<<
 *                         filter_arg = filter_args[j]
 *                         if input_data[filter_arg]:
*/
        __pyx_t_7 = __pyx_v_filter_length;
        __pyx_t_9 = __pyx_t_7;
        for (__pyx_t_10 = 0; __pyx_t_10 < __pyx_t_9; __pyx_t_10+=1) {
          __pyx_v_j = __pyx_t_10;

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":201
 *                     filter_args = distinct_view_descriptor.filter_args
 *                     for j in range(filter_length):
 *                         filter_arg = filter_args[j]             # <<<<<<<<<<<<<<
 *                         if input_data[filter_arg]:
 *                             filtered = False
*/
          __pyx_v_filter_arg = (__pyx_v_filter_args[__pyx_v_j]);

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":202
 *                     for j in range(filter_length):
 *                         filter_arg = filter_args[j]
 *                         if input_data[filter_arg]:             # <<<<<<<<<<<<<<
 *                             filtered = False
 *                             break
*/
          __pyx_t_6 = __Pyx_GetItemInt(((PyObject *)__pyx_v_input_data), __pyx_v_filter_arg, int, 1, __Pyx_PyLong_From_int, 0, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 202, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 202, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__pyx_t_5) {

            /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":203
 *                         filter_arg = filter_args[j]
 *                         if input_data[filter_arg]:
 *                             filtered = False             # <<<<<<<<<<<<<<
 *                             break
 *                 if not filtered:
*/
            __pyx_v_filtered = 0;

            /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":204
 *                         if input_data[filter_arg]:
 *                             filtered = False
 *                             break             # <<<<<<<<<<<<<<
 *                 if not filtered:
 *                     args = distinct_view_descriptor.input_extractor(input_data)
*/
            goto __pyx_L8_break;

            /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":202
 *                     for j in range(filter_length):
 *                         filter_arg = filter_args[j]
 *                         if input_data[filter_arg]:             # <<<<<<<<<<<<<<
 *                             filtered = False
 *                             break
*/
          }
        }
        __pyx_L8_break:;
      }
      __pyx_L6:;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":205
 *                             filtered = False
 *                             break
 *                 if not filtered:             # <<<<<<<<<<<<<<
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]
*/
      __pyx_t_5 = (!__pyx_v_filtered);
      if (__pyx_t_5) {

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":206
 *                             break
 *                 if not filtered:
 *                     args = distinct_view_descriptor.input_extractor(input_data)             # <<<<<<<<<<<<<<
 *                     distinct_data_view = self._distinct_data_views[i]
 *                     if args in distinct_data_view:
*/
        __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_distinct_view_descriptor->input_extractor);
        __pyx_t_11 = __pyx_v_distinct_view_descriptor->input_extractor; 
        __pyx_t_7 = 1;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_11))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_11);
          assert(__pyx_t_4);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_11);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_11, __pyx__function);
          __pyx_t_7 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_input_data)};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_11, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 206, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_v_args, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":207
 *                 if not filtered:
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]             # <<<<<<<<<<<<<<
 *                     if args in distinct_data_view:
 *                         distinct_data_view[args] += 1
*/
        if (unlikely(__pyx_v_self->_distinct_data_views == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 207, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_PyLong_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 207, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = __Pyx_PyDict_GetItem(__pyx_v_self->_distinct_data_views, __pyx_t_6); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 207, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_distinct_data_view, __pyx_t_11);
        __pyx_t_11 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":208
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]
 *                     if args in distinct_data_view:             # <<<<<<<<<<<<<<
 *                         distinct_data_view[args] += 1
 *                     else:
*/
        __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_args, __pyx_v_distinct_data_view, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 208, __pyx_L1_error)
        if (__pyx_t_5) {

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":209
 *                     distinct_data_view = self._distinct_data_views[i]
 *                     if args in distinct_data_view:
 *                         distinct_data_view[args] += 1             # <<<<<<<<<<<<<<
 *                     else:
 *                         distinct_data_view[args] = 1
*/
          __Pyx_INCREF(__pyx_v_args);
          __pyx_t_11 = __pyx_v_args;
          __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_distinct_data_view, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 209, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = __Pyx_PyLong_AddObjC(__pyx_t_6, __pyx_mstate_global->__pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 209, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely((PyObject_SetItem(__pyx_v_distinct_data_view, __pyx_t_11, __pyx_t_4) < 0))) __PYX_ERR(0, 209, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":208
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]
 *                     if args in distinct_data_view:             # <<<<<<<<<<<<<<
 *                         distinct_data_view[args] += 1
 *                     else:
*/
          goto __pyx_L11;
        }

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":211
 *                         distinct_data_view[args] += 1
 *                     else:
 *                         distinct_data_view[args] = 1             # <<<<<<<<<<<<<<
 *             filter_arg = self._filter_args[i]
 *             if filter_arg >= 0 and not input_data[filter_arg]:
*/
        /*else*/ {
          if (unlikely((PyObject_SetItem(__pyx_v_distinct_data_view, __pyx_v_args, __pyx_mstate_global->__pyx_int_1) < 0))) __PYX_ERR(0, 211, __pyx_L1_error)
        }
        __pyx_L11:;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":205
 *                             filtered = False
 *                             break
 *                 if not filtered:             # <<<<<<<<<<<<<<
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]
*/
      }

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":192
 *         cdef InternalRow internal_row
 *         for i in range(self._udf_num):
 *             if i in self._distinct_data_views:             # <<<<<<<<<<<<<<
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":212
 *                     else:
 *                         distinct_data_view[args] = 1
 *             filter_arg = self._filter_args[i]             # <<<<<<<<<<<<<<
 *             if filter_arg >= 0 and not input_data[filter_arg]:
 *                 continue
*/
    __pyx_v_filter_arg = (__pyx_v_self->_filter_args[__pyx_v_i]);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":213
 *                         distinct_data_view[args] = 1
 *             filter_arg = self._filter_args[i]
 *             if filter_arg >= 0 and not input_data[filter_arg]:             # <<<<<<<<<<<<<<
 *                 continue
 *             args = self._input_extractors[i](input_data)
*/
    __pyx_t_12 = (__pyx_v_filter_arg >= 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_5 = __pyx_t_12;
      goto __pyx_L13_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_GetItemInt(((PyObject *)__pyx_v_input_data), __pyx_v_filter_arg, int, 1, __Pyx_PyLong_From_int, 0, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_13 = (!__pyx_t_12);
    __pyx_t_5 = __pyx_t_13;
    __pyx_L13_bool_binop_done:;
    if (__pyx_t_5) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":214
 *             filter_arg = self._filter_args[i]
 *             if filter_arg >= 0 and not input_data[filter_arg]:
 *                 continue             # <<<<<<<<<<<<<<
 *             args = self._input_extractors[i](input_data)
 *             distinct_index = self._distinct_indexes[i]
*/
      goto __pyx_L3_continue;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":213
 *                         distinct_data_view[args] = 1
 *             filter_arg = self._filter_args[i]
 *             if filter_arg >= 0 and not input_data[filter_arg]:             # <<<<<<<<<<<<<<
 *                 continue
 *             args = self._input_extractors[i](input_data)
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":215
 *             if filter_arg >= 0 and not input_data[filter_arg]:
 *                 continue
 *             args = self._input_extractors[i](input_data)             # <<<<<<<<<<<<<<
 *             distinct_index = self._distinct_indexes[i]
 *             if distinct_index >= 0:
*/
    __pyx_t_4 = NULL;
    if (unlikely(__pyx_v_self->_input_extractors == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 215, __pyx_L1_error)
    }
    __Pyx_INCREF(__Pyx_PyList_GET_ITEM(__pyx_v_self->_input_extractors, __pyx_v_i));
    __pyx_t_6 = __Pyx_PyList_GET_ITEM(__pyx_v_self->_input_extractors, __pyx_v_i); 
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_input_data)};
      __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 215, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
    }
    __Pyx_XDECREF_SET(__pyx_v_args, __pyx_t_11);
    __pyx_t_11 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":216
 *                 continue
 *             args = self._input_extractors[i](input_data)
 *             distinct_index = self._distinct_indexes[i]             # <<<<<<<<<<<<<<
 *             if distinct_index >= 0:
 *                 distinct_data_view = self._distinct_data_views[distinct_index]
*/
    __pyx_v_distinct_index = (__pyx_v_self->_distinct_indexes[__pyx_v_i]);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":217
 *             args = self._input_extractors[i](input_data)
 *             distinct_index = self._distinct_indexes[i]
 *             if distinct_index >= 0:             # <<<<<<<<<<<<<<
 *                 distinct_data_view = self._distinct_data_views[distinct_index]
 *                 if args in distinct_data_view:
*/
    __pyx_t_5 = (__pyx_v_distinct_index >= 0);
    if (__pyx_t_5) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":218
 *             distinct_index = self._distinct_indexes[i]
 *             if distinct_index >= 0:
 *                 distinct_data_view = self._distinct_data_views[distinct_index]             # <<<<<<<<<<<<<<
 *                 if args in distinct_data_view:
 *                     if distinct_data_view[args] > 1:
*/
      if (unlikely(__pyx_v_self->_distinct_data_views == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 218, __pyx_L1_error)
      }
      __pyx_t_11 = __Pyx_PyLong_From_int(__pyx_v_distinct_index); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_self->_distinct_data_views, __pyx_t_11); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF_SET(__pyx_v_distinct_data_view, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":219
 *             if distinct_index >= 0:
 *                 distinct_data_view = self._distinct_data_views[distinct_index]
 *                 if args in distinct_data_view:             # <<<<<<<<<<<<<<
 *                     if distinct_data_view[args] > 1:
 *                         continue
*/
      __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_args, __pyx_v_distinct_data_view, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 219, __pyx_L1_error)
      if (likely(__pyx_t_5)) {

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":220
 *                 distinct_data_view = self._distinct_data_views[distinct_index]
 *                 if args in distinct_data_view:
 *                     if distinct_data_view[args] > 1:             # <<<<<<<<<<<<<<
 *                         continue
 *                 else:
*/
        __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_distinct_data_view, __pyx_v_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_11 = PyObject_RichCompare(__pyx_t_6, __pyx_mstate_global->__pyx_int_1, Py_GT); __Pyx_XGOTREF(__pyx_t_11); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_11); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 220, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        if (__pyx_t_5) {

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":221
 *                 if args in distinct_data_view:
 *                     if distinct_data_view[args] > 1:
 *                         continue             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise Exception(
*/
          goto __pyx_L3_continue;

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":220
 *                 distinct_data_view = self._distinct_data_views[distinct_index]
 *                 if args in distinct_data_view:
 *                     if distinct_data_view[args] > 1:             # <<<<<<<<<<<<<<
 *                         continue
 *                 else:
*/
        }

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":219
 *             if distinct_index >= 0:
 *                 distinct_data_view = self._distinct_data_views[distinct_index]
 *                 if args in distinct_data_view:             # <<<<<<<<<<<<<<
 *                     if distinct_data_view[args] > 1:
 *                         continue
*/
        goto __pyx_L16;
      }

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":223
 *                         continue
 *                 else:
 *                     raise Exception(             # <<<<<<<<<<<<<<
 *                         "The args are not in the distinct data view, this should not happen.")
 *             # Transfer InternalRow to Row in row-based operations
*/
      /*else*/ {
        __pyx_t_6 = NULL;
        __Pyx_INCREF((PyObject *)(((PyTypeObject*)PyExc_Exception)));
        __pyx_t_4 = ((PyObject *)(((PyTypeObject*)PyExc_Exception))); 
        __pyx_t_7 = 1;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_6, __pyx_mstate_global->__pyx_kp_u_The_args_are_not_in_the_distinct};
          __pyx_t_11 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 223, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_11);
        }
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(0, 223, __pyx_L1_error)
      }
      __pyx_L16:;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":217
 *             args = self._input_extractors[i](input_data)
 *             distinct_index = self._distinct_indexes[i]
 *             if distinct_index >= 0:             # <<<<<<<<<<<<<<
 *                 distinct_data_view = self._distinct_data_views[distinct_index]
 *                 if args in distinct_data_view:
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":226
 *                         "The args are not in the distinct data view, this should not happen.")
 *             # Transfer InternalRow to Row in row-based operations
 *             if len(args) == 1 and isinstance(args[0], InternalRow):             # <<<<<<<<<<<<<<
 *                 internal_row = <InternalRow> args[0]
 *                 args[0] = internal_row.to_row()
*/
    __pyx_t_14 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_14 == ((Py_ssize_t)-1))) __PYX_ERR(0, 226, __pyx_L1_error)
    __pyx_t_13 = (__pyx_t_14 == 1);
    if (__pyx_t_13) {
    } else {
      __pyx_t_5 = __pyx_t_13;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_args, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __pyx_t_13 = __Pyx_TypeCheck(__pyx_t_11, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow); 
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __pyx_t_5 = __pyx_t_13;
    __pyx_L19_bool_binop_done:;
    if (__pyx_t_5) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":227
 *             # Transfer InternalRow to Row in row-based operations
 *             if len(args) == 1 and isinstance(args[0], InternalRow):
 *                 internal_row = <InternalRow> args[0]             # <<<<<<<<<<<<<<
 *                 args[0] = internal_row.to_row()
 *             self._udfs[i].accumulate(self._accumulators[i], *args)
*/
      __pyx_t_11 = __Pyx_GetItemInt(__pyx_v_args, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 0, 1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_11);
      __pyx_t_4 = __pyx_t_11;
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
      __Pyx_XDECREF_SET(__pyx_v_internal_row, ((struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)__pyx_t_4));
      __pyx_t_4 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":228
 *             if len(args) == 1 and isinstance(args[0], InternalRow):
 *                 internal_row = <InternalRow> args[0]
 *                 args[0] = internal_row.to_row()             # <<<<<<<<<<<<<<
 *             self._udfs[i].accumulate(self._accumulators[i], *args)
 * 
*/
      __pyx_t_4 = ((struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)__pyx_v_internal_row->__pyx_vtab)->to_row(__pyx_v_internal_row, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 228, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_args, 0, __pyx_t_4, long, 1, __Pyx_PyLong_From_long, 0, 0, 0, 1) < 0))) __PYX_ERR(0, 228, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":226
 *                         "The args are not in the distinct data view, this should not happen.")
 *             # Transfer InternalRow to Row in row-based operations
 *             if len(args) == 1 and isinstance(args[0], InternalRow):             # <<<<<<<<<<<<<<
 *                 internal_row = <InternalRow> args[0]
 *                 args[0] = internal_row.to_row()
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":229
 *                 internal_row = <InternalRow> args[0]
 *                 args[0] = internal_row.to_row()
 *             self._udfs[i].accumulate(self._accumulators[i], *args)             # <<<<<<<<<<<<<<
 * 
 *     cdef void retract(self, InternalRow input_data):
*/
    if (unlikely(__pyx_v_self->_udfs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 229, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__Pyx_PyList_GET_ITEM(__pyx_v_self->_udfs, __pyx_v_i), __pyx_mstate_global->__pyx_n_u_accumulate); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__pyx_v_self->_accumulators == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 229, __pyx_L1_error)
    }
    __pyx_t_11 = PyTuple_New(1); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_INCREF(__Pyx_PyList_GET_ITEM(__pyx_v_self->_accumulators, __pyx_v_i));
    __Pyx_GIVEREF(__Pyx_PyList_GET_ITEM(__pyx_v_self->_accumulators, __pyx_v_i));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_11, 0, __Pyx_PyList_GET_ITEM(__pyx_v_self->_accumulators, __pyx_v_i)) != (0)) __PYX_ERR(0, 229, __pyx_L1_error);
    __pyx_t_6 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_15 = PyNumber_Add(__pyx_t_11, __pyx_t_6); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 229, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_L3_continue:;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":183
 *                 PickleCoder())
 * 
 *     cdef void accumulate(self, InternalRow input_data):             # <<<<<<<<<<<<<<
 *         cdef size_t i, j, filter_length
 *         cdef int distinct_index, filter_arg
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 183, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 183, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.accumulate", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_distinct_view_descriptor);
  __Pyx_XDECREF(__pyx_v_distinct_data_view);
  __Pyx_XDECREF((PyObject *)__pyx_v_internal_row);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":231
 *             self._udfs[i].accumulate(self._accumulators[i], *args)
 * 
 *     cdef void retract(self, InternalRow input_data):             # <<<<<<<<<<<<<<
 *         cdef size_t i, j, filter_length
 *         cdef int distinct_index, filter_arg
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_retract(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_data) {
  size_t __pyx_v_i;
  size_t __pyx_v_j;
  size_t __pyx_v_filter_length;
  int __pyx_v_distinct_index;
  int __pyx_v_filter_arg;
  int __pyx_v_filtered;
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor *__pyx_v_distinct_view_descriptor = 0;
  PyObject *__pyx_v_distinct_data_view = 0;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_internal_row = 0;
  PyObject *__pyx_v_args = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  size_t __pyx_t_1;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[18]))
  __Pyx_RefNannySetupContext("retract", 0);
  __Pyx_TraceStartFunc("retract", __pyx_f[0], 231, 0, 0, 0, __PYX_ERR(0, 231, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":238
 *         cdef object distinct_data_view
 *         cdef InternalRow internal_row
 *         for i in range(self._udf_num):             # <<<<<<<<<<<<<<
 *             if i in self._distinct_data_views:
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
*/
  __pyx_t_1 = __pyx_v_self->_udf_num;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":239
 *         cdef InternalRow internal_row
 *         for i in range(self._udf_num):
 *             if i in self._distinct_data_views:             # <<<<<<<<<<<<<<
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length
*/
    __pyx_t_4 = __Pyx_PyLong_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 239, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (unlikely(__pyx_v_self->_distinct_data_views == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 239, __pyx_L1_error)
    }
    __pyx_t_5 = (__Pyx_PyDict_ContainsTF(__pyx_t_4, __pyx_v_self->_distinct_data_views, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 239, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (__pyx_t_5) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":240
 *         for i in range(self._udf_num):
 *             if i in self._distinct_data_views:
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]             # <<<<<<<<<<<<<<
 *                 filter_length = distinct_view_descriptor.filter_length
 *                 if filter_length == 0:
*/
      if (unlikely(__pyx_v_self->_distinct_view_descriptors == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 240, __pyx_L1_error)
      }
      __pyx_t_4 = __Pyx_PyLong_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_self->_distinct_view_descriptors, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!(likely(((__pyx_t_6) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_6, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor))))) __PYX_ERR(0, 240, __pyx_L1_error)
      __Pyx_XDECREF_SET(__pyx_v_distinct_view_descriptor, ((struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor *)__pyx_t_6));
      __pyx_t_6 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":241
 *             if i in self._distinct_data_views:
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length             # <<<<<<<<<<<<<<
 *                 if filter_length == 0:
 *                     filtered = False
*/
      __pyx_t_7 = __pyx_v_distinct_view_descriptor->filter_length;
      __pyx_v_filter_length = __pyx_t_7;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":242
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length
 *                 if filter_length == 0:             # <<<<<<<<<<<<<<
 *                     filtered = False
 *                 else:
*/
      __pyx_t_5 = (__pyx_v_filter_length == 0);
      if (__pyx_t_5) {

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":243
 *                 filter_length = distinct_view_descriptor.filter_length
 *                 if filter_length == 0:
 *                     filtered = False             # <<<<<<<<<<<<<<
 *                 else:
 *                     filtered = True
*/
        __pyx_v_filtered = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":242
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length
 *                 if filter_length == 0:             # <<<<<<<<<<<<<<
 *                     filtered = False
 *                 else:
*/
        goto __pyx_L6;
      }

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":245
 *                     filtered = False
 *                 else:
 *                     filtered = True             # <<<<<<<<<<<<<<
 *                     for j in range(filter_length):
 *                         filter_arg = distinct_view_descriptor.filter_args[j]
*/
      /*else*/ {
        __pyx_v_filtered = 1;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":246
 *                 else:
 *                     filtered = True
 *                     for j in range(filter_length):             # <<<<<<<<<<<<<<
 *                         filter_arg = distinct_view_descriptor.filter_args[j]
 *                         if input_data[filter_arg]:
*/
        __pyx_t_7 = __pyx_v_filter_length;
        __pyx_t_8 = __pyx_t_7;
        for (__pyx_t_9 = 0; __pyx_t_9 < __pyx_t_8; __pyx_t_9+=1) {
          __pyx_v_j = __pyx_t_9;

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":247
 *                     filtered = True
 *                     for j in range(filter_length):
 *                         filter_arg = distinct_view_descriptor.filter_args[j]             # <<<<<<<<<<<<<<
 *                         if input_data[filter_arg]:
 *                             filtered = False
*/
          __pyx_v_filter_arg = (__pyx_v_distinct_view_descriptor->filter_args[__pyx_v_j]);

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":248
 *                     for j in range(filter_length):
 *                         filter_arg = distinct_view_descriptor.filter_args[j]
 *                         if input_data[filter_arg]:             # <<<<<<<<<<<<<<
 *                             filtered = False
 *                             break
*/
          __pyx_t_6 = __Pyx_GetItemInt(((PyObject *)__pyx_v_input_data), __pyx_v_filter_arg, int, 1, __Pyx_PyLong_From_int, 0, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 248, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 248, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (__pyx_t_5) {

            /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":249
 *                         filter_arg = distinct_view_descriptor.filter_args[j]
 *                         if input_data[filter_arg]:
 *                             filtered = False             # <<<<<<<<<<<<<<
 *                             break
 *                 if not filtered:
*/
            __pyx_v_filtered = 0;

            /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":250
 *                         if input_data[filter_arg]:
 *                             filtered = False
 *                             break             # <<<<<<<<<<<<<<
 *                 if not filtered:
 *                     args = distinct_view_descriptor.input_extractor(input_data)
*/
            goto __pyx_L8_break;

            /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":248
 *                     for j in range(filter_length):
 *                         filter_arg = distinct_view_descriptor.filter_args[j]
 *                         if input_data[filter_arg]:             # <<<<<<<<<<<<<<
 *                             filtered = False
 *                             break
*/
          }
        }
        __pyx_L8_break:;
      }
      __pyx_L6:;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":251
 *                             filtered = False
 *                             break
 *                 if not filtered:             # <<<<<<<<<<<<<<
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]
*/
      __pyx_t_5 = (!__pyx_v_filtered);
      if (__pyx_t_5) {

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":252
 *                             break
 *                 if not filtered:
 *                     args = distinct_view_descriptor.input_extractor(input_data)             # <<<<<<<<<<<<<<
 *                     distinct_data_view = self._distinct_data_views[i]
 *                     if args in distinct_data_view:
*/
        __pyx_t_4 = NULL;
        __Pyx_INCREF(__pyx_v_distinct_view_descriptor->input_extractor);
        __pyx_t_10 = __pyx_v_distinct_view_descriptor->input_extractor; 
        __pyx_t_7 = 1;
        #if CYTHON_UNPACK_METHODS
        if (likely(PyMethod_Check(__pyx_t_10))) {
          __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_10);
          assert(__pyx_t_4);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_10);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_10, __pyx__function);
          __pyx_t_7 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_input_data)};
          __pyx_t_6 = __Pyx_PyObject_FastCall(__pyx_t_10, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 252, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
        }
        __Pyx_XDECREF_SET(__pyx_v_args, __pyx_t_6);
        __pyx_t_6 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":253
 *                 if not filtered:
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]             # <<<<<<<<<<<<<<
 *                     if args in distinct_data_view:
 *                         distinct_data_view[args] -= 1
*/
        if (unlikely(__pyx_v_self->_distinct_data_views == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 253, __pyx_L1_error)
        }
        __pyx_t_6 = __Pyx_PyLong_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 253, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_10 = __Pyx_PyDict_GetItem(__pyx_v_self->_distinct_data_views, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 253, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF_SET(__pyx_v_distinct_data_view, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":254
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]
 *                     if args in distinct_data_view:             # <<<<<<<<<<<<<<
 *                         distinct_data_view[args] -= 1
 *                         if distinct_data_view[args] == 0:
*/
        __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_v_args, __pyx_v_distinct_data_view, Py_EQ)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 254, __pyx_L1_error)
        if (__pyx_t_5) {

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":255
 *                     distinct_data_view = self._distinct_data_views[i]
 *                     if args in distinct_data_view:
 *                         distinct_data_view[args] -= 1             # <<<<<<<<<<<<<<
 *                         if distinct_data_view[args] == 0:
 *                             del distinct_data_view[args]
*/
          __Pyx_INCREF(__pyx_v_args);
          __pyx_t_10 = __pyx_v_args;
          __pyx_t_6 = __Pyx_PyObject_GetItem(__pyx_v_distinct_data_view, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 255, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_4 = __Pyx_PyLong_SubtractObjC(__pyx_t_6, __pyx_mstate_global->__pyx_int_1, 1, 1, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 255, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          if (unlikely((PyObject_SetItem(__pyx_v_distinct_data_view, __pyx_t_10, __pyx_t_4) < 0))) __PYX_ERR(0, 255, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":256
 *                     if args in distinct_data_view:
 *                         distinct_data_view[args] -= 1
 *                         if distinct_data_view[args] == 0:             # <<<<<<<<<<<<<<
 *                             del distinct_data_view[args]
 *             filter_arg = self._filter_args[i]
*/
          __pyx_t_10 = __Pyx_PyObject_GetItem(__pyx_v_distinct_data_view, __pyx_v_args); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 256, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_5 = (__Pyx_PyLong_BoolEqObjC(__pyx_t_10, __pyx_mstate_global->__pyx_int_0, 0, 0)); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 256, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          if (__pyx_t_5) {

            /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":257
 *                         distinct_data_view[args] -= 1
 *                         if distinct_data_view[args] == 0:
 *                             del distinct_data_view[args]             # <<<<<<<<<<<<<<
 *             filter_arg = self._filter_args[i]
 *             if filter_arg >= 0 and not input_data[filter_arg]:
*/
            if (unlikely((PyObject_DelItem(__pyx_v_distinct_data_view, __pyx_v_args) < 0))) __PYX_ERR(0, 257, __pyx_L1_error)

            /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":256
 *                     if args in distinct_data_view:
 *                         distinct_data_view[args] -= 1
 *                         if distinct_data_view[args] == 0:             # <<<<<<<<<<<<<<
 *                             del distinct_data_view[args]
 *             filter_arg = self._filter_args[i]
*/
          }

          /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":254
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]
 *                     if args in distinct_data_view:             # <<<<<<<<<<<<<<
 *                         distinct_data_view[args] -= 1
 *                         if distinct_data_view[args] == 0:
*/
        }

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":251
 *                             filtered = False
 *                             break
 *                 if not filtered:             # <<<<<<<<<<<<<<
 *                     args = distinct_view_descriptor.input_extractor(input_data)
 *                     distinct_data_view = self._distinct_data_views[i]
*/
      }

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":239
 *         cdef InternalRow internal_row
 *         for i in range(self._udf_num):
 *             if i in self._distinct_data_views:             # <<<<<<<<<<<<<<
 *                 distinct_view_descriptor = self._distinct_view_descriptors[i]
 *                 filter_length = distinct_view_descriptor.filter_length
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":258
 *                         if distinct_data_view[args] == 0:
 *                             del distinct_data_view[args]
 *             filter_arg = self._filter_args[i]             # <<<<<<<<<<<<<<
 *             if filter_arg >= 0 and not input_data[filter_arg]:
 *                 continue
*/
    __pyx_v_filter_arg = (__pyx_v_self->_filter_args[__pyx_v_i]);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":259
 *                             del distinct_data_view[args]
 *             filter_arg = self._filter_args[i]
 *             if filter_arg >= 0 and not input_data[filter_arg]:             # <<<<<<<<<<<<<<
 *                 continue
 *             args = self._input_extractors[i](input_data)
*/
    __pyx_t_11 = (__pyx_v_filter_arg >= 0);
    if (__pyx_t_11) {
    } else {
      __pyx_t_5 = __pyx_t_11;
      goto __pyx_L14_bool_binop_done;
    }
    __pyx_t_10 = __Pyx_GetItemInt(((PyObject *)__pyx_v_input_data), __pyx_v_filter_arg, int, 1, __Pyx_PyLong_From_int, 0, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_11 < 0))) __PYX_ERR(0, 259, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_12 = (!__pyx_t_11);
    __pyx_t_5 = __pyx_t_12;
    __pyx_L14_bool_binop_done:;
    if (__pyx_t_5) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":260
 *             filter_arg = self._filter_args[i]
 *             if filter_arg >= 0 and not input_data[filter_arg]:
 *                 continue             # <<<<<<<<<<<<<<
 *             args = self._input_extractors[i](input_data)
 *             distinct_index = self._distinct_indexes[i]
*/
      goto __pyx_L3_continue;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":259
 *                             del distinct_data_view[args]
 *             filter_arg = self._filter_args[i]
 *             if filter_arg >= 0 and not input_data[filter_arg]:             # <<<<<<<<<<<<<<
 *                 continue
 *             args = self._input_extractors[i](input_data)
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":261
 *             if filter_arg >= 0 and not input_data[filter_arg]:
 *                 continue
 *             args = self._input_extractors[i](input_data)             # <<<<<<<<<<<<<<
 *             distinct_index = self._distinct_indexes[i]
 *             if distinct_index >= 0 and args in self._distinct_data_views[distinct_index]:
*/
    __pyx_t_4 = NULL;
    if (unlikely(__pyx_v_self->_input_extractors == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 261, __pyx_L1_error)
    }
    __Pyx_INCREF(__Pyx_PyList_GET_ITEM(__pyx_v_self->_input_extractors, __pyx_v_i));
    __pyx_t_6 = __Pyx_PyList_GET_ITEM(__pyx_v_self->_input_extractors, __pyx_v_i); 
    __pyx_t_7 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_6);
      assert(__pyx_t_4);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_6);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_6, __pyx__function);
      __pyx_t_7 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, ((PyObject *)__pyx_v_input_data)};
      __pyx_t_10 = __Pyx_PyObject_FastCall(__pyx_t_6, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (__pyx_t_7*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
    }
    __Pyx_XDECREF_SET(__pyx_v_args, __pyx_t_10);
    __pyx_t_10 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":262
 *                 continue
 *             args = self._input_extractors[i](input_data)
 *             distinct_index = self._distinct_indexes[i]             # <<<<<<<<<<<<<<
 *             if distinct_index >= 0 and args in self._distinct_data_views[distinct_index]:
 *                 continue
*/
    __pyx_v_distinct_index = (__pyx_v_self->_distinct_indexes[__pyx_v_i]);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":263
 *             args = self._input_extractors[i](input_data)
 *             distinct_index = self._distinct_indexes[i]
 *             if distinct_index >= 0 and args in self._distinct_data_views[distinct_index]:             # <<<<<<<<<<<<<<
 *                 continue
 *             # Transfer InternalRow to Row in row-based operations
*/
    __pyx_t_12 = (__pyx_v_distinct_index >= 0);
    if (__pyx_t_12) {
    } else {
      __pyx_t_5 = __pyx_t_12;
      goto __pyx_L17_bool_binop_done;
    }
    if (unlikely(__pyx_v_self->_distinct_data_views == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 263, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyLong_From_int(__pyx_v_distinct_index); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_6 = __Pyx_PyDict_GetItem(__pyx_v_self->_distinct_data_views, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_12 = (__Pyx_PySequence_ContainsTF(__pyx_v_args, __pyx_t_6, Py_EQ)); if (unlikely((__pyx_t_12 < 0))) __PYX_ERR(0, 263, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = __pyx_t_12;
    __pyx_L17_bool_binop_done:;
    if (__pyx_t_5) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":264
 *             distinct_index = self._distinct_indexes[i]
 *             if distinct_index >= 0 and args in self._distinct_data_views[distinct_index]:
 *                 continue             # <<<<<<<<<<<<<<
 *             # Transfer InternalRow to Row in row-based operations
 *             if len(args) == 1 and isinstance(args[0], InternalRow):
*/
      goto __pyx_L3_continue;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":263
 *             args = self._input_extractors[i](input_data)
 *             distinct_index = self._distinct_indexes[i]
 *             if distinct_index >= 0 and args in self._distinct_data_views[distinct_index]:             # <<<<<<<<<<<<<<
 *                 continue
 *             # Transfer InternalRow to Row in row-based operations
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":266
 *                 continue
 *             # Transfer InternalRow to Row in row-based operations
 *             if len(args) == 1 and isinstance(args[0], InternalRow):             # <<<<<<<<<<<<<<
 *                 internal_row = <InternalRow> args[0]
 *                 args[0] = internal_row.to_row()
*/
    __pyx_t_13 = PyObject_Length(__pyx_v_args); if (unlikely(__pyx_t_13 == ((Py_ssize_t)-1))) __PYX_ERR(0, 266, __pyx_L1_error)
    __pyx_t_12 = (__pyx_t_13 == 1);
    if (__pyx_t_12) {
    } else {
      __pyx_t_5 = __pyx_t_12;
      goto __pyx_L20_bool_binop_done;
    }
    __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_args, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_12 = __Pyx_TypeCheck(__pyx_t_6, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow); 
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = __pyx_t_12;
    __pyx_L20_bool_binop_done:;
    if (__pyx_t_5) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":267
 *             # Transfer InternalRow to Row in row-based operations
 *             if len(args) == 1 and isinstance(args[0], InternalRow):
 *                 internal_row = <InternalRow> args[0]             # <<<<<<<<<<<<<<
 *                 args[0] = internal_row.to_row()
 *             self._udfs[i].retract(self._accumulators[i], *args)
*/
      __pyx_t_6 = __Pyx_GetItemInt(__pyx_v_args, 0, long, 1, __Pyx_PyLong_From_long, 0, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 267, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = __pyx_t_6;
      __Pyx_INCREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_internal_row, ((struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)__pyx_t_10));
      __pyx_t_10 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":268
 *             if len(args) == 1 and isinstance(args[0], InternalRow):
 *                 internal_row = <InternalRow> args[0]
 *                 args[0] = internal_row.to_row()             # <<<<<<<<<<<<<<
 *             self._udfs[i].retract(self._accumulators[i], *args)
 * 
*/
      __pyx_t_10 = ((struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)__pyx_v_internal_row->__pyx_vtab)->to_row(__pyx_v_internal_row, 0); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (unlikely((__Pyx_SetItemInt(__pyx_v_args, 0, __pyx_t_10, long, 1, __Pyx_PyLong_From_long, 0, 0, 0, 1) < 0))) __PYX_ERR(0, 268, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":266
 *                 continue
 *             # Transfer InternalRow to Row in row-based operations
 *             if len(args) == 1 and isinstance(args[0], InternalRow):             # <<<<<<<<<<<<<<
 *                 internal_row = <InternalRow> args[0]
 *                 args[0] = internal_row.to_row()
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":269
 *                 internal_row = <InternalRow> args[0]
 *                 args[0] = internal_row.to_row()
 *             self._udfs[i].retract(self._accumulators[i], *args)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void merge(self, object namespace, list accumulators):
*/
    if (unlikely(__pyx_v_self->_udfs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 269, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__Pyx_PyList_GET_ITEM(__pyx_v_self->_udfs, __pyx_v_i), __pyx_mstate_global->__pyx_n_u_retract); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    if (unlikely(__pyx_v_self->_accumulators == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 269, __pyx_L1_error)
    }
    __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_INCREF(__Pyx_PyList_GET_ITEM(__pyx_v_self->_accumulators, __pyx_v_i));
    __Pyx_GIVEREF(__Pyx_PyList_GET_ITEM(__pyx_v_self->_accumulators, __pyx_v_i));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_6, 0, __Pyx_PyList_GET_ITEM(__pyx_v_self->_accumulators, __pyx_v_i)) != (0)) __PYX_ERR(0, 269, __pyx_L1_error);
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_args); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_14 = PyNumber_Add(__pyx_t_6, __pyx_t_4); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_14, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 269, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_L3_continue:;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":231
 *             self._udfs[i].accumulate(self._accumulators[i], *args)
 * 
 *     cdef void retract(self, InternalRow input_data):             # <<<<<<<<<<<<<<
 *         cdef size_t i, j, filter_length
 *         cdef int distinct_index, filter_arg
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 231, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 231, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.retract", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_distinct_view_descriptor);
  __Pyx_XDECREF(__pyx_v_distinct_data_view);
  __Pyx_XDECREF((PyObject *)__pyx_v_internal_row);
  __Pyx_XDECREF(__pyx_v_args);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":271
 *             self._udfs[i].retract(self._accumulators[i], *args)
 * 
 *     cpdef void merge(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         if self._udf_data_views:
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_3merge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_merge(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  PyObject *__pyx_v_index = NULL;
  PyObject *__pyx_v_data_view = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_t_12;
  size_t __pyx_t_13;
  size_t __pyx_t_14;
  size_t __pyx_t_15;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19]))
  __Pyx_RefNannySetupContext("merge", 0);
  __Pyx_TraceStartFunc("merge", __pyx_f[0], 271, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 271, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_merge); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_3merge)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_namespace, __pyx_v_accumulators};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":273
 *     cpdef void merge(self, object namespace, list accumulators):
 *         cdef size_t i
 *         if self._udf_data_views:             # <<<<<<<<<<<<<<
 *             for i in range(len(self._udf_data_views)):
 *                 for index, data_view in self._udf_data_views[i].items():
*/
  __pyx_t_6 = (__pyx_v_self->_udf_data_views != Py_None)&&(__Pyx_PyList_GET_SIZE(__pyx_v_self->_udf_data_views) != 0);
  if (unlikely(((!CYTHON_ASSUME_SAFE_MACROS) && __pyx_t_6 < 0))) __PYX_ERR(0, 273, __pyx_L1_error)
  if (__pyx_t_6) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":274
 *         cdef size_t i
 *         if self._udf_data_views:
 *             for i in range(len(self._udf_data_views)):             # <<<<<<<<<<<<<<
 *                 for index, data_view in self._udf_data_views[i].items():
 *                     data_view.set_current_namespace(namespace)
*/
    __pyx_t_1 = __pyx_v_self->_udf_data_views;
    __Pyx_INCREF(__pyx_t_1);
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 274, __pyx_L1_error)
    }
    __pyx_t_7 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(0, 274, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_8 = __pyx_t_7;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_8; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":275
 *         if self._udf_data_views:
 *             for i in range(len(self._udf_data_views)):
 *                 for index, data_view in self._udf_data_views[i].items():             # <<<<<<<<<<<<<<
 *                     data_view.set_current_namespace(namespace)
 *                     accumulators[i][index] = data_view
*/
      __pyx_t_9 = 0;
      if (unlikely(__pyx_v_self->_udf_data_views == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 275, __pyx_L1_error)
      }
      if (unlikely(__Pyx_PyList_GET_ITEM(__pyx_v_self->_udf_data_views, __pyx_v_i) == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(0, 275, __pyx_L1_error)
      }
      __pyx_t_2 = __Pyx_dict_iterator(__Pyx_PyList_GET_ITEM(__pyx_v_self->_udf_data_views, __pyx_v_i), 0, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_10), (&__pyx_t_11)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 275, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_2;
      __pyx_t_2 = 0;
      while (1) {
        __pyx_t_12 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_10, &__pyx_t_9, &__pyx_t_2, &__pyx_t_4, NULL, __pyx_t_11);
        if (unlikely(__pyx_t_12 == 0)) break;
        if (unlikely(__pyx_t_12 == -1)) __PYX_ERR(0, 275, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_XDECREF_SET(__pyx_v_data_view, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":276
 *             for i in range(len(self._udf_data_views)):
 *                 for index, data_view in self._udf_data_views[i].items():
 *                     data_view.set_current_namespace(namespace)             # <<<<<<<<<<<<<<
 *                     accumulators[i][index] = data_view
 *         for i in range(self._udf_num):
*/
        __pyx_t_2 = __pyx_v_data_view;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_13 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_namespace};
          __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_current_namespace, __pyx_callargs+__pyx_t_13, (2-__pyx_t_13) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 276, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":277
 *                 for index, data_view in self._udf_data_views[i].items():
 *                     data_view.set_current_namespace(namespace)
 *                     accumulators[i][index] = data_view             # <<<<<<<<<<<<<<
 *         for i in range(self._udf_num):
 *             self._udfs[i].merge(self._accumulators[i], [accumulators[i]])
*/
        if (unlikely(__pyx_v_accumulators == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 277, __pyx_L1_error)
        }
        if (unlikely((PyObject_SetItem(__Pyx_PyList_GET_ITEM(__pyx_v_accumulators, __pyx_v_i), __pyx_v_index, __pyx_v_data_view) < 0))) __PYX_ERR(0, 277, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":273
 *     cpdef void merge(self, object namespace, list accumulators):
 *         cdef size_t i
 *         if self._udf_data_views:             # <<<<<<<<<<<<<<
 *             for i in range(len(self._udf_data_views)):
 *                 for index, data_view in self._udf_data_views[i].items():
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":278
 *                     data_view.set_current_namespace(namespace)
 *                     accumulators[i][index] = data_view
 *         for i in range(self._udf_num):             # <<<<<<<<<<<<<<
 *             self._udfs[i].merge(self._accumulators[i], [accumulators[i]])
 * 
*/
  __pyx_t_5 = __pyx_v_self->_udf_num;
  __pyx_t_13 = __pyx_t_5;
  for (__pyx_t_14 = 0; __pyx_t_14 < __pyx_t_13; __pyx_t_14+=1) {
    __pyx_v_i = __pyx_t_14;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":279
 *                     accumulators[i][index] = data_view
 *         for i in range(self._udf_num):
 *             self._udfs[i].merge(self._accumulators[i], [accumulators[i]])             # <<<<<<<<<<<<<<
 * 
 *     cpdef void set_accumulators(self, object namespace, list accumulators):
*/
    if (unlikely(__pyx_v_self->_udfs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 279, __pyx_L1_error)
    }
    __pyx_t_4 = __Pyx_PyList_GET_ITEM(__pyx_v_self->_udfs, __pyx_v_i);
    __Pyx_INCREF(__pyx_t_4);
    if (unlikely(__pyx_v_self->_accumulators == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 279, __pyx_L1_error)
    }
    if (unlikely(__pyx_v_accumulators == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 279, __pyx_L1_error)
    }
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__Pyx_PyList_GET_ITEM(__pyx_v_accumulators, __pyx_v_i));
    __Pyx_GIVEREF(__Pyx_PyList_GET_ITEM(__pyx_v_accumulators, __pyx_v_i));
    if (__Pyx_PyList_SET_ITEM(__pyx_t_2, 0, __Pyx_PyList_GET_ITEM(__pyx_v_accumulators, __pyx_v_i)) != (0)) __PYX_ERR(0, 279, __pyx_L1_error);
    __pyx_t_15 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __Pyx_PyList_GET_ITEM(__pyx_v_self->_accumulators, __pyx_v_i), __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_merge, __pyx_callargs+__pyx_t_15, (3-__pyx_t_15) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":271
 *             self._udfs[i].retract(self._accumulators[i], *args)
 * 
 *     cpdef void merge(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         if self._udf_data_views:
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 271, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 271, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.merge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XDECREF(__pyx_v_data_view);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_3merge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_3merge = {"merge", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_3merge, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_3merge(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  PyObject *__pyx_v_accumulators = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("merge (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_namespace,&__pyx_mstate_global->__pyx_n_u_accumulators,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 271, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 271, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 271, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "merge", 0) < 0) __PYX_ERR(0, 271, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("merge", 1, 2, 2, i); __PYX_ERR(0, 271, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 271, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 271, __pyx_L3_error)
    }
    __pyx_v_namespace = values[0];
    __pyx_v_accumulators = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("merge", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 271, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.merge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_accumulators), (&PyList_Type), 1, "accumulators", 1))) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_2merge(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self), __pyx_v_namespace, __pyx_v_accumulators);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_2merge(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19]))
  __Pyx_RefNannySetupContext("merge", 0);
  __Pyx_TraceStartFunc("merge (wrapper)", __pyx_f[0], 271, 0, 0, 0, __PYX_ERR(0, 271, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_merge(__pyx_v_self, __pyx_v_namespace, __pyx_v_accumulators, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 271, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 271, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.merge", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":281
 *             self._udfs[i].merge(self._accumulators[i], [accumulators[i]])
 * 
 *     cpdef void set_accumulators(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         cdef size_t i, index
 *         if self._udf_data_views and namespace is not None:
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_5set_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_set_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  size_t __pyx_v_index;
  PyObject *__pyx_v_data_view = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  int __pyx_t_12;
  int __pyx_t_13;
  size_t __pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20]))
  __Pyx_RefNannySetupContext("set_accumulators", 0);
  __Pyx_TraceStartFunc("set_accumulators", __pyx_f[0], 281, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 281, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_set_accumulators); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_5set_accumulators)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[3] = {__pyx_t_3, __pyx_v_namespace, __pyx_v_accumulators};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":283
 *     cpdef void set_accumulators(self, object namespace, list accumulators):
 *         cdef size_t i, index
 *         if self._udf_data_views and namespace is not None:             # <<<<<<<<<<<<<<
 *             for i in range(len(self._udf_data_views)):
 *                 for index, data_view in self._udf_data_views[i].items():
*/
  __pyx_t_7 = (__pyx_v_self->_udf_data_views != Py_None)&&(__Pyx_PyList_GET_SIZE(__pyx_v_self->_udf_data_views) != 0);
  if (unlikely(((!CYTHON_ASSUME_SAFE_MACROS) && __pyx_t_7 < 0))) __PYX_ERR(0, 283, __pyx_L1_error)
  if (__pyx_t_7) {
  } else {
    __pyx_t_6 = __pyx_t_7;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_7 = (__pyx_v_namespace != Py_None);
  __pyx_t_6 = __pyx_t_7;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":284
 *         cdef size_t i, index
 *         if self._udf_data_views and namespace is not None:
 *             for i in range(len(self._udf_data_views)):             # <<<<<<<<<<<<<<
 *                 for index, data_view in self._udf_data_views[i].items():
 *                     data_view.set_current_namespace(namespace)
*/
    __pyx_t_1 = __pyx_v_self->_udf_data_views;
    __Pyx_INCREF(__pyx_t_1);
    if (unlikely(__pyx_t_1 == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
      __PYX_ERR(0, 284, __pyx_L1_error)
    }
    __pyx_t_8 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(0, 284, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __pyx_t_8;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_9; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":285
 *         if self._udf_data_views and namespace is not None:
 *             for i in range(len(self._udf_data_views)):
 *                 for index, data_view in self._udf_data_views[i].items():             # <<<<<<<<<<<<<<
 *                     data_view.set_current_namespace(namespace)
 *                     accumulators[i][index] = data_view
*/
      __pyx_t_10 = 0;
      if (unlikely(__pyx_v_self->_udf_data_views == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 285, __pyx_L1_error)
      }
      if (unlikely(__Pyx_PyList_GET_ITEM(__pyx_v_self->_udf_data_views, __pyx_v_i) == Py_None)) {
        PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
        __PYX_ERR(0, 285, __pyx_L1_error)
      }
      __pyx_t_2 = __Pyx_dict_iterator(__Pyx_PyList_GET_ITEM(__pyx_v_self->_udf_data_views, __pyx_v_i), 0, __pyx_mstate_global->__pyx_n_u_items, (&__pyx_t_11), (&__pyx_t_12)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 285, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF(__pyx_t_1);
      __pyx_t_1 = __pyx_t_2;
      __pyx_t_2 = 0;
      while (1) {
        __pyx_t_13 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_11, &__pyx_t_10, &__pyx_t_2, &__pyx_t_4, NULL, __pyx_t_12);
        if (unlikely(__pyx_t_13 == 0)) break;
        if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 285, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_GOTREF(__pyx_t_4);
        __pyx_t_14 = __Pyx_PyLong_As_size_t(__pyx_t_2); if (unlikely((__pyx_t_14 == (size_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 285, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_v_index = __pyx_t_14;
        __Pyx_XDECREF_SET(__pyx_v_data_view, __pyx_t_4);
        __pyx_t_4 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":286
 *             for i in range(len(self._udf_data_views)):
 *                 for index, data_view in self._udf_data_views[i].items():
 *                     data_view.set_current_namespace(namespace)             # <<<<<<<<<<<<<<
 *                     accumulators[i][index] = data_view
 *         self._accumulators = accumulators
*/
        __pyx_t_2 = __pyx_v_data_view;
        __Pyx_INCREF(__pyx_t_2);
        __pyx_t_14 = 0;
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_namespace};
          __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_current_namespace, __pyx_callargs+__pyx_t_14, (2-__pyx_t_14) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 286, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

        /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":287
 *                 for index, data_view in self._udf_data_views[i].items():
 *                     data_view.set_current_namespace(namespace)
 *                     accumulators[i][index] = data_view             # <<<<<<<<<<<<<<
 *         self._accumulators = accumulators
 * 
*/
        if (unlikely(__pyx_v_accumulators == Py_None)) {
          PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
          __PYX_ERR(0, 287, __pyx_L1_error)
        }
        if (unlikely((__Pyx_SetItemInt(__Pyx_PyList_GET_ITEM(__pyx_v_accumulators, __pyx_v_i), __pyx_v_index, __pyx_v_data_view, size_t, 0, __Pyx_PyLong_FromSize_t, 0, 0, 0, 1) < 0))) __PYX_ERR(0, 287, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":283
 *     cpdef void set_accumulators(self, object namespace, list accumulators):
 *         cdef size_t i, index
 *         if self._udf_data_views and namespace is not None:             # <<<<<<<<<<<<<<
 *             for i in range(len(self._udf_data_views)):
 *                 for index, data_view in self._udf_data_views[i].items():
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":288
 *                     data_view.set_current_namespace(namespace)
 *                     accumulators[i][index] = data_view
 *         self._accumulators = accumulators             # <<<<<<<<<<<<<<
 * 
 *     cdef list get_accumulators(self):
*/
  __Pyx_INCREF(__pyx_v_accumulators);
  __Pyx_GIVEREF(__pyx_v_accumulators);
  __Pyx_GOTREF(__pyx_v_self->_accumulators);
  __Pyx_DECREF(__pyx_v_self->_accumulators);
  __pyx_v_self->_accumulators = __pyx_v_accumulators;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":281
 *             self._udfs[i].merge(self._accumulators[i], [accumulators[i]])
 * 
 *     cpdef void set_accumulators(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         cdef size_t i, index
 *         if self._udf_data_views and namespace is not None:
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 281, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 281, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.set_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data_view);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_5set_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_5set_accumulators = {"set_accumulators", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_5set_accumulators, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_5set_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  PyObject *__pyx_v_accumulators = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[2] = {0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_accumulators (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_namespace,&__pyx_mstate_global->__pyx_n_u_accumulators,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 281, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 281, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 281, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "set_accumulators", 0) < 0) __PYX_ERR(0, 281, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 2; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("set_accumulators", 1, 2, 2, i); __PYX_ERR(0, 281, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 2)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 281, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 281, __pyx_L3_error)
    }
    __pyx_v_namespace = values[0];
    __pyx_v_accumulators = ((PyObject*)values[1]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("set_accumulators", 1, 2, 2, __pyx_nargs); __PYX_ERR(0, 281, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.set_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_accumulators), (&PyList_Type), 1, "accumulators", 1))) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_4set_accumulators(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self), __pyx_v_namespace, __pyx_v_accumulators);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_4set_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, PyObject *__pyx_v_accumulators) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20]))
  __Pyx_RefNannySetupContext("set_accumulators", 0);
  __Pyx_TraceStartFunc("set_accumulators (wrapper)", __pyx_f[0], 281, 0, 0, 0, __PYX_ERR(0, 281, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_set_accumulators(__pyx_v_self, __pyx_v_namespace, __pyx_v_accumulators, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 281, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 281, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.set_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":290
 *         self._accumulators = accumulators
 * 
 *     cdef list get_accumulators(self):             # <<<<<<<<<<<<<<
 *         return self._accumulators
 * 
*/

static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_get_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[21]))
  __Pyx_RefNannySetupContext("get_accumulators", 0);
  __Pyx_TraceStartFunc("get_accumulators", __pyx_f[0], 290, 0, 0, 0, __PYX_ERR(0, 290, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":291
 * 
 *     cdef list get_accumulators(self):
 *         return self._accumulators             # <<<<<<<<<<<<<<
 * 
 *     cpdef list create_accumulators(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_accumulators);
  __pyx_r = __pyx_v_self->_accumulators;
  __Pyx_TraceReturnValue(__pyx_r, 1, 0, __PYX_ERR(0, 291, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":290
 *         self._accumulators = accumulators
 * 
 *     cdef list get_accumulators(self):             # <<<<<<<<<<<<<<
 *         return self._accumulators
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 290, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.get_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":293
 *         return self._accumulators
 * 
 *     cpdef list create_accumulators(self):             # <<<<<<<<<<<<<<
 *         return [udf.create_accumulator() for udf in self._udfs]
 * 
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_7create_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_create_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_8genexpr1__pyx_v_udf = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22]))
  __Pyx_RefNannySetupContext("create_accumulators", 0);
  __Pyx_TraceStartFunc("create_accumulators", __pyx_f[0], 293, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 293, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_create_accumulators); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_7create_accumulators)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 293, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 293, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":294
 * 
 *     cpdef list create_accumulators(self):
 *         return [udf.create_accumulator() for udf in self._udfs]             # <<<<<<<<<<<<<<
 * 
 *     cpdef void cleanup(self, object namespace):
*/
  __Pyx_XDECREF(__pyx_r);
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 294, __pyx_L5_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (unlikely(__pyx_v_self->_udfs == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
      __PYX_ERR(0, 294, __pyx_L5_error)
    }
    __pyx_t_2 = __pyx_v_self->_udfs; __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    for (;;) {
      {
        Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 294, __pyx_L5_error)
        #endif
        if (__pyx_t_6 >= __pyx_temp) break;
      }
      __pyx_t_4 = __Pyx_PyList_GetItemRef(__pyx_t_2, __pyx_t_6);
      ++__pyx_t_6;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L5_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_8genexpr1__pyx_v_udf, __pyx_t_4);
      __pyx_t_4 = 0;
      __pyx_t_3 = __pyx_8genexpr1__pyx_v_udf;
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_5 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
        __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_create_accumulator, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
        if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 294, __pyx_L5_error)
        __Pyx_GOTREF(__pyx_t_4);
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_4))) __PYX_ERR(0, 294, __pyx_L5_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_udf); __pyx_8genexpr1__pyx_v_udf = 0;
    goto __pyx_L9_exit_scope;
    __pyx_L5_error:;
    __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_udf); __pyx_8genexpr1__pyx_v_udf = 0;
    goto __pyx_L1_error;
    __pyx_L9_exit_scope:;
  } /* exit inner scope */
  __pyx_r = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;
  __Pyx_TraceReturnValue(__pyx_r, 1, 0, __PYX_ERR(0, 294, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":293
 *         return self._accumulators
 * 
 *     cpdef list create_accumulators(self):             # <<<<<<<<<<<<<<
 *         return [udf.create_accumulator() for udf in self._udfs]
 * 
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 293, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.create_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_8genexpr1__pyx_v_udf);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_7create_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_7create_accumulators = {"create_accumulators", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_7create_accumulators, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_7create_accumulators(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("create_accumulators (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("create_accumulators", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("create_accumulators", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_6create_accumulators(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_6create_accumulators(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22]))
  __Pyx_RefNannySetupContext("create_accumulators", 0);
  __Pyx_TraceStartFunc("create_accumulators (wrapper)", __pyx_f[0], 293, 0, 0, 0, __PYX_ERR(0, 293, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_create_accumulators(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 293, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.create_accumulators", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":296
 *         return [udf.create_accumulator() for udf in self._udfs]
 * 
 *     cpdef void cleanup(self, object namespace):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(len(self._udf_data_views)):
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_9cleanup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_cleanup(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, int __pyx_skip_dispatch) {
  size_t __pyx_v_i;
  PyObject *__pyx_v_data_view = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  size_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23]))
  __Pyx_RefNannySetupContext("cleanup", 0);
  __Pyx_TraceStartFunc("cleanup", __pyx_f[0], 296, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 296, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_cleanup); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_9cleanup)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_namespace};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":298
 *     cpdef void cleanup(self, object namespace):
 *         cdef size_t i
 *         for i in range(len(self._udf_data_views)):             # <<<<<<<<<<<<<<
 *             for data_view in self._udf_data_views[i].values():
 *                 data_view.set_current_namespace(namespace)
*/
  __pyx_t_1 = __pyx_v_self->_udf_data_views;
  __Pyx_INCREF(__pyx_t_1);
  if (unlikely(__pyx_t_1 == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(0, 298, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_PyList_GET_SIZE(__pyx_t_1); if (unlikely(__pyx_t_6 == ((Py_ssize_t)-1))) __PYX_ERR(0, 298, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_7 = __pyx_t_6;
  for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_7; __pyx_t_5+=1) {
    __pyx_v_i = __pyx_t_5;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":299
 *         cdef size_t i
 *         for i in range(len(self._udf_data_views)):
 *             for data_view in self._udf_data_views[i].values():             # <<<<<<<<<<<<<<
 *                 data_view.set_current_namespace(namespace)
 *                 data_view.clear()
*/
    __pyx_t_8 = 0;
    if (unlikely(__pyx_v_self->_udf_data_views == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 299, __pyx_L1_error)
    }
    if (unlikely(__Pyx_PyList_GET_ITEM(__pyx_v_self->_udf_data_views, __pyx_v_i) == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "values");
      __PYX_ERR(0, 299, __pyx_L1_error)
    }
    __pyx_t_2 = __Pyx_dict_iterator(__Pyx_PyList_GET_ITEM(__pyx_v_self->_udf_data_views, __pyx_v_i), 0, __pyx_mstate_global->__pyx_n_u_values, (&__pyx_t_9), (&__pyx_t_10)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 299, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF(__pyx_t_1);
    __pyx_t_1 = __pyx_t_2;
    __pyx_t_2 = 0;
    while (1) {
      __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_9, &__pyx_t_8, NULL, &__pyx_t_2, NULL, __pyx_t_10);
      if (unlikely(__pyx_t_11 == 0)) break;
      if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 299, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_XDECREF_SET(__pyx_v_data_view, __pyx_t_2);
      __pyx_t_2 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":300
 *         for i in range(len(self._udf_data_views)):
 *             for data_view in self._udf_data_views[i].values():
 *                 data_view.set_current_namespace(namespace)             # <<<<<<<<<<<<<<
 *                 data_view.clear()
 * 
*/
      __pyx_t_4 = __pyx_v_data_view;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_12 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __pyx_v_namespace};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_current_namespace, __pyx_callargs+__pyx_t_12, (2-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 300, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":301
 *             for data_view in self._udf_data_views[i].values():
 *                 data_view.set_current_namespace(namespace)
 *                 data_view.clear()             # <<<<<<<<<<<<<<
 * 
 *     cdef void close(self):
*/
      __pyx_t_4 = __pyx_v_data_view;
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_12 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_clear, __pyx_callargs+__pyx_t_12, (1-__pyx_t_12) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 301, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":296
 *         return [udf.create_accumulator() for udf in self._udfs]
 * 
 *     cpdef void cleanup(self, object namespace):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(len(self._udf_data_views)):
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 296, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 296, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.cleanup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_data_view);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_9cleanup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_9cleanup = {"cleanup", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_9cleanup, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_9cleanup(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cleanup (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_namespace,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 296, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 296, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cleanup", 0) < 0) __PYX_ERR(0, 296, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cleanup", 1, 1, 1, i); __PYX_ERR(0, 296, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 296, __pyx_L3_error)
    }
    __pyx_v_namespace = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cleanup", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 296, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.cleanup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_8cleanup(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self), __pyx_v_namespace);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_8cleanup(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23]))
  __Pyx_RefNannySetupContext("cleanup", 0);
  __Pyx_TraceStartFunc("cleanup (wrapper)", __pyx_f[0], 296, 0, 0, 0, __PYX_ERR(0, 296, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_cleanup(__pyx_v_self, __pyx_v_namespace, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 296, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 296, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.cleanup", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":303
 *                 data_view.clear()
 * 
 *     cdef void close(self):             # <<<<<<<<<<<<<<
 *         for udf in self._udfs:
 *             udf.close()
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_close(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self) {
  PyObject *__pyx_v_udf = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[24]))
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_TraceStartFunc("close", __pyx_f[0], 303, 0, 0, 0, __PYX_ERR(0, 303, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":304
 * 
 *     cdef void close(self):
 *         for udf in self._udfs:             # <<<<<<<<<<<<<<
 *             udf.close()
 * 
*/
  if (unlikely(__pyx_v_self->_udfs == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 304, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_self->_udfs; __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_1);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 304, __pyx_L1_error)
      #endif
      if (__pyx_t_2 >= __pyx_temp) break;
    }
    __pyx_t_3 = __Pyx_PyList_GetItemRef(__pyx_t_1, __pyx_t_2);
    ++__pyx_t_2;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 304, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_udf, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":305
 *     cdef void close(self):
 *         for udf in self._udfs:
 *             udf.close()             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_value(self, object namespace):
*/
    __pyx_t_4 = __pyx_v_udf;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_close, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 305, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":304
 * 
 *     cdef void close(self):
 *         for udf in self._udfs:             # <<<<<<<<<<<<<<
 *             udf.close()
 * 
*/
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":303
 *                 data_view.clear()
 * 
 *     cdef void close(self):             # <<<<<<<<<<<<<<
 *         for udf in self._udfs:
 *             udf.close()
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 303, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 303, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_udf);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":307
 *             udf.close()
 * 
 *     cpdef list get_value(self, object namespace):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t*get_value_indexes
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_11get_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_get_value(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace, int __pyx_skip_dispatch) {
  size_t *__pyx_v_get_value_indexes;
  PyObject *__pyx_v_result = 0;
  size_t __pyx_8genexpr2__pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  size_t *__pyx_t_6;
  size_t __pyx_t_7;
  size_t __pyx_t_8;
  size_t __pyx_t_9;
  int __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25]))
  __Pyx_RefNannySetupContext("get_value", 0);
  __Pyx_TraceStartFunc("get_value", __pyx_f[0], 307, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 307, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_get_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_11get_value)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_namespace};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 307, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 307, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":311
 *         cdef size_t*get_value_indexes
 *         cdef list result
 *         get_value_indexes = self._get_value_indexes             # <<<<<<<<<<<<<<
 *         result = [
 *             self._udfs[get_value_indexes[i]].get_value(self._accumulators[get_value_indexes[i]])
*/
  __pyx_t_6 = __pyx_v_self->_get_value_indexes;
  __pyx_v_get_value_indexes = __pyx_t_6;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":312
 *         cdef list result
 *         get_value_indexes = self._get_value_indexes
 *         result = [             # <<<<<<<<<<<<<<
 *             self._udfs[get_value_indexes[i]].get_value(self._accumulators[get_value_indexes[i]])
 *             for i in range(self._get_value_indexes_length)]
*/
  { /* enter inner scope */
    __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 312, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":314
 *         result = [
 *             self._udfs[get_value_indexes[i]].get_value(self._accumulators[get_value_indexes[i]])
 *             for i in range(self._get_value_indexes_length)]             # <<<<<<<<<<<<<<
 *         if self._named_property_extractor:
 *             result.extend(self._named_property_extractor(namespace))
*/
    __pyx_t_5 = __pyx_v_self->_get_value_indexes_length;
    __pyx_t_7 = __pyx_t_5;
    for (__pyx_t_8 = 0; __pyx_t_8 < __pyx_t_7; __pyx_t_8+=1) {
      __pyx_8genexpr2__pyx_v_i = __pyx_t_8;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":313
 *         get_value_indexes = self._get_value_indexes
 *         result = [
 *             self._udfs[get_value_indexes[i]].get_value(self._accumulators[get_value_indexes[i]])             # <<<<<<<<<<<<<<
 *             for i in range(self._get_value_indexes_length)]
 *         if self._named_property_extractor:
*/
      if (unlikely(__pyx_v_self->_udfs == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 313, __pyx_L1_error)
      }
      __pyx_t_4 = __Pyx_PyList_GET_ITEM(__pyx_v_self->_udfs, (__pyx_v_get_value_indexes[__pyx_8genexpr2__pyx_v_i]));
      __Pyx_INCREF(__pyx_t_4);
      if (unlikely(__pyx_v_self->_accumulators == Py_None)) {
        PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
        __PYX_ERR(0, 313, __pyx_L1_error)
      }
      __pyx_t_9 = 0;
      {
        PyObject *__pyx_callargs[2] = {__pyx_t_4, __Pyx_PyList_GET_ITEM(__pyx_v_self->_accumulators, (__pyx_v_get_value_indexes[__pyx_8genexpr2__pyx_v_i]))};
        __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_value, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
        __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
        if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 313, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
      }
      if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 312, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
  } /* exit inner scope */
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":315
 *             self._udfs[get_value_indexes[i]].get_value(self._accumulators[get_value_indexes[i]])
 *             for i in range(self._get_value_indexes_length)]
 *         if self._named_property_extractor:             # <<<<<<<<<<<<<<
 *             result.extend(self._named_property_extractor(namespace))
 *         return result
*/
  __pyx_t_10 = __Pyx_PyObject_IsTrue(__pyx_v_self->_named_property_extractor); if (unlikely((__pyx_t_10 < 0))) __PYX_ERR(0, 315, __pyx_L1_error)
  if (__pyx_t_10) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":316
 *             for i in range(self._get_value_indexes_length)]
 *         if self._named_property_extractor:
 *             result.extend(self._named_property_extractor(namespace))             # <<<<<<<<<<<<<<
 *         return result
 * 
*/
    __pyx_t_2 = NULL;
    __Pyx_INCREF(__pyx_v_self->_named_property_extractor);
    __pyx_t_4 = __pyx_v_self->_named_property_extractor; 
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_namespace};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 316, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_11 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 316, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":315
 *             self._udfs[get_value_indexes[i]].get_value(self._accumulators[get_value_indexes[i]])
 *             for i in range(self._get_value_indexes_length)]
 *         if self._named_property_extractor:             # <<<<<<<<<<<<<<
 *             result.extend(self._named_property_extractor(namespace))
 *         return result
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":317
 *         if self._named_property_extractor:
 *             result.extend(self._named_property_extractor(namespace))
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  __Pyx_TraceReturnValue(__pyx_r, 36, 0, __PYX_ERR(0, 317, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":307
 *             udf.close()
 * 
 *     cpdef list get_value(self, object namespace):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t*get_value_indexes
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 307, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_11get_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_11get_value = {"get_value", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_11get_value, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_11get_value(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_namespace = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_value (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_namespace,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 307, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 307, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "get_value", 0) < 0) __PYX_ERR(0, 307, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("get_value", 1, 1, 1, i); __PYX_ERR(0, 307, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 307, __pyx_L3_error)
    }
    __pyx_v_namespace = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("get_value", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 307, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_10get_value(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self), __pyx_v_namespace);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_10get_value(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, PyObject *__pyx_v_namespace) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25]))
  __Pyx_RefNannySetupContext("get_value", 0);
  __Pyx_TraceStartFunc("get_value (wrapper)", __pyx_f[0], 307, 0, 0, 0, __PYX_ERR(0, 307, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_get_value(__pyx_v_self, __pyx_v_namespace, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 307, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.get_value", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":319
 *         return result
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._filter_args:
 *             free(self._filter_args)
*/

/* Python wrapper */
static void __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_13__dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_13__dealloc__(PyObject *__pyx_v_self) {
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_12__dealloc__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_12__dealloc__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self) {
  __Pyx_TraceDeclarationsFunc
  int __pyx_t_1;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[26]))
  __Pyx_TraceStartFunc("__dealloc__", __pyx_f[0], 319, 0, 0, 0, __PYX_ERR(0, 319, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":320
 * 
 *     def __dealloc__(self):
 *         if self._filter_args:             # <<<<<<<<<<<<<<
 *             free(self._filter_args)
 *         if self._distinct_indexes:
*/
  __pyx_t_1 = (__pyx_v_self->_filter_args != 0);
  if (__pyx_t_1) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":321
 *     def __dealloc__(self):
 *         if self._filter_args:
 *             free(self._filter_args)             # <<<<<<<<<<<<<<
 *         if self._distinct_indexes:
 *             free(self._distinct_indexes)
*/
    free(__pyx_v_self->_filter_args);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":320
 * 
 *     def __dealloc__(self):
 *         if self._filter_args:             # <<<<<<<<<<<<<<
 *             free(self._filter_args)
 *         if self._distinct_indexes:
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":322
 *         if self._filter_args:
 *             free(self._filter_args)
 *         if self._distinct_indexes:             # <<<<<<<<<<<<<<
 *             free(self._distinct_indexes)
 * 
*/
  __pyx_t_1 = (__pyx_v_self->_distinct_indexes != 0);
  if (__pyx_t_1) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":323
 *             free(self._filter_args)
 *         if self._distinct_indexes:
 *             free(self._distinct_indexes)             # <<<<<<<<<<<<<<
 * 
 * cdef class GroupWindowAggFunctionBase:
*/
    free(__pyx_v_self->_distinct_indexes);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":322
 *         if self._filter_args:
 *             free(self._filter_args)
 *         if self._distinct_indexes:             # <<<<<<<<<<<<<<
 *             free(self._distinct_indexes)
 * 
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":319
 *         return result
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         if self._filter_args:
 *             free(self._filter_args)
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 319, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 319, __pyx_L1_error));
  #endif
  __Pyx_WriteUnraisable("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.__dealloc__", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_15__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_15__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_14__reduce_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_14__reduce_cython__(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27]))
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceStartFunc("__reduce_cython__", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
*/
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_mstate_global->__pyx_kp_u_self__distinct_indexes_self__fil, 0, 0);
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_17__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_17__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  CYTHON_UNUSED PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < 0) __PYX_ERR(1, 3, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 3, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 3, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 3, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_16__setstate_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_16__setstate_cython__(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28]))
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceStartFunc("__setstate_cython__", __pyx_f[1], 3, 0, 0, 0, __PYX_ERR(1, 3, __pyx_L1_error));

  /* "(tree fragment)":4
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"             # <<<<<<<<<<<<<<
*/
  __Pyx_Raise(__pyx_builtin_TypeError, __pyx_mstate_global->__pyx_kp_u_self__distinct_indexes_self__fil, 0, 0);
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 3, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":326
 * 
 * cdef class GroupWindowAggFunctionBase:
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  allowed_lateness: int,
 *                  key_selector: RowKeySelector,
*/

/* Python wrapper */
static int __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_allowed_lateness = 0;
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *__pyx_v_key_selector = 0;
  PyObject *__pyx_v_state_backend = 0;
  PyObject *__pyx_v_state_value_coder = 0;
  PyObject *__pyx_v_window_assigner = 0;
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_window_aggregator = 0;
  PyObject *__pyx_v_trigger = 0;
  PyObject *__pyx_v_rowtime_index = 0;
  PyObject *__pyx_v_shift_timezone = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_allowed_lateness,&__pyx_mstate_global->__pyx_n_u_key_selector,&__pyx_mstate_global->__pyx_n_u_state_backend,&__pyx_mstate_global->__pyx_n_u_state_value_coder,&__pyx_mstate_global->__pyx_n_u_window_assigner,&__pyx_mstate_global->__pyx_n_u_window_aggregator,&__pyx_mstate_global->__pyx_n_u_trigger,&__pyx_mstate_global->__pyx_n_u_rowtime_index,&__pyx_mstate_global->__pyx_n_u_shift_timezone,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 326, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 326, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < 0) __PYX_ERR(0, 326, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 9; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 9, 9, i); __PYX_ERR(0, 326, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 9)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 326, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 326, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 326, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 326, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 326, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 326, __pyx_L3_error)
      values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 326, __pyx_L3_error)
      values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 326, __pyx_L3_error)
      values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 326, __pyx_L3_error)
    }
    __pyx_v_allowed_lateness = ((PyObject*)values[0]);
    __pyx_v_key_selector = ((struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *)values[1]);
    __pyx_v_state_backend = values[2];
    __pyx_v_state_value_coder = values[3];
    __pyx_v_window_assigner = values[4];
    __pyx_v_window_aggregator = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)values[5]);
    __pyx_v_trigger = values[6];
    __pyx_v_rowtime_index = ((PyObject*)values[7]);
    __pyx_v_shift_timezone = ((PyObject*)values[8]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 9, 9, __pyx_nargs); __PYX_ERR(0, 326, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_allowed_lateness), (&PyLong_Type), 0, "allowed_lateness", 2))) __PYX_ERR(0, 327, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key_selector), __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector, 0, "key_selector", 0))) __PYX_ERR(0, 328, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_window_aggregator), __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, 0, "window_aggregator", 0))) __PYX_ERR(0, 332, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rowtime_index), (&PyLong_Type), 0, "rowtime_index", 2))) __PYX_ERR(0, 334, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shift_timezone), (&PyUnicode_Type), 0, "shift_timezone", 2))) __PYX_ERR(0, 335, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase___init__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v_allowed_lateness, __pyx_v_key_selector, __pyx_v_state_backend, __pyx_v_state_value_coder, __pyx_v_window_assigner, __pyx_v_window_aggregator, __pyx_v_trigger, __pyx_v_rowtime_index, __pyx_v_shift_timezone);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase___init__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_allowed_lateness, struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *__pyx_v_key_selector, PyObject *__pyx_v_state_backend, PyObject *__pyx_v_state_value_coder, PyObject *__pyx_v_window_assigner, struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v_window_aggregator, PyObject *__pyx_v_trigger, PyObject *__pyx_v_rowtime_index, PyObject *__pyx_v_shift_timezone) {
  int __pyx_r;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[29]))
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_TraceStartFunc("__init__", __pyx_f[0], 326, 0, 0, 0, __PYX_ERR(0, 326, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":336
 *                  rowtime_index: int,
 *                  shift_timezone: str):
 *         self._allowed_lateness = allowed_lateness             # <<<<<<<<<<<<<<
 *         self._key_selector = key_selector
 *         self._state_backend = state_backend
*/
  __pyx_t_1 = __Pyx_PyLong_As_int64_t(__pyx_v_allowed_lateness); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 336, __pyx_L1_error)
  __pyx_v_self->_allowed_lateness = __pyx_t_1;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":337
 *                  shift_timezone: str):
 *         self._allowed_lateness = allowed_lateness
 *         self._key_selector = key_selector             # <<<<<<<<<<<<<<
 *         self._state_backend = state_backend
 *         self._state_value_coder = state_value_coder
*/
  __Pyx_INCREF((PyObject *)__pyx_v_key_selector);
  __Pyx_GIVEREF((PyObject *)__pyx_v_key_selector);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_key_selector);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_key_selector);
  __pyx_v_self->_key_selector = __pyx_v_key_selector;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":338
 *         self._allowed_lateness = allowed_lateness
 *         self._key_selector = key_selector
 *         self._state_backend = state_backend             # <<<<<<<<<<<<<<
 *         self._state_value_coder = state_value_coder
 *         self._window_assigner = window_assigner
*/
  __Pyx_INCREF(__pyx_v_state_backend);
  __Pyx_GIVEREF(__pyx_v_state_backend);
  __Pyx_GOTREF(__pyx_v_self->_state_backend);
  __Pyx_DECREF(__pyx_v_self->_state_backend);
  __pyx_v_self->_state_backend = __pyx_v_state_backend;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":339
 *         self._key_selector = key_selector
 *         self._state_backend = state_backend
 *         self._state_value_coder = state_value_coder             # <<<<<<<<<<<<<<
 *         self._window_assigner = window_assigner
 *         self._window_aggregator = window_aggregator
*/
  __Pyx_INCREF(__pyx_v_state_value_coder);
  __Pyx_GIVEREF(__pyx_v_state_value_coder);
  __Pyx_GOTREF(__pyx_v_self->_state_value_coder);
  __Pyx_DECREF(__pyx_v_self->_state_value_coder);
  __pyx_v_self->_state_value_coder = __pyx_v_state_value_coder;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":340
 *         self._state_backend = state_backend
 *         self._state_value_coder = state_value_coder
 *         self._window_assigner = window_assigner             # <<<<<<<<<<<<<<
 *         self._window_aggregator = window_aggregator
 *         self._rowtime_index = rowtime_index
*/
  __Pyx_INCREF(__pyx_v_window_assigner);
  __Pyx_GIVEREF(__pyx_v_window_assigner);
  __Pyx_GOTREF(__pyx_v_self->_window_assigner);
  __Pyx_DECREF(__pyx_v_self->_window_assigner);
  __pyx_v_self->_window_assigner = __pyx_v_window_assigner;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":341
 *         self._state_value_coder = state_value_coder
 *         self._window_assigner = window_assigner
 *         self._window_aggregator = window_aggregator             # <<<<<<<<<<<<<<
 *         self._rowtime_index = rowtime_index
 *         self._shift_timezone = shift_timezone
*/
  __Pyx_INCREF((PyObject *)__pyx_v_window_aggregator);
  __Pyx_GIVEREF((PyObject *)__pyx_v_window_aggregator);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->_window_aggregator);
  __Pyx_DECREF((PyObject *)__pyx_v_self->_window_aggregator);
  __pyx_v_self->_window_aggregator = __pyx_v_window_aggregator;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":342
 *         self._window_assigner = window_assigner
 *         self._window_aggregator = window_aggregator
 *         self._rowtime_index = rowtime_index             # <<<<<<<<<<<<<<
 *         self._shift_timezone = shift_timezone
 *         self._window_function = None  # type: InternalWindowProcessFunction[K, W]
*/
  __pyx_t_2 = __Pyx_PyLong_As_int(__pyx_v_rowtime_index); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 342, __pyx_L1_error)
  __pyx_v_self->_rowtime_index = __pyx_t_2;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":343
 *         self._window_aggregator = window_aggregator
 *         self._rowtime_index = rowtime_index
 *         self._shift_timezone = shift_timezone             # <<<<<<<<<<<<<<
 *         self._window_function = None  # type: InternalWindowProcessFunction[K, W]
 *         self._internal_timer_service = None  # type: LegacyInternalTimerServiceImpl
*/
  __Pyx_INCREF(__pyx_v_shift_timezone);
  __Pyx_GIVEREF(__pyx_v_shift_timezone);
  __Pyx_GOTREF(__pyx_v_self->_shift_timezone);
  __Pyx_DECREF(__pyx_v_self->_shift_timezone);
  __pyx_v_self->_shift_timezone = __pyx_v_shift_timezone;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":344
 *         self._rowtime_index = rowtime_index
 *         self._shift_timezone = shift_timezone
 *         self._window_function = None  # type: InternalWindowProcessFunction[K, W]             # <<<<<<<<<<<<<<
 *         self._internal_timer_service = None  # type: LegacyInternalTimerServiceImpl
 *         self._window_context = None  # type: WindowContext
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_window_function);
  __Pyx_DECREF(__pyx_v_self->_window_function);
  __pyx_v_self->_window_function = Py_None;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":345
 *         self._shift_timezone = shift_timezone
 *         self._window_function = None  # type: InternalWindowProcessFunction[K, W]
 *         self._internal_timer_service = None  # type: LegacyInternalTimerServiceImpl             # <<<<<<<<<<<<<<
 *         self._window_context = None  # type: WindowContext
 *         self._trigger = trigger
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_internal_timer_service);
  __Pyx_DECREF(__pyx_v_self->_internal_timer_service);
  __pyx_v_self->_internal_timer_service = Py_None;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":346
 *         self._window_function = None  # type: InternalWindowProcessFunction[K, W]
 *         self._internal_timer_service = None  # type: LegacyInternalTimerServiceImpl
 *         self._window_context = None  # type: WindowContext             # <<<<<<<<<<<<<<
 *         self._trigger = trigger
 *         self._trigger_context = None  # type: TriggerContext
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_window_context);
  __Pyx_DECREF(__pyx_v_self->_window_context);
  __pyx_v_self->_window_context = Py_None;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":347
 *         self._internal_timer_service = None  # type: LegacyInternalTimerServiceImpl
 *         self._window_context = None  # type: WindowContext
 *         self._trigger = trigger             # <<<<<<<<<<<<<<
 *         self._trigger_context = None  # type: TriggerContext
 *         self._window_state = self._state_backend.get_value_state("window_state", state_value_coder)
*/
  __Pyx_INCREF(__pyx_v_trigger);
  __Pyx_GIVEREF(__pyx_v_trigger);
  __Pyx_GOTREF(__pyx_v_self->_trigger);
  __Pyx_DECREF(__pyx_v_self->_trigger);
  __pyx_v_self->_trigger = __pyx_v_trigger;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":348
 *         self._window_context = None  # type: WindowContext
 *         self._trigger = trigger
 *         self._trigger_context = None  # type: TriggerContext             # <<<<<<<<<<<<<<
 *         self._window_state = self._state_backend.get_value_state("window_state", state_value_coder)
 * 
*/
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __Pyx_GOTREF(__pyx_v_self->_trigger_context);
  __Pyx_DECREF(__pyx_v_self->_trigger_context);
  __pyx_v_self->_trigger_context = Py_None;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":349
 *         self._trigger = trigger
 *         self._trigger_context = None  # type: TriggerContext
 *         self._window_state = self._state_backend.get_value_state("window_state", state_value_coder)             # <<<<<<<<<<<<<<
 * 
 *     cpdef void open(self, object function_context):
*/
  __pyx_t_4 = __pyx_v_self->_state_backend;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_mstate_global->__pyx_n_u_window_state, __pyx_v_state_value_coder};
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_value_state, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 349, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_window_state);
  __Pyx_DECREF(__pyx_v_self->_window_state);
  __pyx_v_self->_window_state = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":326
 * 
 * cdef class GroupWindowAggFunctionBase:
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  allowed_lateness: int,
 *                  key_selector: RowKeySelector,
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_TraceReturnCValue(__pyx_r, __Pyx_Owned_Py_None, 0, 0, __PYX_ERR(0, 326, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 326, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":351
 *         self._window_state = self._state_backend.get_value_state("window_state", state_value_coder)
 * 
 *     cpdef void open(self, object function_context):             # <<<<<<<<<<<<<<
 *         self._internal_timer_service = LegacyInternalTimerServiceImpl(self._state_backend)
 *         self._window_aggregator.open(
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_3open(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_open(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_function_context, int __pyx_skip_dispatch) {
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30]))
  __Pyx_RefNannySetupContext("open", 0);
  __Pyx_TraceStartFunc("open", __pyx_f[0], 351, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 351, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_open); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_3open)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_function_context};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":352
 * 
 *     cpdef void open(self, object function_context):
 *         self._internal_timer_service = LegacyInternalTimerServiceImpl(self._state_backend)             # <<<<<<<<<<<<<<
 *         self._window_aggregator.open(
 *             PerWindowStateDataViewStore(function_context, self._state_backend))
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_LegacyInternalTimerServiceImpl); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_4))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_4);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_self->_state_backend};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->_internal_timer_service);
  __Pyx_DECREF(__pyx_v_self->_internal_timer_service);
  __pyx_v_self->_internal_timer_service = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":354
 *         self._internal_timer_service = LegacyInternalTimerServiceImpl(self._state_backend)
 *         self._window_aggregator.open(
 *             PerWindowStateDataViewStore(function_context, self._state_backend))             # <<<<<<<<<<<<<<
 * 
 *         if isinstance(self._window_assigner, PanedWindowAssigner):
*/
  __pyx_t_4 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PerWindowStateDataViewStore); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_4);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_function_context, __pyx_v_self->_state_backend};
    __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 354, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":353
 *     cpdef void open(self, object function_context):
 *         self._internal_timer_service = LegacyInternalTimerServiceImpl(self._state_backend)
 *         self._window_aggregator.open(             # <<<<<<<<<<<<<<
 *             PerWindowStateDataViewStore(function_context, self._state_backend))
 * 
*/
  ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self->_window_aggregator->__pyx_vtab)->open(__pyx_v_self->_window_aggregator, __pyx_t_1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 353, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":356
 *             PerWindowStateDataViewStore(function_context, self._state_backend))
 * 
 *         if isinstance(self._window_assigner, PanedWindowAssigner):             # <<<<<<<<<<<<<<
 *             self._window_function = PanedWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
*/
  __pyx_t_1 = __pyx_v_self->_window_assigner;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PanedWindowAssigner); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyObject_IsInstance(__pyx_t_1, __pyx_t_2); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 356, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_6) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":357
 * 
 *         if isinstance(self._window_assigner, PanedWindowAssigner):
 *             self._window_function = PanedWindowProcessFunction(             # <<<<<<<<<<<<<<
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         elif isinstance(self._window_assigner, MergingWindowAssigner):
*/
    __pyx_t_1 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_PanedWindowProcessFunction); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 357, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":358
 *         if isinstance(self._window_assigner, PanedWindowAssigner):
 *             self._window_function = PanedWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)             # <<<<<<<<<<<<<<
 *         elif isinstance(self._window_assigner, MergingWindowAssigner):
 *             self._window_function = MergingWindowProcessFunction(
*/
    __pyx_t_3 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_allowed_lateness); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 358, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_4);
      assert(__pyx_t_1);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_t_3, __pyx_v_self->_window_assigner, ((PyObject *)__pyx_v_self->_window_aggregator)};
      __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (4-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 357, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":357
 * 
 *         if isinstance(self._window_assigner, PanedWindowAssigner):
 *             self._window_function = PanedWindowProcessFunction(             # <<<<<<<<<<<<<<
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         elif isinstance(self._window_assigner, MergingWindowAssigner):
*/
    __Pyx_GIVEREF(__pyx_t_2);
    __Pyx_GOTREF(__pyx_v_self->_window_function);
    __Pyx_DECREF(__pyx_v_self->_window_function);
    __pyx_v_self->_window_function = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":356
 *             PerWindowStateDataViewStore(function_context, self._state_backend))
 * 
 *         if isinstance(self._window_assigner, PanedWindowAssigner):             # <<<<<<<<<<<<<<
 *             self._window_function = PanedWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
*/
    goto __pyx_L3;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":359
 *             self._window_function = PanedWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         elif isinstance(self._window_assigner, MergingWindowAssigner):             # <<<<<<<<<<<<<<
 *             self._window_function = MergingWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator,
*/
  __pyx_t_2 = __pyx_v_self->_window_assigner;
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_MergingWindowAssigner); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = PyObject_IsInstance(__pyx_t_2, __pyx_t_4); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(0, 359, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_6) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":360
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         elif isinstance(self._window_assigner, MergingWindowAssigner):
 *             self._window_function = MergingWindowProcessFunction(             # <<<<<<<<<<<<<<
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator,
 *                 self._state_backend)
*/
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_MergingWindowProcessFunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 360, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":361
 *         elif isinstance(self._window_assigner, MergingWindowAssigner):
 *             self._window_function = MergingWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator,             # <<<<<<<<<<<<<<
 *                 self._state_backend)
 *         else:
*/
    __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_allowed_lateness); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 361, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":362
 *             self._window_function = MergingWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator,
 *                 self._state_backend)             # <<<<<<<<<<<<<<
 *         else:
 *             self._window_function = GeneralWindowProcessFunction(
*/
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_3, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[5] = {__pyx_t_2, __pyx_t_1, __pyx_v_self->_window_assigner, ((PyObject *)__pyx_v_self->_window_aggregator), __pyx_v_self->_state_backend};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_3, __pyx_callargs+__pyx_t_5, (5-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 360, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":360
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         elif isinstance(self._window_assigner, MergingWindowAssigner):
 *             self._window_function = MergingWindowProcessFunction(             # <<<<<<<<<<<<<<
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator,
 *                 self._state_backend)
*/
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->_window_function);
    __Pyx_DECREF(__pyx_v_self->_window_function);
    __pyx_v_self->_window_function = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":359
 *             self._window_function = PanedWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         elif isinstance(self._window_assigner, MergingWindowAssigner):             # <<<<<<<<<<<<<<
 *             self._window_function = MergingWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator,
*/
    goto __pyx_L3;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":364
 *                 self._state_backend)
 *         else:
 *             self._window_function = GeneralWindowProcessFunction(             # <<<<<<<<<<<<<<
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         self._trigger_context = TriggerContext(
*/
  /*else*/ {
    __pyx_t_3 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_GeneralWindowProcessFunction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 364, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":365
 *         else:
 *             self._window_function = GeneralWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)             # <<<<<<<<<<<<<<
 *         self._trigger_context = TriggerContext(
 *             self._trigger, self._internal_timer_service, self._state_backend)
*/
    __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_allowed_lateness); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 365, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_1);
      assert(__pyx_t_3);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
      __pyx_t_5 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[4] = {__pyx_t_3, __pyx_t_2, __pyx_v_self->_window_assigner, ((PyObject *)__pyx_v_self->_window_aggregator)};
      __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_5, (4-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 364, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":364
 *                 self._state_backend)
 *         else:
 *             self._window_function = GeneralWindowProcessFunction(             # <<<<<<<<<<<<<<
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         self._trigger_context = TriggerContext(
*/
    __Pyx_GIVEREF(__pyx_t_4);
    __Pyx_GOTREF(__pyx_v_self->_window_function);
    __Pyx_DECREF(__pyx_v_self->_window_function);
    __pyx_v_self->_window_function = __pyx_t_4;
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":366
 *             self._window_function = GeneralWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         self._trigger_context = TriggerContext(             # <<<<<<<<<<<<<<
 *             self._trigger, self._internal_timer_service, self._state_backend)
 *         self._trigger_context.open()
*/
  __pyx_t_1 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_TriggerContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 366, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":367
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         self._trigger_context = TriggerContext(
 *             self._trigger, self._internal_timer_service, self._state_backend)             # <<<<<<<<<<<<<<
 *         self._trigger_context.open()
 *         self._window_context = WindowContext(
*/
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    assert(__pyx_t_1);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_1);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_2, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[4] = {__pyx_t_1, __pyx_v_self->_trigger, __pyx_v_self->_internal_timer_service, __pyx_v_self->_state_backend};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_2, __pyx_callargs+__pyx_t_5, (4-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 366, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":366
 *             self._window_function = GeneralWindowProcessFunction(
 *                 self._allowed_lateness, self._window_assigner, self._window_aggregator)
 *         self._trigger_context = TriggerContext(             # <<<<<<<<<<<<<<
 *             self._trigger, self._internal_timer_service, self._state_backend)
 *         self._trigger_context.open()
*/
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_trigger_context);
  __Pyx_DECREF(__pyx_v_self->_trigger_context);
  __pyx_v_self->_trigger_context = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":368
 *         self._trigger_context = TriggerContext(
 *             self._trigger, self._internal_timer_service, self._state_backend)
 *         self._trigger_context.open()             # <<<<<<<<<<<<<<
 *         self._window_context = WindowContext(
 *             self, self._trigger_context, self._state_backend, self._state_value_coder,
*/
  __pyx_t_2 = __pyx_v_self->_trigger_context;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_open, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 368, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":369
 *             self._trigger, self._internal_timer_service, self._state_backend)
 *         self._trigger_context.open()
 *         self._window_context = WindowContext(             # <<<<<<<<<<<<<<
 *             self, self._trigger_context, self._state_backend, self._state_value_coder,
 *             self._internal_timer_service, self._window_assigner.is_event_time())
*/
  __pyx_t_2 = NULL;
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_WindowContext); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 369, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":371
 *         self._window_context = WindowContext(
 *             self, self._trigger_context, self._state_backend, self._state_value_coder,
 *             self._internal_timer_service, self._window_assigner.is_event_time())             # <<<<<<<<<<<<<<
 *         self._window_function.open(self._window_context)
 * 
*/
  __pyx_t_7 = __pyx_v_self->_window_assigner;
  __Pyx_INCREF(__pyx_t_7);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_7, NULL};
    __pyx_t_3 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_event_time, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 371, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_5 = 1;
  #if CYTHON_UNPACK_METHODS
  if (unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_1);
    assert(__pyx_t_2);
    PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_2);
    __Pyx_INCREF(__pyx__function);
    __Pyx_DECREF_SET(__pyx_t_1, __pyx__function);
    __pyx_t_5 = 0;
  }
  #endif
  {
    PyObject *__pyx_callargs[7] = {__pyx_t_2, ((PyObject *)__pyx_v_self), __pyx_v_self->_trigger_context, __pyx_v_self->_state_backend, __pyx_v_self->_state_value_coder, __pyx_v_self->_internal_timer_service, __pyx_t_3};
    __pyx_t_4 = __Pyx_PyObject_FastCall(__pyx_t_1, __pyx_callargs+__pyx_t_5, (7-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 369, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":369
 *             self._trigger, self._internal_timer_service, self._state_backend)
 *         self._trigger_context.open()
 *         self._window_context = WindowContext(             # <<<<<<<<<<<<<<
 *             self, self._trigger_context, self._state_backend, self._state_value_coder,
 *             self._internal_timer_service, self._window_assigner.is_event_time())
*/
  __Pyx_GIVEREF(__pyx_t_4);
  __Pyx_GOTREF(__pyx_v_self->_window_context);
  __Pyx_DECREF(__pyx_v_self->_window_context);
  __pyx_v_self->_window_context = __pyx_t_4;
  __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":372
 *             self, self._trigger_context, self._state_backend, self._state_value_coder,
 *             self._internal_timer_service, self._window_assigner.is_event_time())
 *         self._window_function.open(self._window_context)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list process_element(self, InternalRow input_row):
*/
  __pyx_t_1 = __pyx_v_self->_window_function;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_self->_window_context};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_open, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 372, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":351
 *         self._window_state = self._state_backend.get_value_state("window_state", state_value_coder)
 * 
 *     cpdef void open(self, object function_context):             # <<<<<<<<<<<<<<
 *         self._internal_timer_service = LegacyInternalTimerServiceImpl(self._state_backend)
 *         self._window_aggregator.open(
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 351, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 351, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_3open(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_3open = {"open", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_3open, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_3open(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_function_context = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("open (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_function_context,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 351, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 351, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "open", 0) < 0) __PYX_ERR(0, 351, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("open", 1, 1, 1, i); __PYX_ERR(0, 351, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 351, __pyx_L3_error)
    }
    __pyx_v_function_context = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("open", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 351, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_2open(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v_function_context);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_2open(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_function_context) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30]))
  __Pyx_RefNannySetupContext("open", 0);
  __Pyx_TraceStartFunc("open (wrapper)", __pyx_f[0], 351, 0, 0, 0, __PYX_ERR(0, 351, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_open(__pyx_v_self, __pyx_v_function_context, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 351, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 351, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.open", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":374
 *         self._window_function.open(self._window_context)
 * 
 *     cpdef list process_element(self, InternalRow input_row):             # <<<<<<<<<<<<<<
 *         cdef list input_value, current_key, affected_windows, acc, actual_windows, result
 *         cdef int64_t timestamp, seconds, microseconds_of_second, milliseconds
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_5process_element(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_process_element(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_row, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_input_value = 0;
  PyObject *__pyx_v_current_key = 0;
  PyObject *__pyx_v_affected_windows = 0;
  PyObject *__pyx_v_acc = 0;
  PyObject *__pyx_v_actual_windows = 0;
  PyObject *__pyx_v_result = 0;
  int64_t __pyx_v_timestamp;
  int64_t __pyx_v_seconds;
  int64_t __pyx_v_microseconds_of_second;
  int64_t __pyx_v_milliseconds;
  PyObject *__pyx_v_date_time = 0;
  PyObject *__pyx_v_window = 0;
  int __pyx_v_trigger_result;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int64_t __pyx_t_9;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31]))
  __Pyx_RefNannySetupContext("process_element", 0);
  __Pyx_TraceStartFunc("process_element", __pyx_f[0], 374, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 374, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_process_element); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_5process_element)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, ((PyObject *)__pyx_v_input_row)};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 374, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 374, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 374, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":379
 *         cdef object date_time, window
 *         cdef bint trigger_result
 *         input_value = input_row.values             # <<<<<<<<<<<<<<
 *         current_key = self._key_selector.get_key(input_value)
 *         self._state_backend.set_current_key(current_key)
*/
  __pyx_t_1 = __pyx_v_input_row->values;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_v_input_value = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":380
 *         cdef bint trigger_result
 *         input_value = input_row.values
 *         current_key = self._key_selector.get_key(input_value)             # <<<<<<<<<<<<<<
 *         self._state_backend.set_current_key(current_key)
 *         if self._window_assigner.is_event_time():
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *)__pyx_v_self->_key_selector->__pyx_vtab)->get_key(__pyx_v_self->_key_selector, __pyx_v_input_value); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 380, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_current_key = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":381
 *         input_value = input_row.values
 *         current_key = self._key_selector.get_key(input_value)
 *         self._state_backend.set_current_key(current_key)             # <<<<<<<<<<<<<<
 *         if self._window_assigner.is_event_time():
 *             date_time = input_value[self._rowtime_index]
*/
  __pyx_t_2 = __pyx_v_self->_state_backend;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_current_key};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_current_key, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 381, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":382
 *         current_key = self._key_selector.get_key(input_value)
 *         self._state_backend.set_current_key(current_key)
 *         if self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             date_time = input_value[self._rowtime_index]
 *             seconds = int(date_time.replace(tzinfo=datetime.timezone.utc).timestamp())
*/
  __pyx_t_2 = __pyx_v_self->_window_assigner;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_event_time, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 382, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 < 0))) __PYX_ERR(0, 382, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_6) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":383
 *         self._state_backend.set_current_key(current_key)
 *         if self._window_assigner.is_event_time():
 *             date_time = input_value[self._rowtime_index]             # <<<<<<<<<<<<<<
 *             seconds = int(date_time.replace(tzinfo=datetime.timezone.utc).timestamp())
 *             microseconds_of_second = date_time.microsecond
*/
    if (unlikely(__pyx_v_input_value == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(0, 383, __pyx_L1_error)
    }
    __pyx_t_1 = __Pyx_PyList_GET_ITEM(__pyx_v_input_value, __pyx_v_self->_rowtime_index);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v_date_time = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":384
 *         if self._window_assigner.is_event_time():
 *             date_time = input_value[self._rowtime_index]
 *             seconds = int(date_time.replace(tzinfo=datetime.timezone.utc).timestamp())             # <<<<<<<<<<<<<<
 *             microseconds_of_second = date_time.microsecond
 *             milliseconds = seconds * 1000 + microseconds_of_second // 1000
*/
    __pyx_t_3 = __pyx_v_date_time;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_datetime); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_mstate_global->__pyx_n_u_timezone); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_mstate_global->__pyx_n_u_utc); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_3, NULL};
      __pyx_t_8 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_tzinfo, __pyx_t_7, __pyx_t_8, __pyx_callargs+1, 0) < 0) __PYX_ERR(0, 384, __pyx_L1_error)
      __pyx_t_4 = __Pyx_Object_VectorcallMethod_CallFromBuilder(__pyx_mstate_global->__pyx_n_u_replace, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_8);
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_t_2 = __pyx_t_4;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_timestamp, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 384, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_4 = __Pyx_PyNumber_Int(__pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = __Pyx_PyLong_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_9 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 384, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_seconds = __pyx_t_9;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":385
 *             date_time = input_value[self._rowtime_index]
 *             seconds = int(date_time.replace(tzinfo=datetime.timezone.utc).timestamp())
 *             microseconds_of_second = date_time.microsecond             # <<<<<<<<<<<<<<
 *             milliseconds = seconds * 1000 + microseconds_of_second // 1000
 *             timestamp = milliseconds
*/
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_date_time, __pyx_mstate_global->__pyx_n_u_microsecond); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_9 = __Pyx_PyLong_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_9 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 385, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_microseconds_of_second = __pyx_t_9;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":386
 *             seconds = int(date_time.replace(tzinfo=datetime.timezone.utc).timestamp())
 *             microseconds_of_second = date_time.microsecond
 *             milliseconds = seconds * 1000 + microseconds_of_second // 1000             # <<<<<<<<<<<<<<
 *             timestamp = milliseconds
 *         else:
*/
    __pyx_v_milliseconds = ((__pyx_v_seconds * 0x3E8) + (__pyx_v_microseconds_of_second / 0x3E8));

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":387
 *             microseconds_of_second = date_time.microsecond
 *             milliseconds = seconds * 1000 + microseconds_of_second // 1000
 *             timestamp = milliseconds             # <<<<<<<<<<<<<<
 *         else:
 *             timestamp = self._internal_timer_service.current_processing_time()
*/
    __pyx_v_timestamp = __pyx_v_milliseconds;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":382
 *         current_key = self._key_selector.get_key(input_value)
 *         self._state_backend.set_current_key(current_key)
 *         if self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             date_time = input_value[self._rowtime_index]
 *             seconds = int(date_time.replace(tzinfo=datetime.timezone.utc).timestamp())
*/
    goto __pyx_L3;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":389
 *             timestamp = milliseconds
 *         else:
 *             timestamp = self._internal_timer_service.current_processing_time()             # <<<<<<<<<<<<<<
 * 
 *         timestamp = self.to_utc_timestamp_mills(timestamp)
*/
  /*else*/ {
    __pyx_t_1 = __pyx_v_self->_internal_timer_service;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_current_processing_time, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 389, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_t_9 = __Pyx_PyLong_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_9 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 389, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_timestamp = __pyx_t_9;
  }
  __pyx_L3:;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":391
 *             timestamp = self._internal_timer_service.current_processing_time()
 * 
 *         timestamp = self.to_utc_timestamp_mills(timestamp)             # <<<<<<<<<<<<<<
 * 
 *         # the windows which the input row should be placed into
*/
  __pyx_t_9 = ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self->__pyx_vtab)->to_utc_timestamp_mills(__pyx_v_self, __pyx_v_timestamp, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 391, __pyx_L1_error)
  __pyx_v_timestamp = __pyx_t_9;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":394
 * 
 *         # the windows which the input row should be placed into
 *         affected_windows = self._window_function.assign_state_namespace(input_value, timestamp)             # <<<<<<<<<<<<<<
 *         for window in affected_windows:
 *             self._window_state.set_current_namespace(window)
*/
  __pyx_t_1 = __pyx_v_self->_window_function;
  __Pyx_INCREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_timestamp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_1, __pyx_v_input_value, __pyx_t_2};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_assign_state_namespace, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 394, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (!(likely(PyList_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_4))) __PYX_ERR(0, 394, __pyx_L1_error)
  __pyx_v_affected_windows = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":395
 *         # the windows which the input row should be placed into
 *         affected_windows = self._window_function.assign_state_namespace(input_value, timestamp)
 *         for window in affected_windows:             # <<<<<<<<<<<<<<
 *             self._window_state.set_current_namespace(window)
 *             acc = self._window_state.value()  # type: List
*/
  if (unlikely(__pyx_v_affected_windows == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 395, __pyx_L1_error)
  }
  __pyx_t_4 = __pyx_v_affected_windows; __Pyx_INCREF(__pyx_t_4);
  __pyx_t_10 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 395, __pyx_L1_error)
      #endif
      if (__pyx_t_10 >= __pyx_temp) break;
    }
    __pyx_t_2 = __Pyx_PyList_GetItemRef(__pyx_t_4, __pyx_t_10);
    ++__pyx_t_10;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 395, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_window, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":396
 *         affected_windows = self._window_function.assign_state_namespace(input_value, timestamp)
 *         for window in affected_windows:
 *             self._window_state.set_current_namespace(window)             # <<<<<<<<<<<<<<
 *             acc = self._window_state.value()  # type: List
 *             if acc is None:
*/
    __pyx_t_1 = __pyx_v_self->_window_state;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_window};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_current_namespace, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 396, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":397
 *         for window in affected_windows:
 *             self._window_state.set_current_namespace(window)
 *             acc = self._window_state.value()  # type: List             # <<<<<<<<<<<<<<
 *             if acc is None:
 *                 acc = self._window_aggregator.create_accumulators()
*/
    __pyx_t_1 = __pyx_v_self->_window_state;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, NULL};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_value, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 397, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 397, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_acc, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":398
 *             self._window_state.set_current_namespace(window)
 *             acc = self._window_state.value()  # type: List
 *             if acc is None:             # <<<<<<<<<<<<<<
 *                 acc = self._window_aggregator.create_accumulators()
 *             self._window_aggregator.set_accumulators(window, acc)
*/
    __pyx_t_6 = (__pyx_v_acc == ((PyObject*)Py_None));
    if (__pyx_t_6) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":399
 *             acc = self._window_state.value()  # type: List
 *             if acc is None:
 *                 acc = self._window_aggregator.create_accumulators()             # <<<<<<<<<<<<<<
 *             self._window_aggregator.set_accumulators(window, acc)
 * 
*/
      __pyx_t_2 = ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self->_window_aggregator->__pyx_vtab)->create_accumulators(__pyx_v_self->_window_aggregator, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 399, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF_SET(__pyx_v_acc, ((PyObject*)__pyx_t_2));
      __pyx_t_2 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":398
 *             self._window_state.set_current_namespace(window)
 *             acc = self._window_state.value()  # type: List
 *             if acc is None:             # <<<<<<<<<<<<<<
 *                 acc = self._window_aggregator.create_accumulators()
 *             self._window_aggregator.set_accumulators(window, acc)
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":400
 *             if acc is None:
 *                 acc = self._window_aggregator.create_accumulators()
 *             self._window_aggregator.set_accumulators(window, acc)             # <<<<<<<<<<<<<<
 * 
 *             if input_row.is_accumulate_msg():
*/
    ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self->_window_aggregator->__pyx_vtab)->set_accumulators(__pyx_v_self->_window_aggregator, __pyx_v_window, __pyx_v_acc, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 400, __pyx_L1_error)

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":402
 *             self._window_aggregator.set_accumulators(window, acc)
 * 
 *             if input_row.is_accumulate_msg():             # <<<<<<<<<<<<<<
 *                 self._window_aggregator.accumulate(input_row)
 *             else:
*/
    __pyx_t_6 = ((struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)__pyx_v_input_row->__pyx_vtab)->is_accumulate_msg(__pyx_v_input_row); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 402, __pyx_L1_error)
    if (__pyx_t_6) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":403
 * 
 *             if input_row.is_accumulate_msg():
 *                 self._window_aggregator.accumulate(input_row)             # <<<<<<<<<<<<<<
 *             else:
 *                 self._window_aggregator.retract(input_row)
*/
      ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self->_window_aggregator->__pyx_vtab)->accumulate(__pyx_v_self->_window_aggregator, __pyx_v_input_row); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 403, __pyx_L1_error)

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":402
 *             self._window_aggregator.set_accumulators(window, acc)
 * 
 *             if input_row.is_accumulate_msg():             # <<<<<<<<<<<<<<
 *                 self._window_aggregator.accumulate(input_row)
 *             else:
*/
      goto __pyx_L7;
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":405
 *                 self._window_aggregator.accumulate(input_row)
 *             else:
 *                 self._window_aggregator.retract(input_row)             # <<<<<<<<<<<<<<
 *             acc = self._window_aggregator.get_accumulators()
 *             self._window_state.update(acc)
*/
    /*else*/ {
      ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self->_window_aggregator->__pyx_vtab)->retract(__pyx_v_self->_window_aggregator, __pyx_v_input_row); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 405, __pyx_L1_error)
    }
    __pyx_L7:;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":406
 *             else:
 *                 self._window_aggregator.retract(input_row)
 *             acc = self._window_aggregator.get_accumulators()             # <<<<<<<<<<<<<<
 *             self._window_state.update(acc)
 * 
*/
    __pyx_t_2 = ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self->_window_aggregator->__pyx_vtab)->get_accumulators(__pyx_v_self->_window_aggregator); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 406, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_acc, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":407
 *                 self._window_aggregator.retract(input_row)
 *             acc = self._window_aggregator.get_accumulators()
 *             self._window_state.update(acc)             # <<<<<<<<<<<<<<
 * 
 *         # the actual window which the input row is belongs to
*/
    __pyx_t_1 = __pyx_v_self->_window_state;
    __Pyx_INCREF(__pyx_t_1);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_1, __pyx_v_acc};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 407, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":395
 *         # the windows which the input row should be placed into
 *         affected_windows = self._window_function.assign_state_namespace(input_value, timestamp)
 *         for window in affected_windows:             # <<<<<<<<<<<<<<
 *             self._window_state.set_current_namespace(window)
 *             acc = self._window_state.value()  # type: List
*/
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":410
 * 
 *         # the actual window which the input row is belongs to
 *         actual_windows = self._window_function.assign_actual_windows(input_value, timestamp)             # <<<<<<<<<<<<<<
 *         result = []
 *         for window in actual_windows:
*/
  __pyx_t_2 = __pyx_v_self->_window_function;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_timestamp); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_2, __pyx_v_input_value, __pyx_t_1};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_assign_actual_windows, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  if (!(likely(PyList_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_4))) __PYX_ERR(0, 410, __pyx_L1_error)
  __pyx_v_actual_windows = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":411
 *         # the actual window which the input row is belongs to
 *         actual_windows = self._window_function.assign_actual_windows(input_value, timestamp)
 *         result = []             # <<<<<<<<<<<<<<
 *         for window in actual_windows:
 *             self._trigger_context.window = window
*/
  __pyx_t_4 = PyList_New(0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 411, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_v_result = ((PyObject*)__pyx_t_4);
  __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":412
 *         actual_windows = self._window_function.assign_actual_windows(input_value, timestamp)
 *         result = []
 *         for window in actual_windows:             # <<<<<<<<<<<<<<
 *             self._trigger_context.window = window
 *             trigger_result = self._trigger_context.on_element(input_row, timestamp)
*/
  if (unlikely(__pyx_v_actual_windows == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 412, __pyx_L1_error)
  }
  __pyx_t_4 = __pyx_v_actual_windows; __Pyx_INCREF(__pyx_t_4);
  __pyx_t_10 = 0;
  for (;;) {
    {
      Py_ssize_t __pyx_temp = __Pyx_PyList_GET_SIZE(__pyx_t_4);
      #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely((__pyx_temp < 0))) __PYX_ERR(0, 412, __pyx_L1_error)
      #endif
      if (__pyx_t_10 >= __pyx_temp) break;
    }
    __pyx_t_1 = __Pyx_PyList_GetItemRef(__pyx_t_4, __pyx_t_10);
    ++__pyx_t_10;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_window, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":413
 *         result = []
 *         for window in actual_windows:
 *             self._trigger_context.window = window             # <<<<<<<<<<<<<<
 *             trigger_result = self._trigger_context.on_element(input_row, timestamp)
 *             if trigger_result:
*/
    if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_trigger_context, __pyx_mstate_global->__pyx_n_u_window, __pyx_v_window) < 0) __PYX_ERR(0, 413, __pyx_L1_error)

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":414
 *         for window in actual_windows:
 *             self._trigger_context.window = window
 *             trigger_result = self._trigger_context.on_element(input_row, timestamp)             # <<<<<<<<<<<<<<
 *             if trigger_result:
 *                 result.append(self._emit_window_result(current_key, window))
*/
    __pyx_t_2 = __pyx_v_self->_trigger_context;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_8 = __Pyx_PyLong_From_int64_t(__pyx_v_timestamp); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_2, ((PyObject *)__pyx_v_input_row), __pyx_t_8};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_on_element, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 414, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_6 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 414, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_trigger_result = __pyx_t_6;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":415
 *             self._trigger_context.window = window
 *             trigger_result = self._trigger_context.on_element(input_row, timestamp)
 *             if trigger_result:             # <<<<<<<<<<<<<<
 *                 result.append(self._emit_window_result(current_key, window))
 *             self._register_cleanup_timer(window)
*/
    if (__pyx_v_trigger_result) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":416
 *             trigger_result = self._trigger_context.on_element(input_row, timestamp)
 *             if trigger_result:
 *                 result.append(self._emit_window_result(current_key, window))             # <<<<<<<<<<<<<<
 *             self._register_cleanup_timer(window)
 *         return result
*/
      __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self->__pyx_vtab)->_emit_window_result(__pyx_v_self, __pyx_v_current_key, __pyx_v_window)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_11 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(0, 416, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":415
 *             self._trigger_context.window = window
 *             trigger_result = self._trigger_context.on_element(input_row, timestamp)
 *             if trigger_result:             # <<<<<<<<<<<<<<
 *                 result.append(self._emit_window_result(current_key, window))
 *             self._register_cleanup_timer(window)
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":417
 *             if trigger_result:
 *                 result.append(self._emit_window_result(current_key, window))
 *             self._register_cleanup_timer(window)             # <<<<<<<<<<<<<<
 *         return result
 * 
*/
    ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self->__pyx_vtab)->_register_cleanup_timer(__pyx_v_self, __pyx_v_window); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 417, __pyx_L1_error)

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":412
 *         actual_windows = self._window_function.assign_actual_windows(input_value, timestamp)
 *         result = []
 *         for window in actual_windows:             # <<<<<<<<<<<<<<
 *             self._trigger_context.window = window
 *             trigger_result = self._trigger_context.on_element(input_row, timestamp)
*/
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":418
 *                 result.append(self._emit_window_result(current_key, window))
 *             self._register_cleanup_timer(window)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef void process_watermark(self, int64_t watermark):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  __Pyx_TraceReturnValue(__pyx_r, 156, 0, __PYX_ERR(0, 418, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":374
 *         self._window_function.open(self._window_context)
 * 
 *     cpdef list process_element(self, InternalRow input_row):             # <<<<<<<<<<<<<<
 *         cdef list input_value, current_key, affected_windows, acc, actual_windows, result
 *         cdef int64_t timestamp, seconds, microseconds_of_second, milliseconds
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 374, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.process_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_input_value);
  __Pyx_XDECREF(__pyx_v_current_key);
  __Pyx_XDECREF(__pyx_v_affected_windows);
  __Pyx_XDECREF(__pyx_v_acc);
  __Pyx_XDECREF(__pyx_v_actual_windows);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_date_time);
  __Pyx_XDECREF(__pyx_v_window);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_5process_element(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_5process_element = {"process_element", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_5process_element, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_5process_element(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_row = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_element (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_input_row,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 374, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 374, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "process_element", 0) < 0) __PYX_ERR(0, 374, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("process_element", 1, 1, 1, i); __PYX_ERR(0, 374, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 374, __pyx_L3_error)
    }
    __pyx_v_input_row = ((struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)values[0]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("process_element", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 374, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.process_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_input_row), __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow, 1, "input_row", 0))) __PYX_ERR(0, 374, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_4process_element(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v_input_row);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = NULL;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_4process_element(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_input_row) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31]))
  __Pyx_RefNannySetupContext("process_element", 0);
  __Pyx_TraceStartFunc("process_element (wrapper)", __pyx_f[0], 374, 0, 0, 0, __PYX_ERR(0, 374, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_process_element(__pyx_v_self, __pyx_v_input_row, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 374, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.process_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":420
 *         return result
 * 
 *     cpdef void process_watermark(self, int64_t watermark):             # <<<<<<<<<<<<<<
 *         self._internal_timer_service.advance_watermark(watermark)
 * 
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_7process_watermark(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_process_watermark(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int64_t __pyx_v_watermark, int __pyx_skip_dispatch) {
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32]))
  __Pyx_RefNannySetupContext("process_watermark", 0);
  __Pyx_TraceStartFunc("process_watermark", __pyx_f[0], 420, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 420, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_process_watermark); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_7process_watermark)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyLong_From_int64_t(__pyx_v_watermark); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_6 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":421
 * 
 *     cpdef void process_watermark(self, int64_t watermark):
 *         self._internal_timer_service.advance_watermark(watermark)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list on_event_time(self, object timer):
*/
  __pyx_t_2 = __pyx_v_self->_internal_timer_service;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyLong_From_int64_t(__pyx_v_watermark); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_advance_watermark, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 421, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":420
 *         return result
 * 
 *     cpdef void process_watermark(self, int64_t watermark):             # <<<<<<<<<<<<<<
 *         self._internal_timer_service.advance_watermark(watermark)
 * 
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 420, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 420, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.process_watermark", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_7process_watermark(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_7process_watermark = {"process_watermark", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_7process_watermark, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_7process_watermark(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_watermark;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("process_watermark (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_watermark,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 420, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 420, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "process_watermark", 0) < 0) __PYX_ERR(0, 420, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("process_watermark", 1, 1, 1, i); __PYX_ERR(0, 420, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 420, __pyx_L3_error)
    }
    __pyx_v_watermark = __Pyx_PyLong_As_int64_t(values[0]); if (unlikely((__pyx_v_watermark == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 420, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("process_watermark", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 420, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.process_watermark", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_6process_watermark(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v_watermark);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_6process_watermark(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int64_t __pyx_v_watermark) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32]))
  __Pyx_RefNannySetupContext("process_watermark", 0);
  __Pyx_TraceStartFunc("process_watermark (wrapper)", __pyx_f[0], 420, 0, 0, 0, __PYX_ERR(0, 420, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_process_watermark(__pyx_v_self, __pyx_v_watermark, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 420, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 420, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.process_watermark", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":423
 *         self._internal_timer_service.advance_watermark(watermark)
 * 
 *     cpdef list on_event_time(self, object timer):             # <<<<<<<<<<<<<<
 *         cdef list result, key
 *         cdef int64_t timestamp
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_9on_event_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_on_event_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_timer, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_key = 0;
  int64_t __pyx_v_timestamp;
  PyObject *__pyx_v_window = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33]))
  __Pyx_RefNannySetupContext("on_event_time", 0);
  __Pyx_TraceStartFunc("on_event_time", __pyx_f[0], 423, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 423, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_on_event_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_9on_event_time)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_timer};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 423, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 423, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":427
 *         cdef int64_t timestamp
 *         cdef object window
 *         result = []             # <<<<<<<<<<<<<<
 *         timestamp = timer.get_timestamp()
 *         key = timer.get_key()
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 427, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":428
 *         cdef object window
 *         result = []
 *         timestamp = timer.get_timestamp()             # <<<<<<<<<<<<<<
 *         key = timer.get_key()
 *         self._state_backend.set_current_key(key)
*/
  __pyx_t_2 = __pyx_v_timer;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_timestamp, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 428, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_6 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 428, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_timestamp = __pyx_t_6;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":429
 *         result = []
 *         timestamp = timer.get_timestamp()
 *         key = timer.get_key()             # <<<<<<<<<<<<<<
 *         self._state_backend.set_current_key(key)
 *         window = timer.get_namespace()
*/
  __pyx_t_2 = __pyx_v_timer;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_key, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 429, __pyx_L1_error)
  __pyx_v_key = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":430
 *         timestamp = timer.get_timestamp()
 *         key = timer.get_key()
 *         self._state_backend.set_current_key(key)             # <<<<<<<<<<<<<<
 *         window = timer.get_namespace()
 *         self._trigger_context.window = window
*/
  __pyx_t_2 = __pyx_v_self->_state_backend;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_current_key, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 430, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":431
 *         key = timer.get_key()
 *         self._state_backend.set_current_key(key)
 *         window = timer.get_namespace()             # <<<<<<<<<<<<<<
 *         self._trigger_context.window = window
 *         if self._trigger_context.on_event_time(timestamp):
*/
  __pyx_t_2 = __pyx_v_timer;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_namespace, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 431, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_window = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":432
 *         self._state_backend.set_current_key(key)
 *         window = timer.get_namespace()
 *         self._trigger_context.window = window             # <<<<<<<<<<<<<<
 *         if self._trigger_context.on_event_time(timestamp):
 *             # fire
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_trigger_context, __pyx_mstate_global->__pyx_n_u_window, __pyx_v_window) < 0) __PYX_ERR(0, 432, __pyx_L1_error)

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":433
 *         window = timer.get_namespace()
 *         self._trigger_context.window = window
 *         if self._trigger_context.on_event_time(timestamp):             # <<<<<<<<<<<<<<
 *             # fire
 *             result.append(self._emit_window_result(key, window))
*/
  __pyx_t_2 = __pyx_v_self->_trigger_context;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyLong_From_int64_t(__pyx_v_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_on_event_time, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 433, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":435
 *         if self._trigger_context.on_event_time(timestamp):
 *             # fire
 *             result.append(self._emit_window_result(key, window))             # <<<<<<<<<<<<<<
 * 
 *         if self._window_assigner.is_event_time():
*/
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self->__pyx_vtab)->_emit_window_result(__pyx_v_self, __pyx_v_key, __pyx_v_window)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 435, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 435, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":433
 *         window = timer.get_namespace()
 *         self._trigger_context.window = window
 *         if self._trigger_context.on_event_time(timestamp):             # <<<<<<<<<<<<<<
 *             # fire
 *             result.append(self._emit_window_result(key, window))
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":437
 *             result.append(self._emit_window_result(key, window))
 * 
 *         if self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             self._window_function.clean_window_if_needed(window, timestamp)
 *         return result
*/
  __pyx_t_4 = __pyx_v_self->_window_assigner;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_event_time, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":438
 * 
 *         if self._window_assigner.is_event_time():
 *             self._window_function.clean_window_if_needed(window, timestamp)             # <<<<<<<<<<<<<<
 *         return result
 * 
*/
    __pyx_t_4 = __pyx_v_self->_window_function;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_timestamp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 438, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_window, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_clean_window_if_needed, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":437
 *             result.append(self._emit_window_result(key, window))
 * 
 *         if self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             self._window_function.clean_window_if_needed(window, timestamp)
 *         return result
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":439
 *         if self._window_assigner.is_event_time():
 *             self._window_function.clean_window_if_needed(window, timestamp)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef list on_processing_time(self, object timer):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  __Pyx_TraceReturnValue(__pyx_r, 49, 0, __PYX_ERR(0, 439, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":423
 *         self._internal_timer_service.advance_watermark(watermark)
 * 
 *     cpdef list on_event_time(self, object timer):             # <<<<<<<<<<<<<<
 *         cdef list result, key
 *         cdef int64_t timestamp
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 423, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.on_event_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_window);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_9on_event_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_9on_event_time = {"on_event_time", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_9on_event_time, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_9on_event_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timer = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_event_time (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_timer,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 423, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 423, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "on_event_time", 0) < 0) __PYX_ERR(0, 423, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("on_event_time", 1, 1, 1, i); __PYX_ERR(0, 423, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 423, __pyx_L3_error)
    }
    __pyx_v_timer = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("on_event_time", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 423, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.on_event_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_8on_event_time(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v_timer);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_8on_event_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_timer) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33]))
  __Pyx_RefNannySetupContext("on_event_time", 0);
  __Pyx_TraceStartFunc("on_event_time (wrapper)", __pyx_f[0], 423, 0, 0, 0, __PYX_ERR(0, 423, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_on_event_time(__pyx_v_self, __pyx_v_timer, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 423, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.on_event_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":441
 *         return result
 * 
 *     cpdef list on_processing_time(self, object timer):             # <<<<<<<<<<<<<<
 *         cdef list result, key
 *         cdef int64_t timestamp
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_11on_processing_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_on_processing_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_timer, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_result = 0;
  PyObject *__pyx_v_key = 0;
  int64_t __pyx_v_timestamp;
  PyObject *__pyx_v_window = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34]))
  __Pyx_RefNannySetupContext("on_processing_time", 0);
  __Pyx_TraceStartFunc("on_processing_time", __pyx_f[0], 441, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 441, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_on_processing_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_11on_processing_time)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_timer};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 441, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 441, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":445
 *         cdef int64_t timestamp
 *         cdef object window
 *         result = []             # <<<<<<<<<<<<<<
 *         timestamp = timer.get_timestamp()
 *         key = timer.get_key()
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 445, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":446
 *         cdef object window
 *         result = []
 *         timestamp = timer.get_timestamp()             # <<<<<<<<<<<<<<
 *         key = timer.get_key()
 *         self._state_backend.set_current_key(key)
*/
  __pyx_t_2 = __pyx_v_timer;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_timestamp, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 446, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_6 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 446, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_timestamp = __pyx_t_6;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":447
 *         result = []
 *         timestamp = timer.get_timestamp()
 *         key = timer.get_key()             # <<<<<<<<<<<<<<
 *         self._state_backend.set_current_key(key)
 *         window = timer.get_namespace()
*/
  __pyx_t_2 = __pyx_v_timer;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_key, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 447, __pyx_L1_error)
  __pyx_v_key = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":448
 *         timestamp = timer.get_timestamp()
 *         key = timer.get_key()
 *         self._state_backend.set_current_key(key)             # <<<<<<<<<<<<<<
 *         window = timer.get_namespace()
 *         self._trigger_context.window = window
*/
  __pyx_t_2 = __pyx_v_self->_state_backend;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_key};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_set_current_key, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 448, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":449
 *         key = timer.get_key()
 *         self._state_backend.set_current_key(key)
 *         window = timer.get_namespace()             # <<<<<<<<<<<<<<
 *         self._trigger_context.window = window
 *         if self._trigger_context.on_processing_time(timestamp):
*/
  __pyx_t_2 = __pyx_v_timer;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_namespace, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 449, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v_window = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":450
 *         self._state_backend.set_current_key(key)
 *         window = timer.get_namespace()
 *         self._trigger_context.window = window             # <<<<<<<<<<<<<<
 *         if self._trigger_context.on_processing_time(timestamp):
 *             # fire
*/
  if (__Pyx_PyObject_SetAttrStr(__pyx_v_self->_trigger_context, __pyx_mstate_global->__pyx_n_u_window, __pyx_v_window) < 0) __PYX_ERR(0, 450, __pyx_L1_error)

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":451
 *         window = timer.get_namespace()
 *         self._trigger_context.window = window
 *         if self._trigger_context.on_processing_time(timestamp):             # <<<<<<<<<<<<<<
 *             # fire
 *             result.append(self._emit_window_result(key, window))
*/
  __pyx_t_2 = __pyx_v_self->_trigger_context;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_4 = __Pyx_PyLong_From_int64_t(__pyx_v_timestamp); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_4};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_on_processing_time, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 451, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":453
 *         if self._trigger_context.on_processing_time(timestamp):
 *             # fire
 *             result.append(self._emit_window_result(key, window))             # <<<<<<<<<<<<<<
 * 
 *         if not self._window_assigner.is_event_time():
*/
    __pyx_t_1 = ((PyObject *)((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self->__pyx_vtab)->_emit_window_result(__pyx_v_self, __pyx_v_key, __pyx_v_window)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyList_Append(__pyx_v_result, __pyx_t_1); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(0, 453, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":451
 *         window = timer.get_namespace()
 *         self._trigger_context.window = window
 *         if self._trigger_context.on_processing_time(timestamp):             # <<<<<<<<<<<<<<
 *             # fire
 *             result.append(self._emit_window_result(key, window))
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":455
 *             result.append(self._emit_window_result(key, window))
 * 
 *         if not self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             self._window_function.clean_window_if_needed(window, timestamp)
 *         return result
*/
  __pyx_t_4 = __pyx_v_self->_window_assigner;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_event_time, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 455, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 455, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_9 = (!__pyx_t_7);
  if (__pyx_t_9) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":456
 * 
 *         if not self._window_assigner.is_event_time():
 *             self._window_function.clean_window_if_needed(window, timestamp)             # <<<<<<<<<<<<<<
 *         return result
 * 
*/
    __pyx_t_4 = __pyx_v_self->_window_function;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_timestamp); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 456, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_v_window, __pyx_t_2};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_clean_window_if_needed, __pyx_callargs+__pyx_t_5, (3-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 456, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":455
 *             result.append(self._emit_window_result(key, window))
 * 
 *         if not self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             self._window_function.clean_window_if_needed(window, timestamp)
 *         return result
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":457
 *         if not self._window_assigner.is_event_time():
 *             self._window_function.clean_window_if_needed(window, timestamp)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     cpdef int64_t to_utc_timestamp_mills(self, int64_t epoch_mills):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_result);
  __pyx_r = __pyx_v_result;
  __Pyx_TraceReturnValue(__pyx_r, 50, 0, __PYX_ERR(0, 457, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":441
 *         return result
 * 
 *     cpdef list on_processing_time(self, object timer):             # <<<<<<<<<<<<<<
 *         cdef list result, key
 *         cdef int64_t timestamp
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 441, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.on_processing_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_key);
  __Pyx_XDECREF(__pyx_v_window);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_11on_processing_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_11on_processing_time = {"on_processing_time", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_11on_processing_time, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_11on_processing_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_timer = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("on_processing_time (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_timer,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 441, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 441, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "on_processing_time", 0) < 0) __PYX_ERR(0, 441, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("on_processing_time", 1, 1, 1, i); __PYX_ERR(0, 441, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 441, __pyx_L3_error)
    }
    __pyx_v_timer = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("on_processing_time", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 441, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.on_processing_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_10on_processing_time(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v_timer);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_10on_processing_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_timer) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34]))
  __Pyx_RefNannySetupContext("on_processing_time", 0);
  __Pyx_TraceStartFunc("on_processing_time (wrapper)", __pyx_f[0], 441, 0, 0, 0, __PYX_ERR(0, 441, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_on_processing_time(__pyx_v_self, __pyx_v_timer, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 441, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.on_processing_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":459
 *         return result
 * 
 *     cpdef int64_t to_utc_timestamp_mills(self, int64_t epoch_mills):             # <<<<<<<<<<<<<<
 *         if self._shift_timezone == "UTC":
 *             return epoch_mills
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_13to_utc_timestamp_mills(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int64_t __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_to_utc_timestamp_mills(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int64_t __pyx_v_epoch_mills, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_timezone = NULL;
  PyObject *__pyx_v_local_date_time = NULL;
  PyObject *__pyx_v_epoch = NULL;
  int64_t __pyx_r;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int64_t __pyx_t_7;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35]))
  __Pyx_RefNannySetupContext("to_utc_timestamp_mills", 0);
  __Pyx_TraceStartFunc("to_utc_timestamp_mills", __pyx_f[0], 459, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 459, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_to_utc_timestamp_mills); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 459, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_13to_utc_timestamp_mills)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = __Pyx_PyLong_From_int64_t(__pyx_v_epoch_mills); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 459, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __pyx_t_6 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_6 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_5};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 459, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_t_7 = __Pyx_PyLong_As_int64_t(__pyx_t_2); if (unlikely((__pyx_t_7 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 459, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_7;
        __Pyx_TraceReturnCValue(__pyx_r, __Pyx_PyLong_From_int64_t, 0, 0, __PYX_ERR(0, 459, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":460
 * 
 *     cpdef int64_t to_utc_timestamp_mills(self, int64_t epoch_mills):
 *         if self._shift_timezone == "UTC":             # <<<<<<<<<<<<<<
 *             return epoch_mills
 *         else:
*/
  __pyx_t_8 = (__Pyx_PyUnicode_Equals(__pyx_v_self->_shift_timezone, __pyx_mstate_global->__pyx_n_u_UTC, Py_EQ)); if (unlikely((__pyx_t_8 < 0))) __PYX_ERR(0, 460, __pyx_L1_error)
  if (__pyx_t_8) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":461
 *     cpdef int64_t to_utc_timestamp_mills(self, int64_t epoch_mills):
 *         if self._shift_timezone == "UTC":
 *             return epoch_mills             # <<<<<<<<<<<<<<
 *         else:
 *             timezone = pytz.timezone(self._shift_timezone)
*/
    __pyx_r = __pyx_v_epoch_mills;
    __Pyx_TraceReturnCValue(__pyx_r, __Pyx_PyLong_From_int64_t, 6, 0, __PYX_ERR(0, 461, __pyx_L1_error));
    goto __pyx_L0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":460
 * 
 *     cpdef int64_t to_utc_timestamp_mills(self, int64_t epoch_mills):
 *         if self._shift_timezone == "UTC":             # <<<<<<<<<<<<<<
 *             return epoch_mills
 *         else:
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":463
 *             return epoch_mills
 *         else:
 *             timezone = pytz.timezone(self._shift_timezone)             # <<<<<<<<<<<<<<
 *             local_date_time = datetime.datetime.fromtimestamp(epoch_mills / 1000., timezone)\
 *                 .replace(tzinfo=None)
*/
  /*else*/ {
    __pyx_t_2 = NULL;
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_pytz); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_mstate_global->__pyx_n_u_timezone); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 463, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = 1;
    #if CYTHON_UNPACK_METHODS
    if (unlikely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
      assert(__pyx_t_2);
      PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(__pyx__function);
      __Pyx_DECREF_SET(__pyx_t_5, __pyx__function);
      __pyx_t_6 = 0;
    }
    #endif
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_self->_shift_timezone};
      __pyx_t_1 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 463, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_v_timezone = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":464
 *         else:
 *             timezone = pytz.timezone(self._shift_timezone)
 *             local_date_time = datetime.datetime.fromtimestamp(epoch_mills / 1000., timezone)\             # <<<<<<<<<<<<<<
 *                 .replace(tzinfo=None)
 *             epoch = datetime.datetime.utcfromtimestamp(0)
*/
    __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_datetime); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_datetime); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_4 = __pyx_t_9;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_3 = PyFloat_FromDouble((((double)__pyx_v_epoch_mills) / 1000.)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 464, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[3] = {__pyx_t_4, __pyx_t_3, __pyx_v_timezone};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_fromtimestamp, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 464, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __pyx_t_5 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_5);

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":465
 *             timezone = pytz.timezone(self._shift_timezone)
 *             local_date_time = datetime.datetime.fromtimestamp(epoch_mills / 1000., timezone)\
 *                 .replace(tzinfo=None)             # <<<<<<<<<<<<<<
 *             epoch = datetime.datetime.utcfromtimestamp(0)
 *             return int((local_date_time - epoch).total_seconds() * 1000.0)
*/
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2 + ((CYTHON_VECTORCALL) ? 1 : 0)] = {__pyx_t_5, NULL};
      __pyx_t_9 = __Pyx_MakeVectorcallBuilderKwds(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_VectorcallBuilder_AddArg(__pyx_mstate_global->__pyx_n_u_tzinfo, Py_None, __pyx_t_9, __pyx_callargs+1, 0) < 0) __PYX_ERR(0, 465, __pyx_L1_error)
      __pyx_t_1 = __Pyx_Object_VectorcallMethod_CallFromBuilder(__pyx_mstate_global->__pyx_n_u_replace, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET), __pyx_t_9);
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 465, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_v_local_date_time = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":466
 *             local_date_time = datetime.datetime.fromtimestamp(epoch_mills / 1000., timezone)\
 *                 .replace(tzinfo=None)
 *             epoch = datetime.datetime.utcfromtimestamp(0)             # <<<<<<<<<<<<<<
 *             return int((local_date_time - epoch).total_seconds() * 1000.0)
 * 
*/
    __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_datetime); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_mstate_global->__pyx_n_u_datetime); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_2 = __pyx_t_5;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_mstate_global->__pyx_int_0};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_utcfromtimestamp, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 466, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_v_epoch = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":467
 *                 .replace(tzinfo=None)
 *             epoch = datetime.datetime.utcfromtimestamp(0)
 *             return int((local_date_time - epoch).total_seconds() * 1000.0)             # <<<<<<<<<<<<<<
 * 
 *     cpdef list get_timers(self):
*/
    __pyx_t_2 = PyNumber_Subtract(__pyx_v_local_date_time, __pyx_v_epoch); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = __pyx_t_2;
    __Pyx_INCREF(__pyx_t_5);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_total_seconds, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_2 = PyNumber_Multiply(__pyx_t_1, __pyx_mstate_global->__pyx_float_1000_0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyNumber_Int(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyLong_As_int64_t(__pyx_t_1); if (unlikely((__pyx_t_7 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_r = __pyx_t_7;
    __Pyx_TraceReturnCValue(__pyx_r, __Pyx_PyLong_From_int64_t, 33, 0, __PYX_ERR(0, 467, __pyx_L1_error));
    goto __pyx_L0;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":459
 *         return result
 * 
 *     cpdef int64_t to_utc_timestamp_mills(self, int64_t epoch_mills):             # <<<<<<<<<<<<<<
 *         if self._shift_timezone == "UTC":
 *             return epoch_mills
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 459, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.to_utc_timestamp_mills", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_timezone);
  __Pyx_XDECREF(__pyx_v_local_date_time);
  __Pyx_XDECREF(__pyx_v_epoch);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_13to_utc_timestamp_mills(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_13to_utc_timestamp_mills = {"to_utc_timestamp_mills", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_13to_utc_timestamp_mills, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_13to_utc_timestamp_mills(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  int64_t __pyx_v_epoch_mills;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("to_utc_timestamp_mills (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_epoch_mills,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 459, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 459, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "to_utc_timestamp_mills", 0) < 0) __PYX_ERR(0, 459, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("to_utc_timestamp_mills", 1, 1, 1, i); __PYX_ERR(0, 459, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 459, __pyx_L3_error)
    }
    __pyx_v_epoch_mills = __Pyx_PyLong_As_int64_t(values[0]); if (unlikely((__pyx_v_epoch_mills == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 459, __pyx_L3_error)
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("to_utc_timestamp_mills", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 459, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.to_utc_timestamp_mills", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_12to_utc_timestamp_mills(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v_epoch_mills);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_12to_utc_timestamp_mills(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int64_t __pyx_v_epoch_mills) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35]))
  __Pyx_RefNannySetupContext("to_utc_timestamp_mills", 0);
  __Pyx_TraceStartFunc("to_utc_timestamp_mills (wrapper)", __pyx_f[0], 459, 0, 0, 0, __PYX_ERR(0, 459, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_to_utc_timestamp_mills(__pyx_v_self, __pyx_v_epoch_mills, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 459, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 459, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.to_utc_timestamp_mills", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":469
 *             return int((local_date_time - epoch).total_seconds() * 1000.0)
 * 
 *     cpdef list get_timers(self):             # <<<<<<<<<<<<<<
 *         cdef list timers
 *         cdef object timer
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_15get_timers(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_get_timers(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_timers = 0;
  PyObject *__pyx_v_timer = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36]))
  __Pyx_RefNannySetupContext("get_timers", 0);
  __Pyx_TraceStartFunc("get_timers", __pyx_f[0], 469, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 469, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_get_timers); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_15get_timers)) {
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        if (!(likely(PyList_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_2))) __PYX_ERR(0, 469, __pyx_L1_error)
        __pyx_r = ((PyObject*)__pyx_t_2);
        __pyx_t_2 = 0;
        __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(0, 469, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":472
 *         cdef list timers
 *         cdef object timer
 *         timers = []             # <<<<<<<<<<<<<<
 *         for timer in self._internal_timer_service._timers.keys():
 *             timers.append(timer)
*/
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 472, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_timers = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":473
 *         cdef object timer
 *         timers = []
 *         for timer in self._internal_timer_service._timers.keys():             # <<<<<<<<<<<<<<
 *             timers.append(timer)
 *         self._internal_timer_service._timers.clear()
*/
  __pyx_t_6 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_internal_timer_service, __pyx_mstate_global->__pyx_n_u_timers); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(__pyx_t_2 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "keys");
    __PYX_ERR(0, 473, __pyx_L1_error)
  }
  __pyx_t_4 = __Pyx_dict_iterator(__pyx_t_2, 0, __pyx_mstate_global->__pyx_n_u_keys, (&__pyx_t_7), (&__pyx_t_8)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 473, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_4;
  __pyx_t_4 = 0;
  while (1) {
    __pyx_t_9 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_7, &__pyx_t_6, &__pyx_t_4, NULL, NULL, __pyx_t_8);
    if (unlikely(__pyx_t_9 == 0)) break;
    if (unlikely(__pyx_t_9 == -1)) __PYX_ERR(0, 473, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_XDECREF_SET(__pyx_v_timer, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":474
 *         timers = []
 *         for timer in self._internal_timer_service._timers.keys():
 *             timers.append(timer)             # <<<<<<<<<<<<<<
 *         self._internal_timer_service._timers.clear()
 *         return timers
*/
    __pyx_t_10 = __Pyx_PyList_Append(__pyx_v_timers, __pyx_v_timer); if (unlikely(__pyx_t_10 == ((int)-1))) __PYX_ERR(0, 474, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":475
 *         for timer in self._internal_timer_service._timers.keys():
 *             timers.append(timer)
 *         self._internal_timer_service._timers.clear()             # <<<<<<<<<<<<<<
 *         return timers
 * 
*/
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_self->_internal_timer_service, __pyx_mstate_global->__pyx_n_u_timers); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = __pyx_t_2;
  __Pyx_INCREF(__pyx_t_4);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_clear, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":476
 *             timers.append(timer)
 *         self._internal_timer_service._timers.clear()
 *         return timers             # <<<<<<<<<<<<<<
 * 
 *     cpdef void close(self):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_timers);
  __pyx_r = __pyx_v_timers;
  __Pyx_TraceReturnValue(__pyx_r, 19, 0, __PYX_ERR(0, 476, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":469
 *             return int((local_date_time - epoch).total_seconds() * 1000.0)
 * 
 *     cpdef list get_timers(self):             # <<<<<<<<<<<<<<
 *         cdef list timers
 *         cdef object timer
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 469, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.get_timers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_timers);
  __Pyx_XDECREF(__pyx_v_timer);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_15get_timers(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_15get_timers = {"get_timers", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_15get_timers, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_15get_timers(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("get_timers (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("get_timers", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("get_timers", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_14get_timers(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_14get_timers(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36]))
  __Pyx_RefNannySetupContext("get_timers", 0);
  __Pyx_TraceStartFunc("get_timers (wrapper)", __pyx_f[0], 469, 0, 0, 0, __PYX_ERR(0, 469, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_get_timers(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 469, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.get_timers", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":478
 *         return timers
 * 
 *     cpdef void close(self):             # <<<<<<<<<<<<<<
 *         self._window_aggregator.close()
 * 
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_17close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_close(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, int __pyx_skip_dispatch) {
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37]))
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_TraceStartFunc("close", __pyx_f[0], 478, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 478, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_close); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_17close)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 478, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":479
 * 
 *     cpdef void close(self):
 *         self._window_aggregator.close()             # <<<<<<<<<<<<<<
 * 
 *     cdef void _register_cleanup_timer(self, object window):
*/
  ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v_self->_window_aggregator->__pyx_vtab)->close(__pyx_v_self->_window_aggregator); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 479, __pyx_L1_error)

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":478
 *         return timers
 * 
 *     cpdef void close(self):             # <<<<<<<<<<<<<<
 *         self._window_aggregator.close()
 * 
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 478, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 478, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_17close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_17close = {"close", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_17close, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_17close(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("close (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("close", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("close", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_16close(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_16close(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37]))
  __Pyx_RefNannySetupContext("close", 0);
  __Pyx_TraceStartFunc("close (wrapper)", __pyx_f[0], 478, 0, 0, 0, __PYX_ERR(0, 478, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_close(__pyx_v_self, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 478, __pyx_L1_error)
  __pyx_t_1 = __Pyx_void_to_None(NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 478, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.close", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":481
 *         self._window_aggregator.close()
 * 
 *     cdef void _register_cleanup_timer(self, object window):             # <<<<<<<<<<<<<<
 *         cdef int64_t cleanup_time
 *         cleanup_time = self.cleanup_time(window)
*/

static void __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase__register_cleanup_timer(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_window) {
  int64_t __pyx_v_cleanup_time;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[38]))
  __Pyx_RefNannySetupContext("_register_cleanup_timer", 0);
  __Pyx_TraceStartFunc("_register_cleanup_timer", __pyx_f[0], 481, 0, 0, 0, __PYX_ERR(0, 481, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":483
 *     cdef void _register_cleanup_timer(self, object window):
 *         cdef int64_t cleanup_time
 *         cleanup_time = self.cleanup_time(window)             # <<<<<<<<<<<<<<
 *         if cleanup_time == MAX_LONG_VALUE:
 *             return
*/
  __pyx_t_1 = ((struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self->__pyx_vtab)->cleanup_time(__pyx_v_self, __pyx_v_window, 0); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 483, __pyx_L1_error)
  __pyx_v_cleanup_time = __pyx_t_1;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":484
 *         cdef int64_t cleanup_time
 *         cleanup_time = self.cleanup_time(window)
 *         if cleanup_time == MAX_LONG_VALUE:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_cleanup_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_MAX_LONG_VALUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyObject_RichCompare(__pyx_t_2, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":485
 *         cleanup_time = self.cleanup_time(window)
 *         if cleanup_time == MAX_LONG_VALUE:
 *             return             # <<<<<<<<<<<<<<
 * 
 *         if self._window_assigner.is_event_time():
*/
    goto __pyx_L0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":484
 *         cdef int64_t cleanup_time
 *         cleanup_time = self.cleanup_time(window)
 *         if cleanup_time == MAX_LONG_VALUE:             # <<<<<<<<<<<<<<
 *             return
 * 
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":487
 *             return
 * 
 *         if self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             self._trigger_context.register_event_time_timer(cleanup_time)
 *         else:
*/
  __pyx_t_3 = __pyx_v_self->_window_assigner;
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
    __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_event_time, __pyx_callargs+__pyx_t_6, (1-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely((__pyx_t_5 < 0))) __PYX_ERR(0, 487, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (__pyx_t_5) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":488
 * 
 *         if self._window_assigner.is_event_time():
 *             self._trigger_context.register_event_time_timer(cleanup_time)             # <<<<<<<<<<<<<<
 *         else:
 *             self._trigger_context.register_processing_time_timer(cleanup_time)
*/
    __pyx_t_3 = __pyx_v_self->_trigger_context;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_cleanup_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_t_2};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_register_event_time_timer, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 488, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":487
 *             return
 * 
 *         if self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             self._trigger_context.register_event_time_timer(cleanup_time)
 *         else:
*/
    goto __pyx_L4;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":490
 *             self._trigger_context.register_event_time_timer(cleanup_time)
 *         else:
 *             self._trigger_context.register_processing_time_timer(cleanup_time)             # <<<<<<<<<<<<<<
 * 
 *     cpdef int64_t cleanup_time(self, object window):
*/
  /*else*/ {
    __pyx_t_2 = __pyx_v_self->_trigger_context;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyLong_From_int64_t(__pyx_v_cleanup_time); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 490, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_t_3};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_register_processing_time_timer, __pyx_callargs+__pyx_t_6, (2-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 490, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __pyx_L4:;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":481
 *         self._window_aggregator.close()
 * 
 *     cdef void _register_cleanup_timer(self, object window):             # <<<<<<<<<<<<<<
 *         cdef int64_t cleanup_time
 *         cleanup_time = self.cleanup_time(window)
*/

  /* function exit code */
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 481, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 481, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase._register_cleanup_timer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":492
 *             self._trigger_context.register_processing_time_timer(cleanup_time)
 * 
 *     cpdef int64_t cleanup_time(self, object window):             # <<<<<<<<<<<<<<
 *         cdef int64_t cleanup_time
 *         if self._window_assigner.is_event_time():
*/

static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_19cleanup_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static int64_t __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_cleanup_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_window, int __pyx_skip_dispatch) {
  int64_t __pyx_v_cleanup_time;
  int64_t __pyx_r;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  size_t __pyx_t_5;
  int64_t __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39]))
  __Pyx_RefNannySetupContext("cleanup_time", 0);
  __Pyx_TraceStartFunc("cleanup_time", __pyx_f[0], 492, 0, 0, __pyx_skip_dispatch, __PYX_ERR(0, 492, __pyx_L1_error));
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (
  #if !CYTHON_USE_TYPE_SLOTS
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self)) != __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase &&
  __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), Py_TPFLAGS_HAVE_GC))
  #else
  unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0 || __Pyx_PyType_HasFeature(Py_TYPE(((PyObject *)__pyx_v_self)), (Py_TPFLAGS_IS_ABSTRACT | Py_TPFLAGS_HEAPTYPE)))
  #endif
  ) {
    #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    static PY_UINT64_T __pyx_tp_dict_version = __PYX_DICT_VERSION_INIT, __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
    if (unlikely(!__Pyx_object_dict_version_matches(((PyObject *)__pyx_v_self), __pyx_tp_dict_version, __pyx_obj_dict_version))) {
      PY_UINT64_T __pyx_typedict_guard = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      #endif
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_cleanup_time); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 492, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      if (!__Pyx_IsSameCFunction(__pyx_t_1, (void(*)(void)) __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_19cleanup_time)) {
        __pyx_t_3 = NULL;
        __Pyx_INCREF(__pyx_t_1);
        __pyx_t_4 = __pyx_t_1; 
        __pyx_t_5 = 1;
        #if CYTHON_UNPACK_METHODS
        if (unlikely(PyMethod_Check(__pyx_t_4))) {
          __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_4);
          assert(__pyx_t_3);
          PyObject* __pyx__function = PyMethod_GET_FUNCTION(__pyx_t_4);
          __Pyx_INCREF(__pyx_t_3);
          __Pyx_INCREF(__pyx__function);
          __Pyx_DECREF_SET(__pyx_t_4, __pyx__function);
          __pyx_t_5 = 0;
        }
        #endif
        {
          PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v_window};
          __pyx_t_2 = __Pyx_PyObject_FastCall(__pyx_t_4, __pyx_callargs+__pyx_t_5, (2-__pyx_t_5) | (__pyx_t_5*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
          __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
          __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
          if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_2);
        }
        __pyx_t_6 = __Pyx_PyLong_As_int64_t(__pyx_t_2); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        __pyx_r = __pyx_t_6;
        __Pyx_TraceReturnCValue(__pyx_r, __Pyx_PyLong_From_int64_t, 0, 0, __PYX_ERR(0, 492, __pyx_L1_error));
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
        goto __pyx_L0;
      }
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
      __pyx_tp_dict_version = __Pyx_get_tp_dict_version(((PyObject *)__pyx_v_self));
      __pyx_obj_dict_version = __Pyx_get_object_dict_version(((PyObject *)__pyx_v_self));
      if (unlikely(__pyx_typedict_guard != __pyx_tp_dict_version)) {
        __pyx_tp_dict_version = __pyx_obj_dict_version = __PYX_DICT_VERSION_INIT;
      }
      #endif
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      #if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_PYTYPE_LOOKUP && CYTHON_USE_TYPE_SLOTS
    }
    #endif
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":494
 *     cpdef int64_t cleanup_time(self, object window):
 *         cdef int64_t cleanup_time
 *         if self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             cleanup_time = max(0, window.max_timestamp() + self._allowed_lateness)
 *             if cleanup_time >= window.max_timestamp():
*/
  __pyx_t_2 = __pyx_v_self->_window_assigner;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_5 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_is_event_time, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 494, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_7) {

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":495
 *         cdef int64_t cleanup_time
 *         if self._window_assigner.is_event_time():
 *             cleanup_time = max(0, window.max_timestamp() + self._allowed_lateness)             # <<<<<<<<<<<<<<
 *             if cleanup_time >= window.max_timestamp():
 *                 return cleanup_time
*/
    __pyx_t_2 = __pyx_v_window;
    __Pyx_INCREF(__pyx_t_2);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_2, NULL};
      __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_max_timestamp, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    }
    __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_allowed_lateness); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = PyNumber_Add(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_8 = 0;
    __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_4, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_7) {
      __Pyx_INCREF(__pyx_t_4);
      __pyx_t_2 = __pyx_t_4;
    } else {
      __pyx_t_3 = __Pyx_PyLong_From_long(__pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 495, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_2 = __pyx_t_3;
      __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_6 = __Pyx_PyLong_As_int64_t(__pyx_t_2); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_cleanup_time = __pyx_t_6;

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":496
 *         if self._window_assigner.is_event_time():
 *             cleanup_time = max(0, window.max_timestamp() + self._allowed_lateness)
 *             if cleanup_time >= window.max_timestamp():             # <<<<<<<<<<<<<<
 *                 return cleanup_time
 *             else:
*/
    __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_v_cleanup_time); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __pyx_v_window;
    __Pyx_INCREF(__pyx_t_3);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_3, NULL};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_max_timestamp, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 496, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __pyx_t_3 = PyObject_RichCompare(__pyx_t_2, __pyx_t_4, Py_GE); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 496, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (__pyx_t_7) {

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":497
 *             cleanup_time = max(0, window.max_timestamp() + self._allowed_lateness)
 *             if cleanup_time >= window.max_timestamp():
 *                 return cleanup_time             # <<<<<<<<<<<<<<
 *             else:
 *                 return MAX_LONG_VALUE
*/
      __pyx_r = __pyx_v_cleanup_time;
      __Pyx_TraceReturnCValue(__pyx_r, __Pyx_PyLong_From_int64_t, 21, 0, __PYX_ERR(0, 497, __pyx_L1_error));
      goto __pyx_L0;

      /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":496
 *         if self._window_assigner.is_event_time():
 *             cleanup_time = max(0, window.max_timestamp() + self._allowed_lateness)
 *             if cleanup_time >= window.max_timestamp():             # <<<<<<<<<<<<<<
 *                 return cleanup_time
 *             else:
*/
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":499
 *                 return cleanup_time
 *             else:
 *                 return MAX_LONG_VALUE             # <<<<<<<<<<<<<<
 *         else:
 *             return max(0, window.max_timestamp())
*/
    /*else*/ {
      __Pyx_GetModuleGlobalName(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_MAX_LONG_VALUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 499, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = __Pyx_PyLong_As_int64_t(__pyx_t_3); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 499, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_6;
      __Pyx_TraceReturnCValue(__pyx_r, __Pyx_PyLong_From_int64_t, 23, 0, __PYX_ERR(0, 499, __pyx_L1_error));
      goto __pyx_L0;
    }

    /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":494
 *     cpdef int64_t cleanup_time(self, object window):
 *         cdef int64_t cleanup_time
 *         if self._window_assigner.is_event_time():             # <<<<<<<<<<<<<<
 *             cleanup_time = max(0, window.max_timestamp() + self._allowed_lateness)
 *             if cleanup_time >= window.max_timestamp():
*/
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":501
 *                 return MAX_LONG_VALUE
 *         else:
 *             return max(0, window.max_timestamp())             # <<<<<<<<<<<<<<
 * 
 *     cdef InternalRow _emit_window_result(self, list key, object window):
*/
  /*else*/ {
    __pyx_t_4 = __pyx_v_window;
    __Pyx_INCREF(__pyx_t_4);
    __pyx_t_5 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_4, NULL};
      __pyx_t_3 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_max_timestamp, __pyx_callargs+__pyx_t_5, (1-__pyx_t_5) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    }
    __pyx_t_8 = 0;
    __pyx_t_2 = __Pyx_PyLong_From_long(__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_2, Py_GT); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_7 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely((__pyx_t_7 < 0))) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    if (__pyx_t_7) {
      __Pyx_INCREF(__pyx_t_3);
      __pyx_t_4 = __pyx_t_3;
    } else {
      __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 501, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_4 = __pyx_t_1;
      __pyx_t_1 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = __Pyx_PyLong_As_int64_t(__pyx_t_4); if (unlikely((__pyx_t_6 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(0, 501, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_6;
    __Pyx_TraceReturnCValue(__pyx_r, __Pyx_PyLong_From_int64_t, 25, 0, __PYX_ERR(0, 501, __pyx_L1_error));
    goto __pyx_L0;
  }

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":492
 *             self._trigger_context.register_processing_time_timer(cleanup_time)
 * 
 *     cpdef int64_t cleanup_time(self, object window):             # <<<<<<<<<<<<<<
 *         cdef int64_t cleanup_time
 *         if self._window_assigner.is_event_time():
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 492, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.cleanup_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_19cleanup_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_19cleanup_time = {"cleanup_time", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_19cleanup_time, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_19cleanup_time(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v_window = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("cleanup_time (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_window,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 492, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 492, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "cleanup_time", 0) < 0) __PYX_ERR(0, 492, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("cleanup_time", 1, 1, 1, i); __PYX_ERR(0, 492, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 492, __pyx_L3_error)
    }
    __pyx_v_window = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("cleanup_time", 1, 1, 1, __pyx_nargs); __PYX_ERR(0, 492, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.cleanup_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_18cleanup_time(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v_window);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_18cleanup_time(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v_window) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39]))
  __Pyx_RefNannySetupContext("cleanup_time", 0);
  __Pyx_TraceStartFunc("cleanup_time (wrapper)", __pyx_f[0], 492, 0, 0, 0, __PYX_ERR(0, 492, __pyx_L1_error));
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_cleanup_time(__pyx_v_self, __pyx_v_window, 1); if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 492, __pyx_L1_error)
  __pyx_t_2 = __Pyx_PyLong_From_int64_t(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 492, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.cleanup_time", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":503
 *             return max(0, window.max_timestamp())
 * 
 *     cdef InternalRow _emit_window_result(self, list key, object window):             # <<<<<<<<<<<<<<
 *         pass
 * 
*/

static struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase__emit_window_result(CYTHON_UNUSED struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_key, CYTHON_UNUSED PyObject *__pyx_v_window) {
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[40]))
  __Pyx_RefNannySetupContext("_emit_window_result", 0);
  __Pyx_TraceStartFunc("_emit_window_result", __pyx_f[0], 503, 0, 0, 0, __PYX_ERR(0, 503, __pyx_L1_error));

  /* function exit code */
  __pyx_r = ((struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)Py_None); __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue((PyObject *)__pyx_r, 0, 0, __PYX_ERR(0, 503, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 503, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase._emit_window_result", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_21__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_21__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_20__reduce_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_20__reduce_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41]))
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceStartFunc("__reduce_cython__", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._allowed_lateness, self._internal_timer_service, self._key_selector, self._rowtime_index, self._shift_timezone, self._state_backend, self._state_value_coder, self._trigger, self._trigger_context, self._window_aggregator, self._window_assigner, self._window_context, self._window_function, self._window_state)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
*/
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->_allowed_lateness); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_self->_rowtime_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(14); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_internal_timer_service);
  __Pyx_GIVEREF(__pyx_v_self->_internal_timer_service);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->_internal_timer_service) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_key_selector);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_key_selector);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_self->_key_selector)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_shift_timezone);
  __Pyx_GIVEREF(__pyx_v_self->_shift_timezone);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->_shift_timezone) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_state_backend);
  __Pyx_GIVEREF(__pyx_v_self->_state_backend);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->_state_backend) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_state_value_coder);
  __Pyx_GIVEREF(__pyx_v_self->_state_value_coder);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->_state_value_coder) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_trigger);
  __Pyx_GIVEREF(__pyx_v_self->_trigger);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_v_self->_trigger) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_trigger_context);
  __Pyx_GIVEREF(__pyx_v_self->_trigger_context);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_v_self->_trigger_context) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->_window_aggregator);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->_window_aggregator);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 9, ((PyObject *)__pyx_v_self->_window_aggregator)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_window_assigner);
  __Pyx_GIVEREF(__pyx_v_self->_window_assigner);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_v_self->_window_assigner) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_window_context);
  __Pyx_GIVEREF(__pyx_v_self->_window_context);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 11, __pyx_v_self->_window_context) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_window_function);
  __Pyx_GIVEREF(__pyx_v_self->_window_function);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 12, __pyx_v_self->_window_function) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->_window_state);
  __Pyx_GIVEREF(__pyx_v_self->_window_state);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 13, __pyx_v_self->_window_state) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._allowed_lateness, self._internal_timer_service, self._key_selector, self._rowtime_index, self._shift_timezone, self._state_backend, self._state_value_coder, self._trigger, self._trigger_context, self._window_aggregator, self._window_assigner, self._window_context, self._window_function, self._window_state)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
*/
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self._allowed_lateness, self._internal_timer_service, self._key_selector, self._rowtime_index, self._shift_timezone, self._state_backend, self._state_value_coder, self._trigger, self._trigger_context, self._window_aggregator, self._window_assigner, self._window_context, self._window_function, self._window_state)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._allowed_lateness, self._internal_timer_service, self._key_selector, self._rowtime_index, self._shift_timezone, self._state_backend, self._state_value_coder, self._trigger, self._trigger_context, self._window_aggregator, self._window_assigner, self._window_context, self._window_function, self._window_state)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, None), state
*/
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->_internal_timer_service != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->_key_selector) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_shift_timezone != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_state_backend != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_state_value_coder != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_trigger != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_trigger_context != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->_window_aggregator) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_window_assigner != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_window_context != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_window_function != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->_window_state != Py_None);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_GroupWindowAggFun); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_209029984);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_209029984);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_int_209029984) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_TraceReturnValue(__pyx_r, 106, 0, __PYX_ERR(1, 13, __pyx_L1_error));
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, None), state
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GroupWindowAggFunctionBase__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_GroupWindowAggFun); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_209029984);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_209029984);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_int_209029984) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_TraceReturnValue(__pyx_r, 114, 0, __PYX_ERR(1, 15, __pyx_L1_error));
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GroupWindowAggFunctionBase__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_23__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_23__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_22__setstate_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_22__setstate_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42]))
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceStartFunc("__setstate_cython__", __pyx_f[1], 16, 0, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GroupWindowAggFunctionBase__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_GroupWindowAggFunctionBase__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GroupWindowAggFunctionBase__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":507
 * 
 * cdef class GroupWindowAggFunction(GroupWindowAggFunctionBase):
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  allowed_lateness: int,
 *                  key_selector: RowKeySelector,
*/

/* Python wrapper */
static int __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_allowed_lateness = 0;
  struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *__pyx_v_key_selector = 0;
  PyObject *__pyx_v_state_backend = 0;
  PyObject *__pyx_v_state_value_coder = 0;
  PyObject *__pyx_v_window_assigner = 0;
  PyObject *__pyx_v_window_aggregator = 0;
  PyObject *__pyx_v_trigger = 0;
  PyObject *__pyx_v_rowtime_index = 0;
  PyObject *__pyx_v_shift_timezone = 0;
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[9] = {0,0,0,0,0,0,0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return -1;
  #endif
  __pyx_kwvalues = __Pyx_KwValues_VARARGS(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_allowed_lateness,&__pyx_mstate_global->__pyx_n_u_key_selector,&__pyx_mstate_global->__pyx_n_u_state_backend,&__pyx_mstate_global->__pyx_n_u_state_value_coder,&__pyx_mstate_global->__pyx_n_u_window_assigner,&__pyx_mstate_global->__pyx_n_u_window_aggregator,&__pyx_mstate_global->__pyx_n_u_trigger,&__pyx_mstate_global->__pyx_n_u_rowtime_index,&__pyx_mstate_global->__pyx_n_u_shift_timezone,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_VARARGS(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(0, 507, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  9:
        values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  8:
        values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  7:
        values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  6:
        values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  5:
        values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  4:
        values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  3:
        values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 507, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__init__", 0) < 0) __PYX_ERR(0, 507, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 9; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__init__", 1, 9, 9, i); __PYX_ERR(0, 507, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 9)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_VARARGS(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(0, 507, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_VARARGS(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(0, 507, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_VARARGS(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(0, 507, __pyx_L3_error)
      values[3] = __Pyx_ArgRef_VARARGS(__pyx_args, 3);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[3])) __PYX_ERR(0, 507, __pyx_L3_error)
      values[4] = __Pyx_ArgRef_VARARGS(__pyx_args, 4);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[4])) __PYX_ERR(0, 507, __pyx_L3_error)
      values[5] = __Pyx_ArgRef_VARARGS(__pyx_args, 5);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[5])) __PYX_ERR(0, 507, __pyx_L3_error)
      values[6] = __Pyx_ArgRef_VARARGS(__pyx_args, 6);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[6])) __PYX_ERR(0, 507, __pyx_L3_error)
      values[7] = __Pyx_ArgRef_VARARGS(__pyx_args, 7);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[7])) __PYX_ERR(0, 507, __pyx_L3_error)
      values[8] = __Pyx_ArgRef_VARARGS(__pyx_args, 8);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[8])) __PYX_ERR(0, 507, __pyx_L3_error)
    }
    __pyx_v_allowed_lateness = ((PyObject*)values[0]);
    __pyx_v_key_selector = ((struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *)values[1]);
    __pyx_v_state_backend = values[2];
    __pyx_v_state_value_coder = values[3];
    __pyx_v_window_assigner = values[4];
    __pyx_v_window_aggregator = values[5];
    __pyx_v_trigger = values[6];
    __pyx_v_rowtime_index = ((PyObject*)values[7]);
    __pyx_v_shift_timezone = ((PyObject*)values[8]);
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 9, 9, __pyx_nargs); __PYX_ERR(0, 507, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunction.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_allowed_lateness), (&PyLong_Type), 0, "allowed_lateness", 2))) __PYX_ERR(0, 508, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_key_selector), __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector, 0, "key_selector", 0))) __PYX_ERR(0, 509, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_rowtime_index), (&PyLong_Type), 0, "rowtime_index", 2))) __PYX_ERR(0, 515, __pyx_L1_error)
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shift_timezone), (&PyUnicode_Type), 0, "shift_timezone", 2))) __PYX_ERR(0, 516, __pyx_L1_error)
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction___init__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *)__pyx_v_self), __pyx_v_allowed_lateness, __pyx_v_key_selector, __pyx_v_state_backend, __pyx_v_state_value_coder, __pyx_v_window_assigner, __pyx_v_window_aggregator, __pyx_v_trigger, __pyx_v_rowtime_index, __pyx_v_shift_timezone);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  goto __pyx_L7_cleaned_up;
  __pyx_L0:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __pyx_L7_cleaned_up:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction___init__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v_self, PyObject *__pyx_v_allowed_lateness, struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *__pyx_v_key_selector, PyObject *__pyx_v_state_backend, PyObject *__pyx_v_state_value_coder, PyObject *__pyx_v_window_assigner, PyObject *__pyx_v_window_aggregator, PyObject *__pyx_v_trigger, PyObject *__pyx_v_rowtime_index, PyObject *__pyx_v_shift_timezone) {
  int __pyx_r;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  size_t __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[43]))
  __Pyx_RefNannySetupContext("__init__", 0);
  __Pyx_TraceStartFunc("__init__", __pyx_f[0], 507, 0, 0, 0, __PYX_ERR(0, 507, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":517
 *                  rowtime_index: int,
 *                  shift_timezone: str):
 *         super(GroupWindowAggFunction, self).__init__(             # <<<<<<<<<<<<<<
 *             allowed_lateness, key_selector, state_backend, state_value_coder, window_assigner,
 *             window_aggregator, trigger, rowtime_index, shift_timezone)
*/
  __pyx_t_4 = NULL;
  __Pyx_INCREF(__pyx_builtin_super);
  __pyx_t_5 = __pyx_builtin_super; 
  __pyx_t_6 = 1;
  {
    PyObject *__pyx_callargs[3] = {__pyx_t_4, ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction), ((PyObject *)__pyx_v_self)};
    __pyx_t_3 = __Pyx_PyObject_FastCall(__pyx_t_5, __pyx_callargs+__pyx_t_6, (3-__pyx_t_6) | (__pyx_t_6*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
  }
  __pyx_t_2 = __pyx_t_3;
  __Pyx_INCREF(__pyx_t_2);

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":519
 *         super(GroupWindowAggFunction, self).__init__(
 *             allowed_lateness, key_selector, state_backend, state_value_coder, window_assigner,
 *             window_aggregator, trigger, rowtime_index, shift_timezone)             # <<<<<<<<<<<<<<
 *         self._window_aggregator = window_aggregator
 * 
*/
  __pyx_t_6 = 0;
  {
    PyObject *__pyx_callargs[10] = {__pyx_t_2, __pyx_v_allowed_lateness, ((PyObject *)__pyx_v_key_selector), __pyx_v_state_backend, __pyx_v_state_value_coder, __pyx_v_window_assigner, __pyx_v_window_aggregator, __pyx_v_trigger, __pyx_v_rowtime_index, __pyx_v_shift_timezone};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_init, __pyx_callargs+__pyx_t_6, (10-__pyx_t_6) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 517, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":520
 *             allowed_lateness, key_selector, state_backend, state_value_coder, window_assigner,
 *             window_aggregator, trigger, rowtime_index, shift_timezone)
 *         self._window_aggregator = window_aggregator             # <<<<<<<<<<<<<<
 * 
 *     cdef InternalRow _emit_window_result(self, list key, object window):
*/
  __pyx_t_1 = __pyx_v_window_aggregator;
  __Pyx_INCREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase))))) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF((PyObject *)__pyx_v_self->__pyx_base._window_aggregator);
  __Pyx_DECREF((PyObject *)__pyx_v_self->__pyx_base._window_aggregator);
  __pyx_v_self->__pyx_base._window_aggregator = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":507
 * 
 * cdef class GroupWindowAggFunction(GroupWindowAggFunctionBase):
 *     def __init__(self,             # <<<<<<<<<<<<<<
 *                  allowed_lateness: int,
 *                  key_selector: RowKeySelector,
*/

  /* function exit code */
  __pyx_r = 0;
  __Pyx_TraceReturnCValue(__pyx_r, __Pyx_Owned_Py_None, 0, 0, __PYX_ERR(0, 507, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 507, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunction.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pyflink/fn_execution/table/window_aggregate_fast.pyx":522
 *         self._window_aggregator = window_aggregator
 * 
 *     cdef InternalRow _emit_window_result(self, list key, object window):             # <<<<<<<<<<<<<<
 *         cdef list agg_result
 *         cdef InternalRow result_row
*/

static struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction__emit_window_result(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v_self, PyObject *__pyx_v_key, PyObject *__pyx_v_window) {
  PyObject *__pyx_v_agg_result = 0;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_v_result_row = 0;
  struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  size_t __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[44]))
  __Pyx_RefNannySetupContext("_emit_window_result", 0);
  __Pyx_TraceStartFunc("_emit_window_result", __pyx_f[0], 522, 0, 0, 0, __PYX_ERR(0, 522, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":525
 *         cdef list agg_result
 *         cdef InternalRow result_row
 *         self._window_function.prepare_aggregate_accumulator_for_emit(window)             # <<<<<<<<<<<<<<
 *         agg_result = self._window_aggregator.get_value(window)
 *         result_row = join_row(key, agg_result, InternalRowKind.INSERT)
*/
  __pyx_t_2 = __pyx_v_self->__pyx_base._window_function;
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_window};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_prepare_aggregate_accumulator_fo, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 525, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":526
 *         cdef InternalRow result_row
 *         self._window_function.prepare_aggregate_accumulator_for_emit(window)
 *         agg_result = self._window_aggregator.get_value(window)             # <<<<<<<<<<<<<<
 *         result_row = join_row(key, agg_result, InternalRowKind.INSERT)
 *         return result_row
*/
  __pyx_t_2 = ((PyObject *)__pyx_v_self->__pyx_base._window_aggregator);
  __Pyx_INCREF(__pyx_t_2);
  __pyx_t_3 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_2, __pyx_v_window};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_get_value, __pyx_callargs+__pyx_t_3, (2-__pyx_t_3) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 526, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  if (!(likely(PyList_CheckExact(__pyx_t_1))||((__pyx_t_1) == Py_None) || __Pyx_RaiseUnexpectedTypeError("list", __pyx_t_1))) __PYX_ERR(0, 526, __pyx_L1_error)
  __pyx_v_agg_result = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":527
 *         self._window_function.prepare_aggregate_accumulator_for_emit(window)
 *         agg_result = self._window_aggregator.get_value(window)
 *         result_row = join_row(key, agg_result, InternalRowKind.INSERT)             # <<<<<<<<<<<<<<
 *         return result_row
*/
  __pyx_t_1 = ((PyObject *)__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_join_row(__pyx_v_key, __pyx_v_agg_result, __pyx_e_7pyflink_12fn_execution_15coder_impl_fast_INSERT)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 527, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_result_row = ((struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":528
 *         agg_result = self._window_aggregator.get_value(window)
 *         result_row = join_row(key, agg_result, InternalRowKind.INSERT)
 *         return result_row             # <<<<<<<<<<<<<<
*/
  __Pyx_XDECREF((PyObject *)__pyx_r);
  __Pyx_INCREF((PyObject *)__pyx_v_result_row);
  __pyx_r = __pyx_v_result_row;
  __Pyx_TraceReturnValue((PyObject *)__pyx_r, 18, 0, __PYX_ERR(0, 528, __pyx_L1_error));
  goto __pyx_L0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":522
 *         self._window_aggregator = window_aggregator
 * 
 *     cdef InternalRow _emit_window_result(self, list key, object window):             # <<<<<<<<<<<<<<
 *         cdef list agg_result
 *         cdef InternalRow result_row
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(0, 522, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunction._emit_window_result", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_agg_result);
  __Pyx_XDECREF((PyObject *)__pyx_v_result_row);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_3__reduce_cython__ = {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_3__reduce_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  if (unlikely(__pyx_nargs > 0)) { __Pyx_RaiseArgtupleInvalid("__reduce_cython__", 1, 0, 0, __pyx_nargs); return NULL; }
  const Py_ssize_t __pyx_kwds_len = unlikely(__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
  if (unlikely(__pyx_kwds_len < 0)) return NULL;
  if (unlikely(__pyx_kwds_len > 0)) {__Pyx_RejectKeywords("__reduce_cython__", __pyx_kwds); return NULL;}
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_2__reduce_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_2__reduce_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45]))
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);
  __Pyx_TraceStartFunc("__reduce_cython__", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self._allowed_lateness, self._internal_timer_service, self._key_selector, self._rowtime_index, self._shift_timezone, self._state_backend, self._state_value_coder, self._trigger, self._trigger_context, self._window_aggregator, self._window_assigner, self._window_context, self._window_function, self._window_state)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
*/
  __pyx_t_1 = __Pyx_PyLong_From_int64_t(__pyx_v_self->__pyx_base._allowed_lateness); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyLong_From_int(__pyx_v_self->__pyx_base._rowtime_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(14); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._internal_timer_service);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._internal_timer_service);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_v_self->__pyx_base._internal_timer_service) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base._key_selector);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base._key_selector);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, ((PyObject *)__pyx_v_self->__pyx_base._key_selector)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_GIVEREF(__pyx_t_2);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 3, __pyx_t_2) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._shift_timezone);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._shift_timezone);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 4, __pyx_v_self->__pyx_base._shift_timezone) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._state_backend);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._state_backend);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 5, __pyx_v_self->__pyx_base._state_backend) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._state_value_coder);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._state_value_coder);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 6, __pyx_v_self->__pyx_base._state_value_coder) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._trigger);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._trigger);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 7, __pyx_v_self->__pyx_base._trigger) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._trigger_context);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._trigger_context);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 8, __pyx_v_self->__pyx_base._trigger_context) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF((PyObject *)__pyx_v_self->__pyx_base._window_aggregator);
  __Pyx_GIVEREF((PyObject *)__pyx_v_self->__pyx_base._window_aggregator);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 9, ((PyObject *)__pyx_v_self->__pyx_base._window_aggregator)) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._window_assigner);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._window_assigner);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 10, __pyx_v_self->__pyx_base._window_assigner) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._window_context);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._window_context);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 11, __pyx_v_self->__pyx_base._window_context) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._window_function);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._window_function);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 12, __pyx_v_self->__pyx_base._window_function) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __Pyx_INCREF(__pyx_v_self->__pyx_base._window_state);
  __Pyx_GIVEREF(__pyx_v_self->__pyx_base._window_state);
  if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 13, __pyx_v_self->__pyx_base._window_state) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self._allowed_lateness, self._internal_timer_service, self._key_selector, self._rowtime_index, self._shift_timezone, self._state_backend, self._state_value_coder, self._trigger, self._trigger_context, self._window_aggregator, self._window_assigner, self._window_context, self._window_function, self._window_state)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
*/
  __pyx_t_3 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_mstate_global->__pyx_n_u_dict, Py_None); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v__dict = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":7
 *     state = (self._allowed_lateness, self._internal_timer_service, self._key_selector, self._rowtime_index, self._shift_timezone, self._state_backend, self._state_value_coder, self._trigger, self._trigger_context, self._window_aggregator, self._window_assigner, self._window_context, self._window_function, self._window_state)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
  __pyx_t_4 = (__pyx_v__dict != Py_None);
  if (__pyx_t_4) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
*/
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v__dict) != (0)) __PYX_ERR(1, 8, __pyx_L1_error);
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_2));
    __pyx_t_2 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None
*/
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self._allowed_lateness, self._internal_timer_service, self._key_selector, self._rowtime_index, self._shift_timezone, self._state_backend, self._state_value_coder, self._trigger, self._trigger_context, self._window_aggregator, self._window_assigner, self._window_context, self._window_function, self._window_state)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
*/
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, None), state
*/
  /*else*/ {
    __pyx_t_5 = (__pyx_v_self->__pyx_base._internal_timer_service != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base._key_selector) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._shift_timezone != ((PyObject*)Py_None));
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._state_backend != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._state_value_coder != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._trigger != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._trigger_context != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (((PyObject *)__pyx_v_self->__pyx_base._window_aggregator) != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._window_assigner != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._window_context != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._window_function != Py_None);
    if (!__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_5 = (__pyx_v_self->__pyx_base._window_state != Py_None);
    __pyx_t_4 = __pyx_t_5;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_4;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, None), state
 *     else:
*/
  if (__pyx_v_use_setstate) {

    /* "(tree fragment)":13
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None
 *     if use_setstate:
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, state)
*/
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_GroupWindowAggFun_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_209029984);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_209029984);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_int_209029984) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, Py_None) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_t_3) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 13, __pyx_L1_error);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_1;
    __pyx_t_1 = 0;
    __Pyx_TraceReturnValue(__pyx_r, 106, 0, __PYX_ERR(1, 13, __pyx_L1_error));
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self._internal_timer_service is not None or self._key_selector is not None or self._shift_timezone is not None or self._state_backend is not None or self._state_value_coder is not None or self._trigger is not None or self._trigger_context is not None or self._window_aggregator is not None or self._window_assigner is not None or self._window_context is not None or self._window_function is not None or self._window_state is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, None), state
 *     else:
*/
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, None), state
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GroupWindowAggFunction__set_state(self, __pyx_state)
*/
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_GroupWindowAggFun_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self)))) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_int_209029984);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_int_209029984);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_mstate_global->__pyx_int_209029984) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_v_state) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __Pyx_GIVEREF(__pyx_t_3);
    if (__Pyx_PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3) != (0)) __PYX_ERR(1, 15, __pyx_L1_error);
    __pyx_t_1 = 0;
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_TraceReturnValue(__pyx_r, 114, 0, __PYX_ERR(1, 15, __pyx_L1_error));
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunction.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GroupWindowAggFunction__set_state(self, __pyx_state)
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_5__setstate_cython__ = {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_5__setstate_cython__(PyObject *__pyx_v_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[1] = {0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__setstate_cython__", 0) < 0) __PYX_ERR(1, 16, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 1; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, i); __PYX_ERR(1, 16, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 1)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 16, __pyx_L3_error)
    }
    __pyx_v___pyx_state = values[0];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__setstate_cython__", 1, 1, 1, __pyx_nargs); __PYX_ERR(1, 16, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_4__setstate_cython__(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *)__pyx_v_self), __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_4__setstate_cython__(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46]))
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);
  __Pyx_TraceStartFunc("__setstate_cython__", __pyx_f[1], 16, 0, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));

  /* "(tree fragment)":17
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_GroupWindowAggFunction__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
*/
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_GroupWindowAggFunction__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GroupWindowAggFunction__set_state(self, __pyx_state)
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 16, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunction.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_NamespaceAggsHandleFunctionBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_1__pyx_unpickle_NamespaceAggsHandleFunctionBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_1__pyx_unpickle_NamespaceAggsHandleFunctionBase = {"__pyx_unpickle_NamespaceAggsHandleFunctionBase", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_1__pyx_unpickle_NamespaceAggsHandleFunctionBase, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_1__pyx_unpickle_NamespaceAggsHandleFunctionBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_NamespaceAggsHandleFunctionBase (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 1, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_NamespaceAggsHandleFunctionBase", 0) < 0) __PYX_ERR(1, 1, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NamespaceAggsHandleFunctionBase", 1, 3, 3, i); __PYX_ERR(1, 1, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 1, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 1, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 1, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NamespaceAggsHandleFunctionBase", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_NamespaceAggsHandleFunctionBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunctionBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunctionBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47]))
  __Pyx_RefNannySetupContext("__pyx_unpickle_NamespaceAggsHandleFunctionBase", 0);
  __Pyx_TraceStartFunc("__pyx_unpickle_NamespaceAggsHandleFunctionBase", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
*/
  __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[0], Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = NamespaceAggsHandleFunctionBase.__new__(__pyx_type)
*/
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_PickleError);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_mstate_global->__pyx_n_u_PickleError) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = NamespaceAggsHandleFunctionBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
    __pyx_t_3 = __Pyx_PyLong_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
*/
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = NamespaceAggsHandleFunctionBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(<NamespaceAggsHandleFunctionBase> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = NamespaceAggsHandleFunctionBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(<NamespaceAggsHandleFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = NamespaceAggsHandleFunctionBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(<NamespaceAggsHandleFunctionBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(NamespaceAggsHandleFunctionBase __pyx_result, tuple __pyx_state):
*/
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = NamespaceAggsHandleFunctionBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(<NamespaceAggsHandleFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(<NamespaceAggsHandleFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(NamespaceAggsHandleFunctionBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  __Pyx_TraceReturnValue(__pyx_r, 29, 0, __PYX_ERR(1, 10, __pyx_L1_error));
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NamespaceAggsHandleFunctionBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_NamespaceAggsHandleFunctionBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(<NamespaceAggsHandleFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(NamespaceAggsHandleFunctionBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/

static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[48]))
  __Pyx_RefNannySetupContext("__pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state", 0);
  __Pyx_TraceStartFunc("__pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state", __pyx_f[1], 11, 0, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(NamespaceAggsHandleFunctionBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_mstate_global->__pyx_n_u_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(NamespaceAggsHandleFunctionBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_mstate_global->__pyx_n_u_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_t_6;
    __Pyx_INCREF(__pyx_t_5);
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(NamespaceAggsHandleFunctionBase __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(<NamespaceAggsHandleFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(NamespaceAggsHandleFunctionBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_NamespaceAggsHandleFunction(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_3__pyx_unpickle_NamespaceAggsHandleFunction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_3__pyx_unpickle_NamespaceAggsHandleFunction = {"__pyx_unpickle_NamespaceAggsHandleFunction", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_3__pyx_unpickle_NamespaceAggsHandleFunction, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_3__pyx_unpickle_NamespaceAggsHandleFunction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_NamespaceAggsHandleFunction (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 1, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_NamespaceAggsHandleFunction", 0) < 0) __PYX_ERR(1, 1, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NamespaceAggsHandleFunction", 1, 3, 3, i); __PYX_ERR(1, 1, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 1, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 1, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 1, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_NamespaceAggsHandleFunction", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_NamespaceAggsHandleFunction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_2__pyx_unpickle_NamespaceAggsHandleFunction(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_2__pyx_unpickle_NamespaceAggsHandleFunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49]))
  __Pyx_RefNannySetupContext("__pyx_unpickle_NamespaceAggsHandleFunction", 0);
  __Pyx_TraceStartFunc("__pyx_unpickle_NamespaceAggsHandleFunction", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
*/
  __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[0], Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = NamespaceAggsHandleFunction.__new__(__pyx_type)
*/
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_PickleError);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_mstate_global->__pyx_n_u_PickleError) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = NamespaceAggsHandleFunction.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
    __pyx_t_3 = __Pyx_PyLong_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_Incompatible_checksums_0x_x_vs_0, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
*/
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = NamespaceAggsHandleFunction.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NamespaceAggsHandleFunction__set_state(<NamespaceAggsHandleFunction> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = NamespaceAggsHandleFunction.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NamespaceAggsHandleFunction__set_state(<NamespaceAggsHandleFunction> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = NamespaceAggsHandleFunction.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NamespaceAggsHandleFunction__set_state(<NamespaceAggsHandleFunction> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunction__set_state(NamespaceAggsHandleFunction __pyx_result, tuple __pyx_state):
*/
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunction__set_state(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
 *     __pyx_result = NamespaceAggsHandleFunction.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_NamespaceAggsHandleFunction__set_state(<NamespaceAggsHandleFunction> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_NamespaceAggsHandleFunction__set_state(<NamespaceAggsHandleFunction> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_NamespaceAggsHandleFunction__set_state(NamespaceAggsHandleFunction __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  __Pyx_TraceReturnValue(__pyx_r, 29, 0, __PYX_ERR(1, 10, __pyx_L1_error));
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NamespaceAggsHandleFunction(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_NamespaceAggsHandleFunction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_NamespaceAggsHandleFunction__set_state(<NamespaceAggsHandleFunction> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunction__set_state(NamespaceAggsHandleFunction __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/

static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_NamespaceAggsHandleFunction__set_state(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  size_t __pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[50]))
  __Pyx_RefNannySetupContext("__pyx_unpickle_NamespaceAggsHandleFunction__set_state", 0);
  __Pyx_TraceStartFunc("__pyx_unpickle_NamespaceAggsHandleFunction__set_state", __pyx_f[1], 11, 0, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunction__set_state(NamespaceAggsHandleFunction __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_2 == ((Py_ssize_t)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_3 = (__pyx_t_2 > 0);
  if (__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_mstate_global->__pyx_n_u_dict); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_1 = __pyx_t_3;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_NamespaceAggsHandleFunction__set_state(NamespaceAggsHandleFunction __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])             # <<<<<<<<<<<<<<
*/
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_mstate_global->__pyx_n_u_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_5 = __pyx_t_6;
    __Pyx_INCREF(__pyx_t_5);
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 13, __pyx_L1_error)
    }
    __pyx_t_7 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_5, __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)};
      __pyx_t_4 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_7, (2-__pyx_t_7) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunction__set_state(NamespaceAggsHandleFunction __pyx_result, tuple __pyx_state):
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_NamespaceAggsHandleFunction__set_state(<NamespaceAggsHandleFunction> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunction__set_state(NamespaceAggsHandleFunction __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_NamespaceAggsHandleFunction__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GroupWindowAggFunctionBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_5__pyx_unpickle_GroupWindowAggFunctionBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_5__pyx_unpickle_GroupWindowAggFunctionBase = {"__pyx_unpickle_GroupWindowAggFunctionBase", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_5__pyx_unpickle_GroupWindowAggFunctionBase, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_5__pyx_unpickle_GroupWindowAggFunctionBase(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GroupWindowAggFunctionBase (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 1, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_GroupWindowAggFunctionBase", 0) < 0) __PYX_ERR(1, 1, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GroupWindowAggFunctionBase", 1, 3, 3, i); __PYX_ERR(1, 1, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 1, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 1, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 1, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GroupWindowAggFunctionBase", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_GroupWindowAggFunctionBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_4__pyx_unpickle_GroupWindowAggFunctionBase(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_4__pyx_unpickle_GroupWindowAggFunctionBase(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51]))
  __Pyx_RefNannySetupContext("__pyx_unpickle_GroupWindowAggFunctionBase", 0);
  __Pyx_TraceStartFunc("__pyx_unpickle_GroupWindowAggFunctionBase", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc758b60, 0xb13151a, 0xab32e9d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
*/
  __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[1], Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc758b60, 0xb13151a, 0xab32e9d):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
 *     __pyx_result = GroupWindowAggFunctionBase.__new__(__pyx_type)
*/
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_PickleError);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_mstate_global->__pyx_n_u_PickleError) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc758b60, 0xb13151a, 0xab32e9d):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = GroupWindowAggFunctionBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
    __pyx_t_3 = __Pyx_PyLong_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc758b60, 0xb13151a, 0xab32e9d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
*/
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
 *     __pyx_result = GroupWindowAggFunctionBase.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GroupWindowAggFunctionBase__set_state(<GroupWindowAggFunctionBase> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
 *     __pyx_result = GroupWindowAggFunctionBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GroupWindowAggFunctionBase__set_state(<GroupWindowAggFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GroupWindowAggFunctionBase.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GroupWindowAggFunctionBase__set_state(<GroupWindowAggFunctionBase> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunctionBase__set_state(GroupWindowAggFunctionBase __pyx_result, tuple __pyx_state):
*/
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_GroupWindowAggFunctionBase__set_state(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
 *     __pyx_result = GroupWindowAggFunctionBase.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GroupWindowAggFunctionBase__set_state(<GroupWindowAggFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GroupWindowAggFunctionBase__set_state(<GroupWindowAggFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GroupWindowAggFunctionBase__set_state(GroupWindowAggFunctionBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  __Pyx_TraceReturnValue(__pyx_r, 29, 0, __PYX_ERR(1, 10, __pyx_L1_error));
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GroupWindowAggFunctionBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_GroupWindowAggFunctionBase", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GroupWindowAggFunctionBase__set_state(<GroupWindowAggFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunctionBase__set_state(GroupWindowAggFunctionBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
*/

static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_GroupWindowAggFunctionBase__set_state(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[52]))
  __Pyx_RefNannySetupContext("__pyx_unpickle_GroupWindowAggFunctionBase__set_state", 0);
  __Pyx_TraceStartFunc("__pyx_unpickle_GroupWindowAggFunctionBase__set_state", __pyx_f[1], 11, 0, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunctionBase__set_state(GroupWindowAggFunctionBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])
*/
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyLong_As_int64_t(__Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->_allowed_lateness = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_internal_timer_service);
  __Pyx_DECREF(__pyx_v___pyx_result->_internal_timer_service);
  __pyx_v___pyx_result->_internal_timer_service = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 2);
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->_key_selector);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->_key_selector);
  __pyx_v___pyx_result->_key_selector = ((struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *)__pyx_t_2);
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyLong_As_int(__Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->_rowtime_index = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 4);
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_shift_timezone);
  __Pyx_DECREF(__pyx_v___pyx_result->_shift_timezone);
  __pyx_v___pyx_result->_shift_timezone = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 5);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_state_backend);
  __Pyx_DECREF(__pyx_v___pyx_result->_state_backend);
  __pyx_v___pyx_result->_state_backend = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 6);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_state_value_coder);
  __Pyx_DECREF(__pyx_v___pyx_result->_state_value_coder);
  __pyx_v___pyx_result->_state_value_coder = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 7);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_trigger);
  __Pyx_DECREF(__pyx_v___pyx_result->_trigger);
  __pyx_v___pyx_result->_trigger = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 8);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_trigger_context);
  __Pyx_DECREF(__pyx_v___pyx_result->_trigger_context);
  __pyx_v___pyx_result->_trigger_context = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 9);
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->_window_aggregator);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->_window_aggregator);
  __pyx_v___pyx_result->_window_aggregator = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_t_2);
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 10);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_window_assigner);
  __Pyx_DECREF(__pyx_v___pyx_result->_window_assigner);
  __pyx_v___pyx_result->_window_assigner = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 11);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_window_context);
  __Pyx_DECREF(__pyx_v___pyx_result->_window_context);
  __pyx_v___pyx_result->_window_context = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 12);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_window_function);
  __Pyx_DECREF(__pyx_v___pyx_result->_window_function);
  __pyx_v___pyx_result->_window_function = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 13);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->_window_state);
  __Pyx_DECREF(__pyx_v___pyx_result->_window_state);
  __pyx_v___pyx_result->_window_state = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GroupWindowAggFunctionBase__set_state(GroupWindowAggFunctionBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
*/
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 14);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_mstate_global->__pyx_n_u_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])             # <<<<<<<<<<<<<<
*/
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_mstate_global->__pyx_n_u_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __pyx_t_8;
    __Pyx_INCREF(__pyx_t_7);
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_9 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 14)};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GroupWindowAggFunctionBase__set_state(GroupWindowAggFunctionBase __pyx_result, tuple __pyx_state):
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
*/
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GroupWindowAggFunctionBase__set_state(<GroupWindowAggFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunctionBase__set_state(GroupWindowAggFunctionBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_GroupWindowAggFunctionBase__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_GroupWindowAggFunction(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/

/* Python wrapper */
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_7__pyx_unpickle_GroupWindowAggFunction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
); /*proto*/
static PyMethodDef __pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_7__pyx_unpickle_GroupWindowAggFunction = {"__pyx_unpickle_GroupWindowAggFunction", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_7__pyx_unpickle_GroupWindowAggFunction, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_7__pyx_unpickle_GroupWindowAggFunction(PyObject *__pyx_self, 
#if CYTHON_METH_FASTCALL
PyObject *const *__pyx_args, Py_ssize_t __pyx_nargs, PyObject *__pyx_kwds
#else
PyObject *__pyx_args, PyObject *__pyx_kwds
#endif
) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  #if !CYTHON_METH_FASTCALL
  CYTHON_UNUSED Py_ssize_t __pyx_nargs;
  #endif
  CYTHON_UNUSED PyObject *const *__pyx_kwvalues;
  PyObject* values[3] = {0,0,0};
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_GroupWindowAggFunction (wrapper)", 0);
  #if !CYTHON_METH_FASTCALL
  #if CYTHON_ASSUME_SAFE_SIZE
  __pyx_nargs = PyTuple_GET_SIZE(__pyx_args);
  #else
  __pyx_nargs = PyTuple_Size(__pyx_args); if (unlikely(__pyx_nargs < 0)) return NULL;
  #endif
  #endif
  __pyx_kwvalues = __Pyx_KwValues_FASTCALL(__pyx_args, __pyx_nargs);
  {
    PyObject ** const __pyx_pyargnames[] = {&__pyx_mstate_global->__pyx_n_u_pyx_type,&__pyx_mstate_global->__pyx_n_u_pyx_checksum,&__pyx_mstate_global->__pyx_n_u_pyx_state,0};
    const Py_ssize_t __pyx_kwds_len = (__pyx_kwds) ? __Pyx_NumKwargs_FASTCALL(__pyx_kwds) : 0;
    if (unlikely(__pyx_kwds_len) < 0) __PYX_ERR(1, 1, __pyx_L3_error)
    if (__pyx_kwds_len > 0) {
      switch (__pyx_nargs) {
        case  3:
        values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  2:
        values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  1:
        values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
        if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 1, __pyx_L3_error)
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      const Py_ssize_t kwd_pos_args = __pyx_nargs;
      if (__Pyx_ParseKeywords(__pyx_kwds, __pyx_kwvalues, __pyx_pyargnames, 0, values, kwd_pos_args, __pyx_kwds_len, "__pyx_unpickle_GroupWindowAggFunction", 0) < 0) __PYX_ERR(1, 1, __pyx_L3_error)
      for (Py_ssize_t i = __pyx_nargs; i < 3; i++) {
        if (unlikely(!values[i])) { __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GroupWindowAggFunction", 1, 3, 3, i); __PYX_ERR(1, 1, __pyx_L3_error) }
      }
    } else if (unlikely(__pyx_nargs != 3)) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = __Pyx_ArgRef_FASTCALL(__pyx_args, 0);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[0])) __PYX_ERR(1, 1, __pyx_L3_error)
      values[1] = __Pyx_ArgRef_FASTCALL(__pyx_args, 1);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[1])) __PYX_ERR(1, 1, __pyx_L3_error)
      values[2] = __Pyx_ArgRef_FASTCALL(__pyx_args, 2);
      if (!CYTHON_ASSUME_SAFE_MACROS && unlikely(!values[2])) __PYX_ERR(1, 1, __pyx_L3_error)
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyLong_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L6_skip;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_GroupWindowAggFunction", 1, 3, 3, __pyx_nargs); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L6_skip:;
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L3_error:;
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_GroupWindowAggFunction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_6__pyx_unpickle_GroupWindowAggFunction(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  for (Py_ssize_t __pyx_temp=0; __pyx_temp < (Py_ssize_t)(sizeof(values)/sizeof(values[0])); ++__pyx_temp) {
    Py_XDECREF(values[__pyx_temp]);
  }
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_7pyflink_12fn_execution_5table_21window_aggregate_fast_6__pyx_unpickle_GroupWindowAggFunction(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[53]))
  __Pyx_RefNannySetupContext("__pyx_unpickle_GroupWindowAggFunction", 0);
  __Pyx_TraceStartFunc("__pyx_unpickle_GroupWindowAggFunction", __pyx_f[1], 1, 0, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc758b60, 0xb13151a, 0xab32e9d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
*/
  __pyx_t_1 = __Pyx_PyLong_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = (__Pyx_PySequence_ContainsTF(__pyx_t_1, __pyx_mstate_global->__pyx_tuple[1], Py_NE)); if (unlikely((__pyx_t_2 < 0))) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc758b60, 0xb13151a, 0xab32e9d):
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
 *     __pyx_result = GroupWindowAggFunction.__new__(__pyx_type)
*/
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_mstate_global->__pyx_n_u_PickleError);
    __Pyx_GIVEREF(__pyx_mstate_global->__pyx_n_u_PickleError);
    if (__Pyx_PyList_SET_ITEM(__pyx_t_1, 0, __pyx_mstate_global->__pyx_n_u_PickleError) != (0)) __PYX_ERR(1, 5, __pyx_L1_error);
    __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pickle, __pyx_t_1, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PickleError); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_t_1);
    __pyx_v___pyx_PickleError = __pyx_t_1;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum not in (0xc758b60, 0xb13151a, 0xab32e9d):
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum             # <<<<<<<<<<<<<<
 *     __pyx_result = GroupWindowAggFunction.__new__(__pyx_type)
 *     if __pyx_state is not None:
*/
    __pyx_t_3 = __Pyx_PyLong_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = PyUnicode_Format(__pyx_mstate_global->__pyx_kp_u_Incompatible_checksums_0x_x_vs_0_2, __pyx_t_3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_v___pyx_PickleError, __pyx_t_1, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xc758b60, 0xb13151a, 0xab32e9d):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
*/
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
 *     __pyx_result = GroupWindowAggFunction.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GroupWindowAggFunction__set_state(<GroupWindowAggFunction> __pyx_result, __pyx_state)
*/
  __pyx_t_3 = ((PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_t_4 = 0;
  {
    PyObject *__pyx_callargs[2] = {__pyx_t_3, __pyx_v___pyx_type};
    __pyx_t_1 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_new, __pyx_callargs+__pyx_t_4, (2-__pyx_t_4) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  }
  __pyx_v___pyx_result = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
 *     __pyx_result = GroupWindowAggFunction.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GroupWindowAggFunction__set_state(<GroupWindowAggFunction> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  __pyx_t_2 = (__pyx_v___pyx_state != Py_None);
  if (__pyx_t_2) {

    /* "(tree fragment)":9
 *     __pyx_result = GroupWindowAggFunction.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GroupWindowAggFunction__set_state(<GroupWindowAggFunction> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunction__set_state(GroupWindowAggFunction __pyx_result, tuple __pyx_state):
*/
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None) || __Pyx_RaiseUnexpectedTypeError("tuple", __pyx_v___pyx_state))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_1 = __pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_GroupWindowAggFunction__set_state(((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xc758b60, 0xb13151a, 0xab32e9d) = (_allowed_lateness, _internal_timer_service, _key_selector, _rowtime_index, _shift_timezone, _state_backend, _state_value_coder, _trigger, _trigger_context, _window_aggregator, _window_assigner, _window_context, _window_function, _window_state))" % __pyx_checksum
 *     __pyx_result = GroupWindowAggFunction.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_GroupWindowAggFunction__set_state(<GroupWindowAggFunction> __pyx_result, __pyx_state)
 *     return __pyx_result
*/
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_GroupWindowAggFunction__set_state(<GroupWindowAggFunction> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_GroupWindowAggFunction__set_state(GroupWindowAggFunction __pyx_result, tuple __pyx_state):
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
*/
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  __Pyx_TraceReturnValue(__pyx_r, 29, 0, __PYX_ERR(1, 10, __pyx_L1_error));
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GroupWindowAggFunction(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 1, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_GroupWindowAggFunction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_GroupWindowAggFunction__set_state(<GroupWindowAggFunction> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunction__set_state(GroupWindowAggFunction __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
*/

static PyObject *__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast___pyx_unpickle_GroupWindowAggFunction__set_state(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_TraceDeclarationsFunc
  __Pyx_RefNannyDeclarations
  int64_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  size_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_TraceFrameInit(((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[54]))
  __Pyx_RefNannySetupContext("__pyx_unpickle_GroupWindowAggFunction__set_state", 0);
  __Pyx_TraceStartFunc("__pyx_unpickle_GroupWindowAggFunction__set_state", __pyx_f[1], 11, 0, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunction__set_state(GroupWindowAggFunction __pyx_result, tuple __pyx_state):
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])
*/
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyLong_As_int64_t(__Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)); if (unlikely((__pyx_t_1 == ((int64_t)-1)) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base._allowed_lateness = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 1);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._internal_timer_service);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._internal_timer_service);
  __pyx_v___pyx_result->__pyx_base._internal_timer_service = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 2);
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base._key_selector);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base._key_selector);
  __pyx_v___pyx_result->__pyx_base._key_selector = ((struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *)__pyx_t_2);
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyLong_As_int(__Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->__pyx_base._rowtime_index = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 4);
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(PyUnicode_CheckExact(__pyx_t_2))||((__pyx_t_2) == Py_None) || __Pyx_RaiseUnexpectedTypeError("str", __pyx_t_2))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._shift_timezone);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._shift_timezone);
  __pyx_v___pyx_result->__pyx_base._shift_timezone = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 5);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._state_backend);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._state_backend);
  __pyx_v___pyx_result->__pyx_base._state_backend = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 6);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._state_value_coder);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._state_value_coder);
  __pyx_v___pyx_result->__pyx_base._state_value_coder = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 7);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._trigger);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._trigger);
  __pyx_v___pyx_result->__pyx_base._trigger = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 8);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._trigger_context);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._trigger_context);
  __pyx_v___pyx_result->__pyx_base._trigger_context = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 9);
  __Pyx_INCREF(__pyx_t_2);
  if (!(likely(((__pyx_t_2) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_2, __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF((PyObject *)__pyx_v___pyx_result->__pyx_base._window_aggregator);
  __Pyx_DECREF((PyObject *)__pyx_v___pyx_result->__pyx_base._window_aggregator);
  __pyx_v___pyx_result->__pyx_base._window_aggregator = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)__pyx_t_2);
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 10);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._window_assigner);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._window_assigner);
  __pyx_v___pyx_result->__pyx_base._window_assigner = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 11);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._window_context);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._window_context);
  __pyx_v___pyx_result->__pyx_base._window_context = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 12);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._window_function);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._window_function);
  __pyx_v___pyx_result->__pyx_base._window_function = __pyx_t_2;
  __pyx_t_2 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 13);
  __Pyx_INCREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v___pyx_result->__pyx_base._window_state);
  __Pyx_DECREF(__pyx_v___pyx_result->__pyx_base._window_state);
  __pyx_v___pyx_result->__pyx_base._window_state = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_GroupWindowAggFunction__set_state(GroupWindowAggFunction __pyx_result, tuple __pyx_state):
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
*/
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_5 = __Pyx_PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_6 = (__pyx_t_5 > 14);
  if (__pyx_t_6) {
  } else {
    __pyx_t_4 = __pyx_t_6;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_6 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_mstate_global->__pyx_n_u_dict); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = __pyx_t_6;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_4) {

    /* "(tree fragment)":14
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[14])             # <<<<<<<<<<<<<<
*/
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_mstate_global->__pyx_n_u_dict); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __pyx_t_8;
    __Pyx_INCREF(__pyx_t_7);
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_9 = 0;
    {
      PyObject *__pyx_callargs[2] = {__pyx_t_7, __Pyx_PyTuple_GET_ITEM(__pyx_v___pyx_state, 14)};
      __pyx_t_2 = __Pyx_PyObject_FastCallMethod(__pyx_mstate_global->__pyx_n_u_update, __pyx_callargs+__pyx_t_9, (2-__pyx_t_9) | (1*__Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET));
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 14, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_GroupWindowAggFunction__set_state(GroupWindowAggFunction __pyx_result, tuple __pyx_state):
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[14])
*/
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_GroupWindowAggFunction__set_state(<GroupWindowAggFunction> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunction__set_state(GroupWindowAggFunction __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
*/

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_TraceReturnValue(__pyx_r, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  #if CYTHON_USE_SYS_MONITORING
  __Pyx_TraceExceptionUnwind(0, 0);
  #else
  __Pyx_TraceReturnValue(NULL, 0, 0, __PYX_ERR(1, 11, __pyx_L1_error));
  #endif
  __Pyx_AddTraceback("pyflink.fn_execution.table.window_aggregate_fast.__pyx_unpickle_GroupWindowAggFunction__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_PyMonitoring_ExitScope(0);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
/* #### Code section: module_exttypes ### */
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;

static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;
  return o;
}

static void __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase(PyObject *o) {
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && (!PyType_IS_GC(Py_TYPE(o)) || !__Pyx_PyObject_GC_IsFinalized(o))) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  #if CYTHON_USE_TYPE_SLOTS
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static PyMethodDef __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_9__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_11__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase},
  {Py_tp_methods, (void *)__pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase},
  {Py_tp_new, (void *)__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase_spec = {
  "pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunctionBase",
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase_slots,
};
#else

static PyTypeObject __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyflink.fn_execution.table.window_aggregate_fast.""NamespaceAggsHandleFunctionBase", /*tp_name*/
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;

static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *p;
  PyObject *o = __pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase*)__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;
  return o;
}

static PyMethodDef __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction_slots[] = {
  {Py_tp_methods, (void *)__pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction},
  {Py_tp_new, (void *)__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction_spec = {
  "pyflink.fn_execution.table.window_aggregate_fast.NamespaceAggsHandleFunction",
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE,
  __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction_slots,
};
#else

static PyTypeObject __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyflink.fn_execution.table.window_aggregate_fast.""NamespaceAggsHandleFunction", /*tp_name*/
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction;

static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *p;
  PyObject *o = __pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)o);
  p->__pyx_base.__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase*)__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction;
  p->_udfs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_input_extractors = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_named_property_extractor = Py_None; Py_INCREF(Py_None);
  p->_accumulators = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_udf_data_view_specs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_udf_data_views = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_distinct_view_descriptors = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_distinct_data_views = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->named_property_extractor = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction(PyObject *o) {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *p = (struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_13__dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->_udfs);
  Py_CLEAR(p->_input_extractors);
  Py_CLEAR(p->_named_property_extractor);
  Py_CLEAR(p->_accumulators);
  Py_CLEAR(p->_udf_data_view_specs);
  Py_CLEAR(p->_udf_data_views);
  Py_CLEAR(p->_distinct_view_descriptors);
  Py_CLEAR(p->_distinct_data_views);
  Py_CLEAR(p->named_property_extractor);
  __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase(o);
}

static int __pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *p = (struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)o;
  #if !CYTHON_USE_MODULE_STATE
  e = 0;
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction)) {
    traverseproc traverse = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction, tp_traverse, traverseproc);
    if (traverse) { e = traverse(o, v, a); }
  } else
  #endif
  { e = __Pyx_call_next_tp_traverse(o, v, a, __pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction); }
  if (e) return e;
  {
    e = __Pyx_call_type_traverse(o, 0, v, a);
    if (e) return e;
  }
  if (p->_udfs) {
    e = (*v)(p->_udfs, a); if (e) return e;
  }
  if (p->_input_extractors) {
    e = (*v)(p->_input_extractors, a); if (e) return e;
  }
  if (p->_named_property_extractor) {
    e = (*v)(p->_named_property_extractor, a); if (e) return e;
  }
  if (p->_accumulators) {
    e = (*v)(p->_accumulators, a); if (e) return e;
  }
  if (p->_udf_data_view_specs) {
    e = (*v)(p->_udf_data_view_specs, a); if (e) return e;
  }
  if (p->_udf_data_views) {
    e = (*v)(p->_udf_data_views, a); if (e) return e;
  }
  if (p->_distinct_view_descriptors) {
    e = (*v)(p->_distinct_view_descriptors, a); if (e) return e;
  }
  if (p->_distinct_data_views) {
    e = (*v)(p->_distinct_data_views, a); if (e) return e;
  }
  if (p->named_property_extractor) {
    e = (*v)(p->named_property_extractor, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *p = (struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction *)o;
  #if !CYTHON_USE_MODULE_STATE
  if (likely(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction)) {
    inquiry clear = __Pyx_PyType_GetSlot(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction, tp_clear, inquiry);
    if (clear) clear(o);
  } else
  #endif
  { __Pyx_call_next_tp_clear(o, __pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction); }
  tmp = ((PyObject*)p->_udfs);
  p->_udfs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_input_extractors);
  p->_input_extractors = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_named_property_extractor);
  p->_named_property_extractor = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_accumulators);
  p->_accumulators = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_udf_data_view_specs);
  p->_udf_data_view_specs = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_udf_data_views);
  p->_udf_data_views = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_distinct_view_descriptors);
  p->_distinct_view_descriptors = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_distinct_data_views);
  p->_distinct_data_views = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->named_property_extractor);
  p->named_property_extractor = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_15__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_17__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction},
  {Py_tp_methods, (void *)__pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction},
  {Py_tp_init, (void *)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction_spec = {
  "pyflink.fn_execution.table.window_aggregate_fast.SimpleNamespaceAggsHandleFunction",
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction_slots,
};
#else

static PyTypeObject __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyflink.fn_execution.table.window_aggregate_fast.""SimpleNamespaceAggsHandleFunction", /*tp_name*/
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, /*tp_traverse*/
  __pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;

static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *p;
  PyObject *o;
  #if CYTHON_COMPILING_IN_LIMITED_API
  allocfunc alloc_func = (allocfunc)PyType_GetSlot(t, Py_tp_alloc);
  o = alloc_func(t, 0);
  #else
  if (likely(!__Pyx_PyType_HasFeature(t, Py_TPFLAGS_IS_ABSTRACT))) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_mstate_global->__pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  #endif
  p = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)o);
  p->__pyx_vtab = __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;
  p->_shift_timezone = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_key_selector = ((struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *)Py_None); Py_INCREF(Py_None);
  p->_state_value_coder = Py_None; Py_INCREF(Py_None);
  p->_state_backend = Py_None; Py_INCREF(Py_None);
  p->_window_assigner = Py_None; Py_INCREF(Py_None);
  p->_window_aggregator = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)Py_None); Py_INCREF(Py_None);
  p->_trigger = Py_None; Py_INCREF(Py_None);
  p->_window_function = Py_None; Py_INCREF(Py_None);
  p->_internal_timer_service = Py_None; Py_INCREF(Py_None);
  p->_window_context = Py_None; Py_INCREF(Py_None);
  p->_trigger_context = Py_None; Py_INCREF(Py_None);
  p->_window_state = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase(PyObject *o) {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *p = (struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely((PY_VERSION_HEX >= 0x03080000 || __Pyx_PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE)) && __Pyx_PyObject_GetSlot(o, tp_finalize, destructor)) && !__Pyx_PyObject_GC_IsFinalized(o)) {
    if (__Pyx_PyObject_GetSlot(o, tp_dealloc, destructor) == __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase) {
      if (PyObject_CallFinalizerFromDealloc(o)) return;
    }
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->_shift_timezone);
  Py_CLEAR(p->_key_selector);
  Py_CLEAR(p->_state_value_coder);
  Py_CLEAR(p->_state_backend);
  Py_CLEAR(p->_window_assigner);
  Py_CLEAR(p->_window_aggregator);
  Py_CLEAR(p->_trigger);
  Py_CLEAR(p->_window_function);
  Py_CLEAR(p->_internal_timer_service);
  Py_CLEAR(p->_window_context);
  Py_CLEAR(p->_trigger_context);
  Py_CLEAR(p->_window_state);
  #if CYTHON_USE_TYPE_SLOTS
  (*Py_TYPE(o)->tp_free)(o);
  #else
  {
    freefunc tp_free = (freefunc)PyType_GetSlot(Py_TYPE(o), Py_tp_free);
    if (tp_free) tp_free(o);
  }
  #endif
}

static int __pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *p = (struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)o;
  {
    e = __Pyx_call_type_traverse(o, 1, v, a);
    if (e) return e;
  }
  if (p->_key_selector) {
    e = (*v)(((PyObject *)p->_key_selector), a); if (e) return e;
  }
  if (p->_state_value_coder) {
    e = (*v)(p->_state_value_coder, a); if (e) return e;
  }
  if (p->_state_backend) {
    e = (*v)(p->_state_backend, a); if (e) return e;
  }
  if (p->_window_assigner) {
    e = (*v)(p->_window_assigner, a); if (e) return e;
  }
  if (p->_window_aggregator) {
    e = (*v)(((PyObject *)p->_window_aggregator), a); if (e) return e;
  }
  if (p->_trigger) {
    e = (*v)(p->_trigger, a); if (e) return e;
  }
  if (p->_window_function) {
    e = (*v)(p->_window_function, a); if (e) return e;
  }
  if (p->_internal_timer_service) {
    e = (*v)(p->_internal_timer_service, a); if (e) return e;
  }
  if (p->_window_context) {
    e = (*v)(p->_window_context, a); if (e) return e;
  }
  if (p->_trigger_context) {
    e = (*v)(p->_trigger_context, a); if (e) return e;
  }
  if (p->_window_state) {
    e = (*v)(p->_window_state, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *p = (struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *)o;
  tmp = ((PyObject*)p->_key_selector);
  p->_key_selector = ((struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_state_value_coder);
  p->_state_value_coder = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_state_backend);
  p->_state_backend = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_window_assigner);
  p->_window_assigner = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_window_aggregator);
  p->_window_aggregator = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_trigger);
  p->_trigger = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_window_function);
  p->_window_function = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_internal_timer_service);
  p->_internal_timer_service = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_window_context);
  p->_window_context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_trigger_context);
  p->_trigger_context = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_window_state);
  p->_window_state = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_21__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_23__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase_slots[] = {
  {Py_tp_dealloc, (void *)__pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase},
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase},
  {Py_tp_methods, (void *)__pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase},
  {Py_tp_init, (void *)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase_spec = {
  "pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunctionBase",
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase_slots,
};
#else

static PyTypeObject __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyflink.fn_execution.table.window_aggregate_fast.""GroupWindowAggFunctionBase", /*tp_name*/
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, /*tp_traverse*/
  __pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif
static struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction;

static PyObject *__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *p;
  PyObject *o = __pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase*)__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction;
  return o;
}

static PyMethodDef __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction[] = {
  {"__reduce_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_3__reduce_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {"__setstate_cython__", (PyCFunction)(void(*)(void))(__Pyx_PyCFunction_FastCallWithKeywords)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_5__setstate_cython__, __Pyx_METH_FASTCALL|METH_KEYWORDS, 0},
  {0, 0, 0, 0}
};
#if CYTHON_USE_TYPE_SPECS
static PyType_Slot __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction_slots[] = {
  {Py_tp_traverse, (void *)__pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase},
  {Py_tp_clear, (void *)__pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase},
  {Py_tp_methods, (void *)__pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction},
  {Py_tp_init, (void *)__pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_1__init__},
  {Py_tp_new, (void *)__pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction},
  {0, 0},
};
static PyType_Spec __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction_spec = {
  "pyflink.fn_execution.table.window_aggregate_fast.GroupWindowAggFunction",
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction),
  0,
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC,
  __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction_slots,
};
#else

static PyTypeObject __pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction = {
  PyVarObject_HEAD_INIT(0, 0)
  "pyflink.fn_execution.table.window_aggregate_fast.""GroupWindowAggFunction", /*tp_name*/
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  0, /*tp_as_async*/
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, /*tp_traverse*/
  __pyx_tp_clear_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  #if !CYTHON_USE_TYPE_SPECS
  0, /*tp_dictoffset*/
  #endif
  __pyx_pw_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if CYTHON_USE_TP_FINALIZE
  0, /*tp_finalize*/
  #else
  NULL, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1 && (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM >= 0x07030800)
  0, /*tp_vectorcall*/
  #endif
  #if __PYX_NEED_TP_PRINT_SLOT == 1
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030C0000
  0, /*tp_watched*/
  #endif
  #if PY_VERSION_HEX >= 0x030d00A4
  0, /*tp_versions_used*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX >= 0x03090000 && PY_VERSION_HEX < 0x030a0000
  0, /*tp_pypy_flags*/
  #endif
};
#endif

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};
/* #### Code section: initfunc_declarations ### */
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate); /*proto*/
/* #### Code section: init_module ### */

static int __Pyx_modinit_global_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase = &__pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.open = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_open;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.accumulate = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_accumulate;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.retract = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_retract;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.merge = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_merge;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.set_accumulators = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_set_accumulators;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.get_accumulators = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_get_accumulators;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.create_accumulators = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_create_accumulators;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.cleanup = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_cleanup;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase.close = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_close;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase)) __PYX_ERR(0, 48, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase_spec, __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase = &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunctionBase, (PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction = &__pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction.__pyx_base = *__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction.get_value = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_get_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction)) __PYX_ERR(0, 114, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction_spec, __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction = &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction->tp_base = __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction, __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunction, (PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 114, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction = &__pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base = *__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.open = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_open;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.accumulate = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_accumulate;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.retract = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_retract;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.merge = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_merge;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.set_accumulators = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_set_accumulators;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.get_accumulators = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_get_accumulators;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.create_accumulators = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_create_accumulators;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.cleanup = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_cleanup;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.__pyx_base.close = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_close;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction.__pyx_base.get_value = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_get_value;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction)) __PYX_ERR(0, 124, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction_spec, __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction = &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction->tp_base = __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_SimpleNamespaceAggsHandleFunctio, (PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase = &__pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.open = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_open;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.close = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_close;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.process_element = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_process_element;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.process_watermark = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, int64_t, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_process_watermark;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.on_event_time = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_on_event_time;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.on_processing_time = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_on_processing_time;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.get_timers = (PyObject *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_get_timers;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.to_utc_timestamp_mills = (int64_t (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, int64_t, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_to_utc_timestamp_mills;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase.cleanup_time = (int64_t (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, int __pyx_skip_dispatch))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_cleanup_time;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase._register_cleanup_timer = (void (*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase__register_cleanup_timer;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase._emit_window_result = (struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, PyObject *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase__emit_window_result;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase_spec, NULL); if (unlikely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase)) __PYX_ERR(0, 325, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase_spec, __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase = &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase, (PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase) < 0) __PYX_ERR(0, 325, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction = &__pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction.__pyx_base = *__pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;
  __pyx_vtable_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction.__pyx_base._emit_window_result = (struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow *(*)(struct __pyx_obj_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase *, PyObject *, PyObject *))__pyx_f_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction__emit_window_result;
  #if CYTHON_USE_TYPE_SPECS
  __pyx_t_1 = PyTuple_Pack(1, (PyObject *)__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction = (PyTypeObject *) __Pyx_PyType_FromModuleAndSpec(__pyx_m, &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction_spec, __pyx_t_1);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction)) __PYX_ERR(0, 506, __pyx_L1_error)
  if (__Pyx_fix_up_extension_type_from_spec(&__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction_spec, __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction) < 0) __PYX_ERR(0, 506, __pyx_L1_error)
  #else
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction = &__pyx_type_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction;
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction->tp_base = __pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase;
  #endif
  #if !CYTHON_USE_TYPE_SPECS
  if (__Pyx_PyType_Ready(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction) < 0) __PYX_ERR(0, 506, __pyx_L1_error)
  #endif
  #if !CYTHON_COMPILING_IN_LIMITED_API
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction->tp_dictoffset && __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction->tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction->tp_getattro = PyObject_GenericGetAttr;
  }
  #endif
  if (__Pyx_SetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction, __pyx_vtabptr_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction) < 0) __PYX_ERR(0, 506, __pyx_L1_error)
  if (__Pyx_MergeVtables(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction) < 0) __PYX_ERR(0, 506, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunction, (PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction) < 0) __PYX_ERR(0, 506, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject *) __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction) < 0) __PYX_ERR(0, 506, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyflink.fn_execution.stream_fast"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.stream_fast", "LengthPrefixInputStream",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream) __PYX_ERR(2, 21, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream = (struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_LengthPrefixInputStream)) __PYX_ERR(2, 21, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.stream_fast", "LengthPrefixOutputStream",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream) __PYX_ERR(2, 25, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream = (struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_LengthPrefixOutputStream)) __PYX_ERR(2, 25, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_InputStream = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.stream_fast", "InputStream",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_InputStream),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_InputStream) __PYX_ERR(2, 30, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_InputStream = (struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_InputStream*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_InputStream); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_InputStream)) __PYX_ERR(2, 30, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_OutputStream = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.stream_fast", "OutputStream",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_11stream_fast_OutputStream),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_OutputStream) __PYX_ERR(2, 44, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_OutputStream = (struct __pyx_vtabstruct_7pyflink_12fn_execution_11stream_fast_OutputStream*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_11stream_fast_OutputStream); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_11stream_fast_OutputStream)) __PYX_ERR(2, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyflink.fn_execution.coder_impl_fast"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "InternalRow",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InternalRow),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow) __PYX_ERR(3, 31, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InternalRow = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InternalRow*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InternalRow); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InternalRow)) __PYX_ERR(3, 31, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "MaskUtils",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils) __PYX_ERR(3, 39, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_MaskUtils)) __PYX_ERR(3, 39, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "LengthPrefixBaseCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl) __PYX_ERR(3, 50, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LengthPrefixBaseCoderImpl)) __PYX_ERR(3, 50, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "FieldCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl) __PYX_ERR(3, 58, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FieldCoderImpl)) __PYX_ERR(3, 58, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "InputStreamWrapper",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper) __PYX_ERR(3, 64, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InputStreamWrapper)) __PYX_ERR(3, 64, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "IterableCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl) __PYX_ERR(3, 71, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_IterableCoderImpl)) __PYX_ERR(3, 71, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "ValueCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl) __PYX_ERR(3, 75, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_ValueCoderImpl)) __PYX_ERR(3, 75, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "FlattenRowCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl) __PYX_ERR(3, 80, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FlattenRowCoderImpl)) __PYX_ERR(3, 80, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "RowCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl) __PYX_ERR(3, 86, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_RowCoderImpl)) __PYX_ERR(3, 86, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "ArrowCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl) __PYX_ERR(3, 92, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_ArrowCoderImpl)) __PYX_ERR(3, 92, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "OverWindowArrowCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl) __PYX_ERR(3, 101, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_OverWindowArrowCoderImpl)) __PYX_ERR(3, 101, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "TinyIntCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl) __PYX_ERR(3, 105, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TinyIntCoderImpl)) __PYX_ERR(3, 105, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "SmallIntCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl) __PYX_ERR(3, 108, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_SmallIntCoderImpl)) __PYX_ERR(3, 108, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "IntCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl) __PYX_ERR(3, 111, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_IntCoderImpl)) __PYX_ERR(3, 111, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "BigIntCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl) __PYX_ERR(3, 114, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BigIntCoderImpl)) __PYX_ERR(3, 114, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "BooleanCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl) __PYX_ERR(3, 117, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BooleanCoderImpl)) __PYX_ERR(3, 117, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "FloatCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl) __PYX_ERR(3, 120, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_FloatCoderImpl)) __PYX_ERR(3, 120, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "DoubleCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl) __PYX_ERR(3, 123, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DoubleCoderImpl)) __PYX_ERR(3, 123, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "BinaryCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl) __PYX_ERR(3, 126, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BinaryCoderImpl)) __PYX_ERR(3, 126, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "CharCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl) __PYX_ERR(3, 129, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CharCoderImpl)) __PYX_ERR(3, 129, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "BigDecimalCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl) __PYX_ERR(3, 132, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_BigDecimalCoderImpl)) __PYX_ERR(3, 132, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "DecimalCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl) __PYX_ERR(3, 135, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DecimalCoderImpl)) __PYX_ERR(3, 135, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "DateCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl) __PYX_ERR(3, 140, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DateCoderImpl)) __PYX_ERR(3, 140, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "TimeCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl) __PYX_ERR(3, 143, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimeCoderImpl)) __PYX_ERR(3, 143, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "TimestampCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl) __PYX_ERR(3, 146, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimestampCoderImpl)) __PYX_ERR(3, 146, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "LocalZonedTimestampCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl) __PYX_ERR(3, 151, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalZonedTimestampCoderImpl)) __PYX_ERR(3, 151, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "InstantCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl) __PYX_ERR(3, 154, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_InstantCoderImpl)) __PYX_ERR(3, 154, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "CloudPickleCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl) __PYX_ERR(3, 158, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CloudPickleCoderImpl)) __PYX_ERR(3, 158, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "PickleCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl) __PYX_ERR(3, 161, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_PickleCoderImpl)) __PYX_ERR(3, 161, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "GenericArrayCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl) __PYX_ERR(3, 164, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_GenericArrayCoderImpl)) __PYX_ERR(3, 164, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "PrimitiveArrayCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl) __PYX_ERR(3, 167, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_PrimitiveArrayCoderImpl)) __PYX_ERR(3, 167, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "MapCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl) __PYX_ERR(3, 170, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_MapCoderImpl)) __PYX_ERR(3, 170, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "TupleCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl) __PYX_ERR(3, 174, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TupleCoderImpl)) __PYX_ERR(3, 174, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "TimeWindowCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl) __PYX_ERR(3, 178, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_TimeWindowCoderImpl)) __PYX_ERR(3, 178, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "CountWindowCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl) __PYX_ERR(3, 181, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_CountWindowCoderImpl)) __PYX_ERR(3, 181, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "DataViewFilterCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl) __PYX_ERR(3, 184, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_DataViewFilterCoderImpl)) __PYX_ERR(3, 184, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "AvroCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl) __PYX_ERR(3, 188, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_AvroCoderImpl)) __PYX_ERR(3, 188, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "LocalDateCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl) __PYX_ERR(3, 196, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalDateCoderImpl)) __PYX_ERR(3, 196, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "LocalTimeCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl) __PYX_ERR(3, 199, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalTimeCoderImpl)) __PYX_ERR(3, 199, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.coder_impl_fast", "LocalDateTimeCoderImpl",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl) __PYX_ERR(3, 202, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl = (struct __pyx_vtabstruct_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_15coder_impl_fast_LocalDateTimeCoderImpl)) __PYX_ERR(3, 202, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("pyflink.fn_execution.table.aggregate_fast"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "DistinctViewDescriptor",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_DistinctViewDescriptor) __PYX_ERR(4, 21, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "RowKeySelector",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector) __PYX_ERR(4, 26, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RowKeySelector)) __PYX_ERR(4, 26, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "AggsHandleFunctionBase",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase) __PYX_ERR(4, 32, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_AggsHandleFunctionBase)) __PYX_ERR(4, 32, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "SimpleAggsHandleFunctionBase",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase) __PYX_ERR(4, 45, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunctionBase)) __PYX_ERR(4, 45, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "SimpleAggsHandleFunction",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction) __PYX_ERR(4, 57, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleAggsHandleFunction)) __PYX_ERR(4, 57, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "SimpleTableAggsHandleFunction",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction) __PYX_ERR(4, 61, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_SimpleTableAggsHandleFunction)) __PYX_ERR(4, 61, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "RecordCounter",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter) __PYX_ERR(4, 64, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RecordCounter)) __PYX_ERR(4, 64, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "AccumulationRecordCounter",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter) __PYX_ERR(4, 70, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_AccumulationRecordCounter)) __PYX_ERR(4, 70, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "RetractionRecordCounter",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter) __PYX_ERR(4, 73, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_RetractionRecordCounter)) __PYX_ERR(4, 73, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "GroupAggFunctionBase",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase) __PYX_ERR(4, 76, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunctionBase)) __PYX_ERR(4, 76, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "GroupAggFunction",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction) __PYX_ERR(4, 92, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupAggFunction)) __PYX_ERR(4, 92, __pyx_L1_error)
  __pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction = __Pyx_ImportType_3_1_3(__pyx_t_1, "pyflink.fn_execution.table.aggregate_fast", "GroupTableAggFunction",
  #if defined(PYPY_VERSION_NUM) && PYPY_VERSION_NUM < 0x050B0000
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction),
  #elif CYTHON_COMPILING_IN_LIMITED_API
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction),
  #else
  sizeof(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction), __PYX_GET_STRUCT_ALIGNMENT_3_1_3(struct __pyx_obj_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction),
  #endif
  __Pyx_ImportType_CheckSize_Warn_3_1_3); if (!__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction) __PYX_ERR(4, 95, __pyx_L1_error)
  __pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction = (struct __pyx_vtabstruct_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction*)__Pyx_GetVtable(__pyx_mstate->__pyx_ptype_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction); if (unlikely(!__pyx_vtabptr_7pyflink_12fn_execution_5table_14aggregate_fast_GroupTableAggFunction)) __PYX_ERR(4, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __pyx_t_1 = PyImport_ImportModule("pyflink.fn_execution.coder_impl_fast"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (__Pyx_ImportVoidPtr_3_1_3(__pyx_t_1, "ROW_KIND_BIT_SIZE", (void **)&__pyx_vp_7pyflink_12fn_execution_15coder_impl_fast_ROW_KIND_BIT_SIZE, "unsigned char") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_function_import_code(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_window_aggregate_fast(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_window_aggregate_fast},
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  {Py_mod_gil, Py_MOD_GIL_USED},
  #endif
  #if PY_VERSION_HEX >= 0x030C0000 && CYTHON_USE_MODULE_STATE
  {Py_mod_multiple_interpreters, Py_MOD_MULTIPLE_INTERPRETERS_NOT_SUPPORTED},
  #endif
  {0, NULL}
};
#endif

#ifdef __cplusplus
namespace {
  struct PyModuleDef __pyx_moduledef =
  #else
  static struct PyModuleDef __pyx_moduledef =
  #endif
  {
      PyModuleDef_HEAD_INIT,
      "window_aggregate_fast",
      0, /* m_doc */
    #if CYTHON_USE_MODULE_STATE
      sizeof(__pyx_mstatetype), /* m_size */
    #else
      (CYTHON_PEP489_MULTI_PHASE_INIT) ? 0 : -1, /* m_size */
    #endif
      __pyx_methods /* m_methods */,
    #if CYTHON_PEP489_MULTI_PHASE_INIT
      __pyx_moduledef_slots, /* m_slots */
    #else
      NULL, /* m_reload */
    #endif
    #if CYTHON_USE_MODULE_STATE
      __pyx_m_traverse, /* m_traverse */
      __pyx_m_clear, /* m_clear */
      NULL /* m_free */
    #else
      NULL, /* m_traverse */
      NULL, /* m_clear */
      NULL /* m_free */
    #endif
  };
  #ifdef __cplusplus
} /* anonymous namespace */
#endif

/* PyModInitFuncType */
#ifndef CYTHON_NO_PYINIT_EXPORT
  #define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#else
  #ifdef __cplusplus
  #define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
  #else
  #define __Pyx_PyMODINIT_FUNC PyObject *
  #endif
#endif

__Pyx_PyMODINIT_FUNC PyInit_window_aggregate_fast(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_window_aggregate_fast(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
/* ModuleCreationPEP489 */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
static PY_INT64_T __Pyx_GetCurrentInterpreterId(void) {
    {
        PyObject *module = PyImport_ImportModule("_interpreters"); // 3.13+ I think
        if (!module) {
            PyErr_Clear(); // just try the 3.8-3.12 version
            module = PyImport_ImportModule("_xxsubinterpreters");
            if (!module) goto bad;
        }
        PyObject *current = PyObject_CallMethod(module, "get_current", NULL);
        Py_DECREF(module);
        if (!current) goto bad;
        if (PyTuple_Check(current)) {
            PyObject *new_current = PySequence_GetItem(current, 0);
            Py_DECREF(current);
            current = new_current;
            if (!new_current) goto bad;
        }
        long long as_c_int = PyLong_AsLongLong(current);
        Py_DECREF(current);
        return as_c_int;
    }
  bad:
    PySys_WriteStderr("__Pyx_GetCurrentInterpreterId failed. Try setting the C define CYTHON_PEP489_MULTI_PHASE_INIT=0\n");
    return -1;
}
#endif
#if !CYTHON_USE_MODULE_STATE
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    static PY_INT64_T main_interpreter_id = -1;
#if CYTHON_COMPILING_IN_GRAAL
    PY_INT64_T current_id = PyInterpreterState_GetIDFromThreadState(PyThreadState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX >= 0x03090000
    PY_INT64_T current_id = PyInterpreterState_GetID(PyInterpreterState_Get());
#elif CYTHON_COMPILING_IN_LIMITED_API
    PY_INT64_T current_id = __Pyx_GetCurrentInterpreterId();
#else
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
#endif
    if (unlikely(current_id == -1)) {
        return -1;
    }
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return 0;
    } else if (unlikely(main_interpreter_id != current_id)) {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
#endif
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none)
{
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    CYTHON_UNUSED_VAR(def);
    #if !CYTHON_USE_MODULE_STATE
    if (__Pyx_check_single_interpreter())
        return NULL;
    #endif
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_window_aggregate_fast(PyObject *__pyx_pyinit_module)
#endif
{
  int stringtab_initialized = 0;
  #if CYTHON_USE_MODULE_STATE
  int pystate_addmodule_run = 0;
  #endif
  __pyx_mstatetype *__pyx_mstate = NULL;
  __Pyx_TraceDeclarationsFunc
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'window_aggregate_fast' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #else
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_t_1 = __pyx_pyinit_module;
  Py_INCREF(__pyx_t_1);
  #else
  __pyx_t_1 = PyModule_Create(&__pyx_moduledef); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #if CYTHON_USE_MODULE_STATE
  {
    int add_module_result = __Pyx_State_AddModule(__pyx_t_1, &__pyx_moduledef);
    __pyx_t_1 = 0; /* transfer ownership from __pyx_t_1 to "window_aggregate_fast" pseudovariable */
    if (unlikely((add_module_result < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    pystate_addmodule_run = 1;
  }
  #else
  __pyx_m = __pyx_t_1;
  #endif
  #if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
  PyUnstable_Module_SetGIL(__pyx_m, Py_MOD_GIL_USED);
  #endif
  __pyx_mstate = __pyx_mstate_global;
  CYTHON_UNUSED_VAR(__pyx_t_1);
  __pyx_mstate->__pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_mstate->__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_mstate->__pyx_d);
  __pyx_mstate->__pyx_b = __Pyx_PyImport_AddModuleRef(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_mstate->__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_cython_runtime = __Pyx_PyImport_AddModuleRef("cython_runtime"); if (unlikely(!__pyx_mstate->__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_mstate->__pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /* ImportRefnannyAPI */
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif

__Pyx_RefNannySetupContext("PyInit_window_aggregate_fast", 0);
  if (__Pyx_check_binary_version(__PYX_LIMITED_VERSION_HEX, __Pyx_get_runtime_version(), CYTHON_COMPILING_IN_LIMITED_API) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_mstate->__pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_mstate->__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_mstate->__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitConstants(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  stringtab_initialized = 1;
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if 0 || defined(__Pyx_CyFunction_USED) || defined(__Pyx_FusedFunction_USED) || defined(__Pyx_Coroutine_USED) || defined(__Pyx_Generator_USED) || defined(__Pyx_AsyncGen_USED)
  if (__pyx_CommonTypesMetaclass_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init(__pyx_m) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  if (__pyx_module_is_main_pyflink__fn_execution__table__window_aggregate_fast) {
    if (PyObject_SetAttr(__pyx_m, __pyx_mstate_global->__pyx_n_u_name, __pyx_mstate_global->__pyx_n_u_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pyflink.fn_execution.table.window_aggregate_fast")) {
      if (unlikely((PyDict_SetItemString(modules, "pyflink.fn_execution.table.window_aggregate_fast", __pyx_m) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_CreateCodeObjects(__pyx_mstate) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code(__pyx_mstate);
  (void)__Pyx_modinit_variable_export_code(__pyx_mstate);
  (void)__Pyx_modinit_function_export_code(__pyx_mstate);
  if (unlikely((__Pyx_modinit_type_init_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_type_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely((__Pyx_modinit_variable_import_code(__pyx_mstate) < 0))) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_function_import_code(__pyx_mstate);
  /*--- Execution code ---*/
  __Pyx_TraceStartFunc("PyInit_window_aggregate_fast", __pyx_f[0], 1, 0, 0, 0, __PYX_ERR(0, 1, __pyx_L1_error));

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":27
 * from pyflink.fn_execution.coder_impl_fast cimport InternalRowKind
 * 
 * import datetime             # <<<<<<<<<<<<<<
 * from typing import List, Dict
 * 
*/
  __pyx_t_2 = __Pyx_ImportDottedModule(__pyx_mstate_global->__pyx_n_u_datetime, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_datetime, __pyx_t_2) < 0) __PYX_ERR(0, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":28
 * 
 * import datetime
 * from typing import List, Dict             # <<<<<<<<<<<<<<
 * 
 * import pytz
*/
  __pyx_t_2 = __Pyx_PyList_Pack(2, __pyx_mstate_global->__pyx_n_u_List, __pyx_mstate_global->__pyx_n_u_Dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_typing, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_List); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_List, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Dict, __pyx_t_2) < 0) __PYX_ERR(0, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":30
 * from typing import List, Dict
 * 
 * import pytz             # <<<<<<<<<<<<<<
 * 
 * from pyflink.common.constants import MAX_LONG_VALUE
*/
  __pyx_t_3 = __Pyx_ImportDottedModule(__pyx_mstate_global->__pyx_n_u_pytz, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pytz, __pyx_t_3) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":32
 * import pytz
 * 
 * from pyflink.common.constants import MAX_LONG_VALUE             # <<<<<<<<<<<<<<
 * from pyflink.fn_execution.datastream.process.timerservice_impl import LegacyInternalTimerServiceImpl
 * from pyflink.fn_execution.coders import PickleCoder
*/
  __pyx_t_3 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_MAX_LONG_VALUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_common_constants, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_MAX_LONG_VALUE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MAX_LONG_VALUE, __pyx_t_3) < 0) __PYX_ERR(0, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":33
 * 
 * from pyflink.common.constants import MAX_LONG_VALUE
 * from pyflink.fn_execution.datastream.process.timerservice_impl import LegacyInternalTimerServiceImpl             # <<<<<<<<<<<<<<
 * from pyflink.fn_execution.coders import PickleCoder
 * from pyflink.fn_execution.table.state_data_view import DataViewSpec, ListViewSpec, MapViewSpec, \
*/
  __pyx_t_2 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_LegacyInternalTimerServiceImpl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_datastream, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_LegacyInternalTimerServiceImpl); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_LegacyInternalTimerServiceImpl, __pyx_t_2) < 0) __PYX_ERR(0, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":34
 * from pyflink.common.constants import MAX_LONG_VALUE
 * from pyflink.fn_execution.datastream.process.timerservice_impl import LegacyInternalTimerServiceImpl
 * from pyflink.fn_execution.coders import PickleCoder             # <<<<<<<<<<<<<<
 * from pyflink.fn_execution.table.state_data_view import DataViewSpec, ListViewSpec, MapViewSpec, \
 *     PerWindowStateDataViewStore
*/
  __pyx_t_3 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_PickleCoder); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_coders, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PickleCoder); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PickleCoder, __pyx_t_3) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":35
 * from pyflink.fn_execution.datastream.process.timerservice_impl import LegacyInternalTimerServiceImpl
 * from pyflink.fn_execution.coders import PickleCoder
 * from pyflink.fn_execution.table.state_data_view import DataViewSpec, ListViewSpec, MapViewSpec, \             # <<<<<<<<<<<<<<
 *     PerWindowStateDataViewStore
 * from pyflink.fn_execution.table.window_assigner import WindowAssigner, PanedWindowAssigner, \
*/
  __pyx_t_2 = __Pyx_PyList_Pack(4, __pyx_mstate_global->__pyx_n_u_DataViewSpec, __pyx_mstate_global->__pyx_n_u_ListViewSpec, __pyx_mstate_global->__pyx_n_u_MapViewSpec, __pyx_mstate_global->__pyx_n_u_PerWindowStateDataViewStore); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_state, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_DataViewSpec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_DataViewSpec, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_ListViewSpec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ListViewSpec, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_MapViewSpec); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MapViewSpec, __pyx_t_2) < 0) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_PerWindowStateDataViewStore); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PerWindowStateDataViewStore, __pyx_t_2) < 0) __PYX_ERR(0, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":37
 * from pyflink.fn_execution.table.state_data_view import DataViewSpec, ListViewSpec, MapViewSpec, \
 *     PerWindowStateDataViewStore
 * from pyflink.fn_execution.table.window_assigner import WindowAssigner, PanedWindowAssigner, \             # <<<<<<<<<<<<<<
 *     MergingWindowAssigner
 * from pyflink.fn_execution.table.window_context import WindowContext, TriggerContext, K, W
*/
  __pyx_t_3 = __Pyx_PyList_Pack(3, __pyx_mstate_global->__pyx_n_u_WindowAssigner, __pyx_mstate_global->__pyx_n_u_PanedWindowAssigner, __pyx_mstate_global->__pyx_n_u_MergingWindowAssigner); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_WindowAssigner); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_WindowAssigner, __pyx_t_3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PanedWindowAssigner); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PanedWindowAssigner, __pyx_t_3) < 0) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_MergingWindowAssigner); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MergingWindowAssigner, __pyx_t_3) < 0) __PYX_ERR(0, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":39
 * from pyflink.fn_execution.table.window_assigner import WindowAssigner, PanedWindowAssigner, \
 *     MergingWindowAssigner
 * from pyflink.fn_execution.table.window_context import WindowContext, TriggerContext, K, W             # <<<<<<<<<<<<<<
 * from pyflink.fn_execution.table.window_process_function import GeneralWindowProcessFunction, \
 *     InternalWindowProcessFunction, PanedWindowProcessFunction, MergingWindowProcessFunction
*/
  __pyx_t_2 = __Pyx_PyList_Pack(4, __pyx_mstate_global->__pyx_n_u_WindowContext, __pyx_mstate_global->__pyx_n_u_TriggerContext, __pyx_mstate_global->__pyx_n_u_K, __pyx_mstate_global->__pyx_n_u_W); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_2, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_WindowContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_WindowContext, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_TriggerContext); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_TriggerContext, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_K); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_K, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_W); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_W, __pyx_t_2) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":40
 *     MergingWindowAssigner
 * from pyflink.fn_execution.table.window_context import WindowContext, TriggerContext, K, W
 * from pyflink.fn_execution.table.window_process_function import GeneralWindowProcessFunction, \             # <<<<<<<<<<<<<<
 *     InternalWindowProcessFunction, PanedWindowProcessFunction, MergingWindowProcessFunction
 * from pyflink.fn_execution.table.window_trigger import Trigger
*/
  __pyx_t_3 = __Pyx_PyList_Pack(4, __pyx_mstate_global->__pyx_n_u_GeneralWindowProcessFunction, __pyx_mstate_global->__pyx_n_u_InternalWindowProcessFunction, __pyx_mstate_global->__pyx_n_u_PanedWindowProcessFunction, __pyx_mstate_global->__pyx_n_u_MergingWindowProcessFunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_3, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_GeneralWindowProcessFunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_GeneralWindowProcessFunction, __pyx_t_3) < 0) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_InternalWindowProcessFunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_InternalWindowProcessFunction, __pyx_t_3) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_PanedWindowProcessFunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_PanedWindowProcessFunction, __pyx_t_3) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_MergingWindowProcessFunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_MergingWindowProcessFunction, __pyx_t_3) < 0) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":42
 * from pyflink.fn_execution.table.window_process_function import GeneralWindowProcessFunction, \
 *     InternalWindowProcessFunction, PanedWindowProcessFunction, MergingWindowProcessFunction
 * from pyflink.fn_execution.table.window_trigger import Trigger             # <<<<<<<<<<<<<<
 * from pyflink.table.udf import ImperativeAggregateFunction
 * 
*/
  __pyx_t_2 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_Trigger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_4, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_mstate_global->__pyx_n_u_Trigger); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_Trigger, __pyx_t_2) < 0) __PYX_ERR(0, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":43
 *     InternalWindowProcessFunction, PanedWindowProcessFunction, MergingWindowProcessFunction
 * from pyflink.fn_execution.table.window_trigger import Trigger
 * from pyflink.table.udf import ImperativeAggregateFunction             # <<<<<<<<<<<<<<
 * 
 * cdef InternalRow join_row(list left, list right, InternalRowKind row_kind):
*/
  __pyx_t_3 = __Pyx_PyList_Pack(1, __pyx_mstate_global->__pyx_n_u_ImperativeAggregateFunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_2 = __Pyx_Import(__pyx_mstate_global->__pyx_n_u_pyflink_table_udf, __pyx_t_3, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_2, __pyx_mstate_global->__pyx_n_u_ImperativeAggregateFunction); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_ImperativeAggregateFunction, __pyx_t_3) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":72
 *         """
 * 
 *     cpdef void merge(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         """
 *         Merges the other accumulators into current accumulators.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_1merge, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunctionBase_2, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[4])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, __pyx_mstate_global->__pyx_n_u_merge, __pyx_t_2) < 0) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":78
 *         pass
 * 
 *     cpdef void set_accumulators(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         """
 *         Set the current accumulators (saved in a row) which contains the current aggregated results.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_3set_accumulators, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunctionBase_3, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[5])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, __pyx_mstate_global->__pyx_n_u_set_accumulators, __pyx_t_2) < 0) __PYX_ERR(0, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":93
 *         pass
 * 
 *     cpdef list create_accumulators(self):             # <<<<<<<<<<<<<<
 *         """
 *         Initializes the accumulators and save them to an accumulators List.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_5create_accumulators, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunctionBase_4, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[7])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, __pyx_mstate_global->__pyx_n_u_create_accumulators, __pyx_t_2) < 0) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":101
 *         pass
 * 
 *     cpdef void cleanup(self, object namespace):             # <<<<<<<<<<<<<<
 *         """
 *         Cleanup for the retired accumulators state.
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_7cleanup, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunctionBase_5, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[8])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, __pyx_mstate_global->__pyx_n_u_cleanup, __pyx_t_2) < 0) __PYX_ERR(0, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_9__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunctionBase_6, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[10])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunctionBase, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_31NamespaceAggsHandleFunctionBase_11__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunctionBase_7, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[11])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunctionBase, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":115
 * 
 * cdef class NamespaceAggsHandleFunction(NamespaceAggsHandleFunctionBase):
 *     cpdef list get_value(self, object namespace):             # <<<<<<<<<<<<<<
 *         """
 *         Gets the result of the aggregation from the current accumulators and namespace properties
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_1get_value, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunction_get, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[12])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction, __pyx_mstate_global->__pyx_n_u_get_value, __pyx_t_2) < 0) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunction___re, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[13])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_NamespaceAggsHandleFunction, (type(self), 0xe3b0c44, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_NamespaceAggsHandleFunction__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_27NamespaceAggsHandleFunction_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_NamespaceAggsHandleFunction___se, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[14])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_NamespaceAggsHandleFunction, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":271
 *             self._udfs[i].retract(self._accumulators[i], *args)
 * 
 *     cpdef void merge(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         if self._udf_data_views:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_3merge, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SimpleNamespaceAggsHandleFunctio_2, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[19])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, __pyx_mstate_global->__pyx_n_u_merge, __pyx_t_2) < 0) __PYX_ERR(0, 271, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":281
 *             self._udfs[i].merge(self._accumulators[i], [accumulators[i]])
 * 
 *     cpdef void set_accumulators(self, object namespace, list accumulators):             # <<<<<<<<<<<<<<
 *         cdef size_t i, index
 *         if self._udf_data_views and namespace is not None:
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_5set_accumulators, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SimpleNamespaceAggsHandleFunctio_3, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[20])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, __pyx_mstate_global->__pyx_n_u_set_accumulators, __pyx_t_2) < 0) __PYX_ERR(0, 281, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":293
 *         return self._accumulators
 * 
 *     cpdef list create_accumulators(self):             # <<<<<<<<<<<<<<
 *         return [udf.create_accumulator() for udf in self._udfs]
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_7create_accumulators, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SimpleNamespaceAggsHandleFunctio_4, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[22])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, __pyx_mstate_global->__pyx_n_u_create_accumulators, __pyx_t_2) < 0) __PYX_ERR(0, 293, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":296
 *         return [udf.create_accumulator() for udf in self._udfs]
 * 
 *     cpdef void cleanup(self, object namespace):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         for i in range(len(self._udf_data_views)):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_9cleanup, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SimpleNamespaceAggsHandleFunctio_5, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[23])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, __pyx_mstate_global->__pyx_n_u_cleanup, __pyx_t_2) < 0) __PYX_ERR(0, 296, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":307
 *             udf.close()
 * 
 *     cpdef list get_value(self, object namespace):             # <<<<<<<<<<<<<<
 *         cdef size_t i
 *         cdef size_t*get_value_indexes
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_11get_value, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SimpleNamespaceAggsHandleFunctio_6, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[25])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_SimpleNamespaceAggsHandleFunction, __pyx_mstate_global->__pyx_n_u_get_value, __pyx_t_2) < 0) __PYX_ERR(0, 307, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_15__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SimpleNamespaceAggsHandleFunctio_7, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[27])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError, "self._distinct_indexes,self._filter_args,self._get_value_indexes cannot be converted to a Python object for pickling"
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_33SimpleNamespaceAggsHandleFunction_17__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_SimpleNamespaceAggsHandleFunctio_8, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[28])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":351
 *         self._window_state = self._state_backend.get_value_state("window_state", state_value_coder)
 * 
 *     cpdef void open(self, object function_context):             # <<<<<<<<<<<<<<
 *         self._internal_timer_service = LegacyInternalTimerServiceImpl(self._state_backend)
 *         self._window_aggregator.open(
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_3open, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_open, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[30])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_open, __pyx_t_2) < 0) __PYX_ERR(0, 351, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":374
 *         self._window_function.open(self._window_context)
 * 
 *     cpdef list process_element(self, InternalRow input_row):             # <<<<<<<<<<<<<<
 *         cdef list input_value, current_key, affected_windows, acc, actual_windows, result
 *         cdef int64_t timestamp, seconds, microseconds_of_second, milliseconds
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_5process_element, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_proce, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[31])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_process_element, __pyx_t_2) < 0) __PYX_ERR(0, 374, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":420
 *         return result
 * 
 *     cpdef void process_watermark(self, int64_t watermark):             # <<<<<<<<<<<<<<
 *         self._internal_timer_service.advance_watermark(watermark)
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_7process_watermark, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_proce_2, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[32])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_process_watermark, __pyx_t_2) < 0) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":423
 *         self._internal_timer_service.advance_watermark(watermark)
 * 
 *     cpdef list on_event_time(self, object timer):             # <<<<<<<<<<<<<<
 *         cdef list result, key
 *         cdef int64_t timestamp
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_9on_event_time, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_on_ev, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[33])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_on_event_time, __pyx_t_2) < 0) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":441
 *         return result
 * 
 *     cpdef list on_processing_time(self, object timer):             # <<<<<<<<<<<<<<
 *         cdef list result, key
 *         cdef int64_t timestamp
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_11on_processing_time, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_on_pr, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[34])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_on_processing_time, __pyx_t_2) < 0) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":459
 *         return result
 * 
 *     cpdef int64_t to_utc_timestamp_mills(self, int64_t epoch_mills):             # <<<<<<<<<<<<<<
 *         if self._shift_timezone == "UTC":
 *             return epoch_mills
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_13to_utc_timestamp_mills, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_to_ut, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[35])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_to_utc_timestamp_mills, __pyx_t_2) < 0) __PYX_ERR(0, 459, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":469
 *             return int((local_date_time - epoch).total_seconds() * 1000.0)
 * 
 *     cpdef list get_timers(self):             # <<<<<<<<<<<<<<
 *         cdef list timers
 *         cdef object timer
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_15get_timers, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_get_t, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[36])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_get_timers, __pyx_t_2) < 0) __PYX_ERR(0, 469, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":478
 *         return timers
 * 
 *     cpdef void close(self):             # <<<<<<<<<<<<<<
 *         self._window_aggregator.close()
 * 
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_17close, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_close, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[37])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_close, __pyx_t_2) < 0) __PYX_ERR(0, 478, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":492
 *             self._trigger_context.register_processing_time_timer(cleanup_time)
 * 
 *     cpdef int64_t cleanup_time(self, object window):             # <<<<<<<<<<<<<<
 *         cdef int64_t cleanup_time
 *         if self._window_assigner.is_event_time():
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_19cleanup_time, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase_clean, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[39])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_cleanup_time, __pyx_t_2) < 0) __PYX_ERR(0, 492, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_21__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase___red, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[41])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunctionBase, (type(self), 0xc758b60, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GroupWindowAggFunctionBase__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_26GroupWindowAggFunctionBase_23__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunctionBase___set, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[42])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunctionBase, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_3__reduce_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunction___reduce, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[45])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction, __pyx_mstate_global->__pyx_n_u_reduce_cython, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_GroupWindowAggFunction, (type(self), 0xc758b60, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_GroupWindowAggFunction__set_state(self, __pyx_state)
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_22GroupWindowAggFunction_5__setstate_cython__, __Pyx_CYFUNCTION_CCLASS, __pyx_mstate_global->__pyx_n_u_GroupWindowAggFunction___setstat, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[46])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (__Pyx_SetItemOnTypeDict(__pyx_mstate_global->__pyx_ptype_7pyflink_12fn_execution_5table_21window_aggregate_fast_GroupWindowAggFunction, __pyx_mstate_global->__pyx_n_u_setstate_cython, __pyx_t_2) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_NamespaceAggsHandleFunctionBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_1__pyx_unpickle_NamespaceAggsHandleFunctionBase, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NamespaceAggsHand, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[47])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NamespaceAggsHand, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(<NamespaceAggsHandleFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_NamespaceAggsHandleFunctionBase__set_state(NamespaceAggsHandleFunctionBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 0 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[0])
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_3__pyx_unpickle_NamespaceAggsHandleFunction, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NamespaceAggsHand_2, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[49])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_NamespaceAggsHand_2, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_GroupWindowAggFunctionBase(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_5__pyx_unpickle_GroupWindowAggFunctionBase, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_GroupWindowAggFun, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[51])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_GroupWindowAggFun, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_GroupWindowAggFunctionBase__set_state(<GroupWindowAggFunctionBase> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_GroupWindowAggFunctionBase__set_state(GroupWindowAggFunctionBase __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result._allowed_lateness = __pyx_state[0]; __pyx_result._internal_timer_service = __pyx_state[1]; __pyx_result._key_selector = __pyx_state[2]; __pyx_result._rowtime_index = __pyx_state[3]; __pyx_result._shift_timezone = __pyx_state[4]; __pyx_result._state_backend = __pyx_state[5]; __pyx_result._state_value_coder = __pyx_state[6]; __pyx_result._trigger = __pyx_state[7]; __pyx_result._trigger_context = __pyx_state[8]; __pyx_result._window_aggregator = __pyx_state[9]; __pyx_result._window_assigner = __pyx_state[10]; __pyx_result._window_context = __pyx_state[11]; __pyx_result._window_function = __pyx_state[12]; __pyx_result._window_state = __pyx_state[13]
 *     if len(__pyx_state) > 14 and hasattr(__pyx_result, '__dict__'):
*/
  __pyx_t_2 = __Pyx_CyFunction_New(&__pyx_mdef_7pyflink_12fn_execution_5table_21window_aggregate_fast_7__pyx_unpickle_GroupWindowAggFunction, 0, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_GroupWindowAggFun_2, NULL, __pyx_mstate_global->__pyx_n_u_pyflink_fn_execution_table_windo_5, __pyx_mstate_global->__pyx_d, ((PyObject *)__pyx_mstate_global->__pyx_codeobj_tab[53])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_pyx_unpickle_GroupWindowAggFun_2, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pyflink/fn_execution/table/window_aggregate_fast.pyx":1
 * ################################################################################             # <<<<<<<<<<<<<<
 * #  Licensed to the Apache Software Foundation (ASF) under one
 * #  or more contributor license agreements.  See the NOTICE file
*/
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_mstate_global->__pyx_d, __pyx_mstate_global->__pyx_n_u_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_TraceReturnValue(Py_None, 0, 0, __PYX_ERR(0, 1, __pyx_L1_error));
  __Pyx_PyMonitoring_ExitScope(0);

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_TraceException(__pyx_lineno, 0, 0);
  __Pyx_TraceExceptionUnwind(0, 0);
  if (__pyx_m) {
    if (__pyx_mstate->__pyx_d && stringtab_initialized) {
      __Pyx_AddTraceback("init pyflink.fn_execution.table.window_aggregate_fast", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    #if !CYTHON_USE_MODULE_STATE
    Py_CLEAR(__pyx_m);
    #else
    Py_DECREF(__pyx_m);
    if (pystate_addmodule_run) {
      PyObject *tp, *value, *tb;
      PyErr_Fetch(&tp, &value, &tb);
      PyState_RemoveModule(&__pyx_moduledef);
      PyErr_Restore(tp, value, tb);
    }
    #endif
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pyflink.fn_execution.table.window_aggregate_fast");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #else
  return __pyx_m;
  #endif
}
/* #### Code section: pystring_table ### */

typedef struct {
    const char *s;
#if 313 <= 65535
    const unsigned short n;
#elif 313 / 2 < INT_MAX
    const unsigned int n;
#elif 313 / 2 < LONG_MAX
    const unsigned long n;
#else
    const Py_ssize_t n;
#endif
#if 1 <= 31
    const unsigned int encoding : 5;
#elif 1 <= 255
    const unsigned char encoding;
#elif 1 <= 65535
    const unsigned short encoding;
#else
    const Py_ssize_t encoding;
#endif
    const unsigned int is_unicode : 1;
    const unsigned int intern : 1;
} __Pyx_StringTabEntry;
static const char * const __pyx_string_tab_encodings[] = { 0 };
static const __Pyx_StringTabEntry __pyx_string_tab[] = {
  {__pyx_k_, sizeof(__pyx_k_), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_ */
  {__pyx_k_DataViewSpec, sizeof(__pyx_k_DataViewSpec), 0, 1, 1}, /* PyObject cname: __pyx_n_u_DataViewSpec */
  {__pyx_k_Dict, sizeof(__pyx_k_Dict), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Dict */
  {__pyx_k_GeneralWindowProcessFunction, sizeof(__pyx_k_GeneralWindowProcessFunction), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GeneralWindowProcessFunction */
  {__pyx_k_GroupWindowAggFunction, sizeof(__pyx_k_GroupWindowAggFunction), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunction */
  {__pyx_k_GroupWindowAggFunctionBase, sizeof(__pyx_k_GroupWindowAggFunctionBase), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase */
  {__pyx_k_GroupWindowAggFunctionBase___red, sizeof(__pyx_k_GroupWindowAggFunctionBase___red), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase___red */
  {__pyx_k_GroupWindowAggFunctionBase___set, sizeof(__pyx_k_GroupWindowAggFunctionBase___set), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase___set */
  {__pyx_k_GroupWindowAggFunctionBase_clean, sizeof(__pyx_k_GroupWindowAggFunctionBase_clean), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_clean */
  {__pyx_k_GroupWindowAggFunctionBase_close, sizeof(__pyx_k_GroupWindowAggFunctionBase_close), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_close */
  {__pyx_k_GroupWindowAggFunctionBase_get_t, sizeof(__pyx_k_GroupWindowAggFunctionBase_get_t), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_get_t */
  {__pyx_k_GroupWindowAggFunctionBase_on_ev, sizeof(__pyx_k_GroupWindowAggFunctionBase_on_ev), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_on_ev */
  {__pyx_k_GroupWindowAggFunctionBase_on_pr, sizeof(__pyx_k_GroupWindowAggFunctionBase_on_pr), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_on_pr */
  {__pyx_k_GroupWindowAggFunctionBase_open, sizeof(__pyx_k_GroupWindowAggFunctionBase_open), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_open */
  {__pyx_k_GroupWindowAggFunctionBase_proce, sizeof(__pyx_k_GroupWindowAggFunctionBase_proce), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_proce */
  {__pyx_k_GroupWindowAggFunctionBase_proce_2, sizeof(__pyx_k_GroupWindowAggFunctionBase_proce_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_proce_2 */
  {__pyx_k_GroupWindowAggFunctionBase_to_ut, sizeof(__pyx_k_GroupWindowAggFunctionBase_to_ut), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunctionBase_to_ut */
  {__pyx_k_GroupWindowAggFunction___reduce, sizeof(__pyx_k_GroupWindowAggFunction___reduce), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunction___reduce */
  {__pyx_k_GroupWindowAggFunction___setstat, sizeof(__pyx_k_GroupWindowAggFunction___setstat), 0, 1, 1}, /* PyObject cname: __pyx_n_u_GroupWindowAggFunction___setstat */
  {__pyx_k_ImperativeAggregateFunction, sizeof(__pyx_k_ImperativeAggregateFunction), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ImperativeAggregateFunction */
  {__pyx_k_Incompatible_checksums_0x_x_vs_0, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Incompatible_checksums_0x_x_vs_0 */
  {__pyx_k_Incompatible_checksums_0x_x_vs_0_2, sizeof(__pyx_k_Incompatible_checksums_0x_x_vs_0_2), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Incompatible_checksums_0x_x_vs_0_2 */
  {__pyx_k_InternalWindowProcessFunction, sizeof(__pyx_k_InternalWindowProcessFunction), 0, 1, 1}, /* PyObject cname: __pyx_n_u_InternalWindowProcessFunction */
  {__pyx_k_K, sizeof(__pyx_k_K), 0, 1, 1}, /* PyObject cname: __pyx_n_u_K */
  {__pyx_k_LegacyInternalTimerServiceImpl, sizeof(__pyx_k_LegacyInternalTimerServiceImpl), 0, 1, 1}, /* PyObject cname: __pyx_n_u_LegacyInternalTimerServiceImpl */
  {__pyx_k_List, sizeof(__pyx_k_List), 0, 1, 1}, /* PyObject cname: __pyx_n_u_List */
  {__pyx_k_ListViewSpec, sizeof(__pyx_k_ListViewSpec), 0, 1, 1}, /* PyObject cname: __pyx_n_u_ListViewSpec */
  {__pyx_k_MAX_LONG_VALUE, sizeof(__pyx_k_MAX_LONG_VALUE), 0, 1, 1}, /* PyObject cname: __pyx_n_u_MAX_LONG_VALUE */
  {__pyx_k_MapViewSpec, sizeof(__pyx_k_MapViewSpec), 0, 1, 1}, /* PyObject cname: __pyx_n_u_MapViewSpec */
  {__pyx_k_MergingWindowAssigner, sizeof(__pyx_k_MergingWindowAssigner), 0, 1, 1}, /* PyObject cname: __pyx_n_u_MergingWindowAssigner */
  {__pyx_k_MergingWindowProcessFunction, sizeof(__pyx_k_MergingWindowProcessFunction), 0, 1, 1}, /* PyObject cname: __pyx_n_u_MergingWindowProcessFunction */
  {__pyx_k_NamespaceAggsHandleFunction, sizeof(__pyx_k_NamespaceAggsHandleFunction), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunction */
  {__pyx_k_NamespaceAggsHandleFunctionBase, sizeof(__pyx_k_NamespaceAggsHandleFunctionBase), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunctionBase */
  {__pyx_k_NamespaceAggsHandleFunctionBase_2, sizeof(__pyx_k_NamespaceAggsHandleFunctionBase_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunctionBase_2 */
  {__pyx_k_NamespaceAggsHandleFunctionBase_3, sizeof(__pyx_k_NamespaceAggsHandleFunctionBase_3), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunctionBase_3 */
  {__pyx_k_NamespaceAggsHandleFunctionBase_4, sizeof(__pyx_k_NamespaceAggsHandleFunctionBase_4), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunctionBase_4 */
  {__pyx_k_NamespaceAggsHandleFunctionBase_5, sizeof(__pyx_k_NamespaceAggsHandleFunctionBase_5), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunctionBase_5 */
  {__pyx_k_NamespaceAggsHandleFunctionBase_6, sizeof(__pyx_k_NamespaceAggsHandleFunctionBase_6), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunctionBase_6 */
  {__pyx_k_NamespaceAggsHandleFunctionBase_7, sizeof(__pyx_k_NamespaceAggsHandleFunctionBase_7), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunctionBase_7 */
  {__pyx_k_NamespaceAggsHandleFunction___re, sizeof(__pyx_k_NamespaceAggsHandleFunction___re), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunction___re */
  {__pyx_k_NamespaceAggsHandleFunction___se, sizeof(__pyx_k_NamespaceAggsHandleFunction___se), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunction___se */
  {__pyx_k_NamespaceAggsHandleFunction_get, sizeof(__pyx_k_NamespaceAggsHandleFunction_get), 0, 1, 1}, /* PyObject cname: __pyx_n_u_NamespaceAggsHandleFunction_get */
  {__pyx_k_Note_that_Cython_is_deliberately, sizeof(__pyx_k_Note_that_Cython_is_deliberately), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_Note_that_Cython_is_deliberately */
  {__pyx_k_PanedWindowAssigner, sizeof(__pyx_k_PanedWindowAssigner), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PanedWindowAssigner */
  {__pyx_k_PanedWindowProcessFunction, sizeof(__pyx_k_PanedWindowProcessFunction), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PanedWindowProcessFunction */
  {__pyx_k_PerWindowStateDataViewStore, sizeof(__pyx_k_PerWindowStateDataViewStore), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PerWindowStateDataViewStore */
  {__pyx_k_PickleCoder, sizeof(__pyx_k_PickleCoder), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PickleCoder */
  {__pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_PickleError */
  {__pyx_k_SimpleNamespaceAggsHandleFunctio, sizeof(__pyx_k_SimpleNamespaceAggsHandleFunctio), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SimpleNamespaceAggsHandleFunctio */
  {__pyx_k_SimpleNamespaceAggsHandleFunctio_2, sizeof(__pyx_k_SimpleNamespaceAggsHandleFunctio_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SimpleNamespaceAggsHandleFunctio_2 */
  {__pyx_k_SimpleNamespaceAggsHandleFunctio_3, sizeof(__pyx_k_SimpleNamespaceAggsHandleFunctio_3), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SimpleNamespaceAggsHandleFunctio_3 */
  {__pyx_k_SimpleNamespaceAggsHandleFunctio_4, sizeof(__pyx_k_SimpleNamespaceAggsHandleFunctio_4), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SimpleNamespaceAggsHandleFunctio_4 */
  {__pyx_k_SimpleNamespaceAggsHandleFunctio_5, sizeof(__pyx_k_SimpleNamespaceAggsHandleFunctio_5), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SimpleNamespaceAggsHandleFunctio_5 */
  {__pyx_k_SimpleNamespaceAggsHandleFunctio_6, sizeof(__pyx_k_SimpleNamespaceAggsHandleFunctio_6), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SimpleNamespaceAggsHandleFunctio_6 */
  {__pyx_k_SimpleNamespaceAggsHandleFunctio_7, sizeof(__pyx_k_SimpleNamespaceAggsHandleFunctio_7), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SimpleNamespaceAggsHandleFunctio_7 */
  {__pyx_k_SimpleNamespaceAggsHandleFunctio_8, sizeof(__pyx_k_SimpleNamespaceAggsHandleFunctio_8), 0, 1, 1}, /* PyObject cname: __pyx_n_u_SimpleNamespaceAggsHandleFunctio_8 */
  {__pyx_k_The_args_are_not_in_the_distinct, sizeof(__pyx_k_The_args_are_not_in_the_distinct), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_The_args_are_not_in_the_distinct */
  {__pyx_k_Trigger, sizeof(__pyx_k_Trigger), 0, 1, 1}, /* PyObject cname: __pyx_n_u_Trigger */
  {__pyx_k_TriggerContext, sizeof(__pyx_k_TriggerContext), 0, 1, 1}, /* PyObject cname: __pyx_n_u_TriggerContext */
  {__pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_TypeError */
  {__pyx_k_UTC, sizeof(__pyx_k_UTC), 0, 1, 1}, /* PyObject cname: __pyx_n_u_UTC */
  {__pyx_k_W, sizeof(__pyx_k_W), 0, 1, 1}, /* PyObject cname: __pyx_n_u_W */
  {__pyx_k_WindowAssigner, sizeof(__pyx_k_WindowAssigner), 0, 1, 1}, /* PyObject cname: __pyx_n_u_WindowAssigner */
  {__pyx_k_WindowContext, sizeof(__pyx_k_WindowContext), 0, 1, 1}, /* PyObject cname: __pyx_n_u_WindowContext */
  {__pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0}, /* PyObject cname: __pyx_kp_u__2 */
  {__pyx_k_accumulate, sizeof(__pyx_k_accumulate), 0, 1, 1}, /* PyObject cname: __pyx_n_u_accumulate */
  {__pyx_k_accumulators, sizeof(__pyx_k_accumulators), 0, 1, 1}, /* PyObject cname: __pyx_n_u_accumulators */
  {__pyx_k_accumulators_2, sizeof(__pyx_k_accumulators_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_accumulators_2 */
  {__pyx_k_add, sizeof(__pyx_k_add), 0, 1, 1}, /* PyObject cname: __pyx_n_u_add */
  {__pyx_k_add_note, sizeof(__pyx_k_add_note), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_add_note */
  {__pyx_k_advance_watermark, sizeof(__pyx_k_advance_watermark), 0, 1, 1}, /* PyObject cname: __pyx_n_u_advance_watermark */
  {__pyx_k_agg_ddistinct, sizeof(__pyx_k_agg_ddistinct), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_agg_ddistinct */
  {__pyx_k_allowed_lateness, sizeof(__pyx_k_allowed_lateness), 0, 1, 1}, /* PyObject cname: __pyx_n_u_allowed_lateness */
  {__pyx_k_allowed_lateness_2, sizeof(__pyx_k_allowed_lateness_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_allowed_lateness_2 */
  {__pyx_k_assign_actual_windows, sizeof(__pyx_k_assign_actual_windows), 0, 1, 1}, /* PyObject cname: __pyx_n_u_assign_actual_windows */
  {__pyx_k_assign_state_namespace, sizeof(__pyx_k_assign_state_namespace), 0, 1, 1}, /* PyObject cname: __pyx_n_u_assign_state_namespace */
  {__pyx_k_asyncio_coroutines, sizeof(__pyx_k_asyncio_coroutines), 0, 1, 1}, /* PyObject cname: __pyx_n_u_asyncio_coroutines */
  {__pyx_k_class_getitem, sizeof(__pyx_k_class_getitem), 0, 1, 1}, /* PyObject cname: __pyx_n_u_class_getitem */
  {__pyx_k_clean_window_if_needed, sizeof(__pyx_k_clean_window_if_needed), 0, 1, 1}, /* PyObject cname: __pyx_n_u_clean_window_if_needed */
  {__pyx_k_cleanup, sizeof(__pyx_k_cleanup), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cleanup */
  {__pyx_k_cleanup_time, sizeof(__pyx_k_cleanup_time), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cleanup_time */
  {__pyx_k_clear, sizeof(__pyx_k_clear), 0, 1, 1}, /* PyObject cname: __pyx_n_u_clear */
  {__pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 1, 1}, /* PyObject cname: __pyx_n_u_cline_in_traceback */
  {__pyx_k_close, sizeof(__pyx_k_close), 0, 1, 1}, /* PyObject cname: __pyx_n_u_close */
  {__pyx_k_count_star_inserted, sizeof(__pyx_k_count_star_inserted), 0, 1, 1}, /* PyObject cname: __pyx_n_u_count_star_inserted */
  {__pyx_k_create_accumulator, sizeof(__pyx_k_create_accumulator), 0, 1, 1}, /* PyObject cname: __pyx_n_u_create_accumulator */
  {__pyx_k_create_accumulators, sizeof(__pyx_k_create_accumulators), 0, 1, 1}, /* PyObject cname: __pyx_n_u_create_accumulators */
  {__pyx_k_current_processing_time, sizeof(__pyx_k_current_processing_time), 0, 1, 1}, /* PyObject cname: __pyx_n_u_current_processing_time */
  {__pyx_k_datetime, sizeof(__pyx_k_datetime), 0, 1, 1}, /* PyObject cname: __pyx_n_u_datetime */
  {__pyx_k_dealloc, sizeof(__pyx_k_dealloc), 0, 1, 1}, /* PyObject cname: __pyx_n_u_dealloc */
  {__pyx_k_dict, sizeof(__pyx_k_dict), 0, 1, 1}, /* PyObject cname: __pyx_n_u_dict */
  {__pyx_k_dict_2, sizeof(__pyx_k_dict_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_dict_2 */
  {__pyx_k_disable, sizeof(__pyx_k_disable), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_disable */
  {__pyx_k_distinct_data_views, sizeof(__pyx_k_distinct_data_views), 0, 1, 1}, /* PyObject cname: __pyx_n_u_distinct_data_views */
  {__pyx_k_distinct_indexes, sizeof(__pyx_k_distinct_indexes), 0, 1, 1}, /* PyObject cname: __pyx_n_u_distinct_indexes */
  {__pyx_k_distinct_indexes_2, sizeof(__pyx_k_distinct_indexes_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_distinct_indexes_2 */
  {__pyx_k_distinct_view_descriptors, sizeof(__pyx_k_distinct_view_descriptors), 0, 1, 1}, /* PyObject cname: __pyx_n_u_distinct_view_descriptors */
  {__pyx_k_distinct_view_descriptors_2, sizeof(__pyx_k_distinct_view_descriptors_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_distinct_view_descriptors_2 */
  {__pyx_k_element_coder, sizeof(__pyx_k_element_coder), 0, 1, 1}, /* PyObject cname: __pyx_n_u_element_coder */
  {__pyx_k_emit_window_result, sizeof(__pyx_k_emit_window_result), 0, 1, 1}, /* PyObject cname: __pyx_n_u_emit_window_result */
  {__pyx_k_enable, sizeof(__pyx_k_enable), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_enable */
  {__pyx_k_epoch_mills, sizeof(__pyx_k_epoch_mills), 0, 1, 1}, /* PyObject cname: __pyx_n_u_epoch_mills */
  {__pyx_k_field_index, sizeof(__pyx_k_field_index), 0, 1, 1}, /* PyObject cname: __pyx_n_u_field_index */
  {__pyx_k_filter_args, sizeof(__pyx_k_filter_args), 0, 1, 1}, /* PyObject cname: __pyx_n_u_filter_args */
  {__pyx_k_filter_args_2, sizeof(__pyx_k_filter_args_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_filter_args_2 */
  {__pyx_k_fromtimestamp, sizeof(__pyx_k_fromtimestamp), 0, 1, 1}, /* PyObject cname: __pyx_n_u_fromtimestamp */
  {__pyx_k_func, sizeof(__pyx_k_func), 0, 1, 1}, /* PyObject cname: __pyx_n_u_func */
  {__pyx_k_function_context, sizeof(__pyx_k_function_context), 0, 1, 1}, /* PyObject cname: __pyx_n_u_function_context */
  {__pyx_k_gc, sizeof(__pyx_k_gc), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_gc */
  {__pyx_k_get_accumulators, sizeof(__pyx_k_get_accumulators), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_accumulators */
  {__pyx_k_get_key, sizeof(__pyx_k_get_key), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_key */
  {__pyx_k_get_namespace, sizeof(__pyx_k_get_namespace), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_namespace */
  {__pyx_k_get_runtime_context, sizeof(__pyx_k_get_runtime_context), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_runtime_context */
  {__pyx_k_get_state_list_view, sizeof(__pyx_k_get_state_list_view), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_state_list_view */
  {__pyx_k_get_state_map_view, sizeof(__pyx_k_get_state_map_view), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_state_map_view */
  {__pyx_k_get_timers, sizeof(__pyx_k_get_timers), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_timers */
  {__pyx_k_get_timestamp, sizeof(__pyx_k_get_timestamp), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_timestamp */
  {__pyx_k_get_value, sizeof(__pyx_k_get_value), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_value */
  {__pyx_k_get_value_indexes, sizeof(__pyx_k_get_value_indexes), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_value_indexes */
  {__pyx_k_get_value_indexes_length, sizeof(__pyx_k_get_value_indexes_length), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_value_indexes_length */
  {__pyx_k_get_value_state, sizeof(__pyx_k_get_value_state), 0, 1, 1}, /* PyObject cname: __pyx_n_u_get_value_state */
  {__pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 1, 1}, /* PyObject cname: __pyx_n_u_getstate */
  {__pyx_k_i, sizeof(__pyx_k_i), 0, 1, 1}, /* PyObject cname: __pyx_n_u_i */
  {__pyx_k_index_of_count_star, sizeof(__pyx_k_index_of_count_star), 0, 1, 1}, /* PyObject cname: __pyx_n_u_index_of_count_star */
  {__pyx_k_init, sizeof(__pyx_k_init), 0, 1, 1}, /* PyObject cname: __pyx_n_u_init */
  {__pyx_k_initializing, sizeof(__pyx_k_initializing), 0, 1, 1}, /* PyObject cname: __pyx_n_u_initializing */
  {__pyx_k_input_data, sizeof(__pyx_k_input_data), 0, 1, 1}, /* PyObject cname: __pyx_n_u_input_data */
  {__pyx_k_input_extractors, sizeof(__pyx_k_input_extractors), 0, 1, 1}, /* PyObject cname: __pyx_n_u_input_extractors */
  {__pyx_k_input_extractors_2, sizeof(__pyx_k_input_extractors_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_input_extractors_2 */
  {__pyx_k_input_row, sizeof(__pyx_k_input_row), 0, 1, 1}, /* PyObject cname: __pyx_n_u_input_row */
  {__pyx_k_internal_timer_service, sizeof(__pyx_k_internal_timer_service), 0, 1, 1}, /* PyObject cname: __pyx_n_u_internal_timer_service */
  {__pyx_k_is_coroutine, sizeof(__pyx_k_is_coroutine), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_coroutine */
  {__pyx_k_is_event_time, sizeof(__pyx_k_is_event_time), 0, 1, 1}, /* PyObject cname: __pyx_n_u_is_event_time */
  {__pyx_k_isenabled, sizeof(__pyx_k_isenabled), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_isenabled */
  {__pyx_k_items, sizeof(__pyx_k_items), 0, 1, 1}, /* PyObject cname: __pyx_n_u_items */
  {__pyx_k_join_row, sizeof(__pyx_k_join_row), 0, 1, 1}, /* PyObject cname: __pyx_n_u_join_row */
  {__pyx_k_key, sizeof(__pyx_k_key), 0, 1, 1}, /* PyObject cname: __pyx_n_u_key */
  {__pyx_k_key_coder, sizeof(__pyx_k_key_coder), 0, 1, 1}, /* PyObject cname: __pyx_n_u_key_coder */
  {__pyx_k_key_selector, sizeof(__pyx_k_key_selector), 0, 1, 1}, /* PyObject cname: __pyx_n_u_key_selector */
  {__pyx_k_key_selector_2, sizeof(__pyx_k_key_selector_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_key_selector_2 */
  {__pyx_k_keys, sizeof(__pyx_k_keys), 0, 1, 1}, /* PyObject cname: __pyx_n_u_keys */
  {__pyx_k_left, sizeof(__pyx_k_left), 0, 1, 1}, /* PyObject cname: __pyx_n_u_left */
  {__pyx_k_main, sizeof(__pyx_k_main), 0, 1, 1}, /* PyObject cname: __pyx_n_u_main */
  {__pyx_k_max_timestamp, sizeof(__pyx_k_max_timestamp), 0, 1, 1}, /* PyObject cname: __pyx_n_u_max_timestamp */
  {__pyx_k_merge, sizeof(__pyx_k_merge), 0, 1, 1}, /* PyObject cname: __pyx_n_u_merge */
  {__pyx_k_microsecond, sizeof(__pyx_k_microsecond), 0, 1, 1}, /* PyObject cname: __pyx_n_u_microsecond */
  {__pyx_k_module, sizeof(__pyx_k_module), 0, 1, 1}, /* PyObject cname: __pyx_n_u_module */
  {__pyx_k_name, sizeof(__pyx_k_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_name */
  {__pyx_k_named_property_extractor, sizeof(__pyx_k_named_property_extractor), 0, 1, 1}, /* PyObject cname: __pyx_n_u_named_property_extractor */
  {__pyx_k_named_property_extractor_2, sizeof(__pyx_k_named_property_extractor_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_named_property_extractor_2 */
  {__pyx_k_namespace, sizeof(__pyx_k_namespace), 0, 1, 1}, /* PyObject cname: __pyx_n_u_namespace */
  {__pyx_k_new, sizeof(__pyx_k_new), 0, 1, 1}, /* PyObject cname: __pyx_n_u_new */
  {__pyx_k_on_element, sizeof(__pyx_k_on_element), 0, 1, 1}, /* PyObject cname: __pyx_n_u_on_element */
  {__pyx_k_on_event_time, sizeof(__pyx_k_on_event_time), 0, 1, 1}, /* PyObject cname: __pyx_n_u_on_event_time */
  {__pyx_k_on_processing_time, sizeof(__pyx_k_on_processing_time), 0, 1, 1}, /* PyObject cname: __pyx_n_u_on_processing_time */
  {__pyx_k_open, sizeof(__pyx_k_open), 0, 1, 1}, /* PyObject cname: __pyx_n_u_open */
  {__pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pickle */
  {__pyx_k_pop, sizeof(__pyx_k_pop), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pop */
  {__pyx_k_prepare_aggregate_accumulator_fo, sizeof(__pyx_k_prepare_aggregate_accumulator_fo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_prepare_aggregate_accumulator_fo */
  {__pyx_k_process_element, sizeof(__pyx_k_process_element), 0, 1, 1}, /* PyObject cname: __pyx_n_u_process_element */
  {__pyx_k_process_watermark, sizeof(__pyx_k_process_watermark), 0, 1, 1}, /* PyObject cname: __pyx_n_u_process_watermark */
  {__pyx_k_pyflink_common_constants, sizeof(__pyx_k_pyflink_common_constants), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_common_constants */
  {__pyx_k_pyflink_fn_execution_coders, sizeof(__pyx_k_pyflink_fn_execution_coders), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_fn_execution_coders */
  {__pyx_k_pyflink_fn_execution_datastream, sizeof(__pyx_k_pyflink_fn_execution_datastream), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_fn_execution_datastream */
  {__pyx_k_pyflink_fn_execution_table_state, sizeof(__pyx_k_pyflink_fn_execution_table_state), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_fn_execution_table_state */
  {__pyx_k_pyflink_fn_execution_table_windo, sizeof(__pyx_k_pyflink_fn_execution_table_windo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_fn_execution_table_windo */
  {__pyx_k_pyflink_fn_execution_table_windo_2, sizeof(__pyx_k_pyflink_fn_execution_table_windo_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_fn_execution_table_windo_2 */
  {__pyx_k_pyflink_fn_execution_table_windo_3, sizeof(__pyx_k_pyflink_fn_execution_table_windo_3), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_fn_execution_table_windo_3 */
  {__pyx_k_pyflink_fn_execution_table_windo_4, sizeof(__pyx_k_pyflink_fn_execution_table_windo_4), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_fn_execution_table_windo_4 */
  {__pyx_k_pyflink_fn_execution_table_windo_5, sizeof(__pyx_k_pyflink_fn_execution_table_windo_5), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_fn_execution_table_windo_5 */
  {__pyx_k_pyflink_fn_execution_table_windo_6, sizeof(__pyx_k_pyflink_fn_execution_table_windo_6), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_pyflink_fn_execution_table_windo_6 */
  {__pyx_k_pyflink_table_udf, sizeof(__pyx_k_pyflink_table_udf), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyflink_table_udf */
  {__pyx_k_pytz, sizeof(__pyx_k_pytz), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pytz */
  {__pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_PickleError */
  {__pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_checksum */
  {__pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_result */
  {__pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_state */
  {__pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_type */
  {__pyx_k_pyx_unpickle_GroupWindowAggFun, sizeof(__pyx_k_pyx_unpickle_GroupWindowAggFun), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_unpickle_GroupWindowAggFun */
  {__pyx_k_pyx_unpickle_GroupWindowAggFun_2, sizeof(__pyx_k_pyx_unpickle_GroupWindowAggFun_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_unpickle_GroupWindowAggFun_2 */
  {__pyx_k_pyx_unpickle_GroupWindowAggFun_3, sizeof(__pyx_k_pyx_unpickle_GroupWindowAggFun_3), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_unpickle_GroupWindowAggFun_3 */
  {__pyx_k_pyx_unpickle_GroupWindowAggFun_4, sizeof(__pyx_k_pyx_unpickle_GroupWindowAggFun_4), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_unpickle_GroupWindowAggFun_4 */
  {__pyx_k_pyx_unpickle_NamespaceAggsHand, sizeof(__pyx_k_pyx_unpickle_NamespaceAggsHand), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_unpickle_NamespaceAggsHand */
  {__pyx_k_pyx_unpickle_NamespaceAggsHand_2, sizeof(__pyx_k_pyx_unpickle_NamespaceAggsHand_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_unpickle_NamespaceAggsHand_2 */
  {__pyx_k_pyx_unpickle_NamespaceAggsHand_3, sizeof(__pyx_k_pyx_unpickle_NamespaceAggsHand_3), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_unpickle_NamespaceAggsHand_3 */
  {__pyx_k_pyx_unpickle_NamespaceAggsHand_4, sizeof(__pyx_k_pyx_unpickle_NamespaceAggsHand_4), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_unpickle_NamespaceAggsHand_4 */
  {__pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 1, 1}, /* PyObject cname: __pyx_n_u_pyx_vtable */
  {__pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 1, 1}, /* PyObject cname: __pyx_n_u_qualname */
  {__pyx_k_range, sizeof(__pyx_k_range), 0, 1, 1}, /* PyObject cname: __pyx_n_u_range */
  {__pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 1, 1}, /* PyObject cname: __pyx_n_u_reduce */
  {__pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 1, 1}, /* PyObject cname: __pyx_n_u_reduce_cython */
  {__pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 1, 1}, /* PyObject cname: __pyx_n_u_reduce_ex */
  {__pyx_k_register_cleanup_timer, sizeof(__pyx_k_register_cleanup_timer), 0, 1, 1}, /* PyObject cname: __pyx_n_u_register_cleanup_timer */
  {__pyx_k_register_event_time_timer, sizeof(__pyx_k_register_event_time_timer), 0, 1, 1}, /* PyObject cname: __pyx_n_u_register_event_time_timer */
  {__pyx_k_register_processing_time_timer, sizeof(__pyx_k_register_processing_time_timer), 0, 1, 1}, /* PyObject cname: __pyx_n_u_register_processing_time_timer */
  {__pyx_k_remove, sizeof(__pyx_k_remove), 0, 1, 1}, /* PyObject cname: __pyx_n_u_remove */
  {__pyx_k_replace, sizeof(__pyx_k_replace), 0, 1, 1}, /* PyObject cname: __pyx_n_u_replace */
  {__pyx_k_retract, sizeof(__pyx_k_retract), 0, 1, 1}, /* PyObject cname: __pyx_n_u_retract */
  {__pyx_k_right, sizeof(__pyx_k_right), 0, 1, 1}, /* PyObject cname: __pyx_n_u_right */
  {__pyx_k_row_kind, sizeof(__pyx_k_row_kind), 0, 1, 1}, /* PyObject cname: __pyx_n_u_row_kind */
  {__pyx_k_rowtime_index, sizeof(__pyx_k_rowtime_index), 0, 1, 1}, /* PyObject cname: __pyx_n_u_rowtime_index */
  {__pyx_k_rowtime_index_2, sizeof(__pyx_k_rowtime_index_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_rowtime_index_2 */
  {__pyx_k_self, sizeof(__pyx_k_self), 0, 1, 1}, /* PyObject cname: __pyx_n_u_self */
  {__pyx_k_self__distinct_indexes_self__fil, sizeof(__pyx_k_self__distinct_indexes_self__fil), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_self__distinct_indexes_self__fil */
  {__pyx_k_set_accumulators, sizeof(__pyx_k_set_accumulators), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_accumulators */
  {__pyx_k_set_current_key, sizeof(__pyx_k_set_current_key), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_current_key */
  {__pyx_k_set_current_namespace, sizeof(__pyx_k_set_current_namespace), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_current_namespace */
  {__pyx_k_set_name, sizeof(__pyx_k_set_name), 0, 1, 1}, /* PyObject cname: __pyx_n_u_set_name */
  {__pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 1, 1}, /* PyObject cname: __pyx_n_u_setstate */
  {__pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 1, 1}, /* PyObject cname: __pyx_n_u_setstate_cython */
  {__pyx_k_shift_timezone, sizeof(__pyx_k_shift_timezone), 0, 1, 1}, /* PyObject cname: __pyx_n_u_shift_timezone */
  {__pyx_k_shift_timezone_2, sizeof(__pyx_k_shift_timezone_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_shift_timezone_2 */
  {__pyx_k_spec, sizeof(__pyx_k_spec), 0, 1, 1}, /* PyObject cname: __pyx_n_u_spec */
  {__pyx_k_state, sizeof(__pyx_k_state), 0, 1, 1}, /* PyObject cname: __pyx_n_u_state */
  {__pyx_k_state_backend, sizeof(__pyx_k_state_backend), 0, 1, 1}, /* PyObject cname: __pyx_n_u_state_backend */
  {__pyx_k_state_backend_2, sizeof(__pyx_k_state_backend_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_state_backend_2 */
  {__pyx_k_state_data_view_store, sizeof(__pyx_k_state_data_view_store), 0, 1, 1}, /* PyObject cname: __pyx_n_u_state_data_view_store */
  {__pyx_k_state_id, sizeof(__pyx_k_state_id), 0, 1, 1}, /* PyObject cname: __pyx_n_u_state_id */
  {__pyx_k_state_value_coder, sizeof(__pyx_k_state_value_coder), 0, 1, 1}, /* PyObject cname: __pyx_n_u_state_value_coder */
  {__pyx_k_state_value_coder_2, sizeof(__pyx_k_state_value_coder_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_state_value_coder_2 */
  {__pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 1, 0}, /* PyObject cname: __pyx_kp_u_stringsource */
  {__pyx_k_super, sizeof(__pyx_k_super), 0, 1, 1}, /* PyObject cname: __pyx_n_u_super */
  {__pyx_k_temp, sizeof(__pyx_k_temp), 0, 1, 1}, /* PyObject cname: __pyx_n_u_temp */
  {__pyx_k_test, sizeof(__pyx_k_test), 0, 1, 1}, /* PyObject cname: __pyx_n_u_test */
  {__pyx_k_timer, sizeof(__pyx_k_timer), 0, 1, 1}, /* PyObject cname: __pyx_n_u_timer */
  {__pyx_k_timers, sizeof(__pyx_k_timers), 0, 1, 1}, /* PyObject cname: __pyx_n_u_timers */
  {__pyx_k_timestamp, sizeof(__pyx_k_timestamp), 0, 1, 1}, /* PyObject cname: __pyx_n_u_timestamp */
  {__pyx_k_timezone, sizeof(__pyx_k_timezone), 0, 1, 1}, /* PyObject cname: __pyx_n_u_timezone */
  {__pyx_k_to_utc_timestamp_mills, sizeof(__pyx_k_to_utc_timestamp_mills), 0, 1, 1}, /* PyObject cname: __pyx_n_u_to_utc_timestamp_mills */
  {__pyx_k_total_seconds, sizeof(__pyx_k_total_seconds), 0, 1, 1}, /* PyObject cname: __pyx_n_u_total_seconds */
  {__pyx_k_trigger, sizeof(__pyx_k_trigger), 0, 1, 1}, /* PyObject cname: __pyx_n_u_trigger */
  {__pyx_k_trigger_2, sizeof(__pyx_k_trigger_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_trigger_2 */
  {__pyx_k_trigger_context, sizeof(__pyx_k_trigger_context), 0, 1, 1}, /* PyObject cname: __pyx_n_u_trigger_context */
  {__pyx_k_typing, sizeof(__pyx_k_typing), 0, 1, 1}, /* PyObject cname: __pyx_n_u_typing */
  {__pyx_k_tzinfo, sizeof(__pyx_k_tzinfo), 0, 1, 1}, /* PyObject cname: __pyx_n_u_tzinfo */
  {__pyx_k_udf_data_view_specs, sizeof(__pyx_k_udf_data_view_specs), 0, 1, 1}, /* PyObject cname: __pyx_n_u_udf_data_view_specs */
  {__pyx_k_udf_data_view_specs_2, sizeof(__pyx_k_udf_data_view_specs_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_udf_data_view_specs_2 */
  {__pyx_k_udf_data_views, sizeof(__pyx_k_udf_data_views), 0, 1, 1}, /* PyObject cname: __pyx_n_u_udf_data_views */
  {__pyx_k_udf_num, sizeof(__pyx_k_udf_num), 0, 1, 1}, /* PyObject cname: __pyx_n_u_udf_num */
  {__pyx_k_udfs, sizeof(__pyx_k_udfs), 0, 1, 1}, /* PyObject cname: __pyx_n_u_udfs */
  {__pyx_k_udfs_2, sizeof(__pyx_k_udfs_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_udfs_2 */
  {__pyx_k_update, sizeof(__pyx_k_update), 0, 1, 1}, /* PyObject cname: __pyx_n_u_update */
  {__pyx_k_use_setstate, sizeof(__pyx_k_use_setstate), 0, 1, 1}, /* PyObject cname: __pyx_n_u_use_setstate */
  {__pyx_k_utc, sizeof(__pyx_k_utc), 0, 1, 1}, /* PyObject cname: __pyx_n_u_utc */
  {__pyx_k_utcfromtimestamp, sizeof(__pyx_k_utcfromtimestamp), 0, 1, 1}, /* PyObject cname: __pyx_n_u_utcfromtimestamp */
  {__pyx_k_value, sizeof(__pyx_k_value), 0, 1, 1}, /* PyObject cname: __pyx_n_u_value */
  {__pyx_k_value_coder, sizeof(__pyx_k_value_coder), 0, 1, 1}, /* PyObject cname: __pyx_n_u_value_coder */
  {__pyx_k_values, sizeof(__pyx_k_values), 0, 1, 1}, /* PyObject cname: __pyx_n_u_values */
  {__pyx_k_watermark, sizeof(__pyx_k_watermark), 0, 1, 1}, /* PyObject cname: __pyx_n_u_watermark */
  {__pyx_k_window, sizeof(__pyx_k_window), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window */
  {__pyx_k_window_aggregator, sizeof(__pyx_k_window_aggregator), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window_aggregator */
  {__pyx_k_window_aggregator_2, sizeof(__pyx_k_window_aggregator_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window_aggregator_2 */
  {__pyx_k_window_assigner, sizeof(__pyx_k_window_assigner), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window_assigner */
  {__pyx_k_window_assigner_2, sizeof(__pyx_k_window_assigner_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window_assigner_2 */
  {__pyx_k_window_context, sizeof(__pyx_k_window_context), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window_context */
  {__pyx_k_window_function, sizeof(__pyx_k_window_function), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window_function */
  {__pyx_k_window_state, sizeof(__pyx_k_window_state), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window_state */
  {__pyx_k_window_state_2, sizeof(__pyx_k_window_state_2), 0, 1, 1}, /* PyObject cname: __pyx_n_u_window_state_2 */
  {0, 0, 0, 0, 0}
};
/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry const *t, PyObject **target, const char* const* encoding_names);

/* #### Code section: cached_builtins ### */

static int __Pyx_InitCachedBuiltins(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_range); if (!__pyx_builtin_range) __PYX_ERR(0, 144, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 2, __pyx_L1_error)
  __pyx_builtin_super = __Pyx_GetBuiltinName(__pyx_mstate->__pyx_n_u_super); if (!__pyx_builtin_super) __PYX_ERR(0, 517, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cached_constants ### */

static int __Pyx_InitCachedConstants(__pyx_mstatetype *__pyx_mstate) {
  __Pyx_RefNannyDeclarations
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum not in (0xe3b0c44, 0xda39a3e, 0xd41d8cd):             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError, "Incompatible checksums (0x%x vs (0xe3b0c44, 0xda39a3e, 0xd41d8cd) = ())" % __pyx_checksum
*/
  __pyx_mstate_global->__pyx_tuple[0] = PyTuple_Pack(3, __pyx_mstate_global->__pyx_int_238750788, __pyx_mstate_global->__pyx_int_228825662, __pyx_mstate_global->__pyx_int_222419149); if (unlikely(!__pyx_mstate_global->__pyx_tuple[0])) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[0]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[0]);
  __pyx_mstate_global->__pyx_tuple[1] = PyTuple_Pack(3, __pyx_mstate_global->__pyx_int_209029984, __pyx_mstate_global->__pyx_int_185799962, __pyx_mstate_global->__pyx_int_179515037); if (unlikely(!__pyx_mstate_global->__pyx_tuple[1])) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_GIVEREF(__pyx_mstate_global->__pyx_tuple[1]);
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}
/* #### Code section: init_constants ### */

static int __Pyx_InitConstants(__pyx_mstatetype *__pyx_mstate) {
  CYTHON_UNUSED_VAR(__pyx_mstate);
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.type = (PyObject*)&PyDict_Type;
  __pyx_mstate->__pyx_umethod_PyDict_Type_pop.method_name = &__pyx_mstate->__pyx_n_u_pop;
  __pyx_mstate->__pyx_umethod_PyList_Type____add.type = (PyObject*)(&PyList_Type);
  __pyx_mstate->__pyx_umethod_PyList_Type____add.method_name = &__pyx_mstate->__pyx_n_u_add;
  __pyx_mstate->__pyx_umethod_PyList_Type__remove.type = (PyObject*)(&PyList_Type);
  __pyx_mstate->__pyx_umethod_PyList_Type__remove.method_name = &__pyx_mstate->__pyx_n_u_remove;
  if (__Pyx_InitStrings(__pyx_string_tab, __pyx_mstate->__pyx_string_tab, __pyx_string_tab_encodings) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_mstate->__pyx_float_1000_0 = PyFloat_FromDouble(1000.0); if (unlikely(!__pyx_mstate->__pyx_float_1000_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_0 = PyLong_FromLong(0); if (unlikely(!__pyx_mstate->__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_1 = PyLong_FromLong(1); if (unlikely(!__pyx_mstate->__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_179515037 = PyLong_FromLong(179515037L); if (unlikely(!__pyx_mstate->__pyx_int_179515037)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_185799962 = PyLong_FromLong(185799962L); if (unlikely(!__pyx_mstate->__pyx_int_185799962)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_209029984 = PyLong_FromLong(209029984L); if (unlikely(!__pyx_mstate->__pyx_int_209029984)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_222419149 = PyLong_FromLong(222419149L); if (unlikely(!__pyx_mstate->__pyx_int_222419149)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_228825662 = PyLong_FromLong(228825662L); if (unlikely(!__pyx_mstate->__pyx_int_228825662)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_mstate->__pyx_int_238750788 = PyLong_FromLong(238750788L); if (unlikely(!__pyx_mstate->__pyx_int_238750788)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: init_codeobjects ### */
\
        typedef struct {
            unsigned int argcount : 4;
            unsigned int num_posonly_args : 1;
            unsigned int num_kwonly_args : 1;
            unsigned int nlocals : 5;
            unsigned int flags : 10;
            unsigned int first_line : 10;
            unsigned int line_table_length : 13;
        } __Pyx_PyCode_New_function_description;
/* NewCodeObj.proto */
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        const char *line_table,
        PyObject *tuple_dedup_map
);


static int __Pyx_CreateCodeObjects(__pyx_mstatetype *__pyx_mstate) {
  PyObject* tuple_dedup_map = PyDict_New();
  if (unlikely(!tuple_dedup_map)) return -1;
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 45, 19};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_left, __pyx_mstate->__pyx_n_u_right, __pyx_mstate->__pyx_n_u_row_kind};
    __pyx_mstate_global->__pyx_codeobj_tab[0] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_join_row, __pyx_k_at81HA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[0])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 49, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state_data_view_store};
    __pyx_mstate_global->__pyx_codeobj_tab[1] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_open, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[1])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 57, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_input_data};
    __pyx_mstate_global->__pyx_codeobj_tab[2] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_accumulate, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[2])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 65, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_input_data};
    __pyx_mstate_global->__pyx_codeobj_tab[3] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_retract, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[3])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 72, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_namespace, __pyx_mstate->__pyx_n_u_accumulators};
    __pyx_mstate_global->__pyx_codeobj_tab[4] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_merge, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[4])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 78, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_namespace, __pyx_mstate->__pyx_n_u_accumulators};
    __pyx_mstate_global->__pyx_codeobj_tab[5] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_set_accumulators, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[5])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 84, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[6] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_get_accumulators, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[6])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 93, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[7] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_create_accumulators, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[7])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 101, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_namespace};
    __pyx_mstate_global->__pyx_codeobj_tab[8] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_cleanup, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[8])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 107, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[9] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_close, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[9])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 88};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[10] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_k_A_G1F_a_vWA_q_q_q_AWKW___AWKWX, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[10])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16, 11};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[11] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_k_QfA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[11])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 115, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_namespace};
    __pyx_mstate_global->__pyx_codeobj_tab[12] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_get_value, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[12])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 88};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[13] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_k_A_G1F_a_vWA_q_q_q_D_SZZ_D_ST, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[13])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16, 11};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[14] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_k_9, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[14])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {10, 0, 0, 13, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 125, 292};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_udfs, __pyx_mstate->__pyx_n_u_input_extractors, __pyx_mstate->__pyx_n_u_index_of_count_star, __pyx_mstate->__pyx_n_u_count_star_inserted, __pyx_mstate->__pyx_n_u_named_property_extractor, __pyx_mstate->__pyx_n_u_udf_data_view_specs, __pyx_mstate->__pyx_n_u_filter_args, __pyx_mstate->__pyx_n_u_distinct_indexes, __pyx_mstate->__pyx_n_u_distinct_view_descriptors, __pyx_mstate->__pyx_n_u_i, __pyx_mstate->__pyx_n_u_temp, __pyx_mstate->__pyx_n_u_i};
    __pyx_mstate_global->__pyx_codeobj_tab[15] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_init, __pyx_k_A_IQ_L_1D_Q_A_q_G6_j_at_Rq_E_at1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[15])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 16, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 158, 192};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state_data_view_store, __pyx_mstate->__pyx_n_u_udfs_2, __pyx_mstate->__pyx_n_u_udf_num, __pyx_mstate->__pyx_n_u_input_extractors_2, __pyx_mstate->__pyx_n_u_named_property_extractor_2, __pyx_mstate->__pyx_n_u_accumulators_2, __pyx_mstate->__pyx_n_u_udf_data_view_specs_2, __pyx_mstate->__pyx_n_u_udf_data_views, __pyx_mstate->__pyx_n_u_filter_args_2, __pyx_mstate->__pyx_n_u_distinct_indexes_2, __pyx_mstate->__pyx_n_u_distinct_view_descriptors_2, __pyx_mstate->__pyx_n_u_distinct_data_views, __pyx_mstate->__pyx_n_u_named_property_extractor, __pyx_mstate->__pyx_n_u_get_value_indexes, __pyx_mstate->__pyx_n_u_get_value_indexes_length};
    __pyx_mstate_global->__pyx_codeobj_tab[16] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_open, __pyx_k_A_G4q_uA_a_t1_Q_a_a_Q_A_q_0_a_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[16])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 16, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 183, 363};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_input_data, __pyx_mstate->__pyx_n_u_udfs_2, __pyx_mstate->__pyx_n_u_udf_num, __pyx_mstate->__pyx_n_u_input_extractors_2, __pyx_mstate->__pyx_n_u_named_property_extractor_2, __pyx_mstate->__pyx_n_u_accumulators_2, __pyx_mstate->__pyx_n_u_udf_data_view_specs_2, __pyx_mstate->__pyx_n_u_udf_data_views, __pyx_mstate->__pyx_n_u_filter_args_2, __pyx_mstate->__pyx_n_u_distinct_indexes_2, __pyx_mstate->__pyx_n_u_distinct_view_descriptors_2, __pyx_mstate->__pyx_n_u_distinct_data_views, __pyx_mstate->__pyx_n_u_named_property_extractor, __pyx_mstate->__pyx_n_u_get_value_indexes, __pyx_mstate->__pyx_n_u_get_value_indexes_length};
    __pyx_mstate_global->__pyx_codeobj_tab[17] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_accumulate, __pyx_k_A_E_at1_r_D_4_J_1_8_A_q_q_U_1_Qa, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[17])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 16, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 231, 343};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_input_data, __pyx_mstate->__pyx_n_u_udfs_2, __pyx_mstate->__pyx_n_u_udf_num, __pyx_mstate->__pyx_n_u_input_extractors_2, __pyx_mstate->__pyx_n_u_named_property_extractor_2, __pyx_mstate->__pyx_n_u_accumulators_2, __pyx_mstate->__pyx_n_u_udf_data_view_specs_2, __pyx_mstate->__pyx_n_u_udf_data_views, __pyx_mstate->__pyx_n_u_filter_args_2, __pyx_mstate->__pyx_n_u_distinct_indexes_2, __pyx_mstate->__pyx_n_u_distinct_view_descriptors_2, __pyx_mstate->__pyx_n_u_distinct_data_views, __pyx_mstate->__pyx_n_u_named_property_extractor, __pyx_mstate->__pyx_n_u_get_value_indexes, __pyx_mstate->__pyx_n_u_get_value_indexes_length};
    __pyx_mstate_global->__pyx_codeobj_tab[18] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_retract, __pyx_k_A_E_at1_r_D_4_J_1_8_A_q_q_U_1_Qa_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[18])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 271, 111};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_namespace, __pyx_mstate->__pyx_n_u_accumulators};
    __pyx_mstate_global->__pyx_codeobj_tab[19] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_merge, __pyx_k_A_4q_U_3at1_G_ARvQ_31A_1IQ_E_at1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[19])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 281, 86};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_namespace, __pyx_mstate->__pyx_n_u_accumulators};
    __pyx_mstate_global->__pyx_codeobj_tab[20] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_set_accumulators, __pyx_k_A_4_JgQ_U_3at1_G_ARvQ_31A_1IQ_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[20])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 15, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 290, 9};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_udfs_2, __pyx_mstate->__pyx_n_u_udf_num, __pyx_mstate->__pyx_n_u_input_extractors_2, __pyx_mstate->__pyx_n_u_named_property_extractor_2, __pyx_mstate->__pyx_n_u_accumulators_2, __pyx_mstate->__pyx_n_u_udf_data_view_specs_2, __pyx_mstate->__pyx_n_u_udf_data_views, __pyx_mstate->__pyx_n_u_filter_args_2, __pyx_mstate->__pyx_n_u_distinct_indexes_2, __pyx_mstate->__pyx_n_u_distinct_view_descriptors_2, __pyx_mstate->__pyx_n_u_distinct_data_views, __pyx_mstate->__pyx_n_u_named_property_extractor, __pyx_mstate->__pyx_n_u_get_value_indexes, __pyx_mstate->__pyx_n_u_get_value_indexes_length};
    __pyx_mstate_global->__pyx_codeobj_tab[21] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_get_accumulators, __pyx_k_A_t1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[21])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 293, 22};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[22] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_create_accumulators, __pyx_k_A_q_c_WD, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[22])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 296, 54};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_namespace};
    __pyx_mstate_global->__pyx_codeobj_tab[23] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_cleanup, __pyx_k_A_E_as_4q_T_1_G1_q_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[23])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 15, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 303, 18};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_udfs_2, __pyx_mstate->__pyx_n_u_udf_num, __pyx_mstate->__pyx_n_u_input_extractors_2, __pyx_mstate->__pyx_n_u_named_property_extractor_2, __pyx_mstate->__pyx_n_u_accumulators_2, __pyx_mstate->__pyx_n_u_udf_data_view_specs_2, __pyx_mstate->__pyx_n_u_udf_data_views, __pyx_mstate->__pyx_n_u_filter_args_2, __pyx_mstate->__pyx_n_u_distinct_indexes_2, __pyx_mstate->__pyx_n_u_distinct_view_descriptors_2, __pyx_mstate->__pyx_n_u_distinct_data_views, __pyx_mstate->__pyx_n_u_named_property_extractor, __pyx_mstate->__pyx_n_u_get_value_indexes, __pyx_mstate->__pyx_n_u_get_value_indexes_length};
    __pyx_mstate_global->__pyx_codeobj_tab[24] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_close, __pyx_k_A_G4q_vQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[24])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 307, 90};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_namespace};
    __pyx_mstate_global->__pyx_codeobj_tab[25] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_get_value, __pyx_k_A_D_a_Jat_J_U_4q_4q_8_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[25])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 319, 34};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[26] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_dealloc, __pyx_k_A_4q_Q_4q_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[26])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 9};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[27] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_k_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[27])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 3, 9};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[28] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_k_Q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[28])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {10, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 326, 149};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_allowed_lateness, __pyx_mstate->__pyx_n_u_key_selector, __pyx_mstate->__pyx_n_u_state_backend, __pyx_mstate->__pyx_n_u_state_value_coder, __pyx_mstate->__pyx_n_u_window_assigner, __pyx_mstate->__pyx_n_u_window_aggregator, __pyx_mstate->__pyx_n_u_trigger, __pyx_mstate->__pyx_n_u_rowtime_index, __pyx_mstate->__pyx_n_u_shift_timezone};
    __pyx_mstate_global->__pyx_codeobj_tab[29] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_init, __pyx_k_A_Q_a_a_q_q_q_L_T_0_AQQR, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[29])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 351, 238};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_function_context};
    __pyx_mstate_global->__pyx_codeobj_tab[30] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_open, __pyx_k_A_EQd_uA_q_a_Qd_Q_a_t1_q_q_t1_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[30])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 374, 381};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_input_row};
    __pyx_mstate_global->__pyx_codeobj_tab[31] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_process_element, __pyx_k_A_iq_d_O_1A_4_a_1D_c_81G89E_ST_Y, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[31])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 420, 15};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_watermark};
    __pyx_mstate_global->__pyx_codeobj_tab[32] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_process_watermark, __pyx_k_A_6aq, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[32])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 423, 122};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_timer};
    __pyx_mstate_global->__pyx_codeobj_tab[33] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_on_event_time, __pyx_k_A_E_q_e81_O_1A_nA_Zq_4_aq_2_5_4, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[33])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 441, 125};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_timer};
    __pyx_mstate_global->__pyx_codeobj_tab[34] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_on_processing_time, __pyx_k_A_E_q_e81_O_1A_nA_Zq_4_31A_2_5_4, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[34])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 459, 97};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_epoch_mills};
    __pyx_mstate_global->__pyx_codeobj_tab[35] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_to_utc_timestamp_mills, __pyx_k_A_4_1_1_t9AT_hi_Ql_GST_HI_6aq_3b, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[35])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 469, 51};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[36] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_get_timers, __pyx_k_A_IT_1_a_HF_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[36])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 1, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 478, 12};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self};
    __pyx_mstate_global->__pyx_codeobj_tab[37] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_close, __pyx_k_A_vQ, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[37])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 16, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 481, 63};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_window, __pyx_mstate->__pyx_n_u_allowed_lateness_2, __pyx_mstate->__pyx_n_u_rowtime_index_2, __pyx_mstate->__pyx_n_u_shift_timezone_2, __pyx_mstate->__pyx_n_u_key_selector_2, __pyx_mstate->__pyx_n_u_state_value_coder_2, __pyx_mstate->__pyx_n_u_state_backend_2, __pyx_mstate->__pyx_n_u_window_assigner_2, __pyx_mstate->__pyx_n_u_window_aggregator_2, __pyx_mstate->__pyx_n_u_trigger_2, __pyx_mstate->__pyx_n_u_window_function, __pyx_mstate->__pyx_n_u_internal_timer_service, __pyx_mstate->__pyx_n_u_window_context, __pyx_mstate->__pyx_n_u_trigger_context, __pyx_mstate->__pyx_n_u_window_state_2};
    __pyx_mstate_global->__pyx_codeobj_tab[38] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_register_cleanup_timer, __pyx_k_A_t_1_4_a_1A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[38])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 492, 67};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_window};
    __pyx_mstate_global->__pyx_codeobj_tab[39] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_cleanup_time, __pyx_k_A_4_a_as_c_4q_Cv_1_q_q_s_a, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[39])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 17, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 503, 2};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_window, __pyx_mstate->__pyx_n_u_allowed_lateness_2, __pyx_mstate->__pyx_n_u_rowtime_index_2, __pyx_mstate->__pyx_n_u_shift_timezone_2, __pyx_mstate->__pyx_n_u_key_selector_2, __pyx_mstate->__pyx_n_u_state_value_coder_2, __pyx_mstate->__pyx_n_u_state_backend_2, __pyx_mstate->__pyx_n_u_window_assigner_2, __pyx_mstate->__pyx_n_u_window_aggregator_2, __pyx_mstate->__pyx_n_u_trigger_2, __pyx_mstate->__pyx_n_u_window_function, __pyx_mstate->__pyx_n_u_internal_timer_service, __pyx_mstate->__pyx_n_u_window_context, __pyx_mstate->__pyx_n_u_trigger_context, __pyx_mstate->__pyx_n_u_window_state_2};
    __pyx_mstate_global->__pyx_codeobj_tab[40] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_emit_window_result, __pyx_k_A, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[40])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 600};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[41] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_k_T_T_C4GWW_llp_q_C_C_G_G_X_X_q_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[41])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16, 11};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[42] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_k_8_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[42])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {10, 0, 0, 10, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 507, 68};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_allowed_lateness, __pyx_mstate->__pyx_n_u_key_selector, __pyx_mstate->__pyx_n_u_state_backend, __pyx_mstate->__pyx_n_u_state_value_coder, __pyx_mstate->__pyx_n_u_window_assigner, __pyx_mstate->__pyx_n_u_window_aggregator, __pyx_mstate->__pyx_n_u_trigger, __pyx_mstate->__pyx_n_u_rowtime_index, __pyx_mstate->__pyx_n_u_shift_timezone};
    __pyx_mstate_global->__pyx_codeobj_tab[43] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_init, __pyx_k_A_Q_e9A_nO_Na_y_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[43])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 3, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 522, 50};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_key, __pyx_mstate->__pyx_n_u_window};
    __pyx_mstate_global->__pyx_codeobj_tab[44] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_pyflink_fn_execution_table_windo_6, __pyx_mstate->__pyx_n_u_emit_window_result, __pyx_k_A_DAQ_T_Jaq_XQe_a_q, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[44])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {1, 0, 0, 4, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 598};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_state, __pyx_mstate->__pyx_n_u_dict_2, __pyx_mstate->__pyx_n_u_use_setstate};
    __pyx_mstate_global->__pyx_codeobj_tab[45] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_reduce_cython, __pyx_k_T_T_C4GWW_llp_q_C_C_G_G_X_X_q_q_2, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[45])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 16, 11};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_self, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[46] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_setstate_cython, __pyx_k_4AV1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[46])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 79};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_PickleError, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[47] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_NamespaceAggsHand, __pyx_k_hk_A_1_kkmmn_2_1_7_A_Bddrrs_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[47])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 11, 40};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_result, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[48] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_NamespaceAggsHand_3, __pyx_k_s_D_q_a_IWA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[48])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 79};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_PickleError, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[49] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_NamespaceAggsHand_2, __pyx_k_hk_A_1_kkmmn_haq_7_Q_jjk_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[49])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 11, 40};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_result, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[50] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_NamespaceAggsHand_4, __pyx_k_s_D_q_a_IWA, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[50])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 90};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_PickleError, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[51] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_GroupWindowAggFun, __pyx_k_hk_A_1_a_XQa_7_A_ZZhhi_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[51])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 11, 482};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_result, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[52] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_GroupWindowAggFun_3, __pyx_k_A_gghhllx_y_J_J_U_U_V_V_Z_Z_f_f, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[52])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {3, 0, 0, 5, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 1, 90};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_type, __pyx_mstate->__pyx_n_u_pyx_checksum, __pyx_mstate->__pyx_n_u_pyx_state, __pyx_mstate->__pyx_n_u_pyx_PickleError, __pyx_mstate->__pyx_n_u_pyx_result};
    __pyx_mstate_global->__pyx_codeobj_tab[53] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_GroupWindowAggFun_2, __pyx_k_hk_A_1_a_7_8_9RR_a_1, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[53])) goto bad;
  }
  {
    const __Pyx_PyCode_New_function_description descr = {2, 0, 0, 2, (unsigned int)(CO_OPTIMIZED|CO_NEWLOCALS), 11, 482};
    PyObject* const varnames[] = {__pyx_mstate->__pyx_n_u_pyx_result, __pyx_mstate->__pyx_n_u_pyx_state};
    __pyx_mstate_global->__pyx_codeobj_tab[54] = __Pyx_PyCode_New(descr, varnames, __pyx_mstate->__pyx_kp_u_stringsource, __pyx_mstate->__pyx_n_u_pyx_unpickle_GroupWindowAggFun_4, __pyx_k_A_gghhllx_y_J_J_U_U_V_V_Z_Z_f_f, tuple_dedup_map); if (unlikely(!__pyx_mstate_global->__pyx_codeobj_tab[54])) goto bad;
  }
  Py_DECREF(tuple_dedup_map);
  return 0;
  bad:
  Py_DECREF(tuple_dedup_map);
  return -1;
}
/* #### Code section: init_globals ### */

static int __Pyx_InitGlobals(void) {
  /* PythonCompatibility.init */
  if (likely(__Pyx_init_co_variables() == 0)); else

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  /* CachedMethodType.init */
  #if CYTHON_COMPILING_IN_LIMITED_API
{
    PyObject *typesModule=NULL;
    typesModule = PyImport_ImportModule("types");
    if (typesModule) {
        __pyx_mstate_global->__Pyx_CachedMethodType = PyObject_GetAttrString(typesModule, "MethodType");
        Py_DECREF(typesModule);
    }
} // error handling follows
#endif

if (unlikely(PyErr_Occurred())) __PYX_ERR(0, 1, __pyx_L1_error)

  return 0;
  __pyx_L1_error:;
  return -1;
}
/* #### Code section: cleanup_globals ### */
/* #### Code section: cleanup_module ### */
/* #### Code section: main_method ### */
/* #### Code section: utility_code_pragmas ### */
#ifdef _MSC_VER
#pragma warning( push )
/* Warning 4127: conditional expression is constant
 * Cython uses constant conditional expressions to allow in inline functions to be optimized at
 * compile-time, so this warning is not useful
 */
#pragma warning( disable : 4127 )
#endif



/* #### Code section: utility_code_def ### */

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    int result;
    PyObject *exc_type;
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *current_exception = tstate->current_exception;
    if (unlikely(!current_exception)) return 0;
    exc_type = (PyObject*) Py_TYPE(current_exception);
    if (exc_type == err) return 1;
#else
    exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
#endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(exc_type);
    #endif
    if (unlikely(PyTuple_Check(err))) {
        result = __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    } else {
        result = __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
    }
    #if CYTHON_AVOID_BORROWED_REFS
    Py_DECREF(exc_type);
    #endif
    return result;
}
#endif

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject *tmp_value;
    assert(type == NULL || (value != NULL && type == (PyObject*) Py_TYPE(value)));
    if (value) {
        #if CYTHON_COMPILING_IN_CPYTHON
        if (unlikely(((PyBaseExceptionObject*) value)->traceback != tb))
        #endif
            PyException_SetTraceback(value, tb);
    }
    tmp_value = tstate->current_exception;
    tstate->current_exception = value;
    Py_XDECREF(tmp_value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
#else
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#endif
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#if PY_VERSION_HEX >= 0x030C00A6
    PyObject* exc_value;
    exc_value = tstate->current_exception;
    tstate->current_exception = 0;
    *value = exc_value;
    *type = NULL;
    *tb = NULL;
    if (exc_value) {
        *type = (PyObject*) Py_TYPE(exc_value);
        Py_INCREF(*type);
        #if CYTHON_COMPILING_IN_CPYTHON
        *tb = ((PyBaseExceptionObject*) exc_value)->traceback;
        Py_XINCREF(*tb);
        #else
        *tb = PyException_GetTraceback(exc_value);
        #endif
    }
#else
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#endif
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* PyObjectGetAttrStrNoError */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    (void) PyObject_GetOptionalAttr(obj, attr_name, &result);
    return result;
#else
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
#endif
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_b, name);
    if (unlikely(!result) && !PyErr_Occurred()) {
        PyErr_Format(PyExc_NameError,
            "name '%U' is not defined", name);
    }
    return result;
}

/* Profile */
#if CYTHON_PROFILE || CYTHON_TRACE
#if CYTHON_TRACE && !CYTHON_USE_SYS_MONITORING
static int __Pyx_call_line_trace_func(PyThreadState *tstate, PyFrameObject *frame, int line) {
    int ret;
    PyObject *type, *value, *traceback;
    __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
    __Pyx_PyFrame_SetLineNumber(frame, line);
    __Pyx_EnterTracing(tstate);
    ret = tstate->c_tracefunc(tstate->c_traceobj, frame, PyTrace_LINE, NULL);
    __Pyx_LeaveTracing(tstate);
    if (likely(!ret)) {
        __Pyx_ErrRestoreInState(tstate, type, value, traceback);
    } else {
        Py_XDECREF(type);
        Py_XDECREF(value);
        Py_XDECREF(traceback);
    }
    return ret;
}
#endif
CYTHON_UNUSED static PyCodeObject *__Pyx_createFrameCodeObject(const char *funcname, const char *srcfile, int firstlineno) {
    PyCodeObject *py_code = PyCode_NewEmpty(srcfile, funcname, firstlineno);
    if (likely(py_code)) {
        py_code->co_flags |= CO_OPTIMIZED | CO_NEWLOCALS;
    }
    return py_code;
}
#if CYTHON_USE_SYS_MONITORING
CYTHON_UNUSED static int __Pyx__TraceStartFunc(PyMonitoringState *state_array, PyObject *code_obj, int offset, int skip_event) {
    int ret;
    __pyx_monitoring_version_type version = 0;
    ret = PyMonitoring_EnterScope(state_array, &version, __Pyx_MonitoringEventTypes, __Pyx_MonitoringEventTypes_CyFunc_count);
    if (unlikely(ret == -1)) return -1;
    return skip_event ? 0 : PyMonitoring_FirePyStartEvent(&state_array[__Pyx_Monitoring_PY_START], code_obj, offset);
}
CYTHON_UNUSED static int __Pyx__TraceStartGen(PyMonitoringState *state_array, __pyx_monitoring_version_type *version, PyObject *code_obj, int offset) {
    int ret;
    ret = PyMonitoring_EnterScope(state_array, version, __Pyx_MonitoringEventTypes, __Pyx_MonitoringEventTypes_CyGen_count);
    if (unlikely(ret == -1)) return -1;
    return PyMonitoring_FirePyStartEvent(&state_array[__Pyx_Monitoring_PY_START], code_obj, offset);
}
CYTHON_UNUSED static int __Pyx__TraceResumeGen(PyMonitoringState *state_array, __pyx_monitoring_version_type *version, PyObject *code_obj, int offset) {
    int ret;
    ret = PyMonitoring_EnterScope(state_array, version, __Pyx_MonitoringEventTypes, __Pyx_MonitoringEventTypes_CyGen_count);
    if (unlikely(ret == -1)) return -1;
    return PyMonitoring_FirePyResumeEvent(&state_array[__Pyx_Monitoring_PY_RESUME], code_obj, offset);
}
CYTHON_UNUSED static void __Pyx__TraceException(PyMonitoringState *monitoring_state, PyObject *code_obj, int offset, int reraised) {
    if (reraised) {
        (void) PyMonitoring_FireReraiseEvent(monitoring_state, code_obj, offset);
    } else {
        (void) PyMonitoring_FireRaiseEvent(monitoring_state, code_obj, offset);
    }
}
#if CYTHON_TRACE
CYTHON_UNUSED static int __Pyx__TraceLine(PyMonitoringState *monitoring_state, PyObject *code_obj, int line, int offset) {
    int ret;
    PyObject *exc = PyErr_GetRaisedException();
    ret = PyMonitoring_FireLineEvent(monitoring_state, code_obj, offset, line);
    if (exc) PyErr_SetRaisedException(exc);
    return ret;
}
#endif
#else
static int __Pyx_TraceSetupAndCall(PyCodeObject** code,
                                   PyFrameObject** frame,
                                   PyThreadState* tstate,
                                   const char *funcname,
                                   const char *srcfile,
                                   int firstlineno,
                                   int skip_event) {
    if (*frame == NULL || !CYTHON_PROFILE_REUSE_FRAME) {
        int needs_new_code_obj = (*code == NULL);
        if (needs_new_code_obj) {
            *code = __Pyx_createFrameCodeObject(funcname, srcfile, firstlineno);
            if (*code == NULL) return 0;
        }
        *frame = PyFrame_New(
            tstate,                          /*PyThreadState *tstate*/
            *code,                           /*PyCodeObject *code*/
            __pyx_mstate_global->__pyx_d,    /*PyObject *globals*/
            0                                /*PyObject *locals*/
        );
        if (needs_new_code_obj && !CYTHON_PROFILE_REUSE_CODEOBJ)
            Py_CLEAR(*code); // otherwise the reference is owned externally
        if (*frame == NULL) return 0;
        if (CYTHON_TRACE && (*frame)->f_trace == NULL) {
            Py_INCREF(Py_None);
            (*frame)->f_trace = Py_None;
        }
    }
    if (!skip_event) {
        PyObject *type, *value, *traceback;
        int retval = 1;
        __Pyx_PyFrame_SetLineNumber(*frame, firstlineno);
        __Pyx_EnterTracing(tstate);
        __Pyx_ErrFetchInState(tstate, &type, &value, &traceback);
        #if CYTHON_TRACE
        if (tstate->c_tracefunc)
            retval = tstate->c_tracefunc(tstate->c_traceobj, *frame, PyTrace_CALL, NULL) == 0;
        if (retval && tstate->c_profilefunc)
        #endif
            retval = tstate->c_profilefunc(tstate->c_profileobj, *frame, PyTrace_CALL, NULL) == 0;
        __Pyx_LeaveTracing(tstate);
        if (unlikely(!retval)) {
            Py_XDECREF(type);
            Py_XDECREF(value);
            Py_XDECREF(traceback);
            return -1;
        }
        __Pyx_ErrRestoreInState(tstate, type, value, traceback);
    }
    return __Pyx_IsTracing(tstate, 0, 0);
}
#endif
#endif

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL && !CYTHON_VECTORCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject *const *args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject *const *args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
    PyObject *kwdefs;
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object"))) {
        return NULL;
    }
    if (
            co->co_kwonlyargcount == 0 &&
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = Py_TYPE(func)->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = __Pyx_CyOrPyCFunction_GET_FUNCTION(func);
    self = __Pyx_CyOrPyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall(" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectFastCall */
#if PY_VERSION_HEX < 0x03090000 || CYTHON_COMPILING_IN_LIMITED_API
static PyObject* __Pyx_PyObject_FastCall_fallback(PyObject *func, PyObject * const*args, size_t nargs, PyObject *kwargs) {
    PyObject *argstuple;
    PyObject *result = 0;
    size_t i;
    argstuple = PyTuple_New((Py_ssize_t)nargs);
    if (unlikely(!argstuple)) return NULL;
    for (i = 0; i < nargs; i++) {
        Py_INCREF(args[i]);
        if (__Pyx_PyTuple_SET_ITEM(argstuple, (Py_ssize_t)i, args[i]) != (0)) goto bad;
    }
    result = __Pyx_PyObject_Call(func, argstuple, kwargs);
  bad:
    Py_DECREF(argstuple);
    return result;
}
#endif
#if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
  #if PY_VERSION_HEX < 0x03090000
    #define __Pyx_PyVectorcall_Function(callable) _PyVectorcall_Function(callable)
  #elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE vectorcallfunc __Pyx_PyVectorcall_Function(PyObject *callable) {
    PyTypeObject *tp = Py_TYPE(callable);
    #if defined(__Pyx_CyFunction_USED)
    if (__Pyx_CyFunction_CheckExact(callable)) {
        return __Pyx_CyFunction_func_vectorcall(callable);
    }
    #endif
    if (!PyType_HasFeature(tp, Py_TPFLAGS_HAVE_VECTORCALL)) {
        return NULL;
    }
    assert(PyCallable_Check(callable));
    Py_ssize_t offset = tp->tp_vectorcall_offset;
    assert(offset > 0);
    vectorcallfunc ptr;
    memcpy(&ptr, (char *) callable + offset, sizeof(ptr));
    return ptr;
}
  #else
    #define __Pyx_PyVectorcall_Function(callable) PyVectorcall_Function(callable)
  #endif
#endif
static CYTHON_INLINE PyObject* __Pyx_PyObject_FastCallDict(PyObject *func, PyObject *const *args, size_t _nargs, PyObject *kwargs) {
    Py_ssize_t nargs = __Pyx_PyVectorcall_NARGS(_nargs);
#if CYTHON_COMPILING_IN_CPYTHON
    if (nargs == 0 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_NOARGS))
            return __Pyx_PyObject_CallMethO(func, NULL);
    }
    else if (nargs == 1 && kwargs == NULL) {
        if (__Pyx_CyOrPyCFunction_Check(func) && likely( __Pyx_CyOrPyCFunction_GET_FLAGS(func) & METH_O))
            return __Pyx_PyObject_CallMethO(func, args[0]);
    }
#endif
    #if PY_VERSION_HEX < 0x030800B1
    #if CYTHON_FAST_PYCCALL
    if (PyCFunction_Check(func)) {
        if (kwargs) {
            return _PyCFunction_FastCallDict(func, args, nargs, kwargs);
        } else {
            return _PyCFunction_FastCallKeywords(func, args, nargs, NULL);
        }
    }
    if (!kwargs && __Pyx_IS_TYPE(func, &PyMethodDescr_Type)) {
        return _PyMethodDescr_FastCallKeywords(func, args, nargs, NULL);
    }
    #endif
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs);
    }
    #endif
    #endif
    if (kwargs == NULL) {
        #if CYTHON_VECTORCALL && !CYTHON_COMPILING_IN_LIMITED_API
        vectorcallfunc f = __Pyx_PyVectorcall_Function(func);
        if (f) {
            return f(func, args, _nargs, NULL);
        }
        #elif defined(__Pyx_CyFunction_USED) && CYTHON_BACKPORT_VECTORCALL
        if (__Pyx_CyFunction_CheckExact(func)) {
            __pyx_vectorcallfunc f = __Pyx_CyFunction_func_vectorcall(func);
            if (f) return f(func, args, _nargs, NULL);
        }
        #elif CYTHON_COMPILING_IN_LIMITED_API && CYTHON_VECTORCALL
        return PyObject_Vectorcall(func, args, _nargs, NULL);
        #endif
    }
    if (nargs == 0) {
        return __Pyx_PyObject_Call(func, __pyx_mstate_global->__pyx_empty_tuple, kwargs);
    }
    #if PY_VERSION_HEX >= 0x03090000 && !CYTHON_COMPILING_IN_LIMITED_API
    return PyObject_VectorcallDict(func, args, (size_t)nargs, kwargs);
    #else
    return __Pyx_PyObject_FastCall_fallback(func, args, (size_t)nargs, kwargs);
    #endif
}

/* PyObjectCall2Args */
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args[3] = {NULL, arg1, arg2};
    return __Pyx_PyObject_FastCall(function, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* UnpackUnboundCMethod */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *args, PyObject *kwargs) {
    PyObject *result;
    PyObject *selfless_args = PyTuple_GetSlice(args, 1, PyTuple_Size(args));
    if (unlikely(!selfless_args)) return NULL;
    result = PyObject_Call(method, selfless_args, kwargs);
    Py_DECREF(selfless_args);
    return result;
}
#elif CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject **args, Py_ssize_t nargs, PyObject *kwnames) {
        return _PyObject_Vectorcall
            (method, args ? args+1 : NULL, nargs ? nargs-1 : 0, kwnames);
}
#else
static PyObject *__Pyx_SelflessCall(PyObject *method, PyObject *const *args, Py_ssize_t nargs, PyObject *kwnames) {
    return
#if PY_VERSION_HEX < 0x03090000
    _PyObject_Vectorcall
#else
    PyObject_Vectorcall
#endif
        (method, args ? args+1 : NULL, nargs ? (size_t) nargs-1 : 0, kwnames);
}
#endif
static PyMethodDef __Pyx_UnboundCMethod_Def = {
     "CythonUnboundCMethod",
     __PYX_REINTERPRET_FUNCION(PyCFunction, __Pyx_SelflessCall),
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030C0000
     METH_VARARGS | METH_KEYWORDS,
#else
     METH_FASTCALL | METH_KEYWORDS,
#endif
     NULL
};
static int __Pyx_TryUnpackUnboundCMethod(__Pyx_CachedCFunction* target) {
    PyObject *method, *result=NULL;
    method = __Pyx_PyObject_GetAttrStr(target->type, *target->method_name);
    if (unlikely(!method))
        return -1;
    result = method;
#if CYTHON_COMPILING_IN_CPYTHON
    if (likely(__Pyx_TypeCheck(method, &PyMethodDescr_Type)))
    {
        PyMethodDescrObject *descr = (PyMethodDescrObject*) method;
        target->func = descr->d_method->ml_meth;
        target->flag = descr->d_method->ml_flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_STACKLESS);
    } else
#endif
#if CYTHON_COMPILING_IN_PYPY
#else
    if (PyCFunction_Check(method))
#endif
    {
        PyObject *self;
        int self_found;
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        self = PyObject_GetAttrString(method, "__self__");
        if (!self) {
            PyErr_Clear();
        }
#else
        self = PyCFunction_GET_SELF(method);
#endif
        self_found = (self && self != Py_None);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_PYPY
        Py_XDECREF(self);
#endif
        if (self_found) {
            PyObject *unbound_method = PyCFunction_New(&__Pyx_UnboundCMethod_Def, method);
            if (unlikely(!unbound_method)) return -1;
            Py_DECREF(method);
            result = unbound_method;
        }
    }
#if !CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    if (unlikely(target->method)) {
        Py_DECREF(result);
    } else
#endif
    target->method = result;
    return 0;
}

/* CallUnboundCMethod1 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg) {
    int was_initialized =  __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        int flag = cfunc->flag;
        if (flag == METH_O) {
            return __Pyx_CallCFunction(cfunc, self, arg);
        } else if (flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, &arg, 1);
        } else if (flag == (METH_FASTCALL | METH_KEYWORDS)) {
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, &arg, 1, NULL);
        }
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod1(&tmp_cfunc, self, arg);
    }
#endif
    PyObject* result = __Pyx__CallUnboundCMethod1(cfunc, self, arg);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod1(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg){
    PyObject *result = NULL;
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *args = PyTuple_New(1);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg);
        PyTuple_SET_ITEM(args, 0, arg);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
    } else
#endif
    {
        result = __Pyx_PyObject_Call2Args(cfunc->method, self, arg);
    }
    return result;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* TupleAndListFromArray */
#if !CYTHON_COMPILING_IN_CPYTHON && CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    Py_ssize_t i;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    for (i = 0; i < n; i++) {
        if (unlikely(__Pyx_PyTuple_SET_ITEM(res, i, src[i]) < 0)) {
            Py_DECREF(res);
            return NULL;
        }
        Py_INCREF(src[i]);
    }
    return res;
}
#elif CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE void __Pyx_copy_object_array(PyObject *const *CYTHON_RESTRICT src, PyObject** CYTHON_RESTRICT dest, Py_ssize_t length) {
    PyObject *v;
    Py_ssize_t i;
    for (i = 0; i < length; i++) {
        v = dest[i] = src[i];
        Py_INCREF(v);
    }
}
static CYTHON_INLINE PyObject *
__Pyx_PyTuple_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return __Pyx_NewRef(__pyx_mstate_global->__pyx_empty_tuple);
    }
    res = PyTuple_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyTupleObject*)res)->ob_item, n);
    return res;
}
static CYTHON_INLINE PyObject *
__Pyx_PyList_FromArray(PyObject *const *src, Py_ssize_t n)
{
    PyObject *res;
    if (n <= 0) {
        return PyList_New(0);
    }
    res = PyList_New(n);
    if (unlikely(res == NULL)) return NULL;
    __Pyx_copy_object_array(src, ((PyListObject*)res)->ob_item, n);
    return res;
}
#endif

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL ||\
        !(CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS)
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS && (PY_VERSION_HEX < 0x030B0000)
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_COMPILING_IN_GRAAL
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length, length2;
        int kind;
        void *data1, *data2;
        #if !CYTHON_COMPILING_IN_LIMITED_API
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        #endif
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length < 0)) return -1;
        #endif
        length2 = __Pyx_PyUnicode_GET_LENGTH(s2);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(length2 < 0)) return -1;
        #endif
        if (length != length2) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    return (equals == Py_EQ);
return_ne:
    return (equals == Py_NE);
#endif
}

/* fastcall */
#if CYTHON_METH_FASTCALL
static CYTHON_INLINE PyObject * __Pyx_GetKwValue_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues, PyObject *s)
{
    Py_ssize_t i, n = __Pyx_PyTuple_GET_SIZE(kwnames);
    #if !CYTHON_ASSUME_SAFE_SIZE
    if (unlikely(n == -1)) return NULL;
    #endif
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        if (s == namei) return kwvalues[i];
    }
    for (i = 0; i < n; i++)
    {
        PyObject *namei = __Pyx_PyTuple_GET_ITEM(kwnames, i);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!namei)) return NULL;
        #endif
        int eq = __Pyx_PyUnicode_Equals(s, namei, Py_EQ);
        if (unlikely(eq != 0)) {
            if (unlikely(eq < 0)) return NULL;
            return kwvalues[i];
        }
    }
    return NULL;
}
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030d0000 || CYTHON_COMPILING_IN_LIMITED_API
CYTHON_UNUSED static PyObject *__Pyx_KwargsAsDict_FASTCALL(PyObject *kwnames, PyObject *const *kwvalues) {
    Py_ssize_t i, nkwargs;
    PyObject *dict;
#if !CYTHON_ASSUME_SAFE_SIZE
    nkwargs = PyTuple_Size(kwnames);
    if (unlikely(nkwargs < 0)) return NULL;
#else
    nkwargs = PyTuple_GET_SIZE(kwnames);
#endif
    dict = PyDict_New();
    if (unlikely(!dict))
        return NULL;
    for (i=0; i<nkwargs; i++) {
#if !CYTHON_ASSUME_SAFE_MACROS
        PyObject *key = PyTuple_GetItem(kwnames, i);
        if (!key) goto bad;
#else
        PyObject *key = PyTuple_GET_ITEM(kwnames, i);
#endif
        if (unlikely(PyDict_SetItem(dict, key, kwvalues[i]) < 0))
            goto bad;
    }
    return dict;
bad:
    Py_DECREF(dict);
    return NULL;
}
#endif
#endif

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
}

/* CallUnboundCMethod2 */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject *__Pyx_CallUnboundCMethod2(__Pyx_CachedCFunction *cfunc, PyObject *self, PyObject *arg1, PyObject *arg2) {
    int was_initialized = __Pyx_CachedCFunction_GetAndSetInitializing(cfunc);
    if (likely(was_initialized == 2 && cfunc->func)) {
        PyObject *args[2] = {arg1, arg2};
        if (cfunc->flag == METH_FASTCALL) {
            return __Pyx_CallCFunctionFast(cfunc, self, args, 2);
        }
        if (cfunc->flag == (METH_FASTCALL | METH_KEYWORDS))
            return __Pyx_CallCFunctionFastWithKeywords(cfunc, self, args, 2, NULL);
    }
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    else if (unlikely(was_initialized == 1)) {
        __Pyx_CachedCFunction tmp_cfunc = {
#ifndef __cplusplus
            0
#endif
        };
        tmp_cfunc.type = cfunc->type;
        tmp_cfunc.method_name = cfunc->method_name;
        return __Pyx__CallUnboundCMethod2(&tmp_cfunc, self, arg1, arg2);
    }
#endif
    PyObject *result = __Pyx__CallUnboundCMethod2(cfunc, self, arg1, arg2);
    __Pyx_CachedCFunction_SetFinishedInitializing(cfunc);
    return result;
}
#endif
static PyObject* __Pyx__CallUnboundCMethod2(__Pyx_CachedCFunction* cfunc, PyObject* self, PyObject* arg1, PyObject* arg2){
    if (unlikely(!cfunc->func && !cfunc->method) && unlikely(__Pyx_TryUnpackUnboundCMethod(cfunc) < 0)) return NULL;
#if CYTHON_COMPILING_IN_CPYTHON
    if (cfunc->func && (cfunc->flag & METH_VARARGS)) {
        PyObject *result = NULL;
        PyObject *args = PyTuple_New(2);
        if (unlikely(!args)) return NULL;
        Py_INCREF(arg1);
        PyTuple_SET_ITEM(args, 0, arg1);
        Py_INCREF(arg2);
        PyTuple_SET_ITEM(args, 1, arg2);
        if (cfunc->flag & METH_KEYWORDS)
            result = __Pyx_CallCFunctionWithKeywords(cfunc, self, args, NULL);
        else
            result = __Pyx_CallCFunction(cfunc, self, args);
        Py_DECREF(args);
        return result;
    }
#endif
    {
        PyObject *args[4] = {NULL, self, arg1, arg2};
        return __Pyx_PyObject_FastCall(cfunc->method, args+1, 3 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
    }
}

/* ParseKeywords */
static int __Pyx_ValidateDuplicatePosArgs(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char* function_name)
{
    PyObject ** const *name = argnames;
    while (name != first_kw_arg) {
        PyObject *key = **name;
        int found = PyDict_Contains(kwds, key);
        if (unlikely(found)) {
            if (found == 1) __Pyx_RaiseDoubleKeywordsError(function_name, key);
            goto bad;
        }
        name++;
    }
    return 0;
bad:
    return -1;
}
#if CYTHON_USE_UNICODE_INTERNALS
static CYTHON_INLINE int __Pyx_UnicodeKeywordsEqual(PyObject *s1, PyObject *s2) {
    int kind;
    Py_ssize_t len = PyUnicode_GET_LENGTH(s1);
    if (len != PyUnicode_GET_LENGTH(s2)) return 0;
    kind = PyUnicode_KIND(s1);
    if (kind != PyUnicode_KIND(s2)) return 0;
    const void *data1 = PyUnicode_DATA(s1);
    const void *data2 = PyUnicode_DATA(s2);
    return (memcmp(data1, data2, (size_t) len * (size_t) kind) == 0);
}
#endif
static int __Pyx_MatchKeywordArg_str(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    #if CYTHON_USE_UNICODE_INTERNALS
    Py_hash_t key_hash = ((PyASCIIObject*)key)->hash;
    if (unlikely(key_hash == -1)) {
        key_hash = PyObject_Hash(key);
        if (unlikely(key_hash == -1))
            goto bad;
    }
    #endif
    name = first_kw_arg;
    while (*name) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (key_hash == ((PyASCIIObject*)name_str)->hash && __Pyx_UnicodeKeywordsEqual(name_str, key)) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) {
                *index_found = (size_t) (name - argnames);
                return 1;
            }
        }
        #endif
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        PyObject *name_str = **name;
        #if CYTHON_USE_UNICODE_INTERNALS
        if (unlikely(key_hash == ((PyASCIIObject*)name_str)->hash)) {
            if (__Pyx_UnicodeKeywordsEqual(name_str, key))
                goto arg_passed_twice;
        }
        #else
        #if CYTHON_ASSUME_SAFE_SIZE
        if (PyUnicode_GET_LENGTH(name_str) == PyUnicode_GET_LENGTH(key))
        #endif
        {
            if (unlikely(name_str == key)) goto arg_passed_twice;
            int cmp = PyUnicode_Compare(name_str, key);
            if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
            if (cmp == 0) goto arg_passed_twice;
        }
        #endif
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
bad:
    return -1;
}
static int __Pyx_MatchKeywordArg_nostr(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    PyObject ** const *name;
    if (unlikely(!PyUnicode_Check(key))) goto invalid_keyword_type;
    name = first_kw_arg;
    while (*name) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (cmp == 1) {
            *index_found = (size_t) (name - argnames);
            return 1;
        }
        if (unlikely(cmp == -1)) goto bad;
        name++;
    }
    name = argnames;
    while (name != first_kw_arg) {
        int cmp = PyObject_RichCompareBool(**name, key, Py_EQ);
        if (unlikely(cmp != 0)) {
            if (cmp == 1) goto arg_passed_twice;
            else goto bad;
        }
        name++;
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
bad:
    return -1;
}
static CYTHON_INLINE int __Pyx_MatchKeywordArg(
    PyObject *key,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    size_t *index_found,
    const char *function_name)
{
    return likely(PyUnicode_CheckExact(key)) ?
        __Pyx_MatchKeywordArg_str(key, argnames, first_kw_arg, index_found, function_name) :
        __Pyx_MatchKeywordArg_nostr(key, argnames, first_kw_arg, index_found, function_name);
}
static void __Pyx_RejectUnknownKeyword(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject ** const *first_kw_arg,
    const char *function_name)
{
    Py_ssize_t pos = 0;
    PyObject *key = NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(kwds);
    while (PyDict_Next(kwds, &pos, &key, NULL)) {
        PyObject** const *name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (!*name) {
            #if CYTHON_AVOID_BORROWED_REFS
            Py_INCREF(key);
            #endif
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp != 1) {
                if (cmp == 0) {
                    PyErr_Format(PyExc_TypeError,
                        "%s() got an unexpected keyword argument '%U'",
                        function_name, key);
                }
                #if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(key);
                #endif
                break;
            }
            #if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(key);
            #endif
        }
    }
    __Pyx_END_CRITICAL_SECTION();
    assert(PyErr_Occurred());
}
static int __Pyx_ParseKeywordDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t extracted = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    name = first_kw_arg;
    while (*name && num_kwargs > extracted) {
        PyObject * key = **name;
        PyObject *value;
        int found = 0;
        #if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        found = PyDict_GetItemRef(kwds, key, &value);
        #else
        value = PyDict_GetItemWithError(kwds, key);
        if (value) {
            Py_INCREF(value);
            found = 1;
        } else {
            if (unlikely(PyErr_Occurred())) goto bad;
        }
        #endif
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            extracted++;
        }
        name++;
    }
    if (num_kwargs > extracted) {
        if (ignore_unknown_kwargs) {
            if (unlikely(__Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name) == -1))
                goto bad;
        } else {
            __Pyx_RejectUnknownKeyword(kwds, argnames, first_kw_arg, function_name);
            goto bad;
        }
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordDictToDict(
    PyObject *kwds,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject** const *name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    Py_ssize_t len;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
    if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return -1;
#endif
    if (PyDict_Update(kwds2, kwds) < 0) goto bad;
    name = first_kw_arg;
    while (*name) {
        PyObject *key = **name;
        PyObject *value;
#if !CYTHON_COMPILING_IN_LIMITED_API && (PY_VERSION_HEX >= 0x030d00A2 || defined(PyDict_Pop))
        int found = PyDict_Pop(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
        }
#elif __PYX_LIMITED_VERSION_HEX >= 0x030d0000
        int found = PyDict_GetItemRef(kwds2, key, &value);
        if (found) {
            if (unlikely(found < 0)) goto bad;
            values[name-argnames] = value;
            if (unlikely(PyDict_DelItem(kwds2, key) < 0)) goto bad;
        }
#else
    #if CYTHON_COMPILING_IN_CPYTHON
        value = _PyDict_Pop(kwds2, key, kwds2);
    #else
        value = __Pyx_CallUnboundCMethod2(&__pyx_mstate_global->__pyx_umethod_PyDict_Type_pop, kwds2, key, kwds2);
    #endif
        if (value == kwds2) {
            Py_DECREF(value);
        } else {
            if (unlikely(!value)) goto bad;
            values[name-argnames] = value;
        }
#endif
        name++;
    }
    len = PyDict_Size(kwds2);
    if (len > 0) {
        return __Pyx_ValidateDuplicatePosArgs(kwds, argnames, first_kw_arg, function_name);
    } else if (unlikely(len == -1)) {
        goto bad;
    }
    return 0;
bad:
    return -1;
}
static int __Pyx_ParseKeywordsTuple(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    PyObject *key = NULL;
    PyObject** const * name;
    PyObject** const *first_kw_arg = argnames + num_pos_args;
    for (Py_ssize_t pos = 0; pos < num_kwargs; pos++) {
#if CYTHON_AVOID_BORROWED_REFS
        key = __Pyx_PySequence_ITEM(kwds, pos);
#else
        key = __Pyx_PyTuple_GET_ITEM(kwds, pos);
#endif
#if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(!key)) goto bad;
#endif
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            PyObject *value = kwvalues[pos];
            values[name-argnames] = __Pyx_NewRef(value);
        } else {
            size_t index_found = 0;
            int cmp = __Pyx_MatchKeywordArg(key, argnames, first_kw_arg, &index_found, function_name);
            if (cmp == 1) {
                PyObject *value = kwvalues[pos];
                values[index_found] = __Pyx_NewRef(value);
            } else {
                if (unlikely(cmp == -1)) goto bad;
                if (kwds2) {
                    PyObject *value = kwvalues[pos];
                    if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
                } else if (!ignore_unknown_kwargs) {
                    goto invalid_keyword;
                }
            }
        }
        #if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(key);
        key = NULL;
        #endif
    }
    return 0;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    goto bad;
bad:
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(key);
    #endif
    return -1;
}
static int __Pyx_ParseKeywords(
    PyObject *kwds,
    PyObject * const *kwvalues,
    PyObject ** const argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    Py_ssize_t num_kwargs,
    const char* function_name,
    int ignore_unknown_kwargs)
{
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds)))
        return __Pyx_ParseKeywordsTuple(kwds, kwvalues, argnames, kwds2, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
    else if (kwds2)
        return __Pyx_ParseKeywordDictToDict(kwds, argnames, kwds2, values, num_pos_args, function_name);
    else
        return __Pyx_ParseKeywordDict(kwds, argnames, values, num_pos_args, num_kwargs, function_name, ignore_unknown_kwargs);
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    __Pyx_TypeName type_name;
    __Pyx_TypeName obj_type_name;
    PyObject *extra_info = __pyx_mstate_global->__pyx_empty_unicode;
    int from_annotation_subclass = 0;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (!exact) {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    } else if (exact == 2) {
        if (__Pyx_TypeCheck(obj, type)) {
            from_annotation_subclass = 1;
            extra_info = __pyx_mstate_global->__pyx_kp_u_Note_that_Cython_is_deliberately;
        }
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected " __Pyx_FMT_TYPENAME
        ", got " __Pyx_FMT_TYPENAME ")"
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        "%s%U"
#endif
        , name, type_name, obj_type_name
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
        , (from_annotation_subclass ? ". " : ""), extra_info
#endif
        );
#if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    if (exact == 2 && from_annotation_subclass) {
        PyObject *res;
        PyObject *vargs[2];
        vargs[0] = PyErr_GetRaisedException();
        vargs[1] = extra_info;
        res = PyObject_VectorcallMethod(__pyx_mstate_global->__pyx_kp_u_add_note, vargs, 2, NULL);
        Py_XDECREF(res);
        PyErr_SetRaisedException(vargs[0]);
    }
#endif
    __Pyx_DECREF_TypeName(type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RaiseUnexpectedTypeError */
static int
__Pyx_RaiseUnexpectedTypeError(const char *expected, PyObject *obj)
{
    __Pyx_TypeName obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError, "Expected %s, got " __Pyx_FMT_TYPENAME,
                 expected, obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return 0;
}

/* RejectKeywords */
static void __Pyx_RejectKeywords(const char* function_name, PyObject *kwds) {
    PyObject *key = NULL;
    if (CYTHON_METH_FASTCALL && likely(PyTuple_Check(kwds))) {
        key = __Pyx_PySequence_ITEM(kwds, 0);
    } else {
        Py_ssize_t pos = 0;
#if !CYTHON_COMPILING_IN_PYPY || defined(PyArg_ValidateKeywordArguments)
        if (unlikely(!PyArg_ValidateKeywordArguments(kwds))) return;
#endif
        PyDict_Next(kwds, &pos, &key, NULL);
        Py_INCREF(key);
    }
    if (likely(key)) {
        PyErr_Format(PyExc_TypeError,
            "%s() got an unexpected keyword argument '%U'",
            function_name, key);
        Py_DECREF(key);
    }
}

/* GetAttr3 */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
#endif
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r;
#if __PYX_LIMITED_VERSION_HEX >= 0x030d0000
    int res = PyObject_GetOptionalAttr(o, n, &r);
    return (res != 0) ? r : __Pyx_NewRef(d);
#else
  #if CYTHON_USE_TYPE_SLOTS
    if (likely(PyUnicode_Check(n))) {
        r = __Pyx_PyObject_GetAttrStrNoError(o, n);
        if (unlikely(!r) && likely(!PyErr_Occurred())) {
            r = __Pyx_NewRef(d);
        }
        return r;
    }
  #endif
    r = PyObject_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
#endif
}

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(!__pyx_m)) {
        if (!PyErr_Occurred())
            PyErr_SetNone(PyExc_NameError);
        return NULL;
    }
    result = PyObject_GetAttr(__pyx_m, name);
    if (likely(result)) {
        return result;
    }
    PyErr_Clear();
#elif CYTHON_AVOID_BORROWED_REFS || CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    if (unlikely(__Pyx_PyDict_GetItemRef(__pyx_mstate_global->__pyx_d, name, &result) == -1)) PyErr_Clear();
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return result;
    }
#else
    result = _PyDict_GetItem_KnownHash(__pyx_mstate_global->__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_mstate_global->__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyObjectFastCallMethod */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_PyObject_FastCallMethod(PyObject *name, PyObject *const *args, size_t nargsf) {
    PyObject *result;
    PyObject *attr = PyObject_GetAttr(args[0], name);
    if (unlikely(!attr))
        return NULL;
    result = __Pyx_PyObject_FastCall(attr, args+1, nargsf - 1);
    Py_DECREF(attr);
    return result;
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
    PyObject* exc_type;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    exc_type = __Pyx_PyErr_CurrentExceptionType();
    if (unlikely(exc_type)) {
        if (unlikely(!__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)))
            return -1;
        __Pyx_PyErr_Clear();
        return 0;
    }
    return 0;
}

/* PyObjectCallNoArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
    PyObject *arg[2] = {NULL, NULL};
    return __Pyx_PyObject_FastCall(func, arg + 1, 0 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectCallOneArg */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *args[2] = {NULL, arg};
    return __Pyx_PyObject_FastCall(func, args+1, 1 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET);
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    __Pyx_TypeName type_name;
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if defined(Py_TPFLAGS_METHOD_DESCRIPTOR) && Py_TPFLAGS_METHOD_DESCRIPTOR
        if (__Pyx_PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_METHOD_DESCRIPTOR))
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || __Pyx_IS_TYPE(descr, &PyMethodDescr_Type)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (likely(descr != NULL)) {
        *method = descr;
        return 0;
    }
    type_name = __Pyx_PyType_GetFullyQualifiedName(tp);
    PyErr_Format(PyExc_AttributeError,
                 "'" __Pyx_FMT_TYPENAME "' object has no attribute '%U'",
                 type_name, name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
#if CYTHON_VECTORCALL && (__PYX_LIMITED_VERSION_HEX >= 0x030C0000 || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x03090000))
    PyObject *args[1] = {obj};
    (void) __Pyx_PyObject_GetMethod;
    (void) __Pyx_PyObject_CallOneArg;
    (void) __Pyx_PyObject_CallNoArg;
    return PyObject_VectorcallMethod(method_name, args, 1 | PY_VECTORCALL_ARGUMENTS_OFFSET, NULL);
#else
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
#endif
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    }
    return __Pyx_IterFinish();
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else {
      Py_ssize_t size = __Pyx_PyTuple_GET_SIZE(t);
 #if !CYTHON_ASSUME_SAFE_SIZE
      if (unlikely(size < 0)) return;
 #endif
      if (size < index) {
        __Pyx_RaiseNeedMoreValuesError(size);
      } else {
        __Pyx_RaiseTooManyValuesError(index);
      }
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2(
        PyObject* tuple, PyObject** value1, PyObject** value2, int is_tuple, int has_known_size, int decref_tuple) {
    if (likely(is_tuple || PyTuple_Check(tuple))) {
        Py_ssize_t size;
        if (has_known_size) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        size = __Pyx_PyTuple_GET_SIZE(tuple);
        if (likely(size == 2)) {
            return __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple);
        }
        if (size >= 0) {
            __Pyx_UnpackTupleError(tuple, 2);
        }
        return -1;
    } else {
        return __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple);
    }
}
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
    value1 = __Pyx_PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = __Pyx_PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = __Pyx_PyObject_GetIterNextFunc(iter);
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
#if CYTHON_COMPILING_IN_PYPY
#include <string.h>
#endif
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#else
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
#if !CYTHON_COMPILING_IN_PYPY
static CYTHON_INLINE int __Pyx_dict_iter_next_source_is_dict(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem) {
    PyObject *key, *value;
    if (unlikely(orig_length != PyDict_Size(iter_obj))) {
        PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
        return -1;
    }
    if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
        return 0;
    }
    if (pitem) {
        PyObject* tuple = PyTuple_New(2);
        if (unlikely(!tuple)) {
            return -1;
        }
        Py_INCREF(key);
        Py_INCREF(value);
        #if CYTHON_ASSUME_SAFE_MACROS
        PyTuple_SET_ITEM(tuple, 0, key);
        PyTuple_SET_ITEM(tuple, 1, value);
        #else
        if (unlikely(PyTuple_SetItem(tuple, 0, key) < 0)) {
            Py_DECREF(value);
            Py_DECREF(tuple);
            return -1;
        }
        if (unlikely(PyTuple_SetItem(tuple, 1, value) < 0)) {
            Py_DECREF(tuple);
            return -1;
        }
        #endif
        *pitem = tuple;
    } else {
        if (pkey) {
            Py_INCREF(key);
            *pkey = key;
        }
        if (pvalue) {
            Py_INCREF(value);
            *pvalue = value;
        }
    }
    return 1;
}
#endif
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        int result;
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_BEGIN_CRITICAL_SECTION(iter_obj);
#endif
        result = __Pyx_dict_iter_next_source_is_dict(iter_obj, orig_length, ppos, pkey, pvalue, pitem);
#if PY_VERSION_HEX >= 0x030d0000 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_END_CRITICAL_SECTION();
#endif
        return result;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t tuple_size = __Pyx_PyTuple_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(tuple_size < 0)) return -1;
        #endif
        if (unlikely(pos >= tuple_size)) return 0;
        *ppos = pos + 1;
        #if CYTHON_ASSUME_SAFE_MACROS
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        #else
        next_item = PyTuple_GetItem(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        #endif
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        Py_ssize_t list_size = __Pyx_PyList_GET_SIZE(iter_obj);
        #if !CYTHON_ASSUME_SAFE_SIZE
        if (unlikely(list_size < 0)) return -1;
        #endif
        if (unlikely(pos >= list_size)) return 0;
        *ppos = pos + 1;
        #if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
        next_item = PyList_GetItemRef(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        #elif CYTHON_ASSUME_SAFE_MACROS
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
        #else
        next_item = PyList_GetItem(iter_obj, pos);
        if (unlikely(!next_item)) return -1;
        Py_INCREF(next_item);
        #endif
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* DictGetItem */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    if (unlikely(__Pyx_PyDict_GetItemRef(d, key, &value) == 0)) { // no value, no error
        if (unlikely(PyTuple_Check(key))) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args)) {
                PyErr_SetObject(PyExc_KeyError, args);
                Py_DECREF(args);
            }
        } else {
            PyErr_SetObject(PyExc_KeyError, key);
        }
    }
    return value;
}
#endif

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    __Pyx_TypeName obj_type_name;
    __Pyx_TypeName type_name;
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    type_name = __Pyx_PyType_GetFullyQualifiedName(type);
    PyErr_Format(PyExc_TypeError,
                 "Cannot convert " __Pyx_FMT_TYPENAME " to " __Pyx_FMT_TYPENAME,
                 obj_type_name, type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    __Pyx_DECREF_TypeName(type_name);
    return 0;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (unlikely(!j)) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS && !CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            return __Pyx_PyList_GetItemRef(o, n);
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_subscript) {
            PyObject *r, *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return NULL;
            r = mm->mp_subscript(o, key);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return sm->sq_item(o, i);
        }
    }
#else
    if (is_list || !PyMapping_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyLong_FromSsize_t(i));
}

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a, x;
#ifdef HAVE_LONG_LONG
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla, llx;
#endif
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return __Pyx_NewRef(op2);
    }
    if (likely(__Pyx_PyLong_IsCompact(op1))) {
        a = __Pyx_PyLong_CompactValue(op1);
    } else {
        const digit* digits = __Pyx_PyLong_Digits(op1);
        const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
        switch (size) {
            case -2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case -3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case -4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
        }
    }
            x = a + b;
        return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
    long_long:
            llx = lla + llb;
        return PyLong_FromLongLong(llx);
#endif
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
    
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_AddObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) + (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_AddObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_AddObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_AddObjC(op1, op2, inplace);
}
#endif

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject *index) {
    PyObject *runerr = NULL;
    Py_ssize_t key_value;
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        __Pyx_TypeName index_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(index));
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError,
            "cannot fit '" __Pyx_FMT_TYPENAME "' into an index-sized integer", index_type_name);
        __Pyx_DECREF_TypeName(index_type_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem_Slow(PyObject *obj, PyObject *key) {
    __Pyx_TypeName obj_type_name;
    if (likely(PyType_Check(obj))) {
        PyObject *meth = __Pyx_PyObject_GetAttrStrNoError(obj, __pyx_mstate_global->__pyx_n_u_class_getitem);
        if (!meth) {
            PyErr_Clear();
        } else {
            PyObject *result = __Pyx_PyObject_CallOneArg(meth, key);
            Py_DECREF(meth);
            return result;
        }
    }
    obj_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(obj));
    PyErr_Format(PyExc_TypeError,
        "'" __Pyx_FMT_TYPENAME "' object is not subscriptable", obj_type_name);
    __Pyx_DECREF_TypeName(obj_type_name);
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject *key) {
    PyTypeObject *tp = Py_TYPE(obj);
    PyMappingMethods *mm = tp->tp_as_mapping;
    PySequenceMethods *sm = tp->tp_as_sequence;
    if (likely(mm && mm->mp_subscript)) {
        return mm->mp_subscript(obj, key);
    }
    if (likely(sm && sm->sq_item)) {
        return __Pyx_PyObject_GetIndex(obj, key);
    }
    return __Pyx_PyObject_GetItem_Slow(obj, key);
}
#endif

/* RaiseException */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if PY_VERSION_HEX >= 0x030C00A6
        PyException_SetTraceback(value, tb);
#elif CYTHON_FAST_THREAD_STATE
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#else
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}

/* SetItemInt */
static int __Pyx_SetItemInt_Generic(PyObject *o, PyObject *j, PyObject *v) {
    int r;
    if (unlikely(!j)) return -1;
    r = PyObject_SetItem(o, j, v);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE int __Pyx_SetItemInt_Fast(PyObject *o, Py_ssize_t i, PyObject *v, int is_list,
                                               CYTHON_NCP_UNUSED int wraparound, CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = (!wraparound) ? i : ((likely(i >= 0)) ? i : i + PyList_GET_SIZE(o));
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o)))) {
            Py_INCREF(v);
#if CYTHON_AVOID_THREAD_UNSAFE_BORROWED_REFS
            PyList_SetItem(o, n, v);
#else
            PyObject* old = PyList_GET_ITEM(o, n);
            PyList_SET_ITEM(o, n, v);
            Py_DECREF(old);
#endif
            return 1;
        }
    } else {
        PyMappingMethods *mm = Py_TYPE(o)->tp_as_mapping;
        PySequenceMethods *sm = Py_TYPE(o)->tp_as_sequence;
        if (mm && mm->mp_ass_subscript) {
            int r;
            PyObject *key = PyLong_FromSsize_t(i);
            if (unlikely(!key)) return -1;
            r = mm->mp_ass_subscript(o, key, v);
            Py_DECREF(key);
            return r;
        }
        if (likely(sm && sm->sq_ass_item)) {
            if (wraparound && unlikely(i < 0) && likely(sm->sq_length)) {
                Py_ssize_t l = sm->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return -1;
                    PyErr_Clear();
                }
            }
            return sm->sq_ass_item(o, i, v);
        }
    }
#else
    if (is_list || !PyMapping_Check(o))
    {
        return PySequence_SetItem(o, i, v);
    }
#endif
    return __Pyx_SetItemInt_Generic(o, PyLong_FromSsize_t(i), v);
}

/* PyLongBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_Fallback___Pyx_PyLong_SubtractObjC(PyObject *op1, PyObject *op2, int inplace) {
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#if CYTHON_USE_PYLONG_INTERNALS
static PyObject* __Pyx_Unpacked___Pyx_PyLong_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(inplace);
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    long a, x;
#ifdef HAVE_LONG_LONG
    const PY_LONG_LONG llb = intval;
    PY_LONG_LONG lla, llx;
#endif
    if (unlikely(__Pyx_PyLong_IsZero(op1))) {
        return PyLong_FromLong(-intval);
    }
    if (likely(__Pyx_PyLong_IsCompact(op1))) {
        a = __Pyx_PyLong_CompactValue(op1);
    } else {
        const digit* digits = __Pyx_PyLong_Digits(op1);
        const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(op1);
        switch (size) {
            case -2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 2:
                if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                    a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case -3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 3:
                if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                    a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case -4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            case 4:
                if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                    a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                    break;
                #ifdef HAVE_LONG_LONG
                } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                    lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                    goto long_long;
                #endif
                }
                CYTHON_FALLTHROUGH;
            default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
        }
    }
            x = a - b;
        return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
    long_long:
            llx = lla - llb;
        return PyLong_FromLongLong(llx);
#endif
    return __Pyx_Fallback___Pyx_PyLong_SubtractObjC(op1, op2, inplace);
    
    
}
#endif
static PyObject* __Pyx_Float___Pyx_PyLong_SubtractObjC(PyObject *float_val, long intval, int zerodivision_check) {
    CYTHON_UNUSED_VAR(zerodivision_check);
    const long b = intval;
    double a = __Pyx_PyFloat_AS_DOUBLE(float_val);
        double result;
        
        result = ((double)a) - (double)b;
        return PyFloat_FromDouble(result);
}
static CYTHON_INLINE PyObject* __Pyx_PyLong_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(zerodivision_check);
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        return __Pyx_Unpacked___Pyx_PyLong_SubtractObjC(op1, op2, intval, inplace, zerodivision_check);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        return __Pyx_Float___Pyx_PyLong_SubtractObjC(op1, intval, zerodivision_check);
    }
    return __Pyx_Fallback___Pyx_PyLong_SubtractObjC(op1, op2, inplace);
}
#endif

/* PyLongCompare */
static CYTHON_INLINE int __Pyx_PyLong_BoolEqObjC(PyObject *op1, PyObject *op2, long intval, long inplace) {
    CYTHON_MAYBE_UNUSED_VAR(intval);
    CYTHON_UNUSED_VAR(inplace);
    if (op1 == op2) {
        return 1;
    }
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        int unequal;
        unsigned long uintval;
        Py_ssize_t size = __Pyx_PyLong_DigitCount(op1);
        const digit* digits = __Pyx_PyLong_Digits(op1);
        if (intval == 0) {
            return (__Pyx_PyLong_IsZero(op1) == 1);
        } else if (intval < 0) {
            if (__Pyx_PyLong_IsNonNeg(op1))
                return 0;
            intval = -intval;
        } else {
            if (__Pyx_PyLong_IsNeg(op1))
                return 0;
        }
        uintval = (unsigned long) intval;
#if PyLong_SHIFT * 4 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 4)) {
            unequal = (size != 5) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[4] != ((uintval >> (4 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 3 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 3)) {
            unequal = (size != 4) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[3] != ((uintval >> (3 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 2 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 2)) {
            unequal = (size != 3) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK)) | (digits[2] != ((uintval >> (2 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
#if PyLong_SHIFT * 1 < SIZEOF_LONG*8
        if (uintval >> (PyLong_SHIFT * 1)) {
            unequal = (size != 2) || (digits[0] != (uintval & (unsigned long) PyLong_MASK))
                 | (digits[1] != ((uintval >> (1 * PyLong_SHIFT)) & (unsigned long) PyLong_MASK));
        } else
#endif
            unequal = (size != 1) || (((unsigned long) digits[0]) != (uintval & (unsigned long) PyLong_MASK));
        return (unequal == 0);
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = __Pyx_PyFloat_AS_DOUBLE(op1);
        return ((double)a == (double)b);
    }
    return __Pyx_PyObject_IsTrueAndDecref(
        PyObject_RichCompare(op1, op2, Py_EQ));
}

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
    else state = (PyGILState_STATE)0;
    CYTHON_UNUSED_VAR(clineno);
    CYTHON_UNUSED_VAR(lineno);
    CYTHON_UNUSED_VAR(filename);
    CYTHON_MAYBE_UNUSED_VAR(nogil);
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(0);
    }
    ctx = PyUnicode_FromString(name);
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
    if (nogil)
        PyGILState_Release(state);
}

/* PyObjectVectorCallKwBuilder */
#if CYTHON_VECTORCALL
static int __Pyx_VectorcallBuilder_AddArg(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_PyObject_FastCallDict;
    if (__Pyx_PyTuple_SET_ITEM(builder, n, key) != (0)) return -1;
    Py_INCREF(key);
    args[n] = value;
    return 0;
}
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    (void)__Pyx_VectorcallBuilder_AddArgStr;
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return __Pyx_VectorcallBuilder_AddArg(key, value, builder, args, n);
}
static int __Pyx_VectorcallBuilder_AddArgStr(const char *key, PyObject *value, PyObject *builder, PyObject **args, int n) {
    PyObject *pyKey = PyUnicode_FromString(key);
    if (!pyKey) return -1;
    return __Pyx_VectorcallBuilder_AddArg(pyKey, value, builder, args, n);
}
#else // CYTHON_VECTORCALL
CYTHON_UNUSED static int __Pyx_VectorcallBuilder_AddArg_Check(PyObject *key, PyObject *value, PyObject *builder, CYTHON_UNUSED PyObject **args, CYTHON_UNUSED int n) {
    if (unlikely(!PyUnicode_Check(key))) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        return -1;
    }
    return PyDict_SetItem(builder, key, value);
}
#endif

/* PyObjectVectorCallMethodKwBuilder */
#if !CYTHON_VECTORCALL || PY_VERSION_HEX < 0x03090000
static PyObject *__Pyx_Object_VectorcallMethod_CallFromBuilder(PyObject *name, PyObject *const *args, size_t nargsf, PyObject *kwnames) {
    PyObject *result;
    PyObject *obj = PyObject_GetAttr(args[0], name);
    if (unlikely(!obj))
        return NULL;
    result = __Pyx_Object_Vectorcall_CallFromBuilder(obj, args+1, nargsf-1, kwnames);
    Py_DECREF(obj);
    return result;
}
#endif

/* PyObjectSetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE int __Pyx_PyObject_SetAttrStr(PyObject* obj, PyObject* attr_name, PyObject* value) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_setattro))
        return tp->tp_setattro(obj, attr_name, value);
    return PyObject_SetAttr(obj, attr_name, value);
}
#endif

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *module = 0;
    PyObject *empty_dict = 0;
    PyObject *empty_list = 0;
    empty_dict = PyDict_New();
    if (unlikely(!empty_dict))
        goto bad;
    if (level == -1) {
        const char* package_sep = strchr(__Pyx_MODULE_NAME, '.');
        if (package_sep != (0)) {
            module = PyImport_ImportModuleLevelObject(
                name, __pyx_mstate_global->__pyx_d, empty_dict, from_list, 1);
            if (unlikely(!module)) {
                if (unlikely(!PyErr_ExceptionMatches(PyExc_ImportError)))
                    goto bad;
                PyErr_Clear();
            }
        }
        level = 0;
    }
    if (!module) {
        module = PyImport_ImportModuleLevelObject(
            name, __pyx_mstate_global->__pyx_d, empty_dict, from_list, level);
    }
bad:
    Py_XDECREF(empty_dict);
    Py_XDECREF(empty_list);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        const char* module_name_str = 0;
        PyObject* module_name = 0;
        PyObject* module_dot = 0;
        PyObject* full_name = 0;
        PyErr_Clear();
        module_name_str = PyModule_GetName(module);
        if (unlikely(!module_name_str)) { goto modbad; }
        module_name = PyUnicode_FromString(module_name_str);
        if (unlikely(!module_name)) { goto modbad; }
        module_dot = PyUnicode_Concat(module_name, __pyx_mstate_global->__pyx_kp_u_);
        if (unlikely(!module_dot)) { goto modbad; }
        full_name = PyUnicode_Concat(module_dot, name);
        if (unlikely(!full_name)) { goto modbad; }
        #if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
                CYTHON_COMPILING_IN_GRAAL
        {
            PyObject *modules = PyImport_GetModuleDict();
            if (unlikely(!modules))
                goto modbad;
            value = PyObject_GetItem(modules, full_name);
        }
        #else
        value = PyImport_GetModule(full_name);
        #endif
      modbad:
        Py_XDECREF(full_name);
        Py_XDECREF(module_dot);
        Py_XDECREF(module_name);
    }
    if (unlikely(!value)) {
        PyErr_Format(PyExc_ImportError, "cannot import name %S", name);
    }
    return value;
}

/* HasAttr */
#if __PYX_LIMITED_VERSION_HEX < 0x030d0000
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!PyUnicode_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_PyObject_GetAttrStrNoError(o, n);
    if (!r) {
        return (unlikely(PyErr_Occurred())) ? -1 : 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}
#endif

/* CallNextTpTraverse */
static int __Pyx_call_next_tp_traverse(PyObject* obj, visitproc v, void *a, traverseproc current_tp_traverse) {
    PyTypeObject* type = Py_TYPE(obj);
    traverseproc tp_traverse = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc) != current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_traverse = __Pyx_PyType_GetSlot(type, tp_traverse, traverseproc)) == current_tp_traverse)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_traverse)
        return tp_traverse(obj, v, a);
    return 0;
}

/* CallTypeTraverse */
#if !CYTHON_USE_TYPE_SPECS || (!CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x03090000)
#else
static int __Pyx_call_type_traverse(PyObject *o, int always_call, visitproc visit, void *arg) {
    #if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x03090000
    if (__Pyx_get_runtime_version() < 0x03090000) return 0;
    #endif
    if (!always_call) {
        PyTypeObject *base = __Pyx_PyObject_GetSlot(o, tp_base, PyTypeObject*);
        unsigned long flags = PyType_GetFlags(base);
        if (flags & Py_TPFLAGS_HEAPTYPE) {
            return 0;
        }
    }
    Py_VISIT((PyObject*)Py_TYPE(o));
    return 0;
}
#endif

/* CallNextTpClear */
static void __Pyx_call_next_tp_clear(PyObject* obj, inquiry current_tp_clear) {
    PyTypeObject* type = Py_TYPE(obj);
    inquiry tp_clear = NULL;
    while (type && __Pyx_PyType_GetSlot(type, tp_clear, inquiry) != current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    while (type && (tp_clear = __Pyx_PyType_GetSlot(type, tp_clear, inquiry)) == current_tp_clear)
        type = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
    if (type && tp_clear)
        tp_clear(obj);
}

/* LimitedApiGetTypeDict */
#if CYTHON_COMPILING_IN_LIMITED_API
static Py_ssize_t __Pyx_GetTypeDictOffset(void) {
    PyObject *tp_dictoffset_o;
    Py_ssize_t tp_dictoffset;
    tp_dictoffset_o = PyObject_GetAttrString((PyObject*)(&PyType_Type), "__dictoffset__");
    if (unlikely(!tp_dictoffset_o)) return -1;
    tp_dictoffset = PyLong_AsSsize_t(tp_dictoffset_o);
    Py_DECREF(tp_dictoffset_o);
    if (unlikely(tp_dictoffset == 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' doesn't have a dictoffset");
        return -1;
    } else if (unlikely(tp_dictoffset < 0)) {
        PyErr_SetString(
            PyExc_TypeError,
            "'type' has an unexpected negative dictoffset. "
            "Please report this as Cython bug");
        return -1;
    }
    return tp_dictoffset;
}
static PyObject *__Pyx_GetTypeDict(PyTypeObject *tp) {
    static Py_ssize_t tp_dictoffset = 0;
    if (unlikely(tp_dictoffset == 0)) {
        tp_dictoffset = __Pyx_GetTypeDictOffset();
        if (unlikely(tp_dictoffset == -1 && PyErr_Occurred())) {
            tp_dictoffset = 0; // try again next time?
            return NULL;
        }
    }
    return *(PyObject**)((char*)tp + tp_dictoffset);
}
#endif

/* SetItemOnTypeDict */
static int __Pyx__SetItemOnTypeDict(PyTypeObject *tp, PyObject *k, PyObject *v) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_SetItem(tp_dict, k, v);
    if (likely(!result)) {
        PyType_Modified(tp);
        if (unlikely(PyObject_HasAttr(v, __pyx_mstate_global->__pyx_n_u_set_name))) {
            PyObject *setNameResult = PyObject_CallMethodObjArgs(v, __pyx_mstate_global->__pyx_n_u_set_name,  (PyObject *) tp, k, NULL);
            if (!setNameResult) return -1;
            Py_DECREF(setNameResult);
        }
    }
    return result;
}

/* FixUpExtensionType */
static int __Pyx_fix_up_extension_type_from_spec(PyType_Spec *spec, PyTypeObject *type) {
#if __PYX_LIMITED_VERSION_HEX > 0x030900B1
    CYTHON_UNUSED_VAR(spec);
    CYTHON_UNUSED_VAR(type);
    CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
#else
    const PyType_Slot *slot = spec->slots;
    int changed = 0;
#if !CYTHON_COMPILING_IN_LIMITED_API
    while (slot && slot->slot && slot->slot != Py_tp_members)
        slot++;
    if (slot && slot->slot == Py_tp_members) {
#if !CYTHON_COMPILING_IN_CPYTHON
        const
#endif  // !CYTHON_COMPILING_IN_CPYTHON)
            PyMemberDef *memb = (PyMemberDef*) slot->pfunc;
        while (memb && memb->name) {
            if (memb->name[0] == '_' && memb->name[1] == '_') {
                if (strcmp(memb->name, "__weaklistoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_weaklistoffset = memb->offset;
                    changed = 1;
                }
                else if (strcmp(memb->name, "__dictoffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
                    type->tp_dictoffset = memb->offset;
                    changed = 1;
                }
#if CYTHON_METH_FASTCALL
                else if (strcmp(memb->name, "__vectorcalloffset__") == 0) {
                    assert(memb->type == T_PYSSIZET);
                    assert(memb->flags == READONLY);
#if PY_VERSION_HEX >= 0x030800b4
                    type->tp_vectorcall_offset = memb->offset;
#else
                    type->tp_print = (printfunc) memb->offset;
#endif
                    changed = 1;
                }
#endif  // CYTHON_METH_FASTCALL
#if !CYTHON_COMPILING_IN_PYPY
                else if (strcmp(memb->name, "__module__") == 0) {
                    PyObject *descr;
                    assert(memb->type == T_OBJECT);
                    assert(memb->flags == 0 || memb->flags == READONLY);
                    descr = PyDescr_NewMember(type, memb);
                    if (unlikely(!descr))
                        return -1;
                    int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                    Py_DECREF(descr);
                    if (unlikely(set_item_result < 0)) {
                        return -1;
                    }
                    changed = 1;
                }
#endif  // !CYTHON_COMPILING_IN_PYPY
            }
            memb++;
        }
    }
#endif  // !CYTHON_COMPILING_IN_LIMITED_API
#if !CYTHON_COMPILING_IN_PYPY
    slot = spec->slots;
    while (slot && slot->slot && slot->slot != Py_tp_getset)
        slot++;
    if (slot && slot->slot == Py_tp_getset) {
        PyGetSetDef *getset = (PyGetSetDef*) slot->pfunc;
        while (getset && getset->name) {
            if (getset->name[0] == '_' && getset->name[1] == '_' && strcmp(getset->name, "__module__") == 0) {
                PyObject *descr = PyDescr_NewGetSet(type, getset);
                if (unlikely(!descr))
                    return -1;
                #if CYTHON_COMPILING_IN_LIMITED_API
                PyObject *pyname = PyUnicode_FromString(getset->name);
                if (unlikely(!pyname)) {
                    Py_DECREF(descr);
                    return -1;
                }
                int set_item_result = __Pyx_SetItemOnTypeDict(type, pyname, descr);
                Py_DECREF(pyname);
                #else
                CYTHON_UNUSED_VAR(__Pyx__SetItemOnTypeDict);
                int set_item_result = PyDict_SetItem(type->tp_dict, PyDescr_NAME(descr), descr);
                #endif
                Py_DECREF(descr);
                if (unlikely(set_item_result < 0)) {
                    return -1;
                }
                changed = 1;
            }
            ++getset;
        }
    }
#endif  // !CYTHON_COMPILING_IN_PYPY
    if (changed)
        PyType_Modified(type);
#endif  // PY_VERSION_HEX > 0x030900B1
    return 0;
}

/* ValidateBasesTuple */
#if CYTHON_COMPILING_IN_CPYTHON || CYTHON_COMPILING_IN_LIMITED_API || CYTHON_USE_TYPE_SPECS
static int __Pyx_validate_bases_tuple(const char *type_name, Py_ssize_t dictoffset, PyObject *bases) {
    Py_ssize_t i, n;
#if CYTHON_ASSUME_SAFE_SIZE
    n = PyTuple_GET_SIZE(bases);
#else
    n = PyTuple_Size(bases);
    if (unlikely(n < 0)) return -1;
#endif
    for (i = 1; i < n; i++)
    {
        PyTypeObject *b;
#if CYTHON_AVOID_BORROWED_REFS
        PyObject *b0 = PySequence_GetItem(bases, i);
        if (!b0) return -1;
#elif CYTHON_ASSUME_SAFE_MACROS
        PyObject *b0 = PyTuple_GET_ITEM(bases, i);
#else
        PyObject *b0 = PyTuple_GetItem(bases, i);
        if (!b0) return -1;
#endif
        b = (PyTypeObject*) b0;
        if (!__Pyx_PyType_HasFeature(b, Py_TPFLAGS_HEAPTYPE))
        {
            __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
            PyErr_Format(PyExc_TypeError,
                "base class '" __Pyx_FMT_TYPENAME "' is not a heap type", b_name);
            __Pyx_DECREF_TypeName(b_name);
#if CYTHON_AVOID_BORROWED_REFS
            Py_DECREF(b0);
#endif
            return -1;
        }
        if (dictoffset == 0)
        {
            Py_ssize_t b_dictoffset = 0;
#if CYTHON_USE_TYPE_SLOTS
            b_dictoffset = b->tp_dictoffset;
#else
            PyObject *py_b_dictoffset = PyObject_GetAttrString((PyObject*)b, "__dictoffset__");
            if (!py_b_dictoffset) goto dictoffset_return;
            b_dictoffset = PyLong_AsSsize_t(py_b_dictoffset);
            Py_DECREF(py_b_dictoffset);
            if (b_dictoffset == -1 && PyErr_Occurred()) goto dictoffset_return;
#endif
            if (b_dictoffset) {
                {
                    __Pyx_TypeName b_name = __Pyx_PyType_GetFullyQualifiedName(b);
                    PyErr_Format(PyExc_TypeError,
                        "extension type '%.200s' has no __dict__ slot, "
                        "but base type '" __Pyx_FMT_TYPENAME "' has: "
                        "either add 'cdef dict __dict__' to the extension type "
                        "or add '__slots__ = [...]' to the base type",
                        type_name, b_name);
                    __Pyx_DECREF_TypeName(b_name);
                }
#if !CYTHON_USE_TYPE_SLOTS
              dictoffset_return:
#endif
#if CYTHON_AVOID_BORROWED_REFS
                Py_DECREF(b0);
#endif
                return -1;
            }
        }
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(b0);
#endif
    }
    return 0;
}
#endif

/* PyType_Ready */
CYTHON_UNUSED static int __Pyx_PyType_HasMultipleInheritance(PyTypeObject *t) {
    while (t) {
        PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
        if (bases) {
            return 1;
        }
        t = __Pyx_PyType_GetSlot(t, tp_base, PyTypeObject*);
    }
    return 0;
}
static int __Pyx_PyType_Ready(PyTypeObject *t) {
#if CYTHON_USE_TYPE_SPECS || !CYTHON_COMPILING_IN_CPYTHON || defined(PYSTON_MAJOR_VERSION)
    (void)__Pyx_PyObject_CallMethod0;
#if CYTHON_USE_TYPE_SPECS
    (void)__Pyx_validate_bases_tuple;
#endif
    return PyType_Ready(t);
#else
    int r;
    if (!__Pyx_PyType_HasMultipleInheritance(t)) {
        return PyType_Ready(t);
    }
    PyObject *bases = __Pyx_PyType_GetSlot(t, tp_bases, PyObject*);
    if (bases && unlikely(__Pyx_validate_bases_tuple(t->tp_name, t->tp_dictoffset, bases) == -1))
        return -1;
#if !defined(PYSTON_MAJOR_VERSION)
    {
        int gc_was_enabled;
    #if PY_VERSION_HEX >= 0x030A00b1
        gc_was_enabled = PyGC_Disable();
        (void)__Pyx_PyObject_CallMethod0;
    #else
        PyObject *ret, *py_status;
        PyObject *gc = NULL;
        #if (!CYTHON_COMPILING_IN_PYPY || PYPY_VERSION_NUM+0 >= 0x07030400) &&\
                !CYTHON_COMPILING_IN_GRAAL
        gc = PyImport_GetModule(__pyx_mstate_global->__pyx_kp_u_gc);
        #endif
        if (unlikely(!gc)) gc = PyImport_Import(__pyx_mstate_global->__pyx_kp_u_gc);
        if (unlikely(!gc)) return -1;
        py_status = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_isenabled);
        if (unlikely(!py_status)) {
            Py_DECREF(gc);
            return -1;
        }
        gc_was_enabled = __Pyx_PyObject_IsTrue(py_status);
        Py_DECREF(py_status);
        if (gc_was_enabled > 0) {
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_disable);
            if (unlikely(!ret)) {
                Py_DECREF(gc);
                return -1;
            }
            Py_DECREF(ret);
        } else if (unlikely(gc_was_enabled == -1)) {
            Py_DECREF(gc);
            return -1;
        }
    #endif
        t->tp_flags |= Py_TPFLAGS_HEAPTYPE;
#if PY_VERSION_HEX >= 0x030A0000
        t->tp_flags |= Py_TPFLAGS_IMMUTABLETYPE;
#endif
#else
        (void)__Pyx_PyObject_CallMethod0;
#endif
    r = PyType_Ready(t);
#if !defined(PYSTON_MAJOR_VERSION)
        t->tp_flags &= ~Py_TPFLAGS_HEAPTYPE;
    #if PY_VERSION_HEX >= 0x030A00b1
        if (gc_was_enabled)
            PyGC_Enable();
    #else
        if (gc_was_enabled) {
            PyObject *tp, *v, *tb;
            PyErr_Fetch(&tp, &v, &tb);
            ret = __Pyx_PyObject_CallMethod0(gc, __pyx_mstate_global->__pyx_kp_u_enable);
            if (likely(ret || r == -1)) {
                Py_XDECREF(ret);
                PyErr_Restore(tp, v, tb);
            } else {
                Py_XDECREF(tp);
                Py_XDECREF(v);
                Py_XDECREF(tb);
                r = -1;
            }
        }
        Py_DECREF(gc);
    #endif
    }
#endif
    return r;
#endif
}

/* SetVTable */
static int __Pyx_SetVtable(PyTypeObject *type, void *vtable) {
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
    if (unlikely(!ob))
        goto bad;
#if CYTHON_COMPILING_IN_LIMITED_API
    if (unlikely(PyObject_SetAttr((PyObject *) type, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#else
    if (unlikely(PyDict_SetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable, ob) < 0))
#endif
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
static void* __Pyx_GetVtable(PyTypeObject *type) {
    void* ptr;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *ob = PyObject_GetAttr((PyObject *)type, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#else
    PyObject *ob = PyObject_GetItem(type->tp_dict, __pyx_mstate_global->__pyx_n_u_pyx_vtable);
#endif
    if (!ob)
        goto bad;
    ptr = PyCapsule_GetPointer(ob, 0);
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* MergeVTables */
static int __Pyx_MergeVtables(PyTypeObject *type) {
    int i=0;
    Py_ssize_t size;
    void** base_vtables;
    __Pyx_TypeName tp_base_name = NULL;
    __Pyx_TypeName base_name = NULL;
    void* unknown = (void*)-1;
    PyObject* bases = __Pyx_PyType_GetSlot(type, tp_bases, PyObject*);
    int base_depth = 0;
    {
        PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        while (base) {
            base_depth += 1;
            base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
        }
    }
    base_vtables = (void**) PyMem_Malloc(sizeof(void*) * (size_t)(base_depth + 1));
    base_vtables[0] = unknown;
#if CYTHON_COMPILING_IN_LIMITED_API
    size = PyTuple_Size(bases);
    if (size < 0) goto other_failure;
#else
    size = PyTuple_GET_SIZE(bases);
#endif
    for (i = 1; i < size; i++) {
        PyObject *basei;
        void* base_vtable;
#if CYTHON_AVOID_BORROWED_REFS
        basei = PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto other_failure;
#else
        basei = PyTuple_GET_ITEM(bases, i);
#endif
        base_vtable = __Pyx_GetVtable((PyTypeObject*)basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
        if (base_vtable != NULL) {
            int j;
            PyTypeObject* base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
            for (j = 0; j < base_depth; j++) {
                if (base_vtables[j] == unknown) {
                    base_vtables[j] = __Pyx_GetVtable(base);
                    base_vtables[j + 1] = unknown;
                }
                if (base_vtables[j] == base_vtable) {
                    break;
                } else if (base_vtables[j] == NULL) {
                    goto bad;
                }
                base = __Pyx_PyType_GetSlot(base, tp_base, PyTypeObject*);
            }
        }
    }
    PyErr_Clear();
    PyMem_Free(base_vtables);
    return 0;
bad:
    {
        PyTypeObject* basei = NULL;
        PyTypeObject* tp_base = __Pyx_PyType_GetSlot(type, tp_base, PyTypeObject*);
        tp_base_name = __Pyx_PyType_GetFullyQualifiedName(tp_base);
#if CYTHON_AVOID_BORROWED_REFS
        basei = (PyTypeObject*)PySequence_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#elif !CYTHON_ASSUME_SAFE_MACROS
        basei = (PyTypeObject*)PyTuple_GetItem(bases, i);
        if (unlikely(!basei)) goto really_bad;
#else
        basei = (PyTypeObject*)PyTuple_GET_ITEM(bases, i);
#endif
        base_name = __Pyx_PyType_GetFullyQualifiedName(basei);
#if CYTHON_AVOID_BORROWED_REFS
        Py_DECREF(basei);
#endif
    }
    PyErr_Format(PyExc_TypeError,
        "multiple bases have vtable conflict: '" __Pyx_FMT_TYPENAME "' and '" __Pyx_FMT_TYPENAME "'", tp_base_name, base_name);
#if CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
really_bad: // bad has failed!
#endif
    __Pyx_DECREF_TypeName(tp_base_name);
    __Pyx_DECREF_TypeName(base_name);
#if CYTHON_COMPILING_IN_LIMITED_API || CYTHON_AVOID_BORROWED_REFS || !CYTHON_ASSUME_SAFE_MACROS
other_failure:
#endif
    PyMem_Free(base_vtables);
    return -1;
}

/* DelItemOnTypeDict */
static int __Pyx__DelItemOnTypeDict(PyTypeObject *tp, PyObject *k) {
    int result;
    PyObject *tp_dict;
#if CYTHON_COMPILING_IN_LIMITED_API
    tp_dict = __Pyx_GetTypeDict(tp);
    if (unlikely(!tp_dict)) return -1;
#else
    tp_dict = tp->tp_dict;
#endif
    result = PyDict_DelItem(tp_dict, k);
    if (likely(!result)) PyType_Modified(tp);
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStrNoError(meth, __pyx_mstate_global->__pyx_n_u_name);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_getstate = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
    PyObject *getstate = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    getstate = _PyType_Lookup((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
#else
    getstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_getstate);
    if (!getstate && PyErr_Occurred()) {
        goto __PYX_BAD;
    }
#endif
    if (getstate) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_getstate = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
#else
        object_getstate = __Pyx_PyObject_GetAttrStrNoError((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_getstate);
        if (!object_getstate && PyErr_Occurred()) {
            goto __PYX_BAD;
        }
#endif
        if (object_getstate != getstate) {
            goto __PYX_GOOD;
        }
    }
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_mstate_global->__pyx_n_u_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_mstate_global->__pyx_n_u_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_mstate_global->__pyx_n_u_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython);
            if (likely(reduce_cython)) {
                ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_mstate_global->__pyx_n_u_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = __Pyx_SetItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = __Pyx_DelItemOnTypeDict((PyTypeObject*)type_obj, __pyx_mstate_global->__pyx_n_u_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred()) {
        __Pyx_TypeName type_obj_name =
            __Pyx_PyType_GetFullyQualifiedName((PyTypeObject*)type_obj);
        PyErr_Format(PyExc_RuntimeError,
            "Unable to initialize pickling for " __Pyx_FMT_TYPENAME, type_obj_name);
        __Pyx_DECREF_TypeName(type_obj_name);
    }
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
    Py_XDECREF(object_getstate);
    Py_XDECREF(getstate);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType_3_1_3
#define __PYX_HAVE_RT_ImportType_3_1_3
static PyTypeObject *__Pyx_ImportType_3_1_3(PyObject *module, const char *module_name, const char *class_name,
    size_t size, size_t alignment, enum __Pyx_ImportType_CheckSize_3_1_3 check_size)
{
    PyObject *result = 0;
    Py_ssize_t basicsize;
    Py_ssize_t itemsize;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    PyObject *py_itemsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#if !CYTHON_COMPILING_IN_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
    itemsize = ((PyTypeObject *)result)->tp_itemsize;
#else
    if (size == 0) {
        return (PyTypeObject *)result;
    }
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
    py_itemsize = PyObject_GetAttrString(result, "__itemsize__");
    if (!py_itemsize)
        goto bad;
    itemsize = PyLong_AsSsize_t(py_itemsize);
    Py_DECREF(py_itemsize);
    py_itemsize = 0;
    if (itemsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (itemsize) {
        if (size % alignment) {
            alignment = size % alignment;
        }
        if (itemsize < (Py_ssize_t)alignment)
            itemsize = (Py_ssize_t)alignment;
    }
    if ((size_t)(basicsize + itemsize) < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize+itemsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error_3_1_3 &&
            ((size_t)basicsize > size || (size_t)(basicsize + itemsize) < size)) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd-%zd from PyObject",
            module_name, class_name, size, basicsize, basicsize+itemsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn_3_1_3 && (size_t)basicsize > size) {
        if (PyErr_WarnFormat(NULL, 0,
                "%.200s.%.200s size changed, may indicate binary incompatibility. "
                "Expected %zd from C header, got %zd from PyObject",
                module_name, class_name, size, basicsize) < 0) {
            goto bad;
        }
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* ImportDottedModule */
static PyObject *__Pyx__ImportDottedModule_Error(PyObject *name, PyObject *parts_tuple, Py_ssize_t count) {
    PyObject *partial_name = NULL, *slice = NULL, *sep = NULL;
    Py_ssize_t size;
    if (unlikely(PyErr_Occurred())) {
        PyErr_Clear();
    }
#if CYTHON_ASSUME_SAFE_SIZE
    size = PyTuple_GET_SIZE(parts_tuple);
#else
    size = PyTuple_Size(parts_tuple);
    if (size < 0) goto bad;
#endif
    if (likely(size == count)) {
        partial_name = name;
    } else {
        slice = PySequence_GetSlice(parts_tuple, 0, count);
        if (unlikely(!slice))
            goto bad;
        sep = PyUnicode_FromStringAndSize(".", 1);
        if (unlikely(!sep))
            goto bad;
        partial_name = PyUnicode_Join(sep, slice);
    }
    PyErr_Format(
        PyExc_ModuleNotFoundError,
        "No module named '%U'", partial_name);
bad:
    Py_XDECREF(sep);
    Py_XDECREF(slice);
    Py_XDECREF(partial_name);
    return NULL;
}
static PyObject *__Pyx__ImportDottedModule_Lookup(PyObject *name) {
    PyObject *imported_module;
#if (CYTHON_COMPILING_IN_PYPY && PYPY_VERSION_NUM  < 0x07030400) ||\
        CYTHON_COMPILING_IN_GRAAL
    PyObject *modules = PyImport_GetModuleDict();
    if (unlikely(!modules))
        return NULL;
    imported_module = __Pyx_PyDict_GetItemStr(modules, name);
    Py_XINCREF(imported_module);
#else
    imported_module = PyImport_GetModule(name);
#endif
    return imported_module;
}
static PyObject *__Pyx_ImportDottedModule_WalkParts(PyObject *module, PyObject *name, PyObject *parts_tuple) {
    Py_ssize_t i, nparts;
#if CYTHON_ASSUME_SAFE_SIZE
    nparts = PyTuple_GET_SIZE(parts_tuple);
#else
    nparts = PyTuple_Size(parts_tuple);
    if (nparts < 0) return NULL;
#endif
    for (i=1; i < nparts && module; i++) {
        PyObject *part, *submodule;
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        part = PyTuple_GET_ITEM(parts_tuple, i);
#else
        part = __Pyx_PySequence_ITEM(parts_tuple, i);
        if (!part) return NULL;
#endif
        submodule = __Pyx_PyObject_GetAttrStrNoError(module, part);
#if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
        Py_DECREF(part);
#endif
        Py_DECREF(module);
        module = submodule;
    }
    if (unlikely(!module)) {
        return __Pyx__ImportDottedModule_Error(name, parts_tuple, i);
    }
    return module;
}
static PyObject *__Pyx__ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
    PyObject *imported_module;
    PyObject *module = __Pyx_Import(name, NULL, 0);
    if (!parts_tuple || unlikely(!module))
        return module;
    imported_module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(imported_module)) {
        Py_DECREF(module);
        return imported_module;
    }
    PyErr_Clear();
    return __Pyx_ImportDottedModule_WalkParts(module, name, parts_tuple);
}
static PyObject *__Pyx_ImportDottedModule(PyObject *name, PyObject *parts_tuple) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject *module = __Pyx__ImportDottedModule_Lookup(name);
    if (likely(module)) {
        PyObject *spec = __Pyx_PyObject_GetAttrStrNoError(module, __pyx_mstate_global->__pyx_n_u_spec);
        if (likely(spec)) {
            PyObject *unsafe = __Pyx_PyObject_GetAttrStrNoError(spec, __pyx_mstate_global->__pyx_n_u_initializing);
            if (likely(!unsafe || !__Pyx_PyObject_IsTrue(unsafe))) {
                Py_DECREF(spec);
                spec = NULL;
            }
            Py_XDECREF(unsafe);
        }
        if (likely(!spec)) {
            PyErr_Clear();
            return module;
        }
        Py_DECREF(spec);
        Py_DECREF(module);
    } else if (PyErr_Occurred()) {
        PyErr_Clear();
    }
#endif
    return __Pyx__ImportDottedModule(name, parts_tuple);
}

/* ListPack */
static PyObject *__Pyx_PyList_Pack(Py_ssize_t n, ...) {
    va_list va;
    PyObject *l = PyList_New(n);
    va_start(va, n);
    if (unlikely(!l)) goto end;
    for (Py_ssize_t i=0; i<n; ++i) {
        PyObject *arg = va_arg(va, PyObject*);
        Py_INCREF(arg);
        if (__Pyx_PyList_SET_ITEM(l, i, arg) != (0)) {
            Py_CLEAR(l);
            goto end;
        }
    }
    end:
    va_end(va);
    return l;
}

/* FetchSharedCythonModule */
static PyObject *__Pyx_FetchSharedCythonABIModule(void) {
    return __Pyx_PyImport_AddModuleRef(__PYX_ABI_MODULE_NAME);
}

/* dict_setdefault */
static CYTHON_INLINE PyObject *__Pyx_PyDict_SetDefault(PyObject *d, PyObject *key, PyObject *default_value,
                                                       int is_safe_type) {
    PyObject* value;
    CYTHON_MAYBE_UNUSED_VAR(is_safe_type);
#if CYTHON_COMPILING_IN_LIMITED_API
    value = PyObject_CallMethod(d, "setdefault", "OO", key, default_value);
#elif PY_VERSION_HEX >= 0x030d0000
    PyDict_SetDefaultRef(d, key, default_value, &value);
#else
    value = PyDict_SetDefault(d, key, default_value);
    if (unlikely(!value)) return NULL;
    Py_INCREF(value);
#endif
    return value;
}

/* FetchCommonType */
#if __PYX_LIMITED_VERSION_HEX < 0x030C0000
static PyObject* __Pyx_PyType_FromMetaclass(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *result = __Pyx_PyType_FromModuleAndSpec(module, spec, bases);
    if (result && metaclass) {
        PyObject *old_tp = (PyObject*)Py_TYPE(result);
    Py_INCREF((PyObject*)metaclass);
#if __PYX_LIMITED_VERSION_HEX >= 0x03090000
        Py_SET_TYPE(result, metaclass);
#else
        result->ob_type = metaclass;
#endif
        Py_DECREF(old_tp);
    }
    return result;
}
#else
#define __Pyx_PyType_FromMetaclass(me, mo, s, b) PyType_FromMetaclass(me, mo, s, b)
#endif
static int __Pyx_VerifyCachedType(PyObject *cached_type,
                               const char *name,
                               Py_ssize_t expected_basicsize) {
    Py_ssize_t basicsize;
    if (!PyType_Check(cached_type)) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s is not a type object", name);
        return -1;
    }
    if (expected_basicsize == 0) {
        return 0; // size is inherited, nothing useful to check
    }
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_basicsize;
    py_basicsize = PyObject_GetAttrString(cached_type, "__basicsize__");
    if (unlikely(!py_basicsize)) return -1;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = NULL;
    if (unlikely(basicsize == (Py_ssize_t)-1) && PyErr_Occurred()) return -1;
#else
    basicsize = ((PyTypeObject*) cached_type)->tp_basicsize;
#endif
    if (basicsize != expected_basicsize) {
        PyErr_Format(PyExc_TypeError,
            "Shared Cython type %.200s has the wrong size, try recompiling",
            name);
        return -1;
    }
    return 0;
}
static PyTypeObject *__Pyx_FetchCommonTypeFromSpec(PyTypeObject *metaclass, PyObject *module, PyType_Spec *spec, PyObject *bases) {
    PyObject *abi_module = NULL, *cached_type = NULL, *abi_module_dict, *new_cached_type, *py_object_name;
    int get_item_ref_result;
    const char* object_name = strrchr(spec->name, '.');
    object_name = object_name ? object_name+1 : spec->name;
    py_object_name = PyUnicode_FromString(object_name);
    if (!py_object_name) return NULL;
    abi_module = __Pyx_FetchSharedCythonABIModule();
    if (!abi_module) goto done;
    abi_module_dict = PyModule_GetDict(abi_module);
    if (!abi_module_dict) goto done;
    get_item_ref_result = __Pyx_PyDict_GetItemRef(abi_module_dict, py_object_name, &cached_type);
    if (get_item_ref_result == 1) {
        if (__Pyx_VerifyCachedType(
              cached_type,
              object_name,
              spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else if (unlikely(get_item_ref_result == -1)) {
        goto bad;
    }
    CYTHON_UNUSED_VAR(module);
    cached_type = __Pyx_PyType_FromMetaclass(metaclass, abi_module, spec, bases);
    if (unlikely(!cached_type)) goto bad;
    if (unlikely(__Pyx_fix_up_extension_type_from_spec(spec, (PyTypeObject *) cached_type) < 0)) goto bad;
    new_cached_type = __Pyx_PyDict_SetDefault(abi_module_dict, py_object_name, cached_type, 1);
    if (unlikely(new_cached_type != cached_type)) {
        if (unlikely(!new_cached_type)) goto bad;
        Py_DECREF(cached_type);
        cached_type = new_cached_type;
        if (__Pyx_VerifyCachedType(
                cached_type,
                object_name,
                spec->basicsize) < 0) {
            goto bad;
        }
        goto done;
    } else {
        Py_DECREF(new_cached_type);
    }
done:
    Py_XDECREF(abi_module);
    Py_DECREF(py_object_name);
    assert(cached_type == NULL || PyType_Check(cached_type));
    return (PyTypeObject *) cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CommonTypesMetaclass */
static PyObject* __pyx_CommonTypesMetaclass_get_module(CYTHON_UNUSED PyObject *self, CYTHON_UNUSED void* context) {
    return PyUnicode_FromString(__PYX_ABI_MODULE_NAME);
}
static PyGetSetDef __pyx_CommonTypesMetaclass_getset[] = {
    {"__module__", __pyx_CommonTypesMetaclass_get_module, NULL, NULL, NULL},
    {0, 0, 0, 0, 0}
};
static PyType_Slot __pyx_CommonTypesMetaclass_slots[] = {
    {Py_tp_getset, (void *)__pyx_CommonTypesMetaclass_getset},
    {0, 0}
};
static PyType_Spec __pyx_CommonTypesMetaclass_spec = {
    __PYX_TYPE_MODULE_PREFIX "_common_types_metatype",
    0,
    0,
#if PY_VERSION_HEX >= 0x030A0000
    Py_TPFLAGS_IMMUTABLETYPE |
    Py_TPFLAGS_DISALLOW_INSTANTIATION |
#endif
    Py_TPFLAGS_DEFAULT,
    __pyx_CommonTypesMetaclass_slots
};
static int __pyx_CommonTypesMetaclass_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    PyObject *bases = PyTuple_Pack(1, &PyType_Type);
    if (unlikely(!bases)) {
        return -1;
    }
    mstate->__pyx_CommonTypesMetaclassType = __Pyx_FetchCommonTypeFromSpec(NULL, module, &__pyx_CommonTypesMetaclass_spec, bases);
    if (unlikely(mstate->__pyx_CommonTypesMetaclassType == NULL)) {
        return -1;
    }
    return 0;
}

/* PyMethodNew */
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    PyObject *result;
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    #if __PYX_LIMITED_VERSION_HEX >= 0x030C0000
    {
        PyObject *args[] = {func, self};
        result = PyObject_Vectorcall(__pyx_mstate_global->__Pyx_CachedMethodType, args, 2, NULL);
    }
    #else
    result = PyObject_CallFunctionObjArgs(__pyx_mstate_global->__Pyx_CachedMethodType, func, self, NULL);
    #endif
    return result;
}
#else
static PyObject *__Pyx_PyMethod_New(PyObject *func, PyObject *self, PyObject *typ) {
    CYTHON_UNUSED_VAR(typ);
    if (!self)
        return __Pyx_NewRef(func);
    return PyMethod_New(func, self);
}
#endif

/* PyVectorcallFastCallDict */
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static PyObject *__Pyx_PyVectorcall_FastCallDict_kw(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    PyObject *res = NULL;
    PyObject *kwnames;
    PyObject **newargs;
    PyObject **kwvalues;
    Py_ssize_t i, pos;
    size_t j;
    PyObject *key, *value;
    unsigned long keys_are_strings;
    #if !CYTHON_ASSUME_SAFE_SIZE
    Py_ssize_t nkw = PyDict_Size(kw);
    if (unlikely(nkw == -1)) return NULL;
    #else
    Py_ssize_t nkw = PyDict_GET_SIZE(kw);
    #endif
    newargs = (PyObject **)PyMem_Malloc((nargs + (size_t)nkw) * sizeof(args[0]));
    if (unlikely(newargs == NULL)) {
        PyErr_NoMemory();
        return NULL;
    }
    for (j = 0; j < nargs; j++) newargs[j] = args[j];
    kwnames = PyTuple_New(nkw);
    if (unlikely(kwnames == NULL)) {
        PyMem_Free(newargs);
        return NULL;
    }
    kwvalues = newargs + nargs;
    pos = i = 0;
    keys_are_strings = Py_TPFLAGS_UNICODE_SUBCLASS;
    while (PyDict_Next(kw, &pos, &key, &value)) {
        keys_are_strings &=
        #if CYTHON_COMPILING_IN_LIMITED_API
            PyType_GetFlags(Py_TYPE(key));
        #else
            Py_TYPE(key)->tp_flags;
        #endif
        Py_INCREF(key);
        Py_INCREF(value);
        #if !CYTHON_ASSUME_SAFE_MACROS
        if (unlikely(PyTuple_SetItem(kwnames, i, key) < 0)) goto cleanup;
        #else
        PyTuple_SET_ITEM(kwnames, i, key);
        #endif
        kwvalues[i] = value;
        i++;
    }
    if (unlikely(!keys_are_strings)) {
        PyErr_SetString(PyExc_TypeError, "keywords must be strings");
        goto cleanup;
    }
    res = vc(func, newargs, nargs, kwnames);
cleanup:
    Py_DECREF(kwnames);
    for (i = 0; i < nkw; i++)
        Py_DECREF(kwvalues[i]);
    PyMem_Free(newargs);
    return res;
}
static CYTHON_INLINE PyObject *__Pyx_PyVectorcall_FastCallDict(PyObject *func, __pyx_vectorcallfunc vc, PyObject *const *args, size_t nargs, PyObject *kw)
{
    Py_ssize_t kw_size =
        likely(kw == NULL) ?
        0 :
#if !CYTHON_ASSUME_SAFE_SIZE
        PyDict_Size(kw);
#else
        PyDict_GET_SIZE(kw);
#endif
    if (kw_size == 0) {
        return vc(func, args, nargs, NULL);
    }
#if !CYTHON_ASSUME_SAFE_SIZE
    else if (unlikely(kw_size == -1)) {
        return NULL;
    }
#endif
    return __Pyx_PyVectorcall_FastCallDict_kw(func, vc, args, nargs, kw);
}
#endif

/* CythonFunctionShared */
#if CYTHON_COMPILING_IN_LIMITED_API
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunctionNoMethod(PyObject *func, void (*cfunc)(void)) {
    if (__Pyx_CyFunction_Check(func)) {
        return PyCFunction_GetFunction(((__pyx_CyFunctionObject*)func)->func) == (PyCFunction) cfunc;
    } else if (PyCFunction_Check(func)) {
        return PyCFunction_GetFunction(func) == (PyCFunction) cfunc;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if ((PyObject*)Py_TYPE(func) == __pyx_mstate_global->__Pyx_CachedMethodType) {
        int result;
        PyObject *newFunc = PyObject_GetAttr(func, __pyx_mstate_global->__pyx_n_u_func);
        if (unlikely(!newFunc)) {
            PyErr_Clear(); // It's only an optimization, so don't throw an error
            return 0;
        }
        result = __Pyx__IsSameCyOrCFunctionNoMethod(newFunc, cfunc);
        Py_DECREF(newFunc);
        return result;
    }
    return __Pyx__IsSameCyOrCFunctionNoMethod(func, cfunc);
}
#else
static CYTHON_INLINE int __Pyx__IsSameCyOrCFunction(PyObject *func, void (*cfunc)(void)) {
    if (PyMethod_Check(func)) {
        func = PyMethod_GET_FUNCTION(func);
    }
    return __Pyx_CyOrPyCFunction_Check(func) && __Pyx_CyOrPyCFunction_GET_FUNCTION(func) == (PyCFunction) cfunc;
}
#endif
static CYTHON_INLINE void __Pyx__CyFunction_SetClassObj(__pyx_CyFunctionObject* f, PyObject* classobj) {
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    __Pyx_Py_XDECREF_SET(
        __Pyx_CyFunction_GetClassObj(f),
            ((classobj) ? __Pyx_NewRef(classobj) : NULL));
#else
    __Pyx_Py_XDECREF_SET(
        ((PyCMethodObject *) (f))->mm_class,
        (PyTypeObject*)((classobj) ? __Pyx_NewRef(classobj) : NULL));
#endif
}
static PyObject *
__Pyx_CyFunction_get_doc_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_doc == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_doc = PyObject_GetAttrString(op->func, "__doc__");
        if (unlikely(!op->func_doc)) return NULL;
#else
        if (((PyCFunctionObject*)op)->m_ml->ml_doc) {
            op->func_doc = PyUnicode_FromString(((PyCFunctionObject*)op)->m_ml->ml_doc);
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
#endif
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, void *closure) {
    PyObject *result;
    CYTHON_UNUSED_VAR(closure);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_doc_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_doc, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if CYTHON_COMPILING_IN_LIMITED_API
        op->func_name = PyObject_GetAttrString(op->func, "__name__");
#else
        op->func_name = PyUnicode_InternFromString(((PyCFunctionObject*)op)->m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op, void *context)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_name_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_name, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    Py_INCREF(op->func_qualname);
    result = op->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_qualname, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_dict_locked(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    PyObject *result;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_dict_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value, void *context)
{
    CYTHON_UNUSED_VAR(context);
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_dict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(__pyx_CyFunctionObject *op, void *context)
{
    CYTHON_UNUSED_VAR(op);
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op, void *context)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    CYTHON_UNUSED_VAR(context);
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = __Pyx_PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = __Pyx_PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyTuple_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__defaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_tuple, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result = NULL;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_defaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value) {
        value = Py_None;
    } else if (unlikely(value != Py_None && !PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    PyErr_WarnEx(PyExc_RuntimeWarning, "changes to cyfunction.__kwdefaults__ will not "
                 "currently affect the values used in function calls", 1);
    Py_INCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->defaults_kwdict, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (unlikely(__Pyx_CyFunction_init_defaults(op) < 0)) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op, void *context) {
    PyObject* result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_kwdefaults_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    if (!value || value == Py_None) {
        value = NULL;
    } else if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    __Pyx_Py_XDECREF_SET(op->func_annotations, value);
    __Pyx_END_CRITICAL_SECTION();
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations_locked(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    result = __Pyx_CyFunction_get_annotations_locked(op);
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine_value(__pyx_CyFunctionObject *op) {
    int is_coroutine = op->flags & __Pyx_CYFUNCTION_COROUTINE;
    if (is_coroutine) {
        PyObject *is_coroutine_value, *module, *fromlist, *marker = __pyx_mstate_global->__pyx_n_u_is_coroutine;
        fromlist = PyList_New(1);
        if (unlikely(!fromlist)) return NULL;
        Py_INCREF(marker);
#if CYTHON_ASSUME_SAFE_MACROS
        PyList_SET_ITEM(fromlist, 0, marker);
#else
        if (unlikely(PyList_SetItem(fromlist, 0, marker) < 0)) {
            Py_DECREF(marker);
            Py_DECREF(fromlist);
            return NULL;
        }
#endif
        module = PyImport_ImportModuleLevelObject(__pyx_mstate_global->__pyx_n_u_asyncio_coroutines, NULL, NULL, fromlist, 0);
        Py_DECREF(fromlist);
        if (unlikely(!module)) goto ignore;
        is_coroutine_value = __Pyx_PyObject_GetAttrStr(module, marker);
        Py_DECREF(module);
        if (likely(is_coroutine_value)) {
            return is_coroutine_value;
        }
ignore:
        PyErr_Clear();
    }
    return __Pyx_PyBool_FromLong(is_coroutine);
}
static PyObject *
__Pyx_CyFunction_get_is_coroutine(__pyx_CyFunctionObject *op, void *context) {
    PyObject *result;
    CYTHON_UNUSED_VAR(context);
    if (op->func_is_coroutine) {
        return __Pyx_NewRef(op->func_is_coroutine);
    }
    result = __Pyx_CyFunction_get_is_coroutine_value(op);
    if (unlikely(!result))
        return NULL;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    if (op->func_is_coroutine) {
        Py_DECREF(result);
        result = __Pyx_NewRef(op->func_is_coroutine);
    } else {
        op->func_is_coroutine = __Pyx_NewRef(result);
    }
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static void __Pyx_CyFunction_raise_argument_count_error(__pyx_CyFunctionObject *func, const char* message, Py_ssize_t size) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        py_name, message, size);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s (%" CYTHON_FORMAT_SSIZE_T "d given)",
        name, message, size);
#endif
}
static void __Pyx_CyFunction_raise_type_error(__pyx_CyFunctionObject *func, const char* message) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *py_name = __Pyx_CyFunction_get_name(func, NULL);
    if (!py_name) return;
    PyErr_Format(PyExc_TypeError,
        "%.200S() %s",
        py_name, message);
    Py_DECREF(py_name);
#else
    const char* name = ((PyCFunctionObject*)func)->m_ml->ml_name;
    PyErr_Format(PyExc_TypeError,
        "%.200s() %s",
        name, message);
#endif
}
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *
__Pyx_CyFunction_get_module(__pyx_CyFunctionObject *op, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_GetAttrString(op->func, "__module__");
}
static int
__Pyx_CyFunction_set_module(__pyx_CyFunctionObject *op, PyObject* value, void *context) {
    CYTHON_UNUSED_VAR(context);
    return PyObject_SetAttrString(op->func, "__module__", value);
}
#endif
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {"func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {"func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {"__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {"func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {"__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {"func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {"func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {"func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {"func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {"__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {"__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {"_is_coroutine", (getter)__Pyx_CyFunction_get_is_coroutine, 0, 0, 0},
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", (getter)__Pyx_CyFunction_get_module, (setter)__Pyx_CyFunction_set_module, 0, 0},
#endif
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
#if !CYTHON_COMPILING_IN_LIMITED_API
    {"__module__", T_OBJECT, offsetof(PyCFunctionObject, m_module), 0, 0},
#endif
    {"__dictoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_dict), READONLY, 0},
#if CYTHON_METH_FASTCALL
#if CYTHON_BACKPORT_VECTORCALL || CYTHON_COMPILING_IN_LIMITED_API
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_vectorcall), READONLY, 0},
#else
    {"__vectorcalloffset__", T_PYSSIZET, offsetof(PyCFunctionObject, vectorcall), READONLY, 0},
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
    {"__weaklistoffset__", T_PYSSIZET, offsetof(__pyx_CyFunctionObject, func_weakreflist), READONLY, 0},
#else
    {"__weaklistoffset__", T_PYSSIZET, offsetof(PyCFunctionObject, m_weakreflist), READONLY, 0},
#endif
#endif
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, PyObject *args)
{
    PyObject *result = NULL;
    CYTHON_UNUSED_VAR(args);
    __Pyx_BEGIN_CRITICAL_SECTION(m);
    Py_INCREF(m->func_qualname);
    result = m->func_qualname;
    __Pyx_END_CRITICAL_SECTION();
    return result;
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if CYTHON_COMPILING_IN_LIMITED_API
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) (((PyCFunctionObject*)cyfunc)->m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_Init(__pyx_CyFunctionObject *op, PyMethodDef *ml, int flags, PyObject* qualname,
                                       PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
#if !CYTHON_COMPILING_IN_LIMITED_API
    PyCFunctionObject *cf = (PyCFunctionObject*) op;
#endif
    if (unlikely(op == NULL))
        return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    op->func = PyCFunction_NewEx(ml, (PyObject*)op, module);
    if (unlikely(!op->func)) return NULL;
#endif
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
#if !CYTHON_COMPILING_IN_LIMITED_API
    cf->m_ml = ml;
    cf->m_self = (PyObject *) op;
#endif
    Py_XINCREF(closure);
    op->func_closure = closure;
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_XINCREF(module);
    cf->m_module = module;
#endif
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
#if PY_VERSION_HEX < 0x030900B1 || CYTHON_COMPILING_IN_LIMITED_API
    op->func_classobj = NULL;
#else
    ((PyCMethodObject*)op)->mm_class = NULL;
#endif
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    op->func_is_coroutine = NULL;
#if CYTHON_METH_FASTCALL
    switch (ml->ml_flags & (METH_VARARGS | METH_FASTCALL | METH_NOARGS | METH_O | METH_KEYWORDS | METH_METHOD)) {
    case METH_NOARGS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_NOARGS;
        break;
    case METH_O:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_O;
        break;
    case METH_METHOD | METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD;
        break;
    case METH_FASTCALL | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS;
        break;
    case METH_VARARGS | METH_KEYWORDS:
        __Pyx_CyFunction_func_vectorcall(op) = NULL;
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        Py_DECREF(op);
        return NULL;
    }
#endif
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_CLEAR(m->func);
#else
    Py_CLEAR(((PyCFunctionObject*)m)->m_module);
#endif
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
#if PY_VERSION_HEX < 0x030900B1
    Py_CLEAR(__Pyx_CyFunction_GetClassObj(m));
#else
    {
        PyObject *cls = (PyObject*) ((PyCMethodObject *) (m))->mm_class;
        ((PyCMethodObject *) (m))->mm_class = NULL;
        Py_XDECREF(cls);
    }
#endif
#endif
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    Py_CLEAR(m->func_is_coroutine);
    Py_CLEAR(m->defaults);
    return 0;
}
static void __Pyx__CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    __Pyx_PyHeapTypeObject_GC_Del(m);
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    __Pyx__CyFunction_dealloc(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    {
        int e = __Pyx_call_type_traverse((PyObject*)m, 1, visit, arg);
        if (e) return e;
    }
    Py_VISIT(m->func_closure);
#if CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(m->func);
#else
    Py_VISIT(((PyCFunctionObject*)m)->m_module);
#endif
    Py_VISIT(m->func_dict);
    __Pyx_VISIT_CONST(m->func_name);
    __Pyx_VISIT_CONST(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    __Pyx_VISIT_CONST(m->func_code);
#if !CYTHON_COMPILING_IN_LIMITED_API
    Py_VISIT(__Pyx_CyFunction_GetClassObj(m));
#endif
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    Py_VISIT(m->func_is_coroutine);
    Py_VISIT(m->defaults);
    return 0;
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
    PyObject *repr;
    __Pyx_BEGIN_CRITICAL_SECTION(op);
    repr = PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
    __Pyx_END_CRITICAL_SECTION();
    return repr;
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
#if CYTHON_COMPILING_IN_LIMITED_API
    PyObject *f = ((__pyx_CyFunctionObject*)func)->func;
    PyCFunction meth;
    int flags;
    meth = PyCFunction_GetFunction(f);
    if (unlikely(!meth)) return NULL;
    flags = PyCFunction_GetFlags(f);
    if (unlikely(flags < 0)) return NULL;
#else
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    int flags = f->m_ml->ml_flags;
#endif
    Py_ssize_t size;
    switch (flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)(void(*)(void))meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 0))
                return (*meth)(self, NULL);
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes no arguments", size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
#if CYTHON_ASSUME_SAFE_SIZE
            size = PyTuple_GET_SIZE(arg);
#else
            size = PyTuple_Size(arg);
            if (unlikely(size < 0)) return NULL;
#endif
            if (likely(size == 1)) {
                PyObject *result, *arg0;
                #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                arg0 = PyTuple_GET_ITEM(arg, 0);
                #else
                arg0 = __Pyx_PySequence_ITEM(arg, 0); if (unlikely(!arg0)) return NULL;
                #endif
                result = (*meth)(self, arg0);
                #if !(CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS)
                Py_DECREF(arg0);
                #endif
                return result;
            }
            __Pyx_CyFunction_raise_argument_count_error(
                (__pyx_CyFunctionObject*)func,
                "takes exactly one argument", size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags for CyFunction");
        return NULL;
    }
    __Pyx_CyFunction_raise_type_error(
        (__pyx_CyFunctionObject*)func, "takes no keyword arguments");
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *self, *result;
#if CYTHON_COMPILING_IN_LIMITED_API
    self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)func)->func);
    if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
    self = ((PyCFunctionObject*)func)->m_self;
#endif
    result = __Pyx_CyFunction_CallMethod(func, self, arg, kw);
    return result;
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
     __pyx_vectorcallfunc vc = __Pyx_CyFunction_func_vectorcall(cyfunc);
    if (vc) {
#if CYTHON_ASSUME_SAFE_MACROS && CYTHON_ASSUME_SAFE_SIZE
        return __Pyx_PyVectorcall_FastCallDict(func, vc, &PyTuple_GET_ITEM(args, 0), (size_t)PyTuple_GET_SIZE(args), kw);
#else
        (void) &__Pyx_PyVectorcall_FastCallDict;
        return PyVectorcall_Call(func, args, kw);
#endif
    }
#endif
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
#if CYTHON_ASSUME_SAFE_SIZE
        argc = PyTuple_GET_SIZE(args);
#else
        argc = PyTuple_Size(args);
        if (unlikely(argc < 0)) return NULL;
#endif
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            PyErr_Format(PyExc_TypeError,
                         "unbound method %.200S() needs an argument",
                         cyfunc->func_qualname);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
#if CYTHON_METH_FASTCALL && (CYTHON_VECTORCALL || CYTHON_BACKPORT_VECTORCALL)
static CYTHON_INLINE int __Pyx_CyFunction_Vectorcall_CheckArgs(__pyx_CyFunctionObject *cyfunc, Py_ssize_t nargs, PyObject *kwnames)
{
    int ret = 0;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        if (unlikely(nargs < 1)) {
            __Pyx_CyFunction_raise_type_error(
                cyfunc, "needs an argument");
            return -1;
        }
        ret = 1;
    }
    if (unlikely(kwnames) && unlikely(__Pyx_PyTuple_GET_SIZE(kwnames))) {
        __Pyx_CyFunction_raise_type_error(
            cyfunc, "takes no keyword arguments");
        return -1;
    }
    return ret;
}
static PyObject * __Pyx_CyFunction_Vectorcall_NOARGS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 0)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes no arguments", nargs);
        return NULL;
    }
    return meth(self, NULL);
}
static PyObject * __Pyx_CyFunction_Vectorcall_O(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, kwnames)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    if (unlikely(nargs != 1)) {
        __Pyx_CyFunction_raise_argument_count_error(
            cyfunc, "takes exactly one argument", nargs);
        return NULL;
    }
    return meth(self, args[0]);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCFunctionFastWithKeywords)(void(*)(void))meth)(self, args, nargs, kwnames);
}
static PyObject * __Pyx_CyFunction_Vectorcall_FASTCALL_KEYWORDS_METHOD(PyObject *func, PyObject *const *args, size_t nargsf, PyObject *kwnames)
{
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *)func;
    PyTypeObject *cls = (PyTypeObject *) __Pyx_CyFunction_GetClassObj(cyfunc);
#if CYTHON_BACKPORT_VECTORCALL
    Py_ssize_t nargs = (Py_ssize_t)nargsf;
#else
    Py_ssize_t nargs = PyVectorcall_NARGS(nargsf);
#endif
    PyObject *self;
#if CYTHON_COMPILING_IN_LIMITED_API
    PyCFunction meth = PyCFunction_GetFunction(cyfunc->func);
    if (unlikely(!meth)) return NULL;
#else
    PyCFunction meth = ((PyCFunctionObject*)cyfunc)->m_ml->ml_meth;
#endif
    switch (__Pyx_CyFunction_Vectorcall_CheckArgs(cyfunc, nargs, NULL)) {
    case 1:
        self = args[0];
        args += 1;
        nargs -= 1;
        break;
    case 0:
#if CYTHON_COMPILING_IN_LIMITED_API
        self = PyCFunction_GetSelf(((__pyx_CyFunctionObject*)cyfunc)->func);
        if (unlikely(!self) && PyErr_Occurred()) return NULL;
#else
        self = ((PyCFunctionObject*)cyfunc)->m_self;
#endif
        break;
    default:
        return NULL;
    }
    return ((__Pyx_PyCMethod)(void(*)(void))meth)(self, cls, args, (size_t)nargs, kwnames);
}
#endif
static PyType_Slot __pyx_CyFunctionType_slots[] = {
    {Py_tp_dealloc, (void *)__Pyx_CyFunction_dealloc},
    {Py_tp_repr, (void *)__Pyx_CyFunction_repr},
    {Py_tp_call, (void *)__Pyx_CyFunction_CallAsMethod},
    {Py_tp_traverse, (void *)__Pyx_CyFunction_traverse},
    {Py_tp_clear, (void *)__Pyx_CyFunction_clear},
    {Py_tp_methods, (void *)__pyx_CyFunction_methods},
    {Py_tp_members, (void *)__pyx_CyFunction_members},
    {Py_tp_getset, (void *)__pyx_CyFunction_getsets},
    {Py_tp_descr_get, (void *)__Pyx_PyMethod_New},
    {0, 0},
};
static PyType_Spec __pyx_CyFunctionType_spec = {
    __PYX_TYPE_MODULE_PREFIX "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
#ifdef Py_TPFLAGS_METHOD_DESCRIPTOR
    Py_TPFLAGS_METHOD_DESCRIPTOR |
#endif
#if CYTHON_METH_FASTCALL
#if defined(Py_TPFLAGS_HAVE_VECTORCALL)
    Py_TPFLAGS_HAVE_VECTORCALL |
#elif defined(_Py_TPFLAGS_HAVE_VECTORCALL)
    _Py_TPFLAGS_HAVE_VECTORCALL |
#endif
#endif // CYTHON_METH_FASTCALL
#if PY_VERSION_HEX >= 0x030A0000
    Py_TPFLAGS_IMMUTABLETYPE |
#endif
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_BASETYPE,
    __pyx_CyFunctionType_slots
};
static int __pyx_CyFunction_init(PyObject *module) {
    __pyx_mstatetype *mstate = __Pyx_PyModule_GetState(module);
    mstate->__pyx_CyFunctionType = __Pyx_FetchCommonTypeFromSpec(
        mstate->__pyx_CommonTypesMetaclassType, module, &__pyx_CyFunctionType_spec, NULL);
    if (unlikely(mstate->__pyx_CyFunctionType == NULL)) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_InitDefaults(PyObject *func, PyTypeObject *defaults_type) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_CallObject((PyObject*)defaults_type, NULL); // _PyObject_New(defaults_type);
    if (unlikely(!m->defaults))
        return NULL;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* CythonFunction */
static PyObject *__Pyx_CyFunction_New(PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    PyObject *op = __Pyx_CyFunction_Init(
        PyObject_GC_New(__pyx_CyFunctionObject, __pyx_mstate_global->__pyx_CyFunctionType),
        ml, flags, qualname, closure, module, globals, code
    );
    if (likely(op)) {
        PyObject_GC_Track(op);
    }
    return op;
}

/* CLineInTraceback */
#if CYTHON_CLINE_IN_TRACEBACK && CYTHON_CLINE_IN_TRACEBACK_RUNTIME
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    CYTHON_MAYBE_UNUSED_VAR(tstate);
    if (unlikely(!__pyx_mstate_global->__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_mstate_global->__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __Pyx_BEGIN_CRITICAL_SECTION(*cython_runtime_dict);
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_mstate_global->__pyx_n_u_cline_in_traceback))
        Py_XINCREF(use_cline);
        __Pyx_END_CRITICAL_SECTION();
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStrNoError(__pyx_mstate_global->__pyx_cython_runtime, __pyx_mstate_global->__pyx_n_u_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_INCREF(use_cline);
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        (void) PyObject_SetAttr(__pyx_mstate_global->__pyx_cython_runtime, __pyx_mstate_global->__pyx_n_u_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    Py_XDECREF(use_cline);
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static __Pyx_CachedCodeObjectType *__pyx__find_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line) {
    __Pyx_CachedCodeObjectType* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!code_cache->entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if (unlikely(pos >= code_cache->count) || unlikely(code_cache->entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = code_cache->entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static __Pyx_CachedCodeObjectType *__pyx_find_code_object(int code_line) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__find_code_object;
    return NULL; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just miss.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type old_count = __pyx_atomic_incr_acq_rel(&code_cache->accessor_count);
    if (old_count < 0) {
        __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
        return NULL;
    }
#endif
    __Pyx_CachedCodeObjectType *result = __pyx__find_code_object(code_cache, code_line);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_decr_acq_rel(&code_cache->accessor_count);
#endif
    return result;
#endif
}
static void __pyx__insert_code_object(struct __Pyx_CodeObjectCache *code_cache, int code_line, __Pyx_CachedCodeObjectType* code_object)
{
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = code_cache->entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            code_cache->entries = entries;
            code_cache->max_count = 64;
            code_cache->count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(code_cache->entries, code_cache->count, code_line);
    if ((pos < code_cache->count) && unlikely(code_cache->entries[pos].code_line == code_line)) {
        __Pyx_CachedCodeObjectType* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_INCREF(code_object);
        Py_DECREF(tmp);
        return;
    }
    if (code_cache->count == code_cache->max_count) {
        int new_max = code_cache->max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            code_cache->entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        code_cache->entries = entries;
        code_cache->max_count = new_max;
    }
    for (i=code_cache->count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    code_cache->count++;
    Py_INCREF(code_object);
}
static void __pyx_insert_code_object(int code_line, __Pyx_CachedCodeObjectType* code_object) {
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING && !CYTHON_ATOMICS
    (void)__pyx__insert_code_object;
    return; // Most implementation should have atomics. But otherwise, don't make it thread-safe, just fail.
#else
    struct __Pyx_CodeObjectCache *code_cache = &__pyx_mstate_global->__pyx_code_cache;
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_nonatomic_int_type expected = 0;
    if (!__pyx_atomic_int_cmp_exchange(&code_cache->accessor_count, &expected, INT_MIN)) {
        return;
    }
#endif
    __pyx__insert_code_object(code_cache, code_line, code_object);
#if CYTHON_COMPILING_IN_CPYTHON_FREETHREADING
    __pyx_atomic_sub(&code_cache->accessor_count, INT_MIN);
#endif
#endif
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
#if PY_VERSION_HEX >= 0x030b00a6 && !CYTHON_COMPILING_IN_LIMITED_API && !defined(PYPY_VERSION)
  #ifndef Py_BUILD_CORE
    #define Py_BUILD_CORE 1
  #endif
  #include "internal/pycore_frame.h"
#endif
#if CYTHON_COMPILING_IN_LIMITED_API
static PyObject *__Pyx_PyCode_Replace_For_AddTraceback(PyObject *code, PyObject *scratch_dict,
                                                       PyObject *firstlineno, PyObject *name) {
    PyObject *replace = NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_firstlineno", firstlineno))) return NULL;
    if (unlikely(PyDict_SetItemString(scratch_dict, "co_name", name))) return NULL;
    replace = PyObject_GetAttrString(code, "replace");
    if (likely(replace)) {
        PyObject *result = PyObject_Call(replace, __pyx_mstate_global->__pyx_empty_tuple, scratch_dict);
        Py_DECREF(replace);
        return result;
    }
    PyErr_Clear();
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyObject *code_object = NULL, *py_py_line = NULL, *py_funcname = NULL, *dict = NULL;
    PyObject *replace = NULL, *getframe = NULL, *frame = NULL;
    PyObject *exc_type, *exc_value, *exc_traceback;
    int success = 0;
    if (c_line) {
        (void) __pyx_cfilenm;
        (void) __Pyx_CLineForTraceback(__Pyx_PyThreadState_Current, c_line);
    }
    PyErr_Fetch(&exc_type, &exc_value, &exc_traceback);
    code_object = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!code_object) {
        code_object = Py_CompileString("_getframe()", filename, Py_eval_input);
        if (unlikely(!code_object)) goto bad;
        py_py_line = PyLong_FromLong(py_line);
        if (unlikely(!py_py_line)) goto bad;
        py_funcname = PyUnicode_FromString(funcname);
        if (unlikely(!py_funcname)) goto bad;
        dict = PyDict_New();
        if (unlikely(!dict)) goto bad;
        {
            PyObject *old_code_object = code_object;
            code_object = __Pyx_PyCode_Replace_For_AddTraceback(code_object, dict, py_py_line, py_funcname);
            Py_DECREF(old_code_object);
        }
        if (unlikely(!code_object)) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, code_object);
    } else {
        dict = PyDict_New();
    }
    getframe = PySys_GetObject("_getframe");
    if (unlikely(!getframe)) goto bad;
    if (unlikely(PyDict_SetItemString(dict, "_getframe", getframe))) goto bad;
    frame = PyEval_EvalCode(code_object, dict, dict);
    if (unlikely(!frame) || frame == Py_None) goto bad;
    success = 1;
  bad:
    PyErr_Restore(exc_type, exc_value, exc_traceback);
    Py_XDECREF(code_object);
    Py_XDECREF(py_py_line);
    Py_XDECREF(py_funcname);
    Py_XDECREF(dict);
    Py_XDECREF(replace);
    if (success) {
        PyTraceBack_Here(
            (struct _frame*)frame);
    }
    Py_XDECREF(frame);
}
#else
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = NULL;
    PyObject *py_funcname = NULL;
    if (c_line) {
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        if (!py_funcname) goto bad;
        funcname = PyUnicode_AsUTF8(py_funcname);
        if (!funcname) goto bad;
    }
    py_code = PyCode_NewEmpty(filename, funcname, py_line);
    Py_XDECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject *ptype, *pvalue, *ptraceback;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) {
            /* If the code object creation fails, then we should clear the
               fetched exception references and propagate the new exception */
            Py_XDECREF(ptype);
            Py_XDECREF(pvalue);
            Py_XDECREF(ptraceback);
            goto bad;
        }
        __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_mstate_global->__pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}
#endif

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int64_t(int64_t value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int64_t) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int64_t) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int64_t) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int64_t) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int64_t),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int64_t));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int64_t __Pyx_PyLong_As_int64_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int64_t neg_one = (int64_t) -1, const_zero = (int64_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int64_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int64_t) -1;
        val = __Pyx_PyLong_As_int64_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 2 * PyLong_SHIFT)) {
                            return (int64_t) (((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 3 * PyLong_SHIFT)) {
                            return (int64_t) (((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) >= 4 * PyLong_SHIFT)) {
                            return (int64_t) (((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int64_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int64_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int64_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int64_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int64_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (int64_t) ((((((int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int64_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int64_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int64_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) (((int64_t)-1)*(((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int64_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int64_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int64_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (int64_t) ((((((((((int64_t)digits[3]) << PyLong_SHIFT) | (int64_t)digits[2]) << PyLong_SHIFT) | (int64_t)digits[1]) << PyLong_SHIFT) | (int64_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int64_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int64_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int64_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int64_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int64_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int64_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int64_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int64_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int64_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int64_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int64_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int64_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int64_t) 1) << (sizeof(int64_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int64_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int64_t");
    return (int64_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int64_t");
    return (int64_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(int));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyLong_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        int val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyLong_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 2 * PyLong_SHIFT)) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 3 * PyLong_SHIFT)) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) >= 4 * PyLong_SHIFT)) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (int) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(int) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(int, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(int) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(int) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(int) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(int) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(int) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(int) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(int) - 1 > 4 * PyLong_SHIFT)) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(int) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(int) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        int val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (int) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (int) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (int) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (int) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(int) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((int) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(int) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((int) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((int) 1) << (sizeof(int) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (int) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyLong_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        long val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyLong_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 2 * PyLong_SHIFT)) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 3 * PyLong_SHIFT)) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) >= 4 * PyLong_SHIFT)) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (long) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(long) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(long, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(long) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(long) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(long) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(long) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(long) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(long) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(long) - 1 > 4 * PyLong_SHIFT)) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(long) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(long) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        long val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (long) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (long) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (long) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (long) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(long) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((long) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(long) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((long) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((long) 1) << (sizeof(long) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (long) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_enum____pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind neg_one = (enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind) -1, const_zero = (enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(enum __pyx_t_7pyflink_12fn_execution_15coder_impl_fast_InternalRowKind));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyLong_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (unlikely(!PyLong_Check(x))) {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_Long(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyLong_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
    if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
        if (unlikely(__Pyx_PyLong_IsNeg(x))) {
            goto raise_neg_overflow;
        } else if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_upylong, __Pyx_PyLong_CompactValueUnsigned(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_DigitCount(x)) {
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 2 * PyLong_SHIFT)) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 3 * PyLong_SHIFT)) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) >= 4 * PyLong_SHIFT)) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
        }
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030C00A7
        if (unlikely(Py_SIZE(x) < 0)) {
            goto raise_neg_overflow;
        }
#else
        {
            int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
            if (unlikely(result < 0))
                return (size_t) -1;
            if (unlikely(result == 1))
                goto raise_neg_overflow;
        }
#endif
        if ((sizeof(size_t) <= sizeof(unsigned long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
        }
    } else {
#if CYTHON_USE_PYLONG_INTERNALS
        if (__Pyx_PyLong_IsCompact(x)) {
            __PYX_VERIFY_RETURN_INT(size_t, __Pyx_compact_pylong, __Pyx_PyLong_CompactValue(x))
        } else {
            const digit* digits = __Pyx_PyLong_Digits(x);
            assert(__Pyx_PyLong_DigitCount(x) > 1);
            switch (__Pyx_PyLong_SignedDigitCount(x)) {
                case -2:
                    if ((8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if ((8 * sizeof(size_t) > 1 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 2 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if ((8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if ((8 * sizeof(size_t) > 2 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 3 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if ((8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if ((8 * sizeof(size_t) > 3 * PyLong_SHIFT)) {
                        if ((8 * sizeof(unsigned long) > 4 * PyLong_SHIFT)) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if ((8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT)) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
        }
#endif
        if ((sizeof(size_t) <= sizeof(long))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
        } else if ((sizeof(size_t) <= sizeof(PY_LONG_LONG))) {
            __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
        }
    }
    {
        size_t val;
        int ret = -1;
#if PY_VERSION_HEX >= 0x030d00A6 && !CYTHON_COMPILING_IN_LIMITED_API
        Py_ssize_t bytes_copied = PyLong_AsNativeBytes(
            x, &val, sizeof(val), Py_ASNATIVEBYTES_NATIVE_ENDIAN | (is_unsigned ? Py_ASNATIVEBYTES_UNSIGNED_BUFFER | Py_ASNATIVEBYTES_REJECT_NEGATIVE : 0));
        if (unlikely(bytes_copied == -1)) {
        } else if (unlikely(bytes_copied > (Py_ssize_t) sizeof(val))) {
            goto raise_overflow;
        } else {
            ret = 0;
        }
#elif PY_VERSION_HEX < 0x030d0000 && !(CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_LIMITED_API) || defined(_PyLong_AsByteArray)
        int one = 1; int is_little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&val;
        ret = _PyLong_AsByteArray((PyLongObject *)x,
                                    bytes, sizeof(val),
                                    is_little, !is_unsigned);
#else
        PyObject *v;
        PyObject *stepval = NULL, *mask = NULL, *shift = NULL;
        int bits, remaining_bits, is_negative = 0;
        int chunk_size = (sizeof(long) < 8) ? 30 : 62;
        if (likely(PyLong_CheckExact(x))) {
            v = __Pyx_NewRef(x);
        } else {
            v = PyNumber_Long(x);
            if (unlikely(!v)) return (size_t) -1;
            assert(PyLong_CheckExact(v));
        }
        {
            int result = PyObject_RichCompareBool(v, Py_False, Py_LT);
            if (unlikely(result < 0)) {
                Py_DECREF(v);
                return (size_t) -1;
            }
            is_negative = result == 1;
        }
        if (is_unsigned && unlikely(is_negative)) {
            Py_DECREF(v);
            goto raise_neg_overflow;
        } else if (is_negative) {
            stepval = PyNumber_Invert(v);
            Py_DECREF(v);
            if (unlikely(!stepval))
                return (size_t) -1;
        } else {
            stepval = v;
        }
        v = NULL;
        val = (size_t) 0;
        mask = PyLong_FromLong((1L << chunk_size) - 1); if (unlikely(!mask)) goto done;
        shift = PyLong_FromLong(chunk_size); if (unlikely(!shift)) goto done;
        for (bits = 0; bits < (int) sizeof(size_t) * 8 - chunk_size; bits += chunk_size) {
            PyObject *tmp, *digit;
            long idigit;
            digit = PyNumber_And(stepval, mask);
            if (unlikely(!digit)) goto done;
            idigit = PyLong_AsLong(digit);
            Py_DECREF(digit);
            if (unlikely(idigit < 0)) goto done;
            val |= ((size_t) idigit) << bits;
            tmp = PyNumber_Rshift(stepval, shift);
            if (unlikely(!tmp)) goto done;
            Py_DECREF(stepval); stepval = tmp;
        }
        Py_DECREF(shift); shift = NULL;
        Py_DECREF(mask); mask = NULL;
        {
            long idigit = PyLong_AsLong(stepval);
            if (unlikely(idigit < 0)) goto done;
            remaining_bits = ((int) sizeof(size_t) * 8) - bits - (is_unsigned ? 0 : 1);
            if (unlikely(idigit >= (1L << remaining_bits)))
                goto raise_overflow;
            val |= ((size_t) idigit) << bits;
        }
        if (!is_unsigned) {
            if (unlikely(val & (((size_t) 1) << (sizeof(size_t) * 8 - 1))))
                goto raise_overflow;
            if (is_negative)
                val = ~val;
        }
        ret = 0;
    done:
        Py_XDECREF(shift);
        Py_XDECREF(mask);
        Py_XDECREF(stepval);
#endif
        if (unlikely(ret))
            return (size_t) -1;
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyLong_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyLong_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#if defined(HAVE_LONG_LONG) && !CYTHON_COMPILING_IN_PYPY
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyLong_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        unsigned char *bytes = (unsigned char *)&value;
#if !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d00A4
        if (is_unsigned) {
            return PyLong_FromUnsignedNativeBytes(bytes, sizeof(value), -1);
        } else {
            return PyLong_FromNativeBytes(bytes, sizeof(value), -1);
        }
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX < 0x030d0000
        int one = 1; int little = (int)*(unsigned char *)&one;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
#else
        int one = 1; int little = (int)*(unsigned char *)&one;
        PyObject *from_bytes, *result = NULL, *kwds = NULL;
        PyObject *py_bytes = NULL, *order_str = NULL;
        from_bytes = PyObject_GetAttrString((PyObject*)&PyLong_Type, "from_bytes");
        if (!from_bytes) return NULL;
        py_bytes = PyBytes_FromStringAndSize((char*)bytes, sizeof(long));
        if (!py_bytes) goto limited_bad;
        order_str = PyUnicode_FromString(little ? "little" : "big");
        if (!order_str) goto limited_bad;
        {
            PyObject *args[3+(CYTHON_VECTORCALL ? 1 : 0)] = { NULL, py_bytes, order_str };
            if (!is_unsigned) {
                kwds = __Pyx_MakeVectorcallBuilderKwds(1);
                if (!kwds) goto limited_bad;
                if (__Pyx_VectorcallBuilder_AddArgStr("signed", __Pyx_NewRef(Py_True), kwds, args+3, 0) < 0) goto limited_bad;
            }
            result = __Pyx_Object_Vectorcall_CallFromBuilder(from_bytes, args+1, 2 | __Pyx_PY_VECTORCALL_ARGUMENTS_OFFSET, kwds);
        }
        limited_bad:
        Py_XDECREF(kwds);
        Py_XDECREF(order_str);
        Py_XDECREF(py_bytes);
        Py_XDECREF(from_bytes);
        return result;
#endif
    }
}

/* FormatTypeName */
#if CYTHON_COMPILING_IN_LIMITED_API && __PYX_LIMITED_VERSION_HEX < 0x030d0000
static __Pyx_TypeName
__Pyx_PyType_GetFullyQualifiedName(PyTypeObject* tp)
{
    PyObject *module = NULL, *name = NULL, *result = NULL;
    #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
    name = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_qualname);
    #else
    name = PyType_GetQualName(tp);
    #endif
    if (unlikely(name == NULL) || unlikely(!PyUnicode_Check(name))) goto bad;
    module = __Pyx_PyObject_GetAttrStr((PyObject *)tp,
                                               __pyx_mstate_global->__pyx_n_u_module);
    if (unlikely(module == NULL) || unlikely(!PyUnicode_Check(module))) goto bad;
    if (PyUnicode_CompareWithASCIIString(module, "builtins") == 0) {
        result = name;
        name = NULL;
        goto done;
    }
    result = PyUnicode_FromFormat("%U.%U", module, name);
    if (unlikely(result == NULL)) goto bad;
  done:
    Py_XDECREF(name);
    Py_XDECREF(module);
    return result;
  bad:
    PyErr_Clear();
    if (name) {
        result = name;
        name = NULL;
    } else {
        result = __Pyx_NewRef(__pyx_mstate_global->__pyx_kp_u__2);
    }
    goto done;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = __Pyx_PyType_GetSlot(a, tp_base, PyTypeObject*);
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
static CYTHON_INLINE int __Pyx_IsAnySubtype2(PyTypeObject *cls, PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (cls == a || cls == b) return 1;
    mro = cls->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            PyObject *base = PyTuple_GET_ITEM(mro, i);
            if (base == (PyObject *)a || base == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(cls, a) || __Pyx_InBases(cls, b);
}
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    if (exc_type1) {
        return __Pyx_IsAnySubtype2((PyTypeObject*)err, (PyTypeObject*)exc_type1, (PyTypeObject*)exc_type2);
    } else {
        return __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
}
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* GetRuntimeVersion */
static unsigned long __Pyx_get_runtime_version(void) {
#if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
    return Py_Version & ~0xFFUL;
#else
    static unsigned long __Pyx_cached_runtime_version = 0;
    if (__Pyx_cached_runtime_version == 0) {
        const char* rt_version = Py_GetVersion();
        unsigned long version = 0;
        unsigned long factor = 0x01000000UL;
        unsigned int digit = 0;
        int i = 0;
        while (factor) {
            while ('0' <= rt_version[i] && rt_version[i] <= '9') {
                digit = digit * 10 + (unsigned int) (rt_version[i] - '0');
                ++i;
            }
            version += factor * digit;
            if (rt_version[i] != '.')
                break;
            digit = 0;
            factor >>= 8;
            ++i;
        }
        __Pyx_cached_runtime_version = version;
    }
    return __Pyx_cached_runtime_version;
#endif
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(unsigned long ct_version, unsigned long rt_version, int allow_newer) {
    const unsigned long MAJOR_MINOR = 0xFFFF0000UL;
    if ((rt_version & MAJOR_MINOR) == (ct_version & MAJOR_MINOR))
        return 0;
    if (likely(allow_newer && (rt_version & MAJOR_MINOR) > (ct_version & MAJOR_MINOR)))
        return 1;
    {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compile time Python version %d.%d "
                      "of module '%.100s' "
                      "%s "
                      "runtime version %d.%d",
                       (int) (ct_version >> 24), (int) ((ct_version >> 16) & 0xFF),
                       __Pyx_MODULE_NAME,
                       (allow_newer) ? "was newer than" : "does not match",
                       (int) (rt_version >> 24), (int) ((rt_version >> 16) & 0xFF)
       );
        return PyErr_WarnEx(NULL, message, 1);
    }
}

/* VoidPtrImport */
#ifndef __PYX_HAVE_RT_ImportVoidPtr_3_1_3
#define __PYX_HAVE_RT_ImportVoidPtr_3_1_3
static int __Pyx_ImportVoidPtr_3_1_3(PyObject *module, const char *name, void **p, const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    d = PyObject_GetAttrString(module, "__pyx_capi__");
    if (!d)
        goto bad;
#if (defined(Py_LIMITED_API) && Py_LIMITED_API >= 0x030d0000) || (!defined(Py_LIMITED_API) && PY_VERSION_HEX >= 0x030d0000)
    PyDict_GetItemStringRef(d, name, &cobj);
#else
    cobj = PyDict_GetItemString(d, name);
    Py_XINCREF(cobj);
#endif
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C variable %.200s",
                PyModule_GetName(module), name);
        goto bad;
    }
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C variable %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), name, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    *p = PyCapsule_GetPointer(cobj, sig);
    if (!(*p))
        goto bad;
    Py_DECREF(d);
    Py_DECREF(cobj);
    return 0;
bad:
    Py_XDECREF(d);
    Py_XDECREF(cobj);
    return -1;
}
#endif

/* NewCodeObj */
#if CYTHON_COMPILING_IN_LIMITED_API
    static PyObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                       PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                       PyObject *fv, PyObject *cell, PyObject* fn,
                                       PyObject *name, int fline, PyObject *lnos) {
        PyObject *exception_table = NULL;
        PyObject *types_module=NULL, *code_type=NULL, *result=NULL;
        #if __PYX_LIMITED_VERSION_HEX < 0x030b0000
        PyObject *version_info;
        PyObject *py_minor_version = NULL;
        #endif
        long minor_version = 0;
        PyObject *type, *value, *traceback;
        PyErr_Fetch(&type, &value, &traceback);
        #if __PYX_LIMITED_VERSION_HEX >= 0x030b0000
        minor_version = 11;
        #else
        if (!(version_info = PySys_GetObject("version_info"))) goto end;
        if (!(py_minor_version = PySequence_GetItem(version_info, 1))) goto end;
        minor_version = PyLong_AsLong(py_minor_version);
        Py_DECREF(py_minor_version);
        if (minor_version == -1 && PyErr_Occurred()) goto end;
        #endif
        if (!(types_module = PyImport_ImportModule("types"))) goto end;
        if (!(code_type = PyObject_GetAttrString(types_module, "CodeType"))) goto end;
        if (minor_version <= 7) {
            (void)p;
            result = PyObject_CallFunction(code_type, "iiiiiOOOOOOiOOO", a, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else if (minor_version <= 10) {
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOiOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, fline, lnos, fv, cell);
        } else {
            if (!(exception_table = PyBytes_FromStringAndSize(NULL, 0))) goto end;
            result = PyObject_CallFunction(code_type, "iiiiiiOOOOOOOiOOOO", a,p, k, l, s, f, code,
                          c, n, v, fn, name, name, fline, lnos, exception_table, fv, cell);
        }
    end:
        Py_XDECREF(code_type);
        Py_XDECREF(exception_table);
        Py_XDECREF(types_module);
        if (type) {
            PyErr_Restore(type, value, traceback);
        }
        return result;
    }
#elif PY_VERSION_HEX >= 0x030B0000
  static PyCodeObject* __Pyx__PyCode_New(int a, int p, int k, int l, int s, int f,
                                         PyObject *code, PyObject *c, PyObject* n, PyObject *v,
                                         PyObject *fv, PyObject *cell, PyObject* fn,
                                         PyObject *name, int fline, PyObject *lnos) {
    PyCodeObject *result;
    result =
      #if PY_VERSION_HEX >= 0x030C0000
        PyUnstable_Code_NewWithPosOnlyArgs
      #else
        PyCode_NewWithPosOnlyArgs
      #endif
        (a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, name, fline, lnos, __pyx_mstate_global->__pyx_empty_bytes);
    return result;
  }
#elif PY_VERSION_HEX >= 0x030800B2 && !CYTHON_COMPILING_IN_PYPY
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_NewWithPosOnlyArgs(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx__PyCode_New(a, p, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
static PyObject* __Pyx_PyCode_New(
        const __Pyx_PyCode_New_function_description descr,
        PyObject * const *varnames,
        PyObject *filename,
        PyObject *funcname,
        const char *line_table,
        PyObject *tuple_dedup_map
) {
    PyObject *code_obj = NULL, *varnames_tuple_dedup = NULL, *code_bytes = NULL, *line_table_bytes = NULL;
    Py_ssize_t var_count = (Py_ssize_t) descr.nlocals;
    PyObject *varnames_tuple = PyTuple_New(var_count);
    if (unlikely(!varnames_tuple)) return NULL;
    for (Py_ssize_t i=0; i < var_count; i++) {
        Py_INCREF(varnames[i]);
        if (__Pyx_PyTuple_SET_ITEM(varnames_tuple, i, varnames[i]) != (0)) goto done;
    }
    #if CYTHON_COMPILING_IN_LIMITED_API
    varnames_tuple_dedup = PyDict_GetItem(tuple_dedup_map, varnames_tuple);
    if (!varnames_tuple_dedup) {
        if (unlikely(PyDict_SetItem(tuple_dedup_map, varnames_tuple, varnames_tuple) < 0)) goto done;
        varnames_tuple_dedup = varnames_tuple;
    }
    #else
    varnames_tuple_dedup = PyDict_SetDefault(tuple_dedup_map, varnames_tuple, varnames_tuple);
    if (unlikely(!varnames_tuple_dedup)) goto done;
    #endif
    #if CYTHON_AVOID_BORROWED_REFS
    Py_INCREF(varnames_tuple_dedup);
    #endif
    if (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table != NULL
        && !CYTHON_COMPILING_IN_GRAAL) {
        line_table_bytes = PyBytes_FromStringAndSize(line_table, descr.line_table_length);
        if (unlikely(!line_table_bytes)) goto done;
        Py_ssize_t code_len = (descr.line_table_length * 2 + 4) & ~3;
        code_bytes = PyBytes_FromStringAndSize(NULL, code_len);
        if (unlikely(!code_bytes)) goto done;
        char* c_code_bytes = PyBytes_AsString(code_bytes);
        if (unlikely(!c_code_bytes)) goto done;
        memset(c_code_bytes, 0, (size_t) code_len);
    }
    code_obj = (PyObject*) __Pyx__PyCode_New(
        (int) descr.argcount,
        (int) descr.num_posonly_args,
        (int) descr.num_kwonly_args,
        (int) descr.nlocals,
        0,
        (int) descr.flags,
        code_bytes ? code_bytes : __pyx_mstate_global->__pyx_empty_bytes,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        varnames_tuple_dedup,
        __pyx_mstate_global->__pyx_empty_tuple,
        __pyx_mstate_global->__pyx_empty_tuple,
        filename,
        funcname,
        (int) descr.first_line,
        (__PYX_LIMITED_VERSION_HEX >= (0x030b0000) && line_table_bytes) ? line_table_bytes : __pyx_mstate_global->__pyx_empty_bytes
    );
done:
    Py_XDECREF(code_bytes);
    Py_XDECREF(line_table_bytes);
    #if CYTHON_AVOID_BORROWED_REFS
    Py_XDECREF(varnames_tuple_dedup);
    #endif
    Py_DECREF(varnames_tuple);
    return code_obj;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry const *t, PyObject **target, const char* const* encoding_names) {
    while (t->s) {
        PyObject *str;
        if (t->is_unicode) {
            if (t->intern) {
                str = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                str = PyUnicode_Decode(t->s, t->n - 1, encoding_names[t->encoding], NULL);
            } else {
                str = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            str = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        if (!str)
            return -1;
        *target = str;
        if (PyObject_Hash(str) == -1)
            return -1;
        ++t;
        ++target;
    }
    return 0;
}

#include <string.h>
static CYTHON_INLINE Py_ssize_t __Pyx_ssize_strlen(const char *s) {
    size_t len = strlen(s);
    if (unlikely(len > (size_t) PY_SSIZE_T_MAX)) {
        PyErr_SetString(PyExc_OverflowError, "byte string is too long");
        return -1;
    }
    return (Py_ssize_t) len;
}
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return __Pyx_PyUnicode_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE PyObject* __Pyx_PyByteArray_FromString(const char* c_str) {
    Py_ssize_t len = __Pyx_ssize_strlen(c_str);
    if (unlikely(len < 0)) return NULL;
    return PyByteArray_FromStringAndSize(c_str, len);
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if CYTHON_COMPILING_IN_LIMITED_API
    {
        const char* result;
        Py_ssize_t unicode_length;
        CYTHON_MAYBE_UNUSED_VAR(unicode_length); // only for __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        #if __PYX_LIMITED_VERSION_HEX < 0x030A0000
        if (unlikely(PyArg_Parse(o, "s#", &result, length) < 0)) return NULL;
        #else
        result = PyUnicode_AsUTF8AndSize(o, length);
        #endif
        #if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        unicode_length = PyUnicode_GetLength(o);
        if (unlikely(unicode_length < 0)) return NULL;
        if (unlikely(unicode_length != *length)) {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
        #endif
        return result;
    }
#else
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
}
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_UTF8
    if (PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
    if (PyByteArray_Check(o)) {
#if (CYTHON_ASSUME_SAFE_SIZE && CYTHON_ASSUME_SAFE_MACROS) || (CYTHON_COMPILING_IN_PYPY && (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE)))
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
#else
        *length = PyByteArray_Size(o);
        if (*length == -1) return NULL;
        return PyByteArray_AsString(o);
#endif
    } else
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_LongWrongResultType(PyObject* result) {
    __Pyx_TypeName result_type_name = __Pyx_PyType_GetFullyQualifiedName(Py_TYPE(result));
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ").  "
                "The ability to return an instance of a strict subclass of int is deprecated, "
                "and may be removed in a future version of Python.",
                result_type_name)) {
            __Pyx_DECREF_TypeName(result_type_name);
            Py_DECREF(result);
            return NULL;
        }
        __Pyx_DECREF_TypeName(result_type_name);
        return result;
    }
    PyErr_Format(PyExc_TypeError,
                 "__int__ returned non-int (type " __Pyx_FMT_TYPENAME ")",
                 result_type_name);
    __Pyx_DECREF_TypeName(result_type_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_Long(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  PyObject *res = NULL;
  if (likely(PyLong_Check(x)))
      return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  if (likely(m && m->nb_int)) {
      res = m->nb_int(x);
  }
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
      res = PyNumber_Long(x);
  }
#endif
  if (likely(res)) {
      if (unlikely(!PyLong_CheckExact(res))) {
          return __Pyx_PyNumber_LongWrongResultType(res);
      }
  }
  else if (!PyErr_Occurred()) {
      PyErr_SetString(PyExc_TypeError,
                      "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(__Pyx_PyLong_IsCompact(b))) {
        return __Pyx_PyLong_CompactValue(b);
    } else {
      const digit* digits = __Pyx_PyLong_Digits(b);
      const Py_ssize_t size = __Pyx_PyLong_SignedDigitCount(b);
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyLong_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE Py_hash_t __Pyx_PyIndex_AsHash_t(PyObject* o) {
  if (sizeof(Py_hash_t) == sizeof(Py_ssize_t)) {
    return (Py_hash_t) __Pyx_PyIndex_AsSsize_t(o);
  } else {
    Py_ssize_t ival;
    PyObject *x;
    x = PyNumber_Index(o);
    if (!x) return -1;
    ival = PyLong_AsLong(x);
    Py_DECREF(x);
    return ival;
  }
}
static CYTHON_INLINE PyObject *__Pyx_Owned_Py_None(int b) {
    CYTHON_UNUSED_VAR(b);
    return __Pyx_NewRef(Py_None);
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyLong_FromSize_t(size_t ival) {
    return PyLong_FromSize_t(ival);
}


/* MultiPhaseInitModuleState */
#if CYTHON_PEP489_MULTI_PHASE_INIT && CYTHON_USE_MODULE_STATE
#ifndef CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#if (CYTHON_COMPILING_IN_LIMITED_API || PY_VERSION_HEX >= 0x030C0000)
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 1
#else
  #define CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE 0
#endif
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE && !CYTHON_ATOMICS
#error "Module state with PEP489 requires atomics. Currently that's one of\
 C11, C++11, gcc atomic intrinsics or MSVC atomic intrinsics"
#endif
#if !CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
#define __Pyx_ModuleStateLookup_Lock()
#define __Pyx_ModuleStateLookup_Unlock()
#elif !CYTHON_COMPILING_IN_LIMITED_API && PY_VERSION_HEX >= 0x030d0000
static PyMutex __Pyx_ModuleStateLookup_mutex = {0};
#define __Pyx_ModuleStateLookup_Lock() PyMutex_Lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() PyMutex_Unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(__cplusplus) && __cplusplus >= 201103L
#include <mutex>
static std::mutex __Pyx_ModuleStateLookup_mutex;
#define __Pyx_ModuleStateLookup_Lock() __Pyx_ModuleStateLookup_mutex.lock()
#define __Pyx_ModuleStateLookup_Unlock() __Pyx_ModuleStateLookup_mutex.unlock()
#elif defined(__STDC_VERSION__) && (__STDC_VERSION__ > 201112L) && !defined(__STDC_NO_THREADS__)
#include <threads.h>
static mtx_t __Pyx_ModuleStateLookup_mutex;
static once_flag __Pyx_ModuleStateLookup_mutex_once_flag = ONCE_FLAG_INIT;
static void __Pyx_ModuleStateLookup_initialize_mutex(void) {
    mtx_init(&__Pyx_ModuleStateLookup_mutex, mtx_plain);
}
#define __Pyx_ModuleStateLookup_Lock()\
  call_once(&__Pyx_ModuleStateLookup_mutex_once_flag, __Pyx_ModuleStateLookup_initialize_mutex);\
  mtx_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() mtx_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(HAVE_PTHREAD_H)
#include <pthread.h>
static pthread_mutex_t __Pyx_ModuleStateLookup_mutex = PTHREAD_MUTEX_INITIALIZER;
#define __Pyx_ModuleStateLookup_Lock() pthread_mutex_lock(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() pthread_mutex_unlock(&__Pyx_ModuleStateLookup_mutex)
#elif defined(_WIN32)
#include <Windows.h>  // synchapi.h on its own doesn't work
static SRWLOCK __Pyx_ModuleStateLookup_mutex = SRWLOCK_INIT;
#define __Pyx_ModuleStateLookup_Lock() AcquireSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#define __Pyx_ModuleStateLookup_Unlock() ReleaseSRWLockExclusive(&__Pyx_ModuleStateLookup_mutex)
#else
#error "No suitable lock available for CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE.\
 Requires C standard >= C11, or C++ standard >= C++11,\
 or pthreads, or the Windows 32 API, or Python >= 3.13."
#endif
typedef struct {
    int64_t id;
    PyObject *module;
} __Pyx_InterpreterIdAndModule;
typedef struct {
    char interpreter_id_as_index;
    Py_ssize_t count;
    Py_ssize_t allocated;
    __Pyx_InterpreterIdAndModule table[1];
} __Pyx_ModuleStateLookupData;
#define __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE 32
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_int_type __Pyx_ModuleStateLookup_read_counter = 0;
#endif
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static __pyx_atomic_ptr_type __Pyx_ModuleStateLookup_data = 0;
#else
static __Pyx_ModuleStateLookupData* __Pyx_ModuleStateLookup_data = NULL;
#endif
static __Pyx_InterpreterIdAndModule* __Pyx_State_FindModuleStateLookupTableLowerBound(
        __Pyx_InterpreterIdAndModule* table,
        Py_ssize_t count,
        int64_t interpreterId) {
    __Pyx_InterpreterIdAndModule* begin = table;
    __Pyx_InterpreterIdAndModule* end = begin + count;
    if (begin->id == interpreterId) {
        return begin;
    }
    while ((end - begin) > __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
        __Pyx_InterpreterIdAndModule* halfway = begin + (end - begin)/2;
        if (halfway->id == interpreterId) {
            return halfway;
        }
        if (halfway->id < interpreterId) {
            begin = halfway;
        } else {
            end = halfway;
        }
    }
    for (; begin < end; ++begin) {
        if (begin->id >= interpreterId) return begin;
    }
    return begin;
}
static PyObject *__Pyx_State_FindModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return NULL;
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData* data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
    {
        __pyx_atomic_incr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        if (likely(data)) {
            __Pyx_ModuleStateLookupData* new_data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_acquire(&__Pyx_ModuleStateLookup_data);
            if (likely(data == new_data)) {
                goto read_finished;
            }
        }
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
        __Pyx_ModuleStateLookup_Lock();
        __pyx_atomic_incr_relaxed(&__Pyx_ModuleStateLookup_read_counter);
        data = (__Pyx_ModuleStateLookupData*)__pyx_atomic_pointer_load_relaxed(&__Pyx_ModuleStateLookup_data);
        __Pyx_ModuleStateLookup_Unlock();
    }
  read_finished:;
#else
    __Pyx_ModuleStateLookupData* data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_InterpreterIdAndModule* found = NULL;
    if (unlikely(!data)) goto end;
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            found = data->table+interpreter_id;
        }
    } else {
        found = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
    }
  end:
    {
        PyObject *result=NULL;
        if (found && found->id == interpreter_id) {
            result = found->module;
        }
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
        __pyx_atomic_decr_acq_rel(&__Pyx_ModuleStateLookup_read_counter);
#endif
        return result;
    }
}
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
static void __Pyx_ModuleStateLookup_wait_until_no_readers(void) {
    while (__pyx_atomic_load(&__Pyx_ModuleStateLookup_read_counter) != 0);
}
#else
#define __Pyx_ModuleStateLookup_wait_until_no_readers()
#endif
static int __Pyx_State_AddModuleInterpIdAsIndex(__Pyx_ModuleStateLookupData **old_data, PyObject* module, int64_t interpreter_id) {
    Py_ssize_t to_allocate = (*old_data)->allocated;
    while (to_allocate <= interpreter_id) {
        if (to_allocate == 0) to_allocate = 1;
        else to_allocate *= 2;
    }
    __Pyx_ModuleStateLookupData *new_data = *old_data;
    if (to_allocate != (*old_data)->allocated) {
         new_data = (__Pyx_ModuleStateLookupData *)realloc(
            *old_data,
            sizeof(__Pyx_ModuleStateLookupData)+(to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
        if (!new_data) {
            PyErr_NoMemory();
            return -1;
        }
        for (Py_ssize_t i = new_data->allocated; i < to_allocate; ++i) {
            new_data->table[i].id = i;
            new_data->table[i].module = NULL;
        }
        new_data->allocated = to_allocate;
    }
    new_data->table[interpreter_id].module = module;
    if (new_data->count < interpreter_id+1) {
        new_data->count = interpreter_id+1;
    }
    *old_data = new_data;
    return 0;
}
static void __Pyx_State_ConvertFromInterpIdAsIndex(__Pyx_ModuleStateLookupData *data) {
    __Pyx_InterpreterIdAndModule *read = data->table;
    __Pyx_InterpreterIdAndModule *write = data->table;
    __Pyx_InterpreterIdAndModule *end = read + data->count;
    for (; read<end; ++read) {
        if (read->module) {
            write->id = read->id;
            write->module = read->module;
            ++write;
        }
    }
    data->count = write - data->table;
    for (; write<end; ++write) {
        write->id = 0;
        write->module = NULL;
    }
    data->interpreter_id_as_index = 0;
}
static int __Pyx_State_AddModule(PyObject* module, CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    int result = 0;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *old_data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *old_data = __Pyx_ModuleStateLookup_data;
#endif
    __Pyx_ModuleStateLookupData *new_data = old_data;
    if (!new_data) {
        new_data = (__Pyx_ModuleStateLookupData *)calloc(1, sizeof(__Pyx_ModuleStateLookupData));
        if (!new_data) {
            result = -1;
            PyErr_NoMemory();
            goto end;
        }
        new_data->allocated = 1;
        new_data->interpreter_id_as_index = 1;
    }
    __Pyx_ModuleStateLookup_wait_until_no_readers();
    if (new_data->interpreter_id_as_index) {
        if (interpreter_id < __PYX_MODULE_STATE_LOOKUP_SMALL_SIZE) {
            result = __Pyx_State_AddModuleInterpIdAsIndex(&new_data, module, interpreter_id);
            goto end;
        }
        __Pyx_State_ConvertFromInterpIdAsIndex(new_data);
    }
    {
        Py_ssize_t insert_at = 0;
        {
            __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
                new_data->table, new_data->count, interpreter_id);
            assert(lower_bound);
            insert_at = lower_bound - new_data->table;
            if (unlikely(insert_at < new_data->count && lower_bound->id == interpreter_id)) {
                lower_bound->module = module;
                goto end;  // already in table, nothing more to do
            }
        }
        if (new_data->count+1 >= new_data->allocated) {
            Py_ssize_t to_allocate = (new_data->count+1)*2;
            new_data =
                (__Pyx_ModuleStateLookupData*)realloc(
                    new_data,
                    sizeof(__Pyx_ModuleStateLookupData) +
                    (to_allocate-1)*sizeof(__Pyx_InterpreterIdAndModule));
            if (!new_data) {
                result = -1;
                new_data = old_data;
                PyErr_NoMemory();
                goto end;
            }
            new_data->allocated = to_allocate;
        }
        ++new_data->count;
        int64_t last_id = interpreter_id;
        PyObject *last_module = module;
        for (Py_ssize_t i=insert_at; i<new_data->count; ++i) {
            int64_t current_id = new_data->table[i].id;
            new_data->table[i].id = last_id;
            last_id = current_id;
            PyObject *current_module = new_data->table[i].module;
            new_data->table[i].module = last_module;
            last_module = current_module;
        }
    }
  end:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, new_data);
#else
    __Pyx_ModuleStateLookup_data = new_data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return result;
}
static int __Pyx_State_RemoveModule(CYTHON_UNUSED void* dummy) {
    int64_t interpreter_id = PyInterpreterState_GetID(__Pyx_PyInterpreterState_Get());
    if (interpreter_id == -1) return -1;
    __Pyx_ModuleStateLookup_Lock();
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __Pyx_ModuleStateLookupData *data = (__Pyx_ModuleStateLookupData *)
            __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, 0);
#else
    __Pyx_ModuleStateLookupData *data = __Pyx_ModuleStateLookup_data;
#endif
    if (data->interpreter_id_as_index) {
        if (interpreter_id < data->count) {
            data->table[interpreter_id].module = NULL;
        }
        goto done;
    }
    {
        __Pyx_ModuleStateLookup_wait_until_no_readers();
        __Pyx_InterpreterIdAndModule* lower_bound = __Pyx_State_FindModuleStateLookupTableLowerBound(
            data->table, data->count, interpreter_id);
        if (!lower_bound) goto done;
        if (lower_bound->id != interpreter_id) goto done;
        __Pyx_InterpreterIdAndModule *end = data->table+data->count;
        for (;lower_bound<end-1; ++lower_bound) {
            lower_bound->id = (lower_bound+1)->id;
            lower_bound->module = (lower_bound+1)->module;
        }
    }
    --data->count;
    if (data->count == 0) {
        free(data);
        data = NULL;
    }
  done:
#if CYTHON_MODULE_STATE_LOOKUP_THREAD_SAFE
    __pyx_atomic_pointer_exchange(&__Pyx_ModuleStateLookup_data, data);
#else
    __Pyx_ModuleStateLookup_data = data;
#endif
    __Pyx_ModuleStateLookup_Unlock();
    return 0;
}
#endif

/* #### Code section: utility_code_pragmas_end ### */
#ifdef _MSC_VER
#pragma warning( pop )
#endif



/* #### Code section: end ### */
#endif /* Py_PYTHON_H */
